//
//  kintKINT.m
//
//  Copyright (c) 2018 Everett Kaser. All rights reserved.
//

/*********************  Header Files  *********************/

#include <CoreFoundation/CoreFoundation.h>
#import <Cocoa/Cocoa.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include <mach/mach_time.h>
#include <unistd.h>

#import "kintMAIN.h"
#import "kintPMACH.h"
#import "kintView.h"
#import "kintAppDelegate.h"
#import "kDJPEG.h"

#if BETA
	#include <time.h>
	BYTE Expired=1;
	KTIME	betatime;
#endif

CFStringRef	gAppName;

@implementation kintKINT

@end

void WriteDebug(char *p);

KLONG rOfs=1, gOfs=2, bOfs=3;

void RectBMB(PBMB bmD, KLONG x1, KLONG y1, KLONG x2, KLONG y2, DWORD fill, DWORD edge);

typedef struct tagBITMAPFILEHEADER { // bmfh
	WORD    bfType;
	DWORD   bfSize;
	WORD    bfReserved1;
	WORD    bfReserved2;
	DWORD   bfOffBits;
} __attribute__((packed)) BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{ // bmih
	DWORD  biSize;
	int32_t   biWidth;
	int32_t   biHeight;
	WORD   biPlanes;
	WORD   biBitCount;
	DWORD  biCompression;
	DWORD  biSizeImage;
	int32_t   biXPelsPerMeter;
	int32_t   biYPelsPerMeter;
	DWORD  biClrUsed;
	DWORD  biClrImportant;
} __attribute__((packed)) BITMAPINFOHEADER;

typedef struct tagRGBQUAD { // rgbq
	BYTE    rgbBlue;
	BYTE    rgbGreen;
	BYTE    rgbRed;
	BYTE    rgbReserved;
} __attribute__((packed)) RGBQUAD;

typedef struct tagBITMAPINFO { // bmi
	BITMAPINFOHEADER bmiHeader;
	RGBQUAD          bmiColors[1];
} __attribute__((packed)) BITMAPINFO;


KLONG			TdBMW, TdBMH, TsBMW, TsBMH;
KLONG			TdclipX1, TdclipX2, TdclipY1, TdclipY2;
KLONG			TsclipX1, TsclipX2, TsclipY1, TsclipY2;
KLONG			TdstepX, TdstepY, TsstepX, TsstepY, TastepX, TastepY;
BYTE		*Tdpbits, *Tspbits, *Tapbits;

DWORD		LineStyle=0xFFFFFFFF;

DWORD		CurTextColor, CurBackColor;
WORD		CurFont;
WORD		AlphaBlendD=0, AlphaBlendS=255;	// 0-255: 0 keeps NONE, 255 keeps ALL, in between keeps some (set by SysAlphaBlend)

typedef struct {
	DWORD	BkColor;
	DWORD	TextColor;
	WORD	Font;
} OLDMODES;

OLDMODES	Moldie[MAXFONTS];
KLONG			MoldCnt=0;
//#define		MAXFONTS	100
BYTE		*pFontData;	// ptr to malloc'd RAM for fonts
WORD		NumFonts;	// number of fonts in pFontData
UNALIGNED DWORD		*pFontTable;	// ptr to table of offsets to fonts in pFontData
BYTE		FontSelection[MAXFONTS];	// used in prioritizing fonts in SysGetFont

// NOTE: This needs to be FALSE for smooth stretching in SHERLOCK but TRUE for speed in SOLITILE!!!!
#define NoInterpolateStretch   false

#define MAXALLOCCNT 15000
DWORD AllocCnt=0;
void *AllocPtrs[MAXALLOCCNT];

KLONG kRotated=0;

#define MAXTIMERS	16
NSTimer	*Timers[MAXTIMERS];
KLONG		timersCNT=0;

char	*fontNames[8] = {
	"Helvetica",
	"Courier",
	"Helvetica-Bold",
	"Courier-Bold",
	"Helvetica-Oblique",
	"Courier-Oblique",
	"Helvetica-BoldOblique",
	"Courier-BoldOblique",
};
//CGContextRef	kCTX=NULL;
KLONG		BMcnt=1;
BMB 	kBM[MAXBM];


FONT	kFONT[MAXFONTS];
KLONG		FONTcnt=0;

//******************************

KE		mainKE;	// struct for main .KE program
DWORD	KINTmsgs[4*MAXMSGS];
KLONG	KINTmsgadd=0, KINTmsgrem=0;
DWORD	KINTptrmove[4];
BOOL	KINTptrmovePENDING=false;

NSTimer *KINTmsgTimer=NULL;

//struct	_stat	fstats;


#ifdef CHECKWORDALIGN
DWORD savePC;
#define	defgetmemw
#endif
/*
#define	GEEK_TRACE
#define	GEEK_TRACE2
#define	GEEK_TRACE3
*/
#ifdef GEEK_TRACE
#define TBMAX		512
DWORD	TraceBuf[TBMAX], TBptr;
#endif
#ifdef GEEK_TRACE2
#define TBMAX2	512
DWORD	TraceBuf2[TBMAX2], TBptr2;
#endif
#ifdef GEEK_TRACE3
#define TBMAX3	512
DWORD	TraceBuf3[TBMAX3], TBptr3;
#endif

//************************************************************
//* PORTING NOTES
//* -------------
//* 1) In this code, a BYTE is 8 bits, a WORD is 16 bits,
//*		a DWORD is 32 bits, all unsigned.  A 'char' is 8 bits,
//*		a 'short' is 16 bits, a long is 32 bits, all signed.
//* 2) Multi-byte values in the KODEs are stored in little-endian
//*		order (what's used on current INTEL processors, the primary
//*		current target platform.  If you're porting this to a
//*		platform that uses a big-endian processor, you'll need to
//*		VERY carefully modify the multi-byte operations to grab
//*		the bytes in the correct order. See the GETREG/SETREG and
//*		GETMEM/SETMEM macros in pmach.h.
//* 3) This implementation knows that memory pointers are DWORDS
//*		and uses that fact to place them directly into "registers"
//*		R0, R1, and R2.  (The .KE code only saves and loads these
//*		values, NEVER does arithmetic on them.)  If this gets ported
//*		to a machine that uses (for example) 8-byte pointers, then
//*		those pointers would need to be saved in a table or structure,
//*		and a HANDLE (or index) would be placed in R0, R1, and R2
//*		in place of the actual memory pointer.  Then, the actual
//*		memory pointer would have to be fetched anytime it was needed
//*		by the execution of the .KE kodes.  Places where this is
//*		a concern, I've tried to mark with "WARN: PTR4".
//*		The same applies to MS Windows HANDLES (such as HDCs, etc).
//* 4) Wherever possible, I've tried to use as simple, lowest-level
//*		C-library functions as possible in order to keep the code
//*		as portable as possible, so that folks doing ports don't
//*		have to modify any more code than necessary.
//* 5) Kode (at this point) assumes that the screen size in
//*		pixels ranges from 640x480 to 1600x1280.  Larger (or
//*		smaller sizes may be supported, but that gives you a
//*		range that you may have to "scale to" if your actual
//*		display size is greatly different (like 10-20 years
//*		from now :-).
//* 5a) Support for Pocket PC screen size of 240x320 has been added
//*
//************************************************************

//*******************************************************************************
// Global general variables, common to all .KE programs running at any given time
BYTE	Kregs[256]; // 4*32 + 2*48 + 1*32 = 256 (32 DWORDS, 48 WORDS, 32 BYTES)
DWORD	result;	// MSbit, EQ, LSbit (ie, CPU flags for the KCPU, see NOTE above "ExecKE()")
BYTE	signedmath;
BYTE	killpgm=0;

//********************************************************************************
// Global graphics variables, common to all .KE programs running at any given time
DWORD		CurTextColor, CurBackColor;
WORD		CurFont;
WORD		AlphaBlend=255;	// 0-255: 0 keeps NONE, 255 keeps ALL, in between keeps some (set by SysAlphaBlend)

NSMutableData   *data=NULL; // used by bitmap RGBA conversion/manipulation code

BYTE		TmpFname[PATHLEN];

KLONG			ScrClipX1, ScrClipY1, ScrClipX2, ScrClipY2;
KLONG			ScrW, ScrH;

#define USE_CMD_KEY 1

// The idle timer period, in milliseconds
#define IDLE_MS 100

#define FAKE_SCREEN_SIZE	0
#define FAKE_SCREEN_SIZE_W	800
#define FAKE_SCREEN_SIZE_H	600

/*********************  Local Structures  *********************/

/*typedef struct
{
	Movie		mMovie;
	Handle		mDataRef;
	bool		mLooping;
	bool		mPlaying;
	bool		mAutoDelete;
	bool		mIsMusic;
	QTCallBack	qtCallback;
} movieSound_t;
static std::deque<movieSound_t *> sSoundOrphans;
*/

typedef struct
{
	int width;
	int height;
} windowSizes_t;

/*********************  Local Prototypes  *********************/

int wildicmp(const char *wild, const char *string);
bool CreateMacWindow (void);
void BuildPath(char *pMem, BYTE inFolderType, BOOL AppData);
//void StopSound_QT (movieSound_t *inSound);
//static int ConvertCFStringToUnicode (const CFStringRef inString, UniChar *outString);
void Kff(BOOL appdata, BOOL removeExtension);
void FixCSVdates(void);

//************************************************************
// pmach GLOBALS
KLONG		PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex;
WORD	MusicPlaying;
//int		CurCur;
//KE		mainKE;

/*******************  Global Variables ********************/

CFStringRef gAppName;

//WindowRef gMacWindow;
bool windowReady=false;
CGContextRef gCGContext;

bool gInBackground;
bool sIgnoreNextMouseUp;
volatile bool gInPrintLoop;
UInt32 gOSXVersion32;

//EventLoopTimerRef sTimerRef = NULL;
//EventLoopTimerRef sIdleTimerRef = NULL;
//DWORD sTimerPeriod;
//BOOL inSM=FALSE;

BOOL WinHide = false;
BOOL AppHide = false;
BOOL DockHide = false;

char	OSctrl_cmdSTR[32] = "Cmd";
TCHAR	PgmPath[PATHLEN], TmpPath[PATHLEN], AppDataPath[PATHLEN], AppPath[PATHLEN];
BOOL	CopyAppData=TRUE;
//FindFileList	*FirstNextList=NULL;
UInt32	lastModifiers;
WORD	MouseButtonsDown=0;

windowSizes_t windowSizes[] = {
	{ 640, 480 },
	{ 800, 600 },
	{ 1024, 768 },
	{ 1152, 864 },
	{ 1280, 1024 },
	{ 1600, 1200 },
	{ 60000, 60000 }
};


#pragma mark -

BOOL OtherError=FALSE;

//********************************************************************
BOOL KeyMsg(DWORD *cmd, DWORD *arg1, DWORD *arg2, DWORD *arg3) {
	
	DWORD	p, s, d, i;
	
	// first trash all messages we want trashed
	p = KINTmsgrem;
	while( p != KINTmsgadd ) {
		*cmd = KINTmsgs[p*4];
		*arg1=KINTmsgs[p*4+1];
		*arg2=KINTmsgs[p*4+2];
		*arg3=KINTmsgs[p*4+3];
		if( *cmd==KHDR_POINTER_MOVE || *cmd==KHDR_TIMER || *cmd==KHDR_IDLE ) {
			d = p;
			s = p+1;
			if( s >= MAXMSGS ) s = 0;
			while( s!=KINTmsgadd ) {	// trash this message, close the hole
				for(i=0; i<4; i++) KINTmsgs[d*4+i] = KINTmsgs[s*4+i];
				if( ++d >= MAXMSGS ) d = 0;
				if( ++s >= MAXMSGS ) s = 0;
			}
			if( KINTmsgadd==0 ) KINTmsgadd = MAXMSGS-1;
			else --KINTmsgadd;
		} else ++p;
		if( p==MAXMSGS ) p = 0;
	}
	// now find first KEY msg
	p = KINTmsgrem;
	while( p != KINTmsgadd ) {
		*cmd = KINTmsgs[p*4];
		*arg1=KINTmsgs[p*4+1];
		*arg2=KINTmsgs[p*4+2];
		*arg3=KINTmsgs[p*4+3];
		if( *cmd==KHDR_KEY_DOWN || *cmd==KHDR_KEY_UP ) {
			d = p;
			s = p+1;
			if( s >= MAXMSGS ) s = 0;
			while( s!=KINTmsgadd ) {	// trash this message, close the hole
				for(i=0; i<4; i++) KINTmsgs[d*4+i] = KINTmsgs[s*4+i];
				if( ++d >= MAXMSGS ) d = 0;
				if( ++s >= MAXMSGS ) s = 0;
			}
			if( KINTmsgadd==0 ) KINTmsgadd = MAXMSGS-1;
			else --KINTmsgadd;
			return TRUE;
		} else ++p;
		if( p==MAXMSGS ) p = 0;
	}
	return FALSE;
}

//********************************************************************
void AddMsg(WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3)
{
	if( !inKE && (inSleep || !windowReady /*gMacWindow*/ || KINTmsgTimer==NULL) ) {
		KINTnonTimer(cmd, arg1, arg2, arg3);
	} else {
		KINTmsgs[KINTmsgadd*4] = (DWORD)cmd;
		KINTmsgs[KINTmsgadd*4+1] = arg1;
		KINTmsgs[KINTmsgadd*4+2] = arg2;
		KINTmsgs[KINTmsgadd*4+3] = arg3;
		++KINTmsgadd;
		if( KINTmsgadd >= MAXMSGS ) KINTmsgadd = 0;
		if( KINTmsgadd==KINTmsgrem ) {
			--KINTmsgadd;
			if( KINTmsgadd<0 ) KINTmsgadd = MAXMSGS-1;
		}
	}
}

//********************************************************************
void SendNextMsg(void) {
	
	KLONG	mi;
	DWORD	msg, arg1, arg2, arg3;
	
	if( !killpgm && !inKE && !inSleep && runKEmsg==0 ) {
		if( KINTmsgrem!=KINTmsgadd ) {
			mi = 4*(KINTmsgrem++);
			if( KINTmsgrem >= MAXMSGS ) KINTmsgrem = 0;
			msg = KINTmsgs[mi++];
			arg1 = KINTmsgs[mi++];
			arg2 = KINTmsgs[mi++];
			arg3 = KINTmsgs[mi++];
/*
			if( KINTptrmovePENDING ) {
				DWORD	p;
				
				p = KINTmsgrem;
				while( p != KINTmsgadd ) {
					if( KINTmsgs[p*4]==KHDR_POINTER_MOVE ) {	// if found, modify it
						KINTmsgs[p*4+1] = KINTptrmove[1];
						KINTmsgs[p*4+2] = KINTptrmove[2];
						KINTmsgs[p*4+3] = KINTptrmove[3];
						KINTptrmovePENDING = false;
						break;
					} else ++p;	// keep searching
					if( p==MAXMSGS ) p = 0;
				}
				if( KINTptrmovePENDING ) {
					KINTmsgs[KINTmsgadd*4] = KINTptrmove[0];
					KINTmsgs[KINTmsgadd*4+1] = KINTptrmove[1];
					KINTmsgs[KINTmsgadd*4+2] = KINTptrmove[2];
					KINTmsgs[KINTmsgadd*4+3] = KINTptrmove[3];
					++KINTmsgadd;
					if( KINTmsgadd >= MAXMSGS ) KINTmsgadd = 0;
					if( KINTmsgadd==KINTmsgrem ) {
						--KINTmsgadd;
						if( KINTmsgadd<0 ) KINTmsgadd = MAXMSGS-1;
					} else KINTptrmovePENDING = false;
				}
			}
*/
	//NSLog(@"SendNextMsg->KINTnonTimer %d msgrem=%d msgadd=%d",KINTmsgs[mi], KINTmsgrem, KINTmsgadd);
			KINTnonTimer(msg, arg1, arg2, arg3);
		}
	}
}

//*********************************************************************
void PurgeMsgs(void) {
	
	KINTmsgadd = KINTmsgrem = 0;
}

//**************************************************************************
void SetAppDataPath(void) {
	
//	CFURLRef parentURL = NULL;
//	CFURLRef eksURL = NULL;
//	CFURLRef resourceURL = NULL;
	CFURLRef pgmURL = NULL;

	// setup AppPath[]
	if( NULL!=(pgmURL = CFBundleCopyBundleURL (CFBundleGetMainBundle())) ) {
		CFURLGetFileSystemRepresentation(pgmURL, true, (UInt8 *)AppPath, sizeof(AppPath));
		CFRelease(pgmURL);
//		strcat(AppDataPath, "/");
	}
	// setup PgmPath[]
	if( NULL!=(pgmURL = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle())) ) {
		CFURLGetFileSystemRepresentation(pgmURL, true, (UInt8 *)PgmPath, sizeof(PgmPath));
		CFRelease(pgmURL);
		strcat(PgmPath, "/");
	}

	// setup AppDataPath[]
	gAppName = (CFStringRef)CFBundleGetValueForInfoDictionaryKey(CFBundleGetMainBundle(), CFSTR("EKSGameName"));

	// Look in ~/Library/Application Support/Everett Kaser Software/<EKSGameName>"
//	NSString *path = [[NSBundle mainBundle] pathForResource:@"Info" ofType:@"plist"];
//	NSDictionary	*dict = [NSDictionary dictionaryWithContentsOfFile:path];
//	NSString *gameName = [dict objectForKey:@"EKSGameName"];

	NSFileManager* fm = [NSFileManager defaultManager];
	NSURL *dirPath = nil;
	NSArray *appSupportDir = [fm URLsForDirectory:NSApplicationSupportDirectory inDomains:NSUserDomainMask];
	if( [appSupportDir count] > 0 ) {
		dirPath = [[appSupportDir objectAtIndex:0] URLByAppendingPathComponent:@"Everett Kaser Software"];
		NSString *pPath=[[[dirPath path] stringByAppendingString:@"/"] stringByAppendingString:(__bridge NSString * _Nonnull)(gAppName)];
		[pPath getCString:AppDataPath maxLength:sizeof(AppDataPath) encoding:NSUTF8StringEncoding];
		strcat(AppDataPath, "/");
	}
//	NSLog(@"AppDataPath=%s",AppDataPath);
//	NSLog(@"AppPath=%s",AppPath);
//	NSLog(@"PgmPath=%s",PgmPath);

	
	
//	if( resourceURL ) CFRelease(resourceURL);
//	if( eksURL ) CFRelease(eksURL);
//	if( parentURL ) CFRelease(parentURL);
}

//**************************************
/*void LoadSound(long index) {
	
	if( !sndLoaded[index] ) {
		NSString *iName = [NSString stringWithUTF8String:sndNames[index]];
		sndRef[index] = (__bridge CFURLRef)[[NSBundle mainBundle]URLForResource:iName withExtension:@"WAV"];
		AudioServicesCreateSystemSoundID(sndRef[index], &sndID[index]);
		sndLoaded[index] = true;
	}
}
 */

//*************************************************************
//*************************************************************
// HARDWARE-INDEPENDENT GRAPHICS FUNCTIONS
//*************************************************************
//*************************************************************
//static inline void SwapDWORD( DWORD *in )	{	*in = __lwbrx( in, 0 );	}
//static inline void SwapWORD( WORD *in ) 		{	*in = __lhbrx( in, 0 );	}
//static inline void SwapLong( long *in ) 		{	*in = __lwbrx( in, 0 );	}

static inline void BlastPixel (BYTE r, BYTE g, BYTE b, BYTE *dst)
{
	*(dst+1) = r;
	*(dst+2) = g;
	*(dst+3) = b;
}

static inline void GetPixelFromSource (BYTE *r, BYTE *g, BYTE *b, BYTE *src)
{
	*r = *(src+1);
	*g = *(src+2);
	*b = *(src+3);
}


static inline void BlastPixelAlpha (BYTE r, BYTE g, BYTE b, BYTE *dst)
{
	*(dst+1) = (BYTE)(((WORD)(*(dst+1)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
	*(dst+2) = (BYTE)(((WORD)(*(dst+2)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
	*(dst+3) = (BYTE)(((WORD)(*(dst+3)) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
}

static inline void BlastPixelFromSource (BYTE *dst, BYTE *src, bool isAlpha, bool inverted)
{
	BYTE r, g, b;
	GetPixelFromSource (&r, &g, &b, src);
	if (inverted)
	{
		r ^= 0xff;
		g ^= 0xff;
		b ^= 0xff;
	}
	if (isAlpha)
		BlastPixelAlpha (r, g, b, dst);
	else
		BlastPixelAlpha (r, g, b, dst);
}
/*
void SwapBMPFileHeader (BITMAPFILEHEADER *inHeader)
{
	SwapWORD (&inHeader->bfType);
	SwapDWORD (&inHeader->bfSize);
	SwapWORD (&inHeader->bfReserved1);
	SwapWORD (&inHeader->bfReserved2);
	SwapDWORD (&inHeader->bfOffBits);
}

void SwapBMPInfoHeader (BITMAPINFOHEADER *inHeader)
{
	SwapDWORD (&inHeader->biSize);
	SwapLong (&inHeader->biWidth);
	SwapLong (&inHeader->biHeight);
	SwapWORD (&inHeader->biPlanes);
	SwapWORD (&inHeader->biBitCount);
	SwapDWORD (&inHeader->biCompression);
	SwapDWORD (&inHeader->biSizeImage);
	SwapLong (&inHeader->biXPelsPerMeter);
	SwapLong (&inHeader->biYPelsPerMeter);
	SwapDWORD (&inHeader->biClrUsed);
	SwapDWORD (&inHeader->biClrImportant);
}
*/

KLONG	dirtyX1=30000, dirtyY1, dirtyX2=-1, dirtyY2;

//**********************************************************************
void InvalidateScreen(int x1, int y1, int x2, int y2) {
	
	if (windowReady ) {
        if( x1<dirtyX1 ) dirtyX1 = x1;
        if( y1<dirtyY1 ) dirtyY1 = y1;
        if( x2>dirtyX2 ) dirtyX2 = x2;
        if( y2>dirtyY2 ) dirtyY2 = y2;
	}
}

//***************************************************
KLONG SaveBMB(PBMB bm, char *fname, BYTE dir) {
	
	// NOTE: If porting to non-MS Windows platform, use
	//	typedefs for BITMAPFILEHEADER and BITMAPINFO at top of this file
	BITMAPFILEHEADER	bmfh;
	BITMAPINFOHEADER	bmih;
	BYTE				*pB;
	DWORD				hFile;
	KLONG					h, bmw;
	
	if( -1==(hFile=Kopen(fname, O_WRBINNEW, dir)) ) return FALSE;
	
	bmw = 4*((3*bm->physW+3)/4);
	bmfh.bfType = 0x4D42;	// 'BM'
	bmfh.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO)+bmw*bm->physH-4;
	bmfh.bfReserved1 = bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
	
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = bm->physW;
	bmih.biHeight = bm->physH;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;
	bmih.biCompression = 0;
	bmih.biSizeImage = bmw*bm->physH;
	bmih.biXPelsPerMeter = 2947;
	bmih.biYPelsPerMeter = 2947;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	
#if MAC_PORT // LBO 9/16/07 [BS]
//	SwapBMPFileHeader (&bmfh);
#endif
	if( sizeof(BITMAPFILEHEADER)!=Kwrite(hFile, (BYTE *)&bmfh, sizeof(BITMAPFILEHEADER)) ) {
		Kclose(hFile);
		return FALSE;
	}
#if MAC_PORT // LBO 9/16/07 [BS]
//	SwapBMPFileHeader (&bmfh); // unswap
//	SwapBMPInfoHeader (&bmih);
#endif
	if( sizeof(BITMAPINFOHEADER)!=Kwrite(hFile, (BYTE *)&bmih, sizeof(BITMAPINFOHEADER)) ) {
		Kclose(hFile);
		return FALSE;
	}
#if MAC_PORT // LBO 9/16/07 [BS]
//	SwapBMPInfoHeader (&bmih); // unswap
	UInt32 destLineLen = 3*(bm->linelen / kPixSize);// * 3 - 3;
	destLineLen = 4*((bm->physW*3+3)/4);
	BYTE *tempLine = malloc(destLineLen);
#endif
	pB = bm->pBits+(bm->physH-1)*bm->linelen;
	for(h=0; h<bm->physH; h++) {
#if MAC_PORT // LBO 9/16/07 [RGBA]
		BYTE *pS = pB;
		BYTE *pD = tempLine;
		KLONG x;
		for (x = 0; x < bm->linelen / kPixSize; x ++)
		{
			*pD++ = *(pS+3);
			*pD++ = *(pS+2);
			*pD++ = *(pS+1);
			pS += kPixSize;
		}
		if( destLineLen != Kwrite(hFile, tempLine, destLineLen) )
		{
			Kclose(hFile);
			free(tempLine);
			return FALSE;
		}
#else
		if( bm->linelen!=Kwrite(hFile, pB, bm->linelen) ) {
			Kclose(hFile);
			return FALSE;
		}
#endif
		pB -= bm->linelen;
	}
	Kclose(hFile);
#if MAC_PORT // LBO 9/16/07 [RGBA]
	free(tempLine);
#endif
	return TRUE;
}

//***************************************
KLONG CreateBMB(short w, short h, WORD sysBM)
{
	DWORD i;
	
/*    if( w<1 ) {
        w = 1024;//PhysicalScreenW;
    }
    if( h<1 ) {
        h = 768;//PhysicalScreenH;
    }
*/
    for(i=1; i<BMcnt; i++) if( kBM[i].cgImage==NULL && kBM[i].bitmapContext==NULL ) break;
	if( i<MAXBM ) {
		if( i==BMcnt ) {
			kBM[BMcnt].cgImage = NULL;
			kBM[BMcnt++].bitmapContext = NULL;
		}
		kBM[i].linelen = 4*w;	// force line of pixels to take even number of DWORDS
		kBM[i].pBits = NULL;
		kBM[i].pBits = (BYTE *)malloc(kBM[i].linelen*h);
		memset(kBM[i].pBits, 0, kBM[i].linelen*h);
		if( !sysBM ) {
			CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
			CGContextRef	ctx = CGBitmapContextCreate(NULL, abs(w), h, 8, 0, colorSpace, (CGBitmapInfo)kCGImageAlphaPremultipliedLast);
			CGColorSpaceRelease(colorSpace);
			kBM[i].cgImage = CGBitmapContextCreateImage(ctx);
			CGContextRelease(ctx);
		} else {
			CGColorSpaceRef colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
			kBM[i].bitmapContext = CGBitmapContextCreate (kBM[i].pBits, w, h, 8, kBM[i].linelen, colorSpace, kCGImageAlphaNoneSkipFirst);
			CGColorSpaceRelease(colorSpace);
			kBM[i].cgImage = NULL;
		}
		kBM[i].physW = kBM[i].logW = abs(w);
		kBM[i].physH = kBM[i].logH = h;
		
		kBM[i].pAND = NULL;
		kBM[i].clipX1 = kBM[i].clipY1 = 0;
		kBM[i].clipX2 = (KLONG)w;
		kBM[i].clipY2 = (KLONG)h;
		kBM[i].orient = BMO_NORMAL;
		return i;
	}
	return -1;
}

//***************************************
void DeleteBMB(DWORD index)
{
	if( index<BMcnt && (kBM[index].cgImage!=NULL || kBM[index].bitmapContext!=NULL) ) {
		if( kBM[index].bitmapContext!= NULL ) {
			CGContextRelease(kBM[index].bitmapContext);
			kBM[index].bitmapContext = NULL;
		}
		if( kBM[index].cgImage!= NULL ) {
			CGImageRelease(kBM[index].cgImage);
			kBM[index].cgImage = NULL;
		}
		while( BMcnt>1 && kBM[BMcnt-1].bitmapContext==NULL && kBM[BMcnt-1].cgImage==NULL ) --BMcnt;

		if( kBM[index].pBits!=NULL ) {
			free(kBM[index].pBits);
			kBM[index].pBits = NULL;
		}
		if( kBM[index].pAND!=NULL ) {
			free(kBM[index].pBits);
			kBM[index].pAND = NULL;
		}
	}
}
//****************
void CreateBM0() {
	
	DeleteBMB(0);
	ScrW = PhysicalScreenW;
	ScrH = PhysicalScreenH;
	kBM[0].linelen = 4*ScrW;	// force line of pixels to take even number of DWORDS
	kBM[0].pBits = (BYTE *)malloc(kBM[0].linelen*ScrH);
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	kBM[0].cgImage = NULL;
	kBM[0].bitmapContext = CGBitmapContextCreate(kBM[0].pBits, ScrW, ScrH, 8, kBM[0].linelen, colorSpace, (CGBitmapInfo)kCGImageAlphaNoneSkipFirst);
	kBM[0].logW = kBM[0].physW = ScrW;
	kBM[0].logH = kBM[0].physH = ScrH;
	CGColorSpaceRelease(colorSpace);
	ScrClipX1 = 0;
	ScrClipY1 = 0;
	ScrClipX2 = ScrW;
	ScrClipY2 = ScrH;

	kBM[0].pAND = NULL;
	kBM[0].clipX1 = kBM[0].clipY1 = 0;
	kBM[0].clipX2 = (KLONG)ScrW;
	kBM[0].clipY2 = (KLONG)ScrH;
	kBM[0].orient = BMO_NORMAL;
}
//*************************************************************
// Load a MS Windows/OS2 .BMP format file into a device-independent
// memory array.  BMP file can be 8-bit (256 color / palettized)
// or 24-bit (RGB pixel values), but can NOT be compressed, and is
// assumed to be in the "bottom-up" format (bmiHeader.biHeight > 0).
// LoadBMB supports TWO transparency colors.  If first pixel
// coordinates (transX1, transY1) are (-1,-1) then the bitmap has
// no transparency color.  If it's a valid pixel coordinate, then
// transX2 and transY2 can either be (-1,-1) for only ONE transparency
// color OR the coordinates of a second color pixel.
DWORD LoadBMB(char *fname, KLONG transX1, KLONG transY1, KLONG transX2, KLONG transY2, BYTE dir) {
	
	// NOTE: If porting to non-MS Windows platform, use
	//	typedefs for BITMAPFILEHEADER and BITMAPINFO at top of this file
	BITMAPFILEHEADER	bmfh;
	BITMAPINFO			*pbmi;
	DWORD				dwInfoSize, dwBitsSize, BytesRead;
	PBMB				pX;
	BYTE				*pB, *pL, *pO, *pA = NULL;
	DWORD				hFile;
	KLONG				ki, x, y, w, h, bitcnt, r, g, b;
	BYTE				ci, tR1=0, tG1=0, tB1=0, tR2 = '\0', tG2 = '\0', tB2 = '\0', tran;
	DWORD				linlen;
	
	// first try to load the graphics file, and if that fails,
	// then if the file is supposed to be in the GAME folder,
	// add .BMP and try to load again.  This allows games to
	// place .BMP files in the GAME folder and access them,
	// but not have them show up in the users directory listing
	// as a BMP file.
	// Open the file: read access, prohibit write access
	hFile = Kopen(fname, O_RDBIN, dir);
	if( -1 == hFile ) {
		if( dir==DIR_GAME ) {
			strcpy((char *)TmpFname, fname);
			strcat((char *)TmpFname, ".BMP");
			hFile = Kopen((char *)TmpFname, O_RDBIN, dir);
			if( -1 == hFile ) return 0;
		}
	}
	
	// Read in the BITMAPFILEHEADER
	BytesRead = Kread(hFile, (BYTE *)&bmfh, sizeof(BITMAPFILEHEADER));
	if( BytesRead!=sizeof(BITMAPFILEHEADER) || (bmfh.bfType != *(WORD *)"BM") ) {
		Kclose(hFile);
		return 0;
	}
	
#if MAC_PORT // LBO 9/16/07 [BS]
//	SwapBMPFileHeader (&bmfh);
#endif
	
	// Allocate memory for the information structure & read it in
	dwInfoSize = bmfh.bfOffBits - sizeof (BITMAPFILEHEADER);
	
	if( NULL==(pbmi=(struct tagBITMAPINFO *)malloc(dwInfoSize)) ) {
		Kclose(hFile);
		return 0;
	}
	BytesRead = Kread(hFile, (BYTE *)pbmi, dwInfoSize);
	if( BytesRead != dwInfoSize )	{
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	
#if MAC_PORT // LBO 9/16/07 [BS]
//	SwapBMPInfoHeader (&pbmi->bmiHeader);
#endif
	
	if( pbmi->bmiHeader.biBitCount<8 || pbmi->bmiHeader.biCompression!=0 ) {	// don't handle < 256 color bitmaps OR compressed bitmaps
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	
	// Create the DIB
	if( (ki=CreateBMB((KLONG)(pbmi->bmiHeader.biWidth), (KLONG)(pbmi->bmiHeader.biHeight), FALSE))<0 ) {
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	pX = &kBM[ki];
	if( transX1>=0 && transX1<(KLONG)pX->physW && transY1>=0 && transY1<(KLONG)pX->physH ) {
		if( NULL==(pX->pAND=(BYTE *)malloc(pX->physW*pX->physH)) ) {
			DeleteBMB(ki);
			Kclose(hFile);
			free(pbmi);
			return 0;
		}
		tran = 1;	// we're doing a transparent (masked) bitmap
	} else tran = 0;
	
	// allocate temp storage for .BMP file RGB/palette-index bytes
	dwBitsSize = bmfh.bfSize - bmfh.bfOffBits;
	if( NULL==(pB=(BYTE *)malloc(dwBitsSize)) ) {
		DeleteBMB(ki);
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	
	BytesRead = Kread(hFile, pB, dwBitsSize);
	Kclose(hFile);
	if( BytesRead != dwBitsSize ) {
		free(pB);
		DeleteBMB(ki);
		free(pbmi);
		return 0;
	}
	
	w = (KLONG)pbmi->bmiHeader.biWidth;
	h = (KLONG)pbmi->bmiHeader.biHeight;
	bitcnt = pbmi->bmiHeader.biBitCount - 8;	// it's either 8 or 24, so 0 or non-0
	if( pbmi->bmiHeader.biSizeImage==0 ) {
		pbmi->bmiHeader.biSizeImage = 4*((w*(bitcnt?3:1)+3)/4)*h;
	}
	linlen = pbmi->bmiHeader.biSizeImage / h;
	if( tran ) {	// if transparent (masked) bitmap, get transparency color(s)
		if( bitcnt ) {	// 24-bit bitmap
			pL = pB + linlen*(h-transY1-1) + 3*transX1;
			tB1 = *pL++;
			tG1 = *pL++;
			tR1 = *pL++;
			if( transX2>=0 && transX2<(KLONG)pX->physW && transY2>=0 && transY2<(KLONG)pX->physH ) {
				pL = pB + linlen*(h-transY2-1) + 3*transX2;
				tB2 = *pL++;
				tG2 = *pL++;
				tR2 = *pL++;
			} else {
				tB2 = tB1;
				tG2 = tG1;
				tR2 = tR1;
			}
		} else {		// 8-bit bitmap
			pL = pB + linlen*(h-transY1-1) + transX1;
			ci = *pL;
			tR1 = pbmi->bmiColors[ci].rgbRed;
			tG1 = pbmi->bmiColors[ci].rgbGreen;
			tB1 = pbmi->bmiColors[ci].rgbBlue;
			if( transX2>=0 && transX2<(KLONG)pX->physW && transY2>=0 && transY2<(KLONG)pX->physH ) {
				pL = pB + linlen*(h-transY2-1) + transX2;
				ci = *pL;
				tR2 = pbmi->bmiColors[ci].rgbRed;
				tG2 = pbmi->bmiColors[ci].rgbGreen;
				tB2 = pbmi->bmiColors[ci].rgbBlue;
			} else {
				tB2 = tB1;
				tG2 = tG1;
				tR2 = tR1;
			}
		}
	}
	for(y=0; y<h; y++) {
		pL = pB + linlen*(h-y-1);	// common .BMPs are inverted vertically
		pO = pX->pBits + pX->linelen * y;
		if( tran ) pA = pX->pAND + pX->physW * y;
		for(x=0; x<w; x++) {
			if( bitcnt ) {	// 24-bit
				b = *pL++;
				g = *pL++;
				r = *pL++;
			} else {		// 8-bit
				ci = *pL++;
				b = pbmi->bmiColors[ci].rgbBlue;
				g = pbmi->bmiColors[ci].rgbGreen;
				r = pbmi->bmiColors[ci].rgbRed;
			}
			// convert any 0x00818181 (middle-gray) pixels to 0x00808080
			// so that 0x00818181 is guaranteed to never be used in customer-created bitmaps
			if( (tR1!=0x81 || tG1!=0x81 || tB1!=0x81) && r==0x81 && g==0x81 && b==0x81 ) r = g = b = 0x80;
			
#if MAC_PORT // LBO 09/16/07 [RGBA]
			*(pO+1) = r;
			*(pO+2) = g;
			*(pO+3) = b;
//			BlastPixel (r, g, b, pO);
			pO += kPixSize;
#else
			*pO++ = b;
			*pO++ = g;
			*pO++ = r;
#endif
			if( tran ) *pA++ = ((r==tR1 && g==tG1 && b==tB1) || (r==tR2 && g==tG2 && b==tB2)) ? 255 : 0;
		}
	}
	free(pB);
	free(pbmi);
	return ki;
}

//*********************************************************
void CreateMaskBMB(PBMB bm, DWORD trans1, DWORD trans2) {
	
	BYTE	*pD, *pDST, *pA, tR1, tG1, tB1, tR2, tG2, tB2, r, g, b;
	KLONG		w, h;
	
	if( bm->pAND==NULL ) {
		if( NULL==(bm->pAND=(BYTE *)malloc(bm->physW*bm->physH)) ) return;
	}
	tR1 = (BYTE)(trans1 & 0x00FF);
	tG1 = (BYTE)((trans1>>8) & 0x00FF);
	tB1 = (BYTE)((trans1>>16) & 0x00FF);
	tR2 = (BYTE)(trans2 & 0x00FF);
	tG2 = (BYTE)((trans2>>8) & 0x00FF);
	tB2 = (BYTE)((trans2>>16) & 0x00FF);
	pD = bm->pBits;
	pA = bm->pAND;
	for(h=bm->physH; h; pD += bm->linelen, --h) {
		pDST = pD;
		for(w=bm->physW; w; --w) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
			GetPixelFromSource (&r, &g, &b, pDST); pDST += kPixSize;
#else
			b = *pDST++;
			g = *pDST++;
			r = *pDST++;
#endif
			*pA++ = ((r==tR1 && g==tG1 && b==tB1) || (r==tR2 && g==tG2 && b==tB2)) ? 255 : 0;
		}
	}
}


//*********************************************************
void SOerr(PBMB kB) {
	
	char	xxx[32];
	
	sprintf(xxx, "ERR SO: %d %8.8X", kB->orient, (unsigned int)(*((DWORD *)(Kregs+3*4))));
	SystemMessage(xxx);
}

//***************************************************************
void SetupBMD(PBMB *bmD) {
	
	PBMB	bD;
	
	bD = *bmD;
//	if( bD==0 ) bD = kBuf;
	TdBMW = bD->logW;
	TdBMH = bD->logH;
	TdclipX1 = bD->clipX1;
	TdclipY1 = bD->clipY1;
	TdclipX2 = bD->clipX2;
	TdclipY2 = bD->clipY2;
	switch( bD->orient ) {
		case BMO_NORMAL:
			TdstepX	= kPixSize;
			TdstepY = bD->linelen;
			Tdpbits = bD->pBits;
			break;
		case BMO_90:
			TdstepX	= -bD->linelen;
			TdstepY = kPixSize;
			Tdpbits = bD->pBits + (TdBMW-1)*bD->linelen;
			break;
		case BMO_180:
			TdstepX	= -kPixSize;
			TdstepY = -bD->linelen;
			Tdpbits = bD->pBits + (TdBMW-1)*kPixSize + (TdBMH-1)*bD->linelen;
			break;
		case BMO_270:
			TdstepX	= bD->linelen;
			TdstepY = -kPixSize;
			Tdpbits = bD->pBits + (TdBMH-1)*kPixSize;
			break;
		case BMO_FNORMAL:
			TdstepX	= -kPixSize;
			TdstepY = bD->linelen;
			Tdpbits = bD->pBits + (TdBMW-1)*kPixSize;
			break;
		case BMO_F90:
			TdstepX	= -bD->linelen;
			TdstepY = -kPixSize;
			Tdpbits = bD->pBits + (TdBMH-1)*kPixSize + (TdBMW-1)*bD->linelen;
			break;
		case BMO_F180:
			TdstepX	= kPixSize;
			TdstepY = -bD->linelen;
			Tdpbits = bD->pBits + (TdBMH-1)*bD->linelen;
			break;
		case BMO_F270:
			TdstepX	= bD->linelen;
			TdstepY = kPixSize;
			Tdpbits = bD->pBits;
			break;
		default:
		{
			char	tbuf[128];
			sprintf(tbuf,"D WH=(%d,%d) linelen=%d pBits=%p", bD->physW, bD->physH, bD->linelen, bD->pBits);
			SystemMessage(tbuf);
		}
			SOerr(bD);
			return;
	}
	*bmD = bD;
}


void SetupBMS(PBMB *bmS) {
	
	PBMB	bS;
	
	bS = *bmS;
//	if( bS==0 ) bS = kBuf;
	TsBMW = bS->logW;
	TsBMH = bS->logH;
	TsclipX1 = bS->clipX1;
	TsclipY1 = bS->clipY1;
	TsclipX2 = bS->clipX2;
	TsclipY2 = bS->clipY2;
	switch( bS->orient ) {
		case BMO_NORMAL:
			TsstepX	= kPixSize;
			TsstepY = bS->linelen;
			Tspbits = bS->pBits;
			TastepX	= 1;
			TastepY = bS->physW;
			Tapbits = bS->pAND;
			break;
		case BMO_180:
			TsstepX	= -kPixSize;
			TsstepY = -bS->linelen;
			Tspbits = bS->pBits + (TsBMW-1)*kPixSize + (TsBMH-1)*bS->linelen;
			TastepX	= -1;
			TastepY = -bS->physW;
			Tapbits = bS->pAND + (TsBMW-1) + (TsBMH-1)*bS->physW;
			break;
		case BMO_90:
			TsstepX	= -bS->linelen;
			TsstepY = kPixSize;
			Tspbits = bS->pBits + (TsBMW-1)*bS->linelen;
			TastepX	= -bS->physW;
			TastepY = 1;
			Tapbits = bS->pAND + (TsBMW-1)*bS->physW;
			break;
		case BMO_270:
			TsstepX	= bS->linelen;
			TsstepY = -kPixSize;
			Tspbits = bS->pBits + (TsBMH-1)*kPixSize;
			TastepX	= bS->physW;
			TastepY = -1;
			Tapbits = bS->pAND + (TsBMH-1);
			break;
		case BMO_FNORMAL:
			TsstepX	= -kPixSize;
			TsstepY = bS->linelen;
			Tspbits = bS->pBits + (TsBMW-1)*kPixSize;
			TastepX	= -1;
			TastepY = bS->physW;
			Tapbits = bS->pAND + TsBMW-1;
			break;
		case BMO_F90:
			TsstepX	= -bS->linelen;
			TsstepY = -kPixSize;
			Tspbits = bS->pBits + (TsBMH-1)*kPixSize + (TsBMW-1)*bS->linelen;
			TastepX	= bS->physW;
			TastepY = -1;
			Tapbits = bS->pAND + (TsBMH-1) + (TsBMW-1)*bS->physW;
			break;
		case BMO_F180:
			TsstepX	= kPixSize;
			TsstepY = -bS->linelen;
			Tspbits = bS->pBits + (TsBMH-1)*bS->linelen;
			TastepX	= 1;
			TastepY = -bS->physW;
			Tapbits = bS->pAND + (TsBMH-1)*bS->physW;
			break;
		case BMO_F270:
			TsstepX	= bS->linelen;
			TsstepY = kPixSize;
			Tspbits = bS->pBits;
			TastepX	= bS->physW;
			TastepY = 1;
			Tapbits = bS->pAND;
			break;
		default:
		{
			char	tbuf[128];
			sprintf(tbuf,"S WH=(%d,%d) linelen=%d pBits=%p", bS->physW, bS->physH, bS->linelen, bS->pBits);
			SystemMessage(tbuf);
		}
			SOerr(bS);
			return;
	}
	*bmS = bS;
}

//**************************************************************************
// Convert bitmap to gray-scale values
// To convert 24-bit RGB values to gray scale, set each of R,G,B to the
// same value, that value being 29.9% of the original Red, 58.7% of the
// original Green, and 11.4% of the original Blue.
// light is from -255 to +255, and is a "percentage" of previous value added to the previous value for each pixel.
void GrayBMB(PBMB bm, KLONG x, KLONG y, KLONG w, KLONG h, KLONG light, KLONG contrast) {
	
	BYTE	*pD, *pDST, *pA, *pAND;
	KLONG		tw, th, v;
	
	SetupBMS(&bm);
	
	if( x<TsclipX1 ) {
		w -= (TsclipX1-x);
		x = TsclipX1;
	}
	if( y<TsclipY1 ) {
		h -= (TsclipY1-y);
		y = TsclipY1;
	}
	if( (x+w)>=TsclipX2 ) w = TsclipX2-x;
	if( (y+h)>=TsclipY2 ) h = TsclipY2-y;
	if( w<=0 || h<=0 ) return;
	pD = Tspbits + TsstepY*y + TsstepX*x;
	if( bm->pAND ) pA = Tapbits + TastepY*y + TastepX*x;
	else pA = NULL;
	for(th=h; th; pD += TsstepY, --th) {
		pDST = pD;
		pAND = pA;
		for(tw=w; tw; pDST += TsstepX, --tw) {
			if( bm->pAND==NULL || (bm->pAND!=NULL && *pAND==0) ) {
				v = ((KLONG)*pDST*114 + (KLONG)*(pDST+1)*587 + (KLONG)*(pDST+2)*299)/1000;
				if( contrast ) {
					v += ((v-128)*contrast)/255;
				}
				if( light ) {
					v += (light*v)/255;
					if( v>255 ) v = 255;
					if( v<0 ) v = 0;
				}
#if MAC_PORT // LBO 09/16/07 [RGBA]
				*(pDST+1) = v;
				*(pDST+2) = v;
				*(pDST+3) = v;
//				BlastPixel (v, v, v, pDST);
#else
				*pDST = *(pDST+1) = *(pDST+2) = (BYTE)v;
#endif
			}
			if( pAND != NULL ) pAND += TastepX;
		}
		if( pA != NULL ) pA += TastepY;
	}
    if( bm==&kBM[0] ) {
        InvalidateScreen(x, y, x+w, y+h);
    }
}

//*****************************************************************************
// hue and sat are values from 0-255, and if >= 0 simply replace the previous values for each pixel.
// light is from -255 to +255, and is a "percentage" of previous value added to the previous value for each pixel.
void ColorizeBMB(PBMB bm, KLONG x, KLONG y, KLONG w, KLONG h, KLONG hue, KLONG sat, KLONG light, KLONG contrast) {
	
	BYTE	*pD, *pDST, *pA, *pAND;
	KLONG		tw, th, r, g, b;
	KLONG		mx, mn, v, s, hh, rc, gc, bc, f, p, q, t;

	SetupBMS(&bm);
	
	if( x<TsclipX1 ) {
		w -= (TsclipX1-x);
		x = TsclipX1;
	}
	if( y<TsclipY1 ) {
		h -= (TsclipY1-y);
		y = TsclipY1;
	}
	if( (x+w)>=TsclipX2 ) w = TsclipX2-x;
	if( (y+h)>=TsclipY2 ) h = TsclipY2-y;
	if( w<=0 || h<=0 ) return;
	pD = Tspbits + TsstepY*y + TsstepX*x;
	if( bm->pAND ) pA = Tapbits + TastepY*y + TastepX*x;
	else pA = NULL;
	for(th=h; th; pD += TsstepY, --th) {
		pDST = pD;
		pAND = pA;
		for(tw=w; tw; pDST += TsstepX, --tw) {

			if( bm->pAND==NULL || (bm->pAND!=NULL && *pAND==0) ) {
				r = (KLONG)*(pDST+1);
				g = (KLONG)*(pDST+2);
				b = (KLONG)*(pDST+3);
				
				mx= (r>=g && r>=b)?r:((g>=r && g>=b)?g:b);
				mn= (r<=g && r<=b)?r:((g<=r && g<=b)?g:b);
				v = mx;	// value
				if( mx ) s = 255*(mx-mn)/mx;	// saturation
				else s = 0;
				if( s==0 ) hh = 0;
				else {
					rc = 255*(mx-r)/(mx-mn);
					gc = 255*(mx-g)/(mx-mn);
					bc = 255*(mx-b)/(mx-mn);
					
					if( r==mx ) hh = (bc-gc)/6;
					else if( g==mx ) hh = 85+(rc-bc)/6;
					else hh = 171+(gc-rc)/6;
				}
				if( hue>=0 ) hh = hue;
				hh &= 255;
				
				if( sat>=0 ) s = sat;
				if( s<0 ) s = 0;
				s &= 255;
				
				if( contrast ) {
					v += ((v-128)*contrast)/255;
				}
				if( light ) {
					v += (light*v)/255;
				}
				if( v>255 ) v = 255;
				else if( v<0 ) v = 0;
				
				if( s==0 ) {
					*(pDST+1) = r;
					*(pDST+2) = g;
					*(pDST+3) = b;
				} else {
					if( hh>255 ) hh = 0;
					f = 6*(hh % 42);
					hh= hh / 42;
					if( hh>5 || hh<0 ) hh = 0;
					p = (v*(255-s))/255;
					q = (f==0)?v:((v*(255*246-s*f))/(255*246));
					t = (v*(255*246-(s*(246-f))))/(255*266);
//if( th==h && tw==w ) NSLog(@"%d %d %d %d=(%d %d %d %d %d)",hue,sat,light,contrast,v,t,p,q,b);
					switch( hh ) {
						case 0:
							*(pDST+1) = v;
							*(pDST+2) = t;
							*(pDST+3) = p;
							break;
						case 1:
							*(pDST+1) = q;
							*(pDST+2) = v;
							*(pDST+3) = p;
							break;
						case 2:
							*(pDST+1) = p;
							*(pDST+2) = v;
							*(pDST+3) = t;
							break;
						case 3:
							*(pDST+1) = p;
							*(pDST+2) = q;
							*(pDST+3) = v;
							break;
						case 4:
							*(pDST+1) = t;
							*(pDST+2) = p;
							*(pDST+3) = v;
							break;
						case 5:
							*(pDST+1) = v;
							*(pDST+2) = p;
							*(pDST+3) = q;
							break;
					}
				}
			}
			if( pAND != NULL ) pAND += TastepX;
		}
		if( pA != NULL ) pA += TastepY;
	}
    if( bm==&kBM[0] ) {
        InvalidateScreen(x, y, x+w, y+h);
    }
}

//**************************************************************************
// Block-transfer a region from one bitmap to another.
// If either bitmap handle is NULL, redirect it to the screen bitmap
// In the source (masked) bitmap, the pBits memory block has 3 bytes
// per pixel containing the RGB information to be forced into the
// destination.  If 'masked' is TRUE, then the pAND memory block has
// 1 byte per pixel that is 255 to leave the destination untouched at
// that pixel, 0 to force the RGB pBits pixel color.  (If 'masked' is
// FALSE, then the pAND is ignored and all source pixel values are forced
// into the destination.
void BltBMB(PBMB bmD, KLONG xd, KLONG yd, PBMB bmS, KLONG xs, KLONG ys, KLONG ws, KLONG hs, KLONG masked, KLONG invert) {
    
    BYTE	*pDST, *pSRC, *pAND, *pD, *pS, *pA;
    KLONG		x, y, ix;
    
    SetupBMD(&bmD);
    SetupBMS(&bmS);
    if( bmS->pAND==NULL ) masked = 0;	// if no mask, don't try to use it
    
    if( xs < TsclipX1 ) {
        if( (ws -= TsclipX1-xs)<=0 ) return;	// if src is clear off the left of source bitmap
        xd += TsclipX1-xs;
        xs = TsclipX1;
    }
    ix = xs+ws-TsclipX2;
    if( ix>0 ) {
        if( (ws -= ix)<=0 ) return;	// if src is clear off right of src bitmap
    }
    if( xd < TdclipX1 ) {
        if( (ws -= TdclipX1-xd)<=0 ) return;	// if dest is CLEAR off the left of dest bitmap
        xs += TdclipX1-xd;
        xd = TdclipX1;
    }
    ix = xd+ws-TdclipX2;
    if( ix>0 ) {
        if( (ws -= ix)<=0 ) return;	// if dest is clear off right of dst bitmap
    }
    if( ys < TsclipY1 ) {
        if( (hs -= TsclipY1-ys)<=0 ) return;	// if src is clear off top of source bitmap
        yd += TsclipY1-ys;
        ys = TsclipY1;
    }
    ix = ys+hs-TsclipY2;
    if( ix>0 ) {
        if( (hs -= ix)<=0 ) return;	// if src is clear off bottom of src bitmap
    }
    if( yd < TdclipY1 ) {
        if( (hs -= TdclipY1-yd)<=0 ) return;	// if dest is CLEAR off top of dst bitmap
        ys += TdclipY1-yd;
        yd = TdclipY1;
    }
    ix = yd+hs-TdclipY2;
    if( ix>0 ) {
        if( (hs -= ix)<=0 ) return;	// if dst is clear off bottom of dst bitmap
    }
    pS = Tspbits + TsstepY*ys + TsstepX*xs;
    pD = Tdpbits + TdstepY*yd + TdstepX*xd;
    if( AlphaBlendD ) {
        if( masked ) {
            if( invert ) {
                pA = Tapbits + TastepY*ys + TastepX*xs;
                for(y=0; y<hs; y++) {
                    pAND = pA;
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
                        if( *pAND == 0 ) {	// if AND mask has a hole, paint the SRC into the DST
#if MAC_PORT // LBO 09/16/07 [RGBA]
                            BlastPixelFromSource (pDST, pSRC, true, true);
#else
                            *pDST	  = (BYTE)(((WORD)(*pDST)     * AlphaBlendD + (WORD)(0xFF ^ *pSRC)     * AlphaBlendS + 127)/255);
                            *(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(0xFF ^ *(pSRC+1)) * AlphaBlendS + 127)/255);
                            *(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(0xFF ^ *(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
                        }
                        pSRC += TsstepX;
                        pAND += TastepX;
                        pDST += TdstepX;		// move dest ptr to next pixel
                    }
                    pA += TastepY;
                    pS += TsstepY;
                    pD += TdstepY;
                }
            } else {
                pA = Tapbits + TastepY*ys + TastepX*xs;
                for(y=0; y<hs; y++) {
                    pAND = pA;
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
                        if( *pAND == 0 ) {	// if AND mask has a hole, paint the SRC into the DST
#if MAC_PORT // LBO 09/16/07 [RGBA]
                            BlastPixelFromSource (pDST, pSRC, true, false);
#else
                            *pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
                            *(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
                            *(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
                        }
                        pSRC += TsstepX;
                        pAND += TastepX;
                        pDST += TdstepX;		// move dest ptr to next pixel
                    }
                    pA += TastepY;
                    pS += TsstepY;
                    pD += TdstepY;
                }
            }
        } else {
            if( invert ) {
                for(y=0; y<hs; y++) {
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
                        BlastPixelFromSource (pDST, pSRC, true, true);
#else
                        *pDST	  = (BYTE)(((WORD)(   *pDST)  * AlphaBlendD + (WORD)(0xFF ^ *pSRC)     * AlphaBlendS + 127)/255);
                        *(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(0xFF ^ *(pSRC+1)) * AlphaBlendS + 127)/255);
                        *(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(0xFF ^ *(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
                        pSRC += TsstepX;
                        pDST += TdstepX;
                    }
                    pS += TsstepY;
                    pD += TdstepY;
                }
            } else {
                for(y=0; y<hs; y++) {
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
                        BlastPixelFromSource (pDST, pSRC, true, false);
#else
                        *pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
                        *(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
                        *(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
                        pSRC += TsstepX;
                        pDST += TdstepX;
                    }
                    pS += TsstepY;
                    pD += TdstepY;
                }
            }
        }
    } else {
        if( masked ) {
            if( invert ) {
                pA = Tapbits + TastepY*ys + TastepX*xs;
                for(y=0; y<hs; y++) {
                    pAND = pA;
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
                        if( *pAND == 0 ) {	// if AND mask has a hole, paint the SRC into the DST
#if MAC_PORT // LBO 09/16/07 [RGBA]
                            BlastPixelFromSource (pDST, pSRC, false, true);
#else
                            *pDST	  = 0xFF ^ *pSRC;
                            *(pDST+1) = 0xFF ^ *(pSRC+1);
                            *(pDST+2) = 0xFF ^ *(pSRC+2);
#endif
                        }
                        pSRC += TsstepX;
                        pAND += TastepX;
                        pDST += TdstepX;		// move dest ptr to next pixel
                    }
                    pA += TastepY;
                    pS += TsstepY;
                    pD += TdstepY;
                }
            } else {
                pA = Tapbits + TastepY*ys + TastepX*xs;
                for(y=0; y<hs; y++) {
                    pAND = pA;
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
                        if( *pAND == 0 ) {	// if AND mask has a hole, paint the SRC into the DST
#if MAC_PORT // LBO 09/16/07 [RGBA]
                            BlastPixelFromSource (pDST, pSRC, false, false);
#else
                            *pDST = *pSRC;
                            *(pDST+1) = *(pSRC+1);
                            *(pDST+2) = *(pSRC+2);
#endif
                        }
                        pSRC += TsstepX;
                        pAND += TastepX;
                        pDST += TdstepX;		// move dest ptr to next pixel
                    }
                    pA += TastepY;
                    pS += TsstepY;
                    pD += TdstepY;
                }
            }
        } else {
            if( invert ) {
                for(y=0; y<hs; y++) {
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
                        BlastPixelFromSource (pDST, pSRC, false, true);
#else
                        *pDST	  = 0xFF ^ *pSRC;
                        *(pDST+1) = 0xFF ^ *(pSRC+1);
                        *(pDST+2) = 0xFF ^ *(pSRC+2);
#endif
                        pSRC += TsstepX;
                        pDST += TdstepX;
                    }
                    pS += TsstepY;
                    pD += TdstepY;
                }
            } else {
                for(y=0; y<hs; y++) {
                    pSRC = pS;
                    pDST = pD;
                    for(x=0; x<ws; x++) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
                        *(pDST+1) = *(pSRC+1);
                        *(pDST+2) = *(pSRC+2);
                        *(pDST+3) = *(pSRC+3);
#else
                        *pDST = *pSRC;
                        *(pDST+1) = *(pSRC+1);
                        *(pDST+2) = *(pSRC+2);
#endif
                        pSRC += TsstepX;
                        pDST += TdstepX;
                    }
                    pS += TsstepY;
                    pD += TdstepY;
                }
            }
        }
    }
    if( bmD==&kBM[0] ) {
        InvalidateScreen(xd, yd, xd+ws, yd+hs);
     }
}

//**************************************************************************
// Stretch or compress the source bitmap into the destination bitmap
// duplicating or removing pixels (pixel-resize)

// A *LOT* of redundant code in StretchBMB, but it's done to keep the
// routine as fast as possible under varying conditions of AlphaBlending or not,
// stretching or compressing or not. At least you shouldn't have to touch
// this code during ports! :-)

void StretchBMB(PBMB bmD, KLONG xd, KLONG yd, KLONG wd, KLONG hd, PBMB bmS, KLONG xs, KLONG ys, KLONG ws, KLONG hs, KLONG masked) {
	
	BYTE	*pDST, *pSRC, *pAND, *pD, *pS, *pA;
	KLONG		ErrX, ErrY, Ycnt, Xcnt;
	PBMB	bmDO;
	KLONG		Xdest, Ydest;
	
	if( wd==0 || hd==0 || ws==0 || hs==0 ) return;
	if( wd==ws && hd==hs ) {
		BltBMB(bmD, xd, yd, bmS, xs, ys, ws, hs, masked, FALSE);
		return;
	}
	bmDO = bmD;
	SetupBMD(&bmD);
//	if( bmS==0 ) bmS = &kBM[0];
	SetupBMS(&bmS);
	if( bmS->pAND==NULL ) masked = 0;	// if no mask, don't try to use it
	
	if( AlphaBlendD ) {
		if( wd>=ws && hd>=hs ) {	// stretch X and stretch Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=wd; Xcnt; Xcnt--) {
				if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
					Ydest = yd;
					ErrY  = hs>>1;
					pDST  = pD;
					pSRC  = pS;
					pAND  = pA;
					for(Ycnt=hd; Ycnt; Ycnt--) {
						if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							BlastPixelFromSource (pDST, pSRC, true, false);
#else
							*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
							*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
							*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
						}
						ErrY += hs;
						if( ErrY>hd ) {
							pSRC += TsstepY;
							pAND += TastepY;	// when masking, SRC bitmap is memory, and not rotated.
							// if not masking, pAND is never used, so don't care.
							ErrY -= hd;
						}
						++Ydest;
						pDST += TdstepY;
					}
				}
				ErrX += ws;
				if( ErrX>wd ) {
					pS += TsstepX;
					pA += TastepX;
					ErrX -= wd;
				}
				++Xdest;
				pD += TdstepX;
			}
		} else if( ws>=wd && hs>=hd ) {	// compress X and compress Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=ws; Xcnt; Xcnt--) {
				ErrX += wd;
				if( ErrX>ws ) {
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						Ydest = yd;
						ErrY  = hs>>1;
						pDST  = pD;
						pSRC  = pS;
						pAND  = pA;
						for(Ycnt=hs; Ycnt; Ycnt--) {
							ErrY += hd;
							if( ErrY>hs ) {
								if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelFromSource (pDST, pSRC, true, false);
#else
									*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
									*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
									*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
								}
								ErrY -= hs;
								++Ydest;
								pDST += TdstepY;
							}
							pSRC += TsstepY;
							pAND += TastepY;
						}
					}
					ErrX -= ws;
					++Xdest;
					pD += TdstepX;
				}
				pS += TsstepX;
				pA += TastepX;
			}
		} else if( wd>=ws && hd<=hs ) {	// stretch X and compress Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=wd; Xcnt; Xcnt--) {
				if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
					Ydest = yd;
					ErrY  = hs>>1;
					pDST  = pD;
					pSRC  = pS;
					pAND  = pA;
					for(Ycnt=hs; Ycnt; Ycnt--) {
						ErrY += hd;
						if( ErrY>hs ) {
							if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelFromSource (pDST, pSRC, true, false);
#else
								*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
								*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
								*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
							}
							ErrY -= hs;
							++Ydest;
							pDST += TdstepY;
						}
						pSRC += TsstepY;
						pAND += TastepY;
					}
				}
				ErrX += ws;
				if( ErrX>wd ) {
					pS += TsstepX;
					pA += TastepX;
					ErrX -= wd;
				}
				++Xdest;
				pD += TdstepX;
			}
		} else if( ws>=wd && hs<=hd ) {	// compress X and stretch Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=ws; Xcnt; Xcnt--) {
				ErrX += wd;
				if( ErrX>ws ) {
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						Ydest = yd;
						ErrY  = hs>>1;
						pDST  = pD;
						pSRC  = pS;
						pAND  = pA;
						for(Ycnt=hd; Ycnt; Ycnt--) {
							if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelFromSource (pDST, pSRC, true, false);
#else
								*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)(*pSRC) * AlphaBlendS + 127)/255);
								*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)(*(pSRC+1)) * AlphaBlendS + 127)/255);
								*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(*(pSRC+2)) * AlphaBlendS + 127)/255);
#endif
							}
							ErrY += hs;
							if( ErrY>hd ) {
								pSRC += TsstepY;
								pA += TastepY;
								ErrY -= hd;
							}
							++Ydest;
							pDST += TdstepY;
						}
					}
					ErrX -= ws;
					++Xdest;
					pD += TdstepX;
				}
				pS += TsstepX;
				pA += TastepX;
			}
		}
	} else {
		if( wd>=ws && hd>=hs ) {	// stretch X and stretch Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=wd; Xcnt; Xcnt--) {
				if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
					Ydest = yd;
					ErrY  = hs>>1;
					pDST  = pD;
					pSRC  = pS;
					pAND  = pA;
					for(Ycnt=hd; Ycnt; Ycnt--) {
						if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							BlastPixelFromSource (pDST, pSRC, false, false);
#else
							*pDST = *pSRC;
							*(pDST+1) = *(pSRC+1);
							*(pDST+2) = *(pSRC+2);
#endif
						}
						ErrY += hs;
						if( ErrY>hd ) {
							pSRC += TsstepY;
							pAND += TastepY;	// when masking, SRC bitmap is memory, and not rotated.
							// if not masking, pAND is never used, so don't care.
							ErrY -= hd;
						}
						++Ydest;
						pDST += TdstepY;
					}
				}
				ErrX += ws;
				if( ErrX>wd ) {
					pS += TsstepX;
					pA += TastepX;
					ErrX -= wd;
				}
				++Xdest;
				pD += TdstepX;
			}
		} else if( ws>=wd && hs>=hd ) {	// compress X and compress Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=ws; Xcnt; Xcnt--) {
				ErrX += wd;
				if( ErrX>ws ) {
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						Ydest = yd;
						ErrY  = hs>>1;
						pDST  = pD;
						pSRC  = pS;
						pAND  = pA;
						for(Ycnt=hs; Ycnt; Ycnt--) {
							ErrY += hd;
							if( ErrY>hs ) {
								if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelFromSource (pDST, pSRC, false, false);
#else
									*pDST = *pSRC;
									*(pDST+1) = *(pSRC+1);
									*(pDST+2) = *(pSRC+2);
#endif
								}
								ErrY -= hs;
								++Ydest;
								pDST += TdstepY;
							}
							pSRC += TsstepY;
							pAND += TastepY;
						}
					}
					ErrX -= ws;
					++Xdest;
					pD += TdstepX;
				}
				pS += TsstepX;
				pA += TastepX;
			}
		} else if( wd>=ws && hd<=hs ) {	// stretch X and compress Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=wd; Xcnt; Xcnt--) {
				if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
					Ydest = yd;
					ErrY  = hs>>1;
					pDST  = pD;
					pSRC  = pS;
					pAND  = pA;
					for(Ycnt=hs; Ycnt; Ycnt--) {
						ErrY += hd;
						if( ErrY>hs ) {
							if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelFromSource (pDST, pSRC, false, false);
#else
								*pDST = *pSRC;
								*(pDST+1) = *(pSRC+1);
								*(pDST+2) = *(pSRC+2);
#endif
							}
							ErrY -= hs;
							++Ydest;
							pDST += TdstepY;
						}
						pSRC += TsstepY;
						pAND += TastepY;
					}
				}
				ErrX += ws;
				if( ErrX>wd ) {
					pS += TsstepX;
					pA += TastepX;
					ErrX -= wd;
				}
				++Xdest;
				pD += TdstepX;
			}
		} else if( ws>=wd && hs<=hd ) {	// compress X and stretch Y
			Xdest = xd;
			ErrX  = ws>>1;
			pD = Tdpbits + TdstepY*yd + TdstepX*xd;
			pS = Tspbits + TsstepY*ys + TsstepX*xs;
			pA = Tapbits + TastepY*ys + TastepX*xs;
			for(Xcnt=ws; Xcnt; Xcnt--) {
				ErrX += wd;
				if( ErrX>ws ) {
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						Ydest = yd;
						ErrY  = hs>>1;
						pDST  = pD;
						pSRC  = pS;
						pAND  = pA;
						for(Ycnt=hd; Ycnt; Ycnt--) {
							if( Ydest>=TdclipY1 && Ydest<TdclipY2 && (!masked || (masked && *pAND==0)) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelFromSource (pDST, pSRC, false, false);
#else
								*pDST = *pSRC;
								*(pDST+1) = *(pSRC+1);
								*(pDST+2) = *(pSRC+2);
#endif
							}
							ErrY += hs;
							if( ErrY>hd ) {
								pSRC += TsstepY;
								pA += TastepY;
								ErrY -= hd;
							}
							++Ydest;
							pDST += TdstepY;
						}
					}
					ErrX -= ws;
					++Xdest;
					pD += TdstepX;
				}
				pS += TsstepX;
				pA += TastepX;
			}
		}
	}
    if( bmD==&kBM[0] ) {
        InvalidateScreen(xd, yd, xd+wd, yd+hd);
    }
}

//**************************************************************************
// Compress the source bitmap into the destination bitmap
// averaging source pixels to generate destination pixels (smooth resize)

// NOTE!!!! Does NOT do Alpha-blending or masked bitmaps!!!

#define	TRANPOINT	0.5

void CompressBMB(PBMB bmD, KLONG xd, KLONG yd, KLONG wd, KLONG hd, PBMB bmS, KLONG xs, KLONG ys, KLONG ws, KLONG hs, KLONG masked) {
	
	BYTE	*pDST, *pSRC, *pD, *pS, *pSS, *pA, *pAND, *pAA;
	KLONG	ki, Ycnt, Xcnt, Ydest, Xdest, srcXcnt, srcYcnt, srcXcnt2, srcYcnt2;
	PBMB	bmDO;
	double	wPPP, hPPP, wP, hP;	// pixels-per-pixels, ratio of ws/wd and hs/hd
	double	fYfirst, fXfirst, fysoff, fxsoff, r, g, b, d, m, a, s, transd;
	double	fY2, fX2, hMag, wMag;
	
	if( wd<=0 || hd<=0 || ws<=0 || hs<=0 ) {
		return;
	}
	if( wd<ws && hd>hs ) {
		ki = CreateBMB((WORD)wd, (WORD)hs, FALSE);
		CompressBMB(&kBM[ki], 0, 0, wd, hs, bmS, xs, ys, ws, hs, masked);
		CompressBMB(bmD, xd, yd, wd, hd, &kBM[ki], 0, 0, wd, hs, masked);
		DeleteBMB(ki);
		return;
	} else if( wd>ws && hd<hs ) {
		ki = CreateBMB((WORD)ws, (WORD)hd, FALSE);
		CompressBMB(&kBM[ki], 0, 0, ws, hd, bmS, xs, ys, ws, hs, masked);
		CompressBMB(bmD, xd, yd, wd, hd, &kBM[ki], 0, 0, ws, hd, masked);
		DeleteBMB(ki);
		return;
	}
	if( wd==ws && hd==hs ) {
		BltBMB(bmD, xd, yd, bmS, xs, ys, ws, hs, masked, FALSE);
		return;
	}
	bmDO = bmD;
	SetupBMD(&bmD);
//	if( bmS==0 ) bmS = kBuf;
	SetupBMS(&bmS);
	
	if( bmS->pAND==NULL ) masked = 0;	// if no mask, don't try to use it
	// ************* STREEEETTTTTTTTTTTTTCCCCCCCCCCCCHHHHHHHHHHHHHHHH
	if( wd>=ws && hd>=hs ) {	// stretch X and stretch Y
		wPPP = (double)ws/(double)wd;
		hPPP = (double)hs/(double)hd;
		// wPPP and hPPP are the fractional width and height (respectively) of a SRC pixel which maps to a full DST pixel
		wMag = (double)wd/(double)ws;
		hMag = (double)hd/(double)hs;
		// wMag and hMag are the amounts by which the fractional part of a SRC pixel must be magnified to make a full DST pixel
		
		pD = Tdpbits + TdstepY*yd + TdstepX*xd;
		pS = Tspbits + TsstepY*ys + TsstepX*xs;
		if( masked ) pA = Tapbits + TastepY*ys + TastepX*xs;
		else pA = NULL;
		
		hP = 0;
		// hP and wP are the Y (height) and X (width) offsets into the current source pixel of the remaining portion of that
		//	source pixel that will be used in the next DST pixel(s).  hP and wP are always >=0 and <1.0.
		
		// fYfirst is the portion of the "first" SRC pixel that will be used, multiplied by hMag.
		// fY2 is the portion of the "second" SRC pixel (if any) that will be used, multiplied by hMag.
		// ditto for fXfirst and fX2.
		
		for(Ycnt=srcYcnt=0, Ydest=yd; Ycnt<hd; Ycnt++, Ydest++) {
			while( hP>=1.0 ) {
				hP -= 1.0;
				pS += TsstepY;
				if( pA != NULL ) pA += TastepY;
				++srcYcnt;
			}
			fY2 = hP+hPPP-1.0;
			if( fY2<=0 ) {
				fYfirst = hPPP*hMag;
				fY2 = 0;
			} else {
				fYfirst = (hPPP-fY2)*hMag;
				fY2 *= hMag;
			}
			
			pDST = pD + TdstepY*Ycnt;
			pSRC = pS;
			pAND = pA;
			
			wP = 0;
			if( Ydest>=TdclipY1 && Ydest<TdclipY2 ) {
				for(Xcnt=srcXcnt=0, Xdest=xd; Xcnt<wd; Xcnt++, Xdest++) {
					while( wP>=1.0 ) {
						wP -= 1.0;
						pSRC += TsstepX;
						if( pAND != NULL ) pAND += TastepX;
						++srcXcnt;
					}
					fX2 = wP+wPPP-1.0;
					if( fX2<=0 ) {
						fXfirst = wPPP*wMag;
						fX2 = 0;
					} else {
						fXfirst = (wPPP-fX2)*wMag;
						fX2 *= wMag;
					}
					
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						if( fY2==0 || (srcYcnt+1)>=hs ) {
							// Y dst is controlled entirely by 1 src row
							if( fX2==0 || (srcXcnt+1)>=ws ) {
								// Y dst is controlled entirely by 1 src col (ie, one src pixel)
								if( !masked || pAND==NULL || (masked && *pAND==0) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelFromSource (pDST, pSRC, false, false);
									if( *(pDST+1)==0x81 && *(pDST+2)==0x81 && *(pDST+3)==0x81 )	// don't allow stretching to generate mask color
										{*(pDST+1) = 0x80; *(pDST+2) = 0x80; *(pDST+3) = 0x80;}
//										BlastPixel (0x80, 0x80, 0x80, pDST);
#else
									*pDST = *pSRC;
									*(pDST+1) = *(pSRC+1);
									*(pDST+2) = *(pSRC+2);
									if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
										*pDST = 0x80;
										*(pDST+1) = 0x80;
										*(pDST+2) = 0x80;
									}
#endif
								}
							} else {
								// Y dst is controlled by 2 src cols (pixels) on same src row
								if( !masked || pAND==NULL ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixel ((BYTE)((double)(*(pSRC+1))*fXfirst + (double)(*(pSRC+TsstepX+1))*fX2),
												(BYTE)((double)(*(pSRC+2))*fXfirst + (double)(*(pSRC+TsstepX+2))*fX2),
												(BYTE)((double)(*(pSRC+3))*fXfirst + (double)(*(pSRC+TsstepX+3))*fX2),
												pDST);
									if( *(pDST+1)==0x81 && *(pDST+2)==0x81 && *(pDST+3)==0x81 )	// don't allow stretching to generate mask color
										BlastPixel (0x80, 0x80, 0x80, pDST);
#else
									*pDST = (BYTE)((double)(*pSRC)*fXfirst + (double)(*(pSRC+TsstepX))*fX2);
									*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fXfirst + (double)(*(pSRC+TsstepX+1))*fX2);
									*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fXfirst + (double)(*(pSRC+TsstepX+2))*fX2);
									if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
										*pDST = 0x80;
										*(pDST+1) = 0x80;
										*(pDST+2) = 0x80;
									}
#endif
								} else if( (*pAND + *(pAND+TastepX))==0 ) {
									r = 1.0;
									s = 0;
									if( srcYcnt ) {
										s += ((*(pAND-TastepY))?(0.5*fXfirst):0);
										s += ((*(pAND-TastepY+TastepX))?(0.5*fX2):0);
										r += 0.5;
									}
									if( srcYcnt<hs-1 ) {
										s += ((*(pAND+TastepY))?(0.5*fXfirst):0);
										s += ((*(pAND+TastepY+TastepX))?(0.5*fX2):0);
										r += 0.5;
									}
									if( s/r<=TRANPOINT ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
										BlastPixel ((BYTE)((double)(*(pSRC+1))*fXfirst + (double)(*(pSRC+TsstepX+1))*fX2),
													(BYTE)((double)(*(pSRC+2))*fXfirst + (double)(*(pSRC+TsstepX+2))*fX2),
													(BYTE)((double)(*(pSRC+3))*fXfirst + (double)(*(pSRC+TsstepX+3))*fX2),
													pDST);
										if( *(pDST+1)==0x81 && *(pDST+2)==0x81 && *(pDST+3)==0x81 )	// don't allow stretching to generate mask color
											BlastPixel (0x80, 0x80, 0x80, pDST);
#else
										*pDST = (BYTE)((double)(*pSRC)*fXfirst + (double)(*(pSRC+TsstepX))*fX2);
										*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fXfirst + (double)(*(pSRC+TsstepX+1))*fX2);
										*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fXfirst + (double)(*(pSRC+TsstepX+2))*fX2);
										if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
											*pDST = 0x80;
											*(pDST+1) = 0x80;
											*(pDST+2) = 0x80;
										}
#endif
									}
								} else {
									r = 1.0;
									s = 0;
									if( srcYcnt ) {
										s += ((*(pAND-TastepY))?(0.5*fXfirst):0);
										s += ((*(pAND-TastepY+TastepX))?(0.5*fX2):0);
										r += 0.5;
									}
									if( srcYcnt<hs-1 ) {
										s += ((*(pAND+TastepY))?(0.5*fXfirst):0);
										s += ((*(pAND+TastepY+TastepX))?(0.5*fX2):0);
										r += 0.5;
									}
									s += ((*pAND)?fXfirst:0);
									s += ((*(pAND+TastepX))?fX2:0);
									if( s/r<=TRANPOINT ) {
										if( *pAND==0 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
											BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*fXfirst),
														(BYTE)((double)(*(pSRC+gOfs))*fXfirst),
														(BYTE)((double)(*(pSRC+bOfs))*fXfirst),
														pDST);
											if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
												BlastPixel (0x80, 0x80, 0x80, pDST);
#else
											*pDST = (BYTE)((double)(*pSRC)*fXfirst);
											*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fXfirst);
											*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fXfirst);
											if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
												*pDST = 0x80;
												*(pDST+1) = 0x80;
												*(pDST+2) = 0x80;
											}
#endif
										} else if( *(pAND+TastepX)==0 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
											BlastPixel ((BYTE)((double)(*(pSRC+TsstepX+rOfs))*fX2),
														(BYTE)((double)(*(pSRC+TsstepX+gOfs))*fX2),
														(BYTE)((double)(*(pSRC+TsstepX+bOfs))*fX2),
														pDST);
											if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
												BlastPixel (0x80, 0x80, 0x80, pDST);
#else
											*pDST = (BYTE)((double)(*(pSRC+TsstepX))*fX2);
											*(pDST+1) = (BYTE)((double)(*(pSRC+TsstepX+1))*fX2);
											*(pDST+2) = (BYTE)((double)(*(pSRC+TsstepX+2))*fX2);
											if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
												*pDST = 0x80;
												*(pDST+1) = 0x80;
												*(pDST+2) = 0x80;
											}
#endif
										}
									}
								}
							}
						} else {
							// fY2 != 0
							if( fX2==0 || (srcXcnt+1)>=ws  ) {
								if( !masked || pAND==NULL ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*fYfirst + (double)(*(pSRC+TsstepY+rOfs))*fY2),
												(BYTE)((double)(*(pSRC+gOfs))*fYfirst + (double)(*(pSRC+TsstepY+gOfs))*fY2),
												(BYTE)((double)(*(pSRC+bOfs))*fYfirst + (double)(*(pSRC+TsstepY+bOfs))*fY2),
												pDST);
#else
									*pDST = (BYTE)((double)(*pSRC)*fYfirst + (double)(*(pSRC+TsstepY))*fY2);
									*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fYfirst + (double)(*(pSRC+TsstepY+1))*fY2);
									*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fYfirst + (double)(*(pSRC+TsstepY+2))*fY2);
#endif
								} else if( (*pAND + *(pAND+TastepY))==0 ) {
									r = 1.0;
									s = 0;
									if( srcXcnt ) {
										s += ((*(pAND-TastepX))?(0.5*fYfirst):0);
										s += ((*(pAND+TastepY-TastepX))?(0.5*fY2):0);
										r += 0.5;
									}
									if( srcXcnt<ws-1 ) {
										s += ((*(pAND+TastepX))?(0.5*fYfirst):0);
										s += ((*(pAND+TastepY+TastepX))?(0.5*fY2):0);
										r += 0.5;
									}
									if( s/r<=TRANPOINT ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
										BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*fYfirst + (double)(*(pSRC+TsstepY+rOfs))*fY2),
													(BYTE)((double)(*(pSRC+gOfs))*fYfirst + (double)(*(pSRC+TsstepY+gOfs))*fY2),
													(BYTE)((double)(*(pSRC+bOfs))*fYfirst + (double)(*(pSRC+TsstepY+bOfs))*fY2),
													pDST);
										if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
											BlastPixel (0x80, 0x80, 0x80, pDST);
#else
										*pDST = (BYTE)((double)(*pSRC)*fYfirst + (double)(*(pSRC+TsstepY))*fY2);
										*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fYfirst + (double)(*(pSRC+TsstepY+1))*fY2);
										*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fYfirst + (double)(*(pSRC+TsstepY+2))*fY2);
										if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
											*pDST = 0x80;
											*(pDST+1) = 0x80;
											*(pDST+2) = 0x80;
										}
#endif
									}
								} else {
									r = 1.0;
									s = 0;
									if( srcXcnt ) {
										s += ((*(pAND-TastepX))?(0.5*fYfirst):0);
										s += ((*(pAND+TastepY-TastepX))?(0.5*fY2):0);
										r += 0.5;
									}
									if( srcXcnt<ws-1 ) {
										s += ((*(pAND+TastepX))?(0.5*fYfirst):0);
										s += ((*(pAND+TastepY+TastepX))?(0.5*fY2):0);
										r += 0.5;
									}
									s += ((*pAND)?fYfirst:0);
									s += ((*(pAND+TastepY))?fY2:0);
									if( s/r<=TRANPOINT ) {
										if( *pAND==0 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
											BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*fYfirst),
														(BYTE)((double)(*(pSRC+gOfs))*fYfirst),
														(BYTE)((double)(*(pSRC+bOfs))*fYfirst),
														pDST);
											if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
												BlastPixel (0x80, 0x80, 0x80, pDST);
#else
											*pDST = (BYTE)((double)(*pSRC)*fYfirst);
											*(pDST+1) = (BYTE)((double)(*(pSRC+1))*fYfirst);
											*(pDST+2) = (BYTE)((double)(*(pSRC+2))*fYfirst);
											if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
												*pDST = 0x80;
												*(pDST+1) = 0x80;
												*(pDST+2) = 0x80;
											}
#endif
										} else if( *(pAND+TastepY)==0 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
											BlastPixel ((BYTE)((double)(*(pSRC+TsstepY+rOfs))*fY2),
														(BYTE)((double)(*(pSRC+TsstepY+gOfs))*fY2),
														(BYTE)((double)(*(pSRC+TsstepY+bOfs))*fY2),
														pDST);
											if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
												BlastPixel (0x80, 0x80, 0x80, pDST);
#else
											*pDST = (BYTE)((double)(*(pSRC+TsstepY))*fY2);
											*(pDST+1) = (BYTE)((double)(*(pSRC+TsstepY+1))*fY2);
											*(pDST+2) = (BYTE)((double)(*(pSRC+TsstepY+2))*fY2);
											if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
												*pDST = 0x80;
												*(pDST+1) = 0x80;
												*(pDST+2) = 0x80;
											}
#endif
										}
									}
								}
							} else {
								// fY2 !=0 && fX2 !=0
								long	tc;
								
								r = fXfirst*fYfirst;// top-left piece
								g = fX2*fYfirst;// top-right piece
								b = fXfirst*fY2;// bottom-left piece
								m = fX2*fY2;// bottom-right piece
								s = 0;
								tc= 0;
								if( masked && pAND!=NULL ) {
									if( *pAND ) {
										s += r;
										r = 0;
										++tc;
									}
									if( srcYcnt+1<hs && *(pAND+TastepY) ) {
										s += b;
										b = 0;
										++tc;
									}
									if( srcXcnt+1<ws && *(pAND+TastepX) ) {
										s += g;
										g = 0;
										++tc;
									}
									if( srcXcnt+1<ws && srcYcnt+1<hs && *(pAND+TastepY+TastepX) ) {
										s += m;
										m = 0;
										++tc;
									}
									if( s<=TRANPOINT ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
										BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*r + (double)(*(pSRC+TsstepY+rOfs))*b + (double)(*(pSRC+TsstepX+rOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+rOfs))*m),
													(BYTE)((double)(*(pSRC+gOfs))*r + (double)(*(pSRC+TsstepY+gOfs))*b + (double)(*(pSRC+TsstepX+gOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+gOfs))*m),
													(BYTE)((double)(*(pSRC+bOfs))*r + (double)(*(pSRC+TsstepY+bOfs))*b + (double)(*(pSRC+TsstepX+bOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+bOfs))*m),
													pDST);
										if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
											BlastPixel (0x80, 0x80, 0x80, pDST);
#else
										*pDST = (BYTE)((double)(*pSRC)*r + (double)(*(pSRC+TsstepY))*b + (double)(*(pSRC+TsstepX))*g + (double)(*(pSRC+TsstepY+TsstepX))*m);
										*(pDST+1) = (BYTE)((double)(*(pSRC+1))*r + (double)(*(pSRC+TsstepY+1))*b + (double)(*(pSRC+TsstepX+1))*g + (double)(*(pSRC+TsstepY+TsstepX+1))*m);
										*(pDST+2) = (BYTE)((double)(*(pSRC+2))*r + (double)(*(pSRC+TsstepY+2))*b + (double)(*(pSRC+TsstepX+2))*g + (double)(*(pSRC+TsstepY+TsstepX+2))*m);
										if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
											*pDST = 0x80;
											*(pDST+1) = 0x80;
											*(pDST+2) = 0x80;
										}
#endif
									}
								} else {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixel ((BYTE)((double)(*(pSRC+rOfs))*r + (double)(*(pSRC+TsstepY+rOfs))*b + (double)(*(pSRC+TsstepX+rOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+rOfs))*m),
												(BYTE)((double)(*(pSRC+gOfs))*r + (double)(*(pSRC+TsstepY+gOfs))*b + (double)(*(pSRC+TsstepX+gOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+gOfs))*m),
												(BYTE)((double)(*(pSRC+bOfs))*r + (double)(*(pSRC+TsstepY+bOfs))*b + (double)(*(pSRC+TsstepX+bOfs))*g + (double)(*(pSRC+TsstepY+TsstepX+bOfs))*m),
												pDST);
									if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
										BlastPixel (0x80, 0x80, 0x80, pDST);
#else
									*pDST = (BYTE)((double)(*pSRC)*r + (double)(*(pSRC+TsstepY))*b + (double)(*(pSRC+TsstepX))*g + (double)(*(pSRC+TsstepY+TsstepX))*m);
									*(pDST+1) = (BYTE)((double)(*(pSRC+1))*r + (double)(*(pSRC+TsstepY+1))*b + (double)(*(pSRC+TsstepX+1))*g + (double)(*(pSRC+TsstepY+TsstepX+1))*m);
									*(pDST+2) = (BYTE)((double)(*(pSRC+2))*r + (double)(*(pSRC+TsstepY+2))*b + (double)(*(pSRC+TsstepX+2))*g + (double)(*(pSRC+TsstepY+TsstepX+2))*m);
									if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't allow stretching to generate mask color
										*pDST = 0x80;
										*(pDST+1) = 0x80;
										*(pDST+2) = 0x80;
									}
#endif
								}
							}
						}
					}
					wP += wPPP;
					pDST += TdstepX;
				}
			}
			hP += hPPP;
		}
		// ************* COMPRESS
	} else if( ws>=wd && hs>=hd ) {	// compress X and compress Y
		wPPP = (double)ws/(double)wd;
		hPPP = (double)hs/(double)hd;
		
		pD = Tdpbits + TdstepY*yd + TdstepX*xd;
		pS = Tspbits + TsstepY*ys + TsstepX*xs;
		if( masked ) pA = Tapbits + TastepY*ys + TastepX*xs;
		else pA = NULL;
		
		for(Ycnt=0, Ydest=yd; Ycnt<hd; Ycnt++, Ydest++) {
			pDST = pD + TdstepY*Ycnt;
			if( Ydest>=TdclipY1 && Ydest<TdclipY2 ) {
				srcYcnt = (KLONG)floor((double)(hPPP*(double)Ycnt));
				for(Xcnt=0, Xdest=xd; Xcnt<wd; Xcnt++, Xdest++) {
					fYfirst = 1.0 - modf((hPPP*(double)Ycnt), &fysoff);	// how much of the first Y pixel line to use for this dest pixel
					srcXcnt = (KLONG)floor((double)(wPPP*(double)Xcnt));
					pSRC = pS + TsstepY*srcYcnt + TsstepX*srcXcnt;
					pAND = pA;
					if( pAND != NULL ) pAND += TastepY*srcYcnt + TastepX*srcXcnt;
					a = r = g = b = d = transd = 0.0;
					srcYcnt2 = srcYcnt;
					for(hP=hPPP; hP>0 && srcYcnt2<hs; ) {
						fXfirst = 1.0 - modf((wPPP*(double)Xcnt), &fxsoff);
						pSS = pSRC;
						pAA = pAND;
						srcXcnt2 = srcXcnt;
						for(wP=wPPP; wP>0 && srcXcnt2<ws; ) {
							m = fYfirst * fXfirst;
							if( !masked || pAA==NULL || (masked && *pAA==0) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								b += (double)(*(pSS+bOfs)) * m;
								g += (double)(*(pSS+gOfs)) * m;
								r += (double)(*(pSS+rOfs)) * m;
#else
								b += (double)(*pSS) * m;
								g += (double)(*(pSS+1)) * m;
								r += (double)(*(pSS+2)) * m;
#endif
								
								if( masked && pAA!=NULL ) a += (double)(*pAA) * m;
								d += m;
							} else {
								transd += m;
							}
							pSS += TsstepX;
							if( pAA!=NULL ) pAA += TastepX;
							++srcXcnt2;
							wP -= fXfirst;
							if( wP>=1.0 ) fXfirst = 1.0;
							else fXfirst = wP;
						}
						pSRC += TsstepY;
						if( pAND!=NULL ) pAND += TastepY;
						++srcYcnt2;
						hP -= fYfirst;
						if( hP>=1.0 ) fYfirst = 1.0;
						else fYfirst = hP;
					}
					if( Xdest>=TdclipX1 && Xdest<TdclipX2 ) {
						if( d>transd && (!masked || ((BYTE)(double)(a/d))<128) ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							BlastPixel ((BYTE)(double)(r/d), (BYTE)(double)(g/d), (BYTE)(double)(b/d), pDST);
							if( *(pDST+rOfs)==0x81 && *(pDST+gOfs)==0x81 && *(pDST+bOfs)==0x81 )	// don't allow stretching to generate mask color
								BlastPixel (0x80, 0x80, 0x80, pDST);
#else
							*pDST = (BYTE)(double)(b/d);
							*(pDST+1) = (BYTE)(double)(g/d);
							*(pDST+2) = (BYTE)(double)(r/d);
							if( *pDST==0x81 && *(pDST+1)==0x81 && *(pDST+2)==0x81 ) {	// don't create mask color via stretching
								*pDST = 0x80;
								*(pDST+1) = 0x80;
								*(pDST+2) = 0x80;
							}
#endif
						}
					}
					pDST += TdstepX;
				}
			}
		}
	}
	
    if( bmD==&kBM[0] ) {
        InvalidateScreen(xd, yd, xd+wd, yd+hd);
    }
}

//**************************************************************************
// Plot a point of a specific color in the destination bitmap
void PointBMB(PBMB bmD, KLONG x, KLONG y, DWORD clr) {
	
	BYTE	*pDST;
	BYTE r, g, b;
	
	SetupBMD(&bmD);
	if( x<TdclipX1 || x>=TdclipX2 || y<TdclipY1 || y>=TdclipY2 ) return;
	pDST = Tdpbits + TdstepY*y + TdstepX*x;
	r = (BYTE)(clr & 0x000000FF);
	g = (BYTE)((clr & 0x0000FF00) >> 8);
	b = (BYTE)((clr & 0x00FF0000) >> 16);
	if( AlphaBlendD ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
		BlastPixelAlpha (r, g, b, pDST);
#else
		*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)((clr & 0x0FF0000)>>16) * AlphaBlendS + 127)/255);
		*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)((clr & 0x0FF00)>>8) * AlphaBlendS + 127)/255);
		*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)(clr & 0x0FF) * AlphaBlendS + 127)/255);
#endif
	} else {
#if MAC_PORT // LBO 09/16/07 [RGBA]
		*(pDST+1) = r;
		*(pDST+2) = g;
		*(pDST+3) = b;
//		BlastPixel (r, g, b, pDST);
#else
		*pDST++ = (BYTE)((clr & 0x0FF0000)>>16);
		*pDST++ = (BYTE)((clr & 0x0FF00)>>8);
		*pDST++ = (BYTE)(clr & 0x0FF);
#endif
	}
    if( bmD==&kBM[0] ) {
        InvalidateScreen(x, y, x+1, y+1);
    }
}

///************************************************************
void StringFindReplace(BYTE *src, BYTE *strF, BYTE *strR) {
	
	int		s, d, f, r;

	for(s=d=0; src[s]; ) {
		if( src[s]==strF[0] ) {
			for(f=0; strF[f] && src[s+f]; f++) if( strF[f]!=src[s+f] ) break;
			if( strF[f]==0 ) {
				for(r=0; strR[r]; r++) src[d+r] = strR[r];
				s += f;
				d += r;
			} else {
				src[d++] = src[s++];
			}
		} else {
			src[d++] = src[s++];
		}
	}
	src[d] = 0;
}

void MacStringFixup(BYTE *inString) {
	
#if USE_CMD_KEY
	StringFindReplace (inString, (BYTE*)"Ctl", (BYTE*)"Cmd");
	StringFindReplace (inString, (BYTE*)"CTL", (BYTE*)"CMD");
	StringFindReplace (inString, (BYTE*)"Ctrl", (BYTE*)"Cmd");
	StringFindReplace (inString, (BYTE*)"CTRL", (BYTE*)"CMD");
#endif
}

	//*****************************************************
// Get the width, cell-height, and ascending (non-descending) height
// of a specified string, using the currently selected font.
void TextSizeBMB(BYTE *str, WORD *w, WORD *h, WORD *a) {
	
	WORD	*pW;
	UNALIGNED DWORD	*pD;
	BYTE	c, i;
	
#if MAC_PORT
	MacStringFixup (str);
#endif
	pW = (WORD *)(pFontData + pFontTable[CurFont]);
	*h = *pW++;	// height of cells
	*a = *pW++;	// AscendingHeight of cells
	i  = (BYTE)(*pW++) & 4;	// italic?
	pD = (DWORD *)pW;	// pD is now a pointer to the table of character pointers
	*w  = 0;
	while( *str ) {
		c = *str++;
		if( !(c & 0x60) ) c = 127;	// map non-printing CTL characters into 127
		c -= (c & 0x80)?64:32;	// map characters 32-127 and 160-255 into 0-191
		if( !i || (i && *str) ) *w += (WORD)(*(pFontData + pD[c] + 3));	// character width on non-italic and italic-non-last characters
		else *w += (WORD)(*(pFontData + pD[c]));	// cell width on italic-last characters
	}
}

//**************************************************************************
// Output text to the destination bitmap using the currently selected font.
void TextBMB(PBMB bmD, KLONG x, KLONG y, BYTE *str) {
	
	UNALIGNED BYTE	*pDST, *pC, *pCsx;
	BYTE	r, g, b, bbits;
	UNALIGNED DWORD	*pF, dwbits;
	WORD	wbits, bw, bh, ba;
	KLONG		sx, ey, cellH, cw, ch, c, toppad, cwtmp, cwmax, charw, tx, ty;
	PBMB	bmDO;
	
#if MAC_PORT
	MacStringFixup (str);
#endif
	bmDO = bmD;
	SetupBMD(&bmD);
	if( !(CurBackColor & 0x80000000) ) {
		TextSizeBMB(str, &bw, &bh, &ba);
		if( bw && bh ) {
			sx = Kmax(x, TdclipX1);	// clip background rectangle
			cw = Kmin(x+bw, TdclipX2);
			ch = Kmax(y, TdclipY1);
			ey = Kmin(y+bh, TdclipY2);
			if( sx<cw && ch<ey ) RectBMB(bmDO, sx, ch, cw, ey, CurBackColor, -1);
		}
	}
	sx = x;	// save starting x for possible InvalidateScreen()
	pF = (DWORD *)(pFontData + pFontTable[CurFont]);
	cellH = *((WORD *)pF);	// height of cells
	ey = y+cellH;
	pF = (DWORD *)(((BYTE *)pF)+6);	// pF is now a pointer to the table of character pointers
	
	r = (BYTE)(CurTextColor & 0x000000FF);
	g = (BYTE)((CurTextColor & 0x0000FF00) >> 8);
	b = (BYTE)((CurTextColor & 0x00FF0000) >> 16);
	for(; *str; x+=charw) {	// charw gets set a few lines further down
		pCsx = Tdpbits + TdstepY*y + TdstepX*x;
		c = (KLONG)(KSHORT)(WORD)(*str++);
		if( !(c & 0x60) ) c = 127;	// map non-printing CTL characters into 127
		c -= (c & 0x80)?64:32;	// map characters 32-127 and 160-255 into 0-191
		pC = pFontData + pF[c];	// pointer to font data for this character
		cw = (KLONG)(*pC++);		// # of non-space columns of pixels in char
		toppad = (KLONG)(*pC++);	// # of blank lines of pixels at top of char
		ch = (KLONG)(*pC++);		// # of non-blank lines of pixel data
		charw = (KLONG)(*pC++);	// # of non-space AND space columns of pixels in char
		
		pDST = (pCsx += TdstepY*toppad);
		
		if( x>=TdclipX1 && (x+charw)<=TdclipX2 && y>=TdclipY1 && ey<=TdclipY2 ) {	// character is NOT clipped
			if( AlphaBlendD ) {
				while( ch-- ) {
					cwtmp = cw;
					while( cwtmp>24 ) {
						cwmax = Kmin(32, cwtmp);
						cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
						dwbits = GETMEMD(pC);
#else
						dwbits = *((UNALIGNED DWORD *)pC);
#endif
						pC += 4;
						while( cwmax-- ) {
							if( dwbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelAlpha (r, g, b, pDST);
#else
								*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
								*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
								*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
							}
							pDST += TdstepX;
							dwbits >>= 1;
						}
					}
					while( cwtmp>8 ) {
						cwmax = Kmin(16, cwtmp);
						cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
						wbits = GETMEMW(pC);
#else
						wbits = *((UNALIGNED WORD *)pC);
#endif
						pC += 2;
						while( cwmax-- ) {
							if( wbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelAlpha (r, g, b, pDST);
#else
								*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
								*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
								*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
							}
							pDST += TdstepX;
							wbits >>= 1;
						}
					}
					if( cwtmp ) {
						bbits = *pC++;
						while( cwtmp-- ) {
							if( bbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								BlastPixelAlpha (r, g, b, pDST);
#else
								*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
								*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
								*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
							}
							pDST += TdstepX;
							bbits >>= 1;
						}
					}
					pDST = (pCsx += TdstepY);
				}
			} else {
				while( ch-- ) {
					cwtmp = cw;
					while( cwtmp>24 ) {
						cwmax = Kmin(32, cwtmp);
						cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
						dwbits = GETMEMD(pC);
#else
						dwbits = *((UNALIGNED DWORD *)pC);
#endif
						pC += 4;
						while( cwmax-- ) {
							if( dwbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								*(pDST+1) = r;
								*(pDST+2) = g;
								*(pDST+3) = b;
//								BlastPixel (r, g, b, pDST);
#else
								*pDST = b;
								*(pDST+1) = g;
								*(pDST+2) = r;
#endif
							}
							pDST += TdstepX;
							dwbits >>= 1;
						}
					}
					while( cwtmp>8 ) {
						cwmax = Kmin(16, cwtmp);
						cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
						wbits = GETMEMW(pC);
#else
						wbits = *((UNALIGNED WORD *)pC);
#endif
						pC += 2;
						while( cwmax-- ) {
							if( wbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								*(pDST+1) = r;
								*(pDST+2) = g;
								*(pDST+3) = b;
//								BlastPixel (r, g, b, pDST);
#else
								*pDST = b;
								*(pDST+1) = g;
								*(pDST+2) = r;
#endif
							}
							pDST += TdstepX;
							wbits >>= 1;
						}
					}
					if( cwtmp ) {
						bbits = *pC++;
						while( cwtmp-- ) {
							if( bbits & 1 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
								*(pDST+1) = r;
								*(pDST+2) = g;
								*(pDST+3) = b;
//								BlastPixel (r, g, b, pDST);
#else
								*pDST = b;
								*(pDST+1) = g;
								*(pDST+2) = r;
#endif
							}
							pDST += TdstepX;
							bbits >>= 1;
						}
					}
					pDST = (pCsx += TdstepY);
				}
			}
		} else { // character is at least partially clipped
			if( (x+charw)<=TdclipX1 || x>=TdclipX2 || ey<=TdclipY1 || y>=TdclipY2 ) continue;	// character is completely clipped
			ty = y+toppad;
			if( AlphaBlendD ) {
				while( ty<TdclipY2 && ch-- ) {
					if( ty>=TdclipY1 ) {
						tx = x;
						cwtmp = cw;
						while( cwtmp>24 ) {
							cwmax = Kmin(32, cwtmp);
							cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
							dwbits = GETMEMD(pC);
#else
							dwbits = *((UNALIGNED DWORD *)pC);
#endif
							pC += 4;
							while( cwmax-- ) {
								if( (dwbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelAlpha (r, g, b, pDST);
#else
									*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
									*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
									*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
								}
								++tx;
								pDST += TdstepX;
								dwbits >>= 1;
							}
						}
						while( cwtmp>8 ) {
							cwmax = Kmin(16, cwtmp);
							cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
							wbits = GETMEMW(pC);
#else
							wbits = *((UNALIGNED WORD *)pC);
#endif
							pC += 2;
							while( cwmax-- ) {
								if( (wbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelAlpha (r, g, b, pDST);
#else
									*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
									*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
									*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
								}
								++tx;
								pDST += TdstepX;
								wbits >>= 1;
							}
						}
						if( cwtmp ) {
							bbits = *pC++;
							while( cwtmp-- ) {
								if( (bbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									BlastPixelAlpha (r, g, b, pDST);
#else
									*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
									*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
									*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
								}
								++tx;
								pDST += TdstepX;
								bbits >>= 1;
							}
						}
					} else pC += (cw+7)/8;
					++ty;
					pDST = (pCsx += TdstepY);
				}
			} else {
				while( ty<TdclipY2 && ch-- ) {
					if( ty>=TdclipY1 ) {
						tx = x;
						cwtmp = cw;
						while( cwtmp>24 ) {
							cwmax = Kmin(32, cwtmp);
							cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
							dwbits = GETMEMD(pC);
#else
							dwbits = *((UNALIGNED DWORD *)pC);
#endif
							pC += 4;
							while( cwmax-- ) {
								if( (dwbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									*(pDST+1) = r;
									*(pDST+2) = g;
									*(pDST+3) = b;
//									BlastPixel (r, g, b, pDST);
#else
									*pDST = b;
									*(pDST+1) = g;
									*(pDST+2) = r;
#endif
								}
								++tx;
								pDST += TdstepX;
								dwbits >>= 1;
							}
						}
						while( cwtmp>8 ) {
							cwmax = Kmin(16, cwtmp);
							cwtmp -= cwmax;
#if MAC_PORT // LBO 09/16/07 [BS]
							wbits = GETMEMW(pC);
#else
							wbits = *((UNALIGNED WORD *)pC);
#endif
							pC += 2;
							while( cwmax-- ) {
								if( (wbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									*(pDST+1) = r;
									*(pDST+2) = g;
									*(pDST+3) = b;
//									BlastPixel (r, g, b, pDST);
#else
									*pDST = b;
									*(pDST+1) = g;
									*(pDST+2) = r;
#endif
								}
								++tx;
								pDST += TdstepX;
								wbits >>= 1;
							}
						}
						if( cwtmp ) {
							bbits = *pC++;
							while( cwtmp-- ) {
								if( (bbits & 1) && tx>=TdclipX1 && tx<TdclipX2 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
									*(pDST+1) = r;
									*(pDST+2) = g;
									*(pDST+3) = b;
//									BlastPixel (r, g, b, pDST);
#else
									*pDST = b;
									*(pDST+1) = g;
									*(pDST+2) = r;
#endif
								}
								++tx;
								pDST += TdstepX;
								bbits >>= 1;
							}
						}
					} else pC += (cw+7)/8;
					++ty;
					pDST = (pCsx += TdstepY);
				}
			}
		}
	}
	if( bmD==&kBM[0] ) {
		InvalidateScreen(sx, y, x, ey);
	}
}

//**************************************************************************
// Draw a line between two points on the destination bitmap
void LineBMB(PBMB bmD, KLONG x1, KLONG y1, KLONG x2, KLONG y2, DWORD clr) {
	
	BYTE	*pDST, r, g, b;
	KLONG		x, y, d=0, w, s=0;
	PBMB	bmDO;
	
	bmDO = bmD;
	SetupBMD(&bmD);
	r = (BYTE)(clr & 0x000000FF);
	g = (BYTE)((clr & 0x0000FF00) >> 8);
	b = (BYTE)((clr & 0x00FF0000) >> 16);
	if( x1==x2 ) {	// VERTICAL line
		if( x1<TdclipX1 || x1>=TdclipX2 ) return;
		// if end-points are same, just plot single point
		if( y1<y2 ) {
			d = 1;
			if( y1>=TdclipY2 || y2<TdclipY1 ) return;
			if( y1<TdclipY1 ) y1 = TdclipY1;
			if( y2>=TdclipY2 ) y2 = TdclipY2-1;
			w = y2-y1+1;
			s = TdstepY;
		} else if( y1>y2 ) {
			d = -1;
			if( y2>=TdclipY2 || y1<TdclipY1 ) return;
			if( y2<TdclipY1 ) y2 = TdclipY1;
			if( y1>=TdclipY2 ) y1 = TdclipY2-1;
			w = y1-y2+1;
			s = -TdstepY;
		} else w = 0;
		if( w==0 ) {
			PointBMB(bmDO, x1, y1, clr);
			return;
		}
		pDST = Tdpbits + TdstepY*y1 + TdstepX*x1;
		if( AlphaBlendD ) {
			for(y=y1; w--; y += d) {
				if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
					BlastPixelAlpha (r, g, b, pDST);
#else
					*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
					*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
					*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
		} else {
			for(y=y1; w--; y += d) {
				if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
					*(pDST+1) = r;
					*(pDST+2) = g;
					*(pDST+3) = b;
//					BlastPixel (r, g, b, pDST);
#else
					*pDST = b;
					*(pDST+1) = g;
					*(pDST+2) = r;
#endif
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
		}
		++x2;	// for invalidate
	} else if( y1==y2 ) {	// HORIZONTAL line
		if( y1<TdclipY1 || y1>=TdclipY2 ) return;
		// if end-points are same, just plot single point
		if( x1<x2 ) {
			d = 1;
			if( x1>=TdclipX2 || x2<TdclipX1 ) return;
			if( x1<TdclipX1 ) x1 = TdclipX1;
			if( x2>=TdclipX2 ) x2 = TdclipX2-1;
			w = x2-x1+1;
			s = TdstepX;
		} else if( x1>x2 ) {
			d = -1;
			if( x2>=TdclipX2 || x1<TdclipX1 ) return;
			if( x2<TdclipX1 ) x2 = TdclipX1;
			if( x1>=TdclipX2 ) x1 = TdclipX2-1;
			w = x1-x2+1;
			s = -TdstepX;
		} else w = 0;
		if( w==0 ) {
			PointBMB(bmDO, x1, y1, clr);
			return;
		}
		pDST = Tdpbits + TdstepY*y1 + TdstepX*x1;
		if( AlphaBlendD ) {
			for(x=x1; w--; x += d) {
				if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
					BlastPixelAlpha (r, g, b, pDST);
#else
					*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
					*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
					*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
		} else {
			for(x=x1; w--; x += d) {
				if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
					*(pDST+1) = r;
					*(pDST+2) = g;
					*(pDST+3) = b;
//					BlastPixel (r, g, b, pDST);
#else
					*pDST = b;
					*(pDST+1) = g;
					*(pDST+2) = r;
#endif
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
		}
		++y2;	// for invalidate
	} else {	// DIAGONAL line
		KLONG dX, dY, Xincr, Yincr, dPr, dPru, P, Ax, Ay;
		
		pDST = Tdpbits + TdstepY*y1 + TdstepX*x1;
		if( x1 > x2 ) {
			Xincr = -1;
			TdstepX = -TdstepX;
		} else Xincr = 1;
		
		if( y1 > y2 ) {
			Yincr = -1;
			TdstepY = -TdstepY;
		} else Yincr = 1;
		Ax = x1;
		Ay = y1;
		dX = abs(x2-x1);
		dY = abs(y2-y1);
		if( dX >= dY ) {
			dPr = dY<<1;
			dPru = dPr - (dX<<1);
			P = dPr - dX;
			if( AlphaBlendD ) {
				for(; dX >= 0; dX--) {
					if( Ax>=TdclipX1 && Ax<TdclipX2 && Ay>=TdclipY1 && Ay<TdclipY2 ) {
						if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							BlastPixelAlpha (r, g, b, pDST);
#else
							*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
							*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
							*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ay += Yincr;
						pDST += TdstepY;
						P += dPru;
					} else P += dPr;
					Ax += Xincr;
					pDST += TdstepX;
				}
			} else {
				for(; dX >= 0; dX--) {
					if( Ax>=TdclipX1 && Ax<TdclipX2 && Ay>=TdclipY1 && Ay<TdclipY2 ) {
						if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							*(pDST+1) = r;
							*(pDST+2) = g;
							*(pDST+3) = b;
//							BlastPixel (r, g, b, pDST);
#else
							*pDST = b;
							*(pDST+1) = g;
							*(pDST+2) = r;
#endif
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ay += Yincr;
						pDST += TdstepY;
						P += dPru;
					} else P += dPr;
					Ax += Xincr;
					pDST += TdstepX;
				}
			}
		} else {
			dPr = dX<<1;
			dPru = dPr - (dY<<1);
			P = dPr - dY;
			if( AlphaBlendD ) {
				for(; dY >= 0; dY--) {
					if( Ax>=TdclipX1 && Ax<TdclipX2 && Ay>=TdclipY1 && Ay<TdclipY2 ) {
						if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							BlastPixelAlpha (r, g, b, pDST);
#else
							*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
							*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
							*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ax += Xincr;
						pDST += TdstepX;
						P += dPru;
					} else P += dPr;
					Ay += Yincr;
					pDST += TdstepY;
				}
			} else {
				for(; dY >= 0; dY--) {
					if( Ax>=TdclipX1 && Ax<TdclipX2 && Ay>=TdclipY1 && Ay<TdclipY2 ) {
						if( LineStyle & 0x80000000 ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
							*(pDST+1) = r;
							*(pDST+2) = g;
							*(pDST+3) = b;
//							BlastPixel (r, g, b, pDST);
#else
							*pDST = b;
							*(pDST+1) = g;
							*(pDST+2) = r;
#endif
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ax += Xincr;
						pDST += TdstepX;
						P += dPru;
					} else P += dPr;
					Ay += Yincr;
					pDST += TdstepY;
				}
			}
		}
	}
    if( bmD==&kBM[0] ) {
        InvalidateScreen(x1, y1, x2, y2);
    }
}

//**************************************************************************
// Draw a rectangle in the destination bitmap.  The rectangle may be
// SOLID, OUTLINED, or BOTH, depending upon the values of 'fill' and 'edge'.
// if 'fill' is not -1, fill the rectangle with that color
// if 'edge' is not -1, outline the rectangle with that color
void RectBMB(PBMB bmD, KLONG x1, KLONG y1, KLONG x2, KLONG y2, DWORD fill, DWORD edge) {
	
	BYTE	*pDST, *pD, r, g, b;
	KLONG		x, y, w, h, ww, edged;
	PBMB	bmDO;
	
	if( x1>x2 ) {
		x = x1;
		x1 = x2;
		x2 = x;
	}
	if( y1>y2 ) {
		y = y1;
		y1 = y2;
		y2 = y;
	}
	bmDO = bmD;
	SetupBMD(&bmD);
	edged = (edge==-1)?0:1;
	if( fill != -1 ) {
		x = Kmax(TdclipX1, x1+edged);
		y = Kmax(TdclipY1, y1+edged);
		w = Kmin(TdclipX2, x2-edged)-x;
		h = Kmin(TdclipY2, y2-edged)-y;
		if( w>0 && h>0 ) {
			r = (BYTE)(fill & 0x000000FF);
			g = (BYTE)((fill & 0x0000FF00) >> 8);
			b = (BYTE)((fill & 0x00FF0000) >> 16);
			pD = Tdpbits + TdstepY*y + TdstepX*x;
			if( AlphaBlendD ) {
				while( h-- ) {
					pDST = pD;
					for(ww=w; ww--; ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
						BlastPixelAlpha (r, g, b, pDST);
#else
						*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
						*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
						*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
						pDST += TdstepX;
					}
					pD += TdstepY;
				}
			} else {
				while( h-- ) {
					pDST = pD;
					for(ww=w; ww--; ) {
#if MAC_PORT // LBO 09/16/07 [RGBA]
						*(pDST+1) = r;
						*(pDST+2) = g;
						*(pDST+3) = b;
//						BlastPixel (r, g, b, pDST);
#else
						*pDST = b;
						*(pDST+1) = g;
						*(pDST+2) = r;
#endif
						pDST += TdstepX;
					}
					pD += TdstepY;
				}
			}
		}
	}
	if( edged ) {
		--x2;
		--y2;
		LineBMB(bmDO, x1, y1, x2, y1, edge);
		LineBMB(bmDO, x2, y1, x2, y2, edge);
		LineBMB(bmDO, x2, y2, x1, y2, edge);
		LineBMB(bmDO, x1, y2, x1, y1, edge);
		++x2;
		++y2;
	}
    if( bmD==&kBM[0] ) {
        InvalidateScreen(x1, y1, x2, y2);
    }
}

//**************************************************************************
// Fills a rectangle with a gradient hue, over a vertical range of lightness
// and a horizontal range of saturation.
void RectGradientBMB(PBMB bmD, KLONG xo, KLONG yo, KLONG wo, KLONG ho, BYTE hue, BYTE satmin, BYTE satmax, BYTE lightmin, BYTE lightmax) {
	
	BYTE	*pDST, *pD, r, g, b;
	KLONG		x, y, w, h, cy, cx, dsat, dlight, hs, ws;
	PBMB	bmDO;
	double	li, sa, hu;
	double	f, p, q, t;
	double	i;
	
	bmDO = bmD;
	SetupBMD(&bmD);
	if( ho < 0 ) {ho = -ho; hs = 1;}
	else hs = 0;
	if( wo < 0 ) {wo = -wo; ws = 1;}
	else ws = 0;
	//	x = Kmax(TdclipX1, xo);
	//	y = Kmax(TdclipY1, yo);
	//	w = Kmin(TdclipX2, xo+wo)-x;
	//	h = Kmin(TdclipY2, yo+ho)-y;
	x = xo;
	y = yo;
	w = wo;
	h = ho;
	if( w>0 && h>0 ) {
		pD = Tdpbits + TdstepY*y + TdstepX*x;
		dsat = satmax-satmin;
		dlight = lightmax-lightmin;
		if( AlphaBlendD ) {
			for(cy=0; cy<h; cy++) {
				pDST = pD;
				li = (double)(hs?(h-cy-1):cy)/(double)(h-1);
				for(cx=0; cx<w; cx++) {
					if( xo+cx>=TdclipX1 && xo+cx<TdclipX2 && yo+cy>=TdclipY1 && yo+cy<TdclipY2 ) {
						sa = (double)(ws?(w-cx-1):cx)/(double)(w-1);
						hu = (double)(255-hue)*((double)360.0/(double)255.0);
						if( sa==0 ) r = g = b = (BYTE)(255.0*li);
						else {
							hu /= 60.0;
							i = floor(hu);
							f = hu-i;
							p = li*(1.0-sa);
							q = li*(1.0-sa*f);
							t = li*(1.0-sa*(1.0-f));
							switch( (KLONG)i ) {
								case 0:
									r = (BYTE)(255.0*li);
									b = (BYTE)(255.0*t);
									g = (BYTE)(255.0*p);
									break;
								case 1:
									r = (BYTE)(255.0*q);
									b = (BYTE)(255.0*li);
									g = (BYTE)(255.0*p);
									break;
								case 2:
									r = (BYTE)(255.0*p);
									b = (BYTE)(255.0*li);
									g = (BYTE)(255.0*t);
									break;
								case 3:
									r = (BYTE)(255.0*p);
									b = (BYTE)(255.0*q);
									g = (BYTE)(255.0*li);
									break;
								case 4:
									r = (BYTE)(255.0*t);
									b = (BYTE)(255.0*p);
									g = (BYTE)(255.0*li);
									break;
								default:
									r = (BYTE)(255.0*li);
									b = (BYTE)(255.0*p);
									g = (BYTE)(255.0*q);
									break;
							}
						}
#if MAC_PORT // LBO 09/16/07 [RGBA]
						BlastPixelAlpha (r, g, b, pDST);
#else
						*pDST = (BYTE)(((WORD)(*pDST) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);
						*(pDST+1) = (BYTE)(((WORD)(*(pDST+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);
						*(pDST+2) = (BYTE)(((WORD)(*(pDST+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);
#endif
					}
					pDST += TdstepX;
				}
				pD += TdstepY;
			}
		} else {
			for(cy=0; cy<h; cy++) {
				pDST = pD;
				li = (double)(hs?(h-cy-1):cy)/(double)(h-1);
				for(cx=0; cx<w; cx++) {
					if( xo+cx>=TdclipX1 && xo+cx<TdclipX2 && yo+cy>=TdclipY1 && yo+cy<TdclipY2 ) {
						sa = (double)(ws?(w-cx-1):cx)/(double)(w-1);
						hu = (double)(255-hue)*((double)360.0/(double)256.0);
						if( sa==0 ) r = g = b = (BYTE)(255.0*li);
						else {
							hu /= 60.0;
							i = floor(hu);
							f = hu-i;
							p = li*(1.0-sa);
							q = li*(1.0-sa*f);
							t = li*(1.0-sa*(1.0-f));
							switch( (KLONG)i ) {
								case 0:
									r = (BYTE)(255.0*li);
									b = (BYTE)(255.0*t);
									g = (BYTE)(255.0*p);
									break;
								case 1:
									r = (BYTE)(255.0*q);
									b = (BYTE)(255.0*li);
									g = (BYTE)(255.0*p);
									break;
								case 2:
									r = (BYTE)(255.0*p);
									b = (BYTE)(255.0*li);
									g = (BYTE)(255.0*t);
									break;
								case 3:
									r = (BYTE)(255.0*p);
									b = (BYTE)(255.0*q);
									g = (BYTE)(255.0*li);
									break;
								case 4:
									r = (BYTE)(255.0*t);
									b = (BYTE)(255.0*p);
									g = (BYTE)(255.0*li);
									break;
								default:
									r = (BYTE)(255.0*li);
									b = (BYTE)(255.0*p);
									g = (BYTE)(255.0*q);
									break;
							}
						}
#if MAC_PORT // LBO 09/16/07 [RGBA]
						*(pDST+1) = r;
						*(pDST+2) = g;
						*(pDST+3) = b;
//						BlastPixel (r, g, b, pDST);
#else
						*pDST = b;
						*(pDST+1) = g;
						*(pDST+2) = r;
#endif
					}
					pDST += TdstepX;
				}
				pD += TdstepY;
			}
		}
	}
    if( bmD==&kBM[0] ) {
        InvalidateScreen(xo, yo, xo+wo, yo+ho);
    }
}

//**************************************************************************
// Draw an oval in the destination bitmap.  The oval may be
// SOLID, OUTLINED, or BOTH, depending upon the values of 'fill' and 'edge'.
// if 'fill' is not -1, fill the rectangle with that color
// if 'edge' is not -1, outline the rectangle with that color

#define OvalPoint(pD, x, y, r, g, b) {\
if( x>=TdclipX1 && x<TdclipX2 && y>=TdclipY1 && y<TdclipY2 ) {\
if( AlphaBlendD ) {\
*pD = (BYTE)(((WORD)(*pD) * AlphaBlendD + (WORD)b * AlphaBlendS + 127)/255);\
*(pD+1) = (BYTE)(((WORD)(*(pD+1)) * AlphaBlendD + (WORD)g * AlphaBlendS + 127)/255);\
*(pD+2) = (BYTE)(((WORD)(*(pD+2)) * AlphaBlendD + (WORD)r * AlphaBlendS + 127)/255);\
} else {\
*pD = b;\
*(pD+1) = g;\
*(pD+2) = r;\
}\
}\
}

void OvalBMB(PBMB bmD, KLONG x1, KLONG y1, KLONG x2, KLONG y2, DWORD fill, DWORD edge) {
	
	BYTE	*pDSTnw, *pDSTne, *pDSTsw, *pDSTse, *pDfn, *pDfs, re, ge, be, rf, gf, bf;
	KLONG		x, y, dx, dy, a_mul, b_mul, tt, xw, xe, yn, ys, xf;
	KLONG		row, as, bs, as2, bs2, as4, bs4, dd, xo, yo, bs6, as6;
	PBMB	bmDO;
	DWORD	edge_color;
	
	edge_color = (edge==-1)?fill:edge;
	if( x1==x2 || y1==y2 ) {
		LineBMB(bmD, x1, y1, x2, y2, edge_color);
		return;
	}
	if( x1>x2 ) {
		x = x1;
		x1 = x2;
		x2 = x;
	}
	if( y1>y2 ) {
		y = y1;
		y1 = y2;
		y2 = y;
	}
	bmDO = bmD;
	SetupBMD(&bmD);
	
	dx = x2-x1;
	dy = y2-y1;
	xo = (dx+1) & 1;		// whether to add an extra line down center of ellipse
	yo = (dy+1) & 1;		// whether to add an extra line across center of ellipse
	x  = x1+(dx-xo)/2;				// center of ellipse
	y  = y1+(dy-yo)/2;
	as4 = (dx-1-xo)*(dx-1-xo);	// 4*A^2
	as2 = as4/2;				// 2*A^2
	as  = as2/2;				// A^2
	bs4 = (dy-1-yo)*(dy-1-yo);	// 4*B^2
	bs2 = bs4/2;				// 2*B^2
	bs  = bs2/2;				// B^2
	bs6 = bs4+bs2;
	as6 = as4+as2;
	
	row = (dy-yo)/2;		// starting y-offset from center
	a_mul= as4*row;			// amount to adjust decision variable when step in Y
	b_mul= 0;				// amount to adjust decision variable when step in X
	dd  = as - as2*row + bs2;	// initial value of decision variable
	tt  = bs - as4*row + as2;
	
	rf = (BYTE)(fill & 0x000000FF);
	gf = (BYTE)((fill & 0x0000FF00) >> 8);
	bf = (BYTE)((fill & 0x00FF0000) >> 16);
	re = (BYTE)(edge_color & 0x000000FF);
	ge = (BYTE)((edge_color & 0x0000FF00) >> 8);
	be = (BYTE)((edge_color & 0x00FF0000) >> 16);
	
	xw = x;
	xe = x+xo;
	yn = y-row;
	ys = y+row+yo;
	pDSTnw = Tdpbits + TdstepY*yn + TdstepX*xw;
	pDSTne = Tdpbits + TdstepY*yn + TdstepX*xe;
	pDSTsw = Tdpbits + TdstepY*ys + TdstepX*xw;
	pDSTse = Tdpbits + TdstepY*ys + TdstepX*xe;
	
	OvalPoint(pDSTnw, xw, yn, re, ge, be);
	if( xw!=xe ) OvalPoint(pDSTne, xe, yn, re, ge, be);
	if( yn!=ys ) {
		OvalPoint(pDSTsw, xw, ys, re, ge, be);
		if( xw!=xe ) OvalPoint(pDSTse, xe, ys, re, ge, be);
	}
	do {
		if( dd<0 ) {
			dd += b_mul + bs6;
			tt += b_mul + bs4;
			b_mul += bs4;
			pDSTnw -= TdstepX;
			pDSTne += TdstepX;
			pDSTsw -= TdstepX;
			pDSTse += TdstepX;
			--xw;
			++xe;
		} else if( tt<0 ) {
			dd += b_mul + bs6 - a_mul + as4;
			tt += b_mul + bs4 - a_mul + as6;
			a_mul -= as4;
			b_mul += bs4;
			--row;
			++yn;
			--ys;
			xf = xw;
			--xw;
			++xe;
			pDSTnw += TdstepY-TdstepX;
			pDSTne += TdstepY+TdstepX;
			pDSTsw -= TdstepY+TdstepX;
			pDSTse -= TdstepY-TdstepX;
			
			pDfn = pDSTnw+TdstepX;
			pDfs = pDSTsw+TdstepX;
			if( fill!=-1 ) {
				while( pDfn!=pDSTne ) {
					OvalPoint(pDfn, xf, yn, rf, gf, bf);
					if( yn!=ys ) OvalPoint(pDfs, xf, ys, rf, gf, bf);
					pDfn += TdstepX;
					pDfs += TdstepX;
					++xf;
				}
			}
		} else {
			dd -= a_mul-as4;
			tt -= a_mul-as6;
			a_mul -= as4;
			--row;
			++yn;
			--ys;
			xf = xw+1;
			pDSTnw += TdstepY;
			pDSTne += TdstepY;
			pDSTsw -= TdstepY;
			pDSTse -= TdstepY;
			
			pDfn = pDSTnw+TdstepX;
			pDfs = pDSTsw+TdstepX;
			if( fill!=-1 && ys>=yn ) {
				while( xf<xe ) {
					OvalPoint(pDfn, xf, yn, rf, gf, bf);
					if( yn!=ys ) OvalPoint(pDfs, xf, ys, rf, gf, bf);
					pDfn += TdstepX;
					pDfs += TdstepX;
					++xf;
				}
			}
		}
		if( ys>=yn ) {	// never write the same dot twice in case AlphaBlend is active
			OvalPoint(pDSTnw, xw, yn, re, ge, be);
			if( xw!=xe) OvalPoint(pDSTne, xe, yn, re, ge, be);
			if( ys>yn ) {
				OvalPoint(pDSTsw, xw, ys, re, ge, be);
				if( xw!=xe ) OvalPoint(pDSTse, xe, ys, re, ge, be);
			}
		}
	} while( ys>yn );
	
    if( bmD==&kBM[0] ) {
        InvalidateScreen(x1, y1, x2+1, y2+1);
    }
}


//******************************************************************
DWORD StartTimer(DWORD timval) {
	
	// create periodic timer
	//	if( TimerID==0 ) TimerID = SetTimer(hWndMain, 1, (UINT)timval, 0);
	KLONG		i;
    
	for(i=0; i<timersCNT; i++) if( Timers[i]==NULL ) break;
	if( i==MAXTIMERS ) return -1;
	if( i>=timersCNT ) ++timersCNT;
	Timers[i] = [NSTimer scheduledTimerWithTimeInterval:((CGFloat)timval)/1000.0 target:pKV selector: @selector(OnTimer:) userInfo:nil repeats:YES];
	return i;
}

//******************************************************************
void StopTimer(DWORD t) {
	
	//	if( TimerID!=0 ) KillTimer(hWndMain, TimerID);
	//	TimerID = 0;
	if( t<timersCNT && Timers[t]!=NULL ) {
		[Timers[t] invalidate];
		Timers[t] = NULL;
	}
}

char SMbuf[1024];
DWORD smFLAG;
//********************************************************************
// Display the NULL-terminated message in 'pMem' using OS resources.
void SystemMessage(char *pMem) {
	
    if( smFLAG ) NSLog(@"smFLAG");
	if( strlen(pMem)>sizeof(SMbuf)-1 ) strncpy(SMbuf, pMem, sizeof(SMbuf)-1);
	else strcpy(SMbuf, pMem);
	
    if( NSThread.currentThread.isMainThread ) {
        smFLAG = false;
        NSAlert    *alert = [[NSAlert alloc] init];
        [alert addButtonWithTitle: @"Sigh"];
        [alert setMessageText:CFBridgingRelease(CFStringCreateWithCString(NULL, SMbuf, kCFStringEncodingMacRoman))];
        [alert setAlertStyle: NSAlertStyleWarning];
        [alert runModal];
        smFLAG = false;
    } else {
        smFLAG = true;
        [[NSOperationQueue mainQueue] addOperationWithBlock:^{
            NSAlert	*alert = [[NSAlert alloc] init];
            [alert addButtonWithTitle: @"Sigh"];
            [alert setMessageText:CFBridgingRelease(CFStringCreateWithCString(NULL, SMbuf, kCFStringEncodingMacRoman))];
            [alert setAlertStyle: NSAlertStyleWarning];
            [alert runModal];
            smFLAG = false;
        }];
    }
    while( smFLAG )
        ;
}

//*********************************************************************
// store the appropriate End-Of-Line sequence into buffer 'pMem' and
// return the number of bytes written to 'pMem'.
KLONG GetSystemEOL(BYTE *pMem) {
	
	*pMem++ = '\n';
	return 1;
}

//*********************************************************
// don't allow a program to access anything above the game directory,
// so don't allow absolute paths or .. or // or such.
char *GetFileName(char *fname, char *xname) {
	/*
	 char	*p;
	 
	 if( fname[0]=='\\' ) return NULL; // fail miserably if they try absolute path, only allow SUB-dirs
	 strcpy(PartPath, fname);
	 for(p=PartPath; *p; p++) {
	 if( (*p=='.' && *(p+1)=='.')
	 || (*p=='/' && *(p+1)=='/')
	 || (*p==':')
	 ) return NULL;	// fail miserably if they try some tricky directory BS, only allow SUB-dirs
	 if( *p=='/' ) *p = '\\';	// DOS/Windows path conversion
	 }
	 strcat(PartPath, xname);	// stick desired file extension on end
	 return PartPath;
	 */
	return "";
}
//****************************************************************
// This function can return as much (well, up to 512 bytes including NUL)
// or as little information as you want regarding the current OS.  This
// information is ONLY used for display purposes within the Help, and
// does NOT affect the program execution in any other way.
void KGetOSversion(BYTE *pMem) {

	if( @available(macOS 10.10, *) ) {
		NSOperatingSystemVersion ver = [[NSProcessInfo processInfo] operatingSystemVersion];
		sprintf((char *)pMem, "KINT %d.%d - Mac OS X %d.%d.%d", (int)INTERP_MAJOR_REV, (int)INTERP_MINOR_REV, (int)ver.majorVersion, (int)ver.minorVersion, (int)ver.patchVersion);
			// as written, this should handle up to OS X 99.99.99
			//		gOSXVersion32 = ((majorVersion / 10) << 20) |
			//		((majorVersion % 10) << 16) |
			//		((minorVersion / 10) << 12) |
			//		((minorVersion % 10) << 8) |
			//		((bugFixVersion / 10) << 4) |
			//		((bugFixVersion % 10) << 0);
	}
}

//******************************************************************
void SetPhysicalScreenSize (void)
{
#if FAKE_SCREEN_SIZE
	int width = FAKE_SCREEN_SIZE_W;
	int height = FAKE_SCREEN_SIZE_H;
#else
	// Get the resolution of the main display
	CGRect mainRect = CGDisplayBounds (kCGDirectMainDisplay);
	
	int height = mainRect.size.height;
	int width = mainRect.size.width;
#endif
	
	PhysicalScreenSizeIndex = 0;
	PhysicalScreenW = width;
	PhysicalScreenH = height;
	
	// Try to find the largest windowSize that fits on this display
	for(PhysicalScreenSizeIndex=0; windowSizes[PhysicalScreenSizeIndex].width <= kBM[0].logW && windowSizes[PhysicalScreenSizeIndex].height <= kBM[0].logH; PhysicalScreenSizeIndex++) ;
	if( PhysicalScreenSizeIndex ) --PhysicalScreenSizeIndex;
}

//****************************************
void InitScreen(long minW, long minH)
{
	int			redraw;
	WORD		o, ow, oh;
	
	if( minW==-1 ) {
		ow = kBM[0].logW;
		oh = kBM[0].logH;
		minW = minH = 0;
	} else {
		ow = (WORD)minW;
		oh = (WORD)minH;
	}
	
	if( kBM[0].bitmapContext!=NULL || kBM[0].cgImage!=NULL ) {
		DeleteBMB(0);
		redraw = 1;
	} else redraw = 0;
	
	SetPhysicalScreenSize();
	
	if( PhysicalScreenW > minW ) minW = PhysicalScreenW;
	if( PhysicalScreenH > minH ) minH = PhysicalScreenH;
	
	CreateBM0();
	
	// initialize the ENTIRE screen bitmap to black
	o = kBM[0].orient;
	kBM[0].orient = BMO_NORMAL;
	RectBMB(&kBM[0], 0, 0, kBM[0].physW, kBM[0].physH, 0x00000000, (DWORD)-1);
	kBM[0].orient = o;
	
	kBM[0].logW = ow;
	kBM[0].logH = oh;
	
	if( kBM[0].logW == 0 ) {
		kBM[0].logW = PhysicalScreenW;
		kBM[0].logH = PhysicalScreenH;
	}
	
//	if( redraw && KEloaded ) {
	if( redraw && KEloaded>1 ) {
		// call the SCREEN_CHANGE code in the KE module
//		NSLog(@"sending SCREEN_CHANGE in InitScreen()");
		AddMsg(KHDR_SCREEN_CHANGE, 0, 0, 0);
		SendNextMsg();
	}
}


//************************************************************
//* Read in .KE, allocate memory for it, initialize ptrs
//*		Return TRUE if successful, FALSE if not.
//*
KLONG	LoadKEFile() {
	
	DWORD	fh, readcnt;
	BYTE	*pB;
	DWORD	remains, zerosize;
	KE		*pKE;
	
	// Open .KE file
	if( -1==(fh=Kopen("main.ke", O_RDBIN, DIR_GAME)) ) {
		SystemMessage(".KE file not found");
		return FALSE;
	}
	// Read .KE header
	if( 20!=(readcnt=Kread(fh, (BYTE *)TmpFname, 20)) ) {
		SystemMessage("Error reading .KE file");
		Kclose(fh);
		return FALSE;
	}
	if( strcmp((char *)TmpFname, "KASMB") ) {
		SystemMessage("Not a .KE file");
		Kclose(fh);
		return FALSE;
	}
#if BETA
	if( !((BYTE)(TmpFname[KHDR_MAJOR_INTERP_REV]) & 0x80) ||  (BYTE)(TmpFname[KHDR_MAJOR_INTERP_REV])>INTERP_MAJOR_REV || ((BYTE)(TmpFname[KHDR_MAJOR_INTERP_REV])==INTERP_MAJOR_REV && (BYTE)(TmpFname[KHDR_MINOR_INTERP_REV])>INTERP_MINOR_REV) ) {
		//char xxx[128];
		//sprintf(xxx, "KE:%u.%u EXE:%ud.%ud", TmpFname[KHDR_MAJOR_INTERP_REV], TmpFname[KHDR_MINOR_INTERP_REV], INTERP_MAJOR_REV, INTERP_MINOR_REV);
		//SystemMessage(xxx);
		SystemMessage("This .KE file requires a newer KINT.EXE");
		Kclose(fh);
		return FALSE;
	}
#else
	if( ((BYTE)TmpFname[KHDR_MAJOR_INTERP_REV] & 0x80)
	   || ((BYTE)TmpFname[KHDR_MAJOR_INTERP_REV] > INTERP_MAJOR_REV)
	   || ((BYTE)TmpFname[KHDR_MAJOR_INTERP_REV]==INTERP_MAJOR_REV
		   && (BYTE)TmpFname[KHDR_MINOR_INTERP_REV]>INTERP_MINOR_REV) ) {
		   SystemMessage("This .KE file requires a newer KINT.EXE");
		   Kclose(fh);
		   return FALSE;
	   }
#endif
	
	// Get size of initialized code/data in file
	remains = GETMEMD(TmpFname+KHDR_CODE_SIZE);
	zerosize = GETMEMD(TmpFname+KHDR_ZERODATA_SIZE);
	pKE = &mainKE;
	// Save pointer to base of stack
	pKE->StackBase = pKE->reg4 = pKE->reg5 = remains + zerosize;
	// Get size of memory required (init code/data size + uninit data size + stack size)
	pKE->memlen = remains + zerosize + GETMEMD(TmpFname+KHDR_STACK_SIZE);
	// Allocate .KE memory
    if( NULL==(pKE->pKode=(BYTE *)malloc(pKE->memlen)) ) {
		SystemMessage("Error allocating .KE memory");
		Kclose(fh);
		return FALSE;
	}
    AllocPtrs[0] = pKE->pKode;
    AllocCnt = 1;
	// Copy first 20 bytes that we already read, and setup ptr for rest of reading
	memcpy(pKE->pKode, TmpFname, 20);
	pB = pKE->pKode + 20;
	remains -= 20;
	// Read rest of .KE file into memory
	readcnt = Kread(fh, pB, remains);
	Kclose(fh);
	if( readcnt!=remains ) return FALSE;
	// Initialize uninitialized area to 0.
	// Make sure that your memset takes at least a 4-byte value for "length".
	// Otherwise, you'll need to modify this next line to properly zero-out
	// all of the ZERODATA area, which may be longer than 64K bytes.
	memset(pKE->pKode+GETMEMD(TmpFname+KHDR_CODE_SIZE), 0, zerosize);
	
#if BETA
	Expired =  (*((DWORD *)(pKE->pKode+5*sizeof(DWORD)))!=BETAKEY);

	KGetTime(&betatime);
	if( betatime.y > BETA_YEAR || (betatime.y==BETA_YEAR && (betatime.mo > BETA_MONTH || (betatime.mo==BETA_MONTH && betatime.d >= BETA_DAY))) ) Expired = 1;
#endif

/*	if( NULL==(KINTmsgs=(DWORD *)malloc(MAXMSGS*16)) ) {
		SystemMessage("Failed to allocate MSG buf");
		exit(0);
	}
*/
	return TRUE;
}

//***************************************************************
//* This routine is called to send a "message" to a .KE module
//* entry point, with:
//* R32 = cmd (offset into pKode...see KHDR_ #defines in pmach.h)
//*	R15 = arg1
//*	R14 = arg2
//*	R13 = arg3
//*
bool inKE=false;


// DON'T CALL THIS DIRECTLY EXCEPT FROM ForceMsg, KINTnonTimer, and runKE in kintView.m!!!!!!!

#define LISTMSGS 0
#if LISTMSGS
char *msgStr[]={
	"SCREEN_CHANGE",
	"ALPHAKEY_DOWN",
	"ALPHAKEY_UP",
	"BEGIN_STARTUP",
	"END_STARTUP",
	"WINDOW_HIDDEN",
	"WINDOW_MOVE",
	"WINDOW_SIZE",
	"POINTER_MOVE",
	"KEY_DOWN",
	"KEY_UP",
	"CHAR",
	"TIMER",
	"IDLE",
	"MUSIC_DONE",
	"QUERY_QUIT",
	"SHUT_DOWN"
};
#endif
DWORD   CURcmd;

WORD KEmsg(KE *pKE, WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	WORD KEretval;
	
	if( inKE ) NSLog(@"inKE KEmsg");
	if( KEloaded<2 ) return 0;

    while( dirtyX1<dirtyX2 ) InvDirty();
    
	inKE = true;
CURcmd = cmd;
    SETREGD(REG_0, 0);//pKE->pKode);	// init code block base register
#if BETA
	if( !Expired || cmd==KHDR_BEGIN_STARTUP || cmd==KHDR_END_STARTUP ) {
		SETREGD(REG_3, GETMEMD(pKE->pKode+cmd));	// init PC
	}
#else
	SETREGD(REG_3, GETMEMD(pKE->pKode+cmd));	// init PC
#endif
	SETREGD(REG_4, pKE->reg4);	// init stack base frame ptr
	SETREGD(REG_5, pKE->reg5);	// init stack top ptr
	
	SETREGW(REG_32, cmd);		// set cmd
	SETREGD(REG_15, arg1);		//    and args into CPU regs for Kode
	SETREGD(REG_14, arg2);
	SETREGD(REG_13, arg3);

#if LISTMSGS
	{
		int	i;
		DWORD	hFile;
		
		i = (cmd-60)/4;
		if( i!=12 && i!=13 ) {
			if( -1==(hFile=Kopen("msgs.txt", O_WRBIN, DIR_GAME)) ) {
				hFile = Kopen("msgs.txt", O_WRBINNEW, DIR_GAME);
			}
			Kseek(hFile, 0, SEEK_END);
			Kwrite(hFile, msgStr[i], strlen(msgStr[i]));
			Kwrite(hFile, "\r\n", 2);
			Kclose(hFile);
		}
	}
#endif
	// call K interpreter
/*	{
		char	dBuf[128];
		sprintf(dBuf,"ExecKE(%d)", cmd);
		Dwrite(dBuf);
	}
*/
	if( KEloaded>1 ) KEretval = ExecKE(pKE);
//	if( KEloaded ) KEretval = ExecKE(pKE);
	else KEretval = 0;
	
	// save KE specific CPU registers
	pKE->reg4 = GETREGD(REG_4);
	pKE->reg5 = GETREGD(REG_5);
	
	// if Kode did a TERM, kill the program, but ONLY do it ONCE!
	// (The Kode may call TERM more than once due to "shut down" Kode calls)
	runKEmsg = 0;
	inKE = false;
	if( killpgm==1 ) {
		killpgm |= 2;
		KillProgram();
	}
	if( !killpgm && dirtyX1<dirtyX2 ) {
		InvDirty();
	}
	return KEretval;
}

//************************************************
//* The K interpreter
//*		Event Kode must put return value into R39 before HALTing
//*
//* NOTE: "result" contains the KCPU "flags" for equality,
//*		least-significant bit, and most-significant bit.
//*		The "result" flags are set by these instructions:
//*			SHL SHR ROL ROR SUB B2I "RET 0" and "RET 1"
//*			AND XOR CMP INC DEC NOT NEG TST FCMP
//*		When the operation is a WORD or BYTE, the result
//*		is first cast to a SHORT or a CHAR before being
//*		cast to a DWORD in order to make sure that the
//*		most-significant bit gets propagated correctly.
//*
//*		The instructions that are affected by the 'signedmath' flag are:
//*			SHR, DIV, MOD, SUB, CMP
//*		The f--- floating point opcodes are ALWAYS signed operations.

//DWORD opcnts[256];
//DWORD minstack=64*1024;
//DWORD entryr7, entryr3;

WORD ExecKE(KE *pKE) {
	
    double  d1;
	WORD	*PC, opcode;
	BYTE	*pMem=NULL, *pKode;
	DWORD	dw, dw2 = 0, dw3, dw4;
    QWORD   qw, qw2=0;
	WORD	w, w2=0, dr;
    BYTE	b=0;//1245+1240;
	BYTE	rev;
	float	f1;
    KFf4    kf;

    //entryr7 = GETREGD(REG_7);
	//entryr3 = GETREGD(REG_3);
	
	pKode = pKE->pKode;
 	PC = (WORD *)(pKode + GETREGD(REG_3));	// get local ptr to PC code
	
	for( ;; ) {
		
#ifdef GEEK_TRACE3
		TraceBuf3[TBptr3++] = (DWORD)((BYTE *)PC-pKode);
		if( TBptr3>=TBMAX3 ) TBptr3 = 0;
/*			if( *((DWORD *)(pKode+0x70)) != 0x0276 ) {
			pKode = pKode;
		}
*/
#endif
		opcode = GETMEMW(PC);
		
#ifdef CHECKWORDALIGN
		savePC = (BYTE *)PC-pKode;
#endif
/*
        dw = (DWORD)((BYTE*)PC - (BYTE*)pKode);
        if( dw==0x49a12 || dw==0x499DE || dw==0x499F6) {
            dw2 = 0;
        }
*/
		PC++;
		switch( (opcode>>8) & 0x00FF ) {
			case 0x00:	// unused
			case 0x01:	// unused
			case 0x02:	// unused
				break;
			case 0x03:	// SHL b,b
				dr = opcode & 0x00FF;
				result = (DWORD)(SETREGB(dr, GETREGB(dr)<<(BYTE)(GETREGB(GETMEMW(PC)) & 0x0007)));
				PC++;
				result |= (result<<24);
				break;
			case 0x04:	// SHL w,b
				dr = opcode & 0x00FF;
				result = (DWORD)(SETREGW(dr, GETREGW(dr)<<(BYTE)(GETREGB(GETMEMW(PC)) & 0x000F)));
				PC++;
				result |= (result<<16);
				break;
			case 0x05:	// SHL d,b
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr)<<(BYTE)(GETREGB(GETMEMW(PC)) & 0x001F));
				PC++;
				break;
			case 0x06:	// SHR b,b
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)(char)(SETREGB(dr, (char)(GETREGB(dr))>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x07)));
				else result = (DWORD)(char)(SETREGB(dr, GETREGB(dr)>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x07)));
				PC++;
				result |= (result<<24);
				break;
			case 0x07:	// SHR w,b
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)(KSHORT)(SETREGW(dr, (KSHORT)(GETREGW(dr))>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x000F)));
				else result = (DWORD)(KSHORT)(SETREGW(dr, GETREGW(dr)>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x000F)));
				PC++;
				result |= (result<<16);
				break;
			case 0x08:	// SHR d,b
				dr = opcode & 0x00FF;
				if( signedmath ) result = SETREGD(dr, (KLONG)(GETREGD(dr))>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x001F));
				else result = SETREGD(dr, GETREGD(dr)>>(BYTE)(GETREGB(GETMEMW(PC)) & 0x001F));
				PC++;
				break;
			case 0x09: // ROL b,b
				dr = opcode & 0x00FF;
				w = (WORD)(GETREGB(GETMEMW(PC)) & 0x07);
				PC++;
				b = GETREGB(dr);
				while( w-- ) {
					if( b & 0x80 ) b = (b << 1) | 1;
					else b <<= 1;
				}
				result = (DWORD)(char)SETREGB(dr, b);
				result |= (result<<24);
				break;
			case 0x0A:	// ROL w,b
				dr = opcode & 0x00FF;
				b = GETREGB(GETMEMW(PC)) & 0x1F;
				PC++;
				w = GETREGW(dr);
				while( b-- ) {
					if( w & 0x8000 ) w = (w<<1) | 1;
					else w <<= 1;
				}
				result = (DWORD)(KSHORT)SETREGW(dr, w);
// changed at 4.9 from <<24 to <<16 so MSbit gets set
				result |= (result<<16);
				break;
			case 0x0B:	// ROL d,b
				dr = opcode & 0x00FF;
				dw = GETREGD(dr);
				w = GETREGB(GETMEMW(PC)) & 0x1F;
				PC++;
				while( w-- ) {
					dw2 = dw;
					dw <<= 1;
					if( dw2 & 0x80000000 ) dw |= 1;
				}
				result = SETREGD(dr, dw);
				break;
			case 0x0C:	// ROR b,b
				dr = opcode & 0x00FF;
				w = (WORD)(GETREGB(GETMEMW(PC)) & 0x07);
				PC++;
				b = GETREGB(dr);
				while( w-- ) {
					if( b & 1 ) b = (b >> 1) | 0x80;
					else b >>= 1;
				}
				result = (DWORD)(char)SETREGB(dr, b);
				result |= (result<<24);
				break;
			case 0x0D:	// ROR w,b
				dr = opcode & 0x00FF;
				b = GETREGB(GETMEMW(PC)) & 0x001F;
				PC++;
				w = GETREGW(dr);
				while( b-- ) {
					if( w & 1 ) w = (w >> 1) | 0x8000;
					else w >>= 1;
				}
				result = (DWORD)(KSHORT)SETREGW(dr, w);
				result |= (result<<16);
				break;
			case 0x0E:	// ROR d,b
				dr = opcode & 0x00FF;
				dw = GETREGD(dr);
				w = GETREGB(GETMEMW(PC)) & 0x1F;
				PC++;
				while( w-- ) {
					dw2 = dw;
					dw >>= 1;
					if( dw2 & 1 ) dw |= 0x80000000;
				}
				result = SETREGD(dr, dw);
				break;
			case 0x0F:	// ADD b,b
				dr = opcode & 0x00FF;
				SETREGB(dr, GETREGB(dr)+GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x10:	// ADD w,w
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr)+GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x11:	// ADD d,d
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr)+GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x12:	// SUB b,b
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)((char)GETREGB(dr)))-((KLONG)((char)GETREGB(GETMEMW(PC))))));
				else result = (DWORD)GETREGB(dr)-(DWORD)GETREGB(GETMEMW(PC));
				PC++;
				SETREGB(dr, (BYTE)result);
				break;
			case 0x13:	// SUB w,w
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)((KSHORT)GETREGW(dr)))-((KLONG)((KSHORT)GETREGW(GETMEMW(PC))))));
				else result = (DWORD)GETREGW(dr)-(DWORD)GETREGW(GETMEMW(PC));
				PC++;
				SETREGW(dr, (WORD)result);
				break;
			case 0x14:	// SUB d,d
				dr = opcode & 0x00FF;
				if( signedmath ) result = SETREGD(dr, (KLONG)(GETREGD(dr)) - (KLONG)GETREGD(GETMEMW(PC)));
				else result = SETREGD(dr, GETREGD(dr) - GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x15:	// MUL b,b
				dr = opcode & 0x00FF;
				SETREGB(dr, GETREGB(dr)*GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x16:	// MUL w,w
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr)*GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x17:	// MUL d,d
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr)*GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x18:	// DIV b,b
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))/(char)GETREGB(GETMEMW(PC)));
				else SETREGB(dr, GETREGB(dr)/GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x19:	// DIV w,w
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGW(dr, (KSHORT)(GETREGW(dr))/(KSHORT)GETREGW(GETMEMW(PC)));
				else SETREGW(dr, GETREGW(dr)/GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x1A:	// DIV d,d
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGD(dr, (KLONG)(GETREGD(dr))/(KLONG)(GETREGD(GETMEMW(PC))));
				else SETREGD(dr, GETREGD(dr)/GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x1B:	// MOD b,b
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))%(char)GETREGB(GETMEMW(PC)));
				else SETREGB(dr, GETREGB(dr)%GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x1C:	// MOD w,w
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGW(dr, (KSHORT)(GETREGW(dr))%(KSHORT)GETREGW(GETMEMW(PC)));
				else SETREGW(dr, GETREGW(dr)%GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x1D:	// MOD d,d
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGD(dr, (KLONG)(GETREGD(dr))%(KLONG)(GETREGD(GETMEMW(PC))));
				else SETREGD(dr, GETREGD(dr)%GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x1E:	// OR b,b
				dr = opcode & 0x00FF;
				SETREGB(dr, GETREGB(dr) | GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x1F:	// OR w,w
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr) | GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x20:	// OR d,d
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr) | GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x21:	// AND b,b
				dr = opcode & 0x00FF;
				result = (DWORD)((KLONG)(char)GETREGB(dr) & (KLONG)(char)GETREGB(GETMEMW(PC)));
				PC++;
				SETREGB(dr, (BYTE)result);
				break;
			case 0x22:	// AND w,w
				dr = opcode & 0x00FF;
				result = (DWORD)((KLONG)(KSHORT)GETREGW(dr) & (KLONG)(KSHORT)GETREGW(GETMEMW(PC)));
				PC++;
				SETREGW(dr, (WORD)result);
				break;
			case 0x23:	// AND d,d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) & GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x24:	// XOR b,b
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr) ^ GETREGB(GETMEMW(PC)));
				PC++;
				result |= (result<<24);
				break;
			case 0x25:	// XOR w,w
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, GETREGW(dr) ^ GETREGW(GETMEMW(PC)));
				PC++;
				result |= (result<<16);
				break;
			case 0x26:	// XOR d,d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) ^ GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x27:	// CMP b,b
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)(char)GETREGB(opcode & 0x00FF))-((KLONG)(char)(GETREGB(GETMEMW(PC))))));
				else result = (DWORD)((DWORD)((KLONG)((KSHORT)((char)((BYTE)GETREGB(opcode & 0x00FF)))))-(DWORD)((KLONG)((KSHORT)((char)((BYTE)GETREGB(GETMEMW(PC)))))));
				PC++;
				break;
			case 0x28:	// CMP w,w
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)(KSHORT)GETREGW(opcode & 0x00FF))-((KLONG)(KSHORT)(GETREGW(GETMEMW(PC))))));
				else result = (DWORD)GETREGW(opcode & 0x00FF)-(DWORD)GETREGW(GETMEMW(PC));
				PC++;
				break;
			case 0x29:	// CMP d,d
				if( signedmath ) result = (KLONG)(GETREGD(opcode & 0x00FF)) - (KLONG)(GETREGD(GETMEMW(PC)));
				else result = (DWORD)GETREGD(opcode & 0x00FF) - (DWORD)GETREGD(GETMEMW(PC));
				PC++;
				break;
			case 0x2A:	// TST b,b
				result = (DWORD)(char)(GETREGB(opcode & 0x00FF) & GETREGB(GETMEMW(PC)));
				PC++;
				result |= (result<<24);
				break;
			case 0x2B:	// TST w,w
				result = (DWORD)(KSHORT)(GETREGW(opcode & 0x00FF) & GETREGW(GETMEMW(PC)));
				PC++;
				result |= (result<<16);
				break;
			case 0x2C:	// TST d,d
				result = GETREGD(opcode & 0x00FF) & GETREGD(GETMEMW(PC));
				PC++;
				break;
			case 0x2D:	// XCH b,b
				dr = opcode & 0x00FF;
				b = GETREGB(dr);
				w = GETMEMW(PC);
				PC++;
				SETREGB(dr, GETREGB(w));
				SETREGB(w, b);
				break;
			case 0x2E:	// XCH w,w
				dr = opcode & 0x00FF;
				w2 = GETREGW(dr);
				w = GETMEMW(PC);
				PC++;
				SETREGW(dr, GETREGW(w));
				SETREGW(w, w2);
				break;
			case 0x2F:	// XCH d,d
				dr = opcode & 0x00FF;
				dw = GETREGD(dr);
				w = GETMEMW(PC);
				PC++;
				SETREGD(dr, GETREGD(w));
				SETREGD(w, dw);
				break;
			case 0x30:	// LDR b,b
				SETREGB(opcode & 0x00FF, GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x31:	// LDR w,w
				SETREGW(opcode & 0x00FF, GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x32:	// LDR d,d
				SETREGD(opcode & 0x00FF, GETREGD(GETMEMW(PC)));
				PC++;
				break;
			case 0x33:	// LDE d,b
				SETREGD(opcode & 0x00FF, (DWORD)(BYTE)GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x34:	// LDE d,w
				SETREGD(opcode & 0x00FF, (DWORD)(WORD)GETREGW(GETMEMW(PC)));
				PC++;
				break;
			case 0x35:	// LDE w,b
				SETREGW(opcode & 0x00FF, (WORD)GETREGB(GETMEMW(PC)));
				PC++;
				break;
			case 0x36:	// ABIT w,b
				dr = opcode & 0x00FF;
				if( dr<=REG_96 ) {
					w2 = GETMEMW(PC);
					PC++;
					w  = GETREGW(dr);
					for(; w; w>>=1, ++w2) if( w & 1 ) SETREGB(w2, 1+GETREGB(w2));
				}
				break;
			case 0x37:	// ABIT b,b
				w2 = GETMEMW(PC);
				PC++;
				b  = GETREGB(opcode & 0x00FF);
				for(; b; b>>=1, ++w2) if( b & 1 ) SETREGB(w2, 1+GETREGB(w2));
				break;
			case 0x38:	// LD0 b,d
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x39:	// LD0 w,d
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)))));
				PC++;
				break;
			case 0x3A:	// LD0 d,d
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x3B:	// LD1 b,d
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x3C:	// LD1 w,d
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x3D:	// LD1 d,d
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x3E:	// LD2 b,d
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x3F:	// LD2 w,d
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x40:	// LD2 d,d
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x41:	// LDS b,d
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x42:	// LDS w,d
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC)))));
				PC++;
				break;
			case 0x43:	// LDS d,d
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC))));
				PC++;
				break;
			case 0x44:	// ST0 b,d
				SETMEMB(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGB(opcode & 0x00FF));
				PC++;
				break;
			case 0x45:	// ST0 w,d
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))), GETREGW(opcode & 0x00FF));
				PC++;
				break;
			case 0x46:	// ST0 d,d
				SETMEMD(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGD(opcode & 0x00FF));
				PC++;
				break;
			case 0x47:	// ST1 b,d
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGB(opcode & 0x00FF));
				PC++;
				break;
			case 0x48:	// ST1 w,d
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGW(opcode & 0x00FF));
				PC++;
				break;
			case 0x49:	// ST1 d,d
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGD(opcode & 0x00FF));
				PC++;
				break;
			case 0x4A:	// ST2 b,d
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGB(opcode & 0x00FF));
				PC++;
				break;
			case 0x4B:	// ST2 w,d
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGW(opcode & 0x00FF));
				PC++;
				break;
			case 0x4C:	// ST2 d,d
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGD(opcode & 0x00FF));
				PC++;
				break;
			case 0x4D:	// STS b,d
				SETMEMB(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGB(opcode & 0x00FF));
				PC++;
				break;
			case 0x4E:	// STS w,d
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC))), GETREGW(opcode & 0x00FF));
				PC++;
				break;
			case 0x4F:	// STS d,d
				SETMEMD(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(GETMEMW(PC)), GETREGD(opcode & 0x00FF));
				PC++;
				break;
			case 0x50:	// PU0 b,d
				w = GETMEMW(PC);
				PC++;
				SETMEMB(pKode + (long)(KLONG)GETREGD(w), GETREGB(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+1);
				break;
			case 0x51:	// PU0 w,d
				w = GETMEMW(PC);
				PC++;
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w)), GETREGW(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+2);
				break;
			case 0x52:	// PU0 d,d
				w = GETMEMW(PC);
				PC++;
				SETMEMD(pKode + (long)(KLONG)GETREGD(w), GETREGD(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+4);
				break;
			case 0x53:	// PO0 b,d
				w = GETMEMW(PC);
				PC++;
				SETREGD(w, GETREGD(w)-1);
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(w)));
				break;
			case 0x54:	// PO0 w,d
				w = GETMEMW(PC);
				PC++;
				SETREGD(w, GETREGD(w)-2);
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w))));
				break;
			case 0x55:	// PO0 d,d
				w = GETMEMW(PC);
				PC++;
				SETREGD(w, GETREGD(w)-4);
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(w)));
				break;
			case 0x56:	// memory ops
			case 0x57:	// memory ops
				/*
				 56/57op
				 0101011cciiiaaaa
				 cc = chunk size (0=byte, 1=word, 2=dword)
				 iii = instruction:
				 000	LDM
				 001 ADD
				 010 SUB
				 011 CMP
				 100 TST
				 101 AND
				 110 OR
				 111 XOR
				 aaaa = addressing mode
				 WORD1	WORD2	WORD3	WORD4	WORD5	WORD6	WORD7	WORD8		regs	meof	imm
				 0000	regl,[regh]			56/57op reg/reg													2		0		0
				 0001	regl,[regh+off]		56/57op reg/reg	off----------									2		1		0
				 0010	regl,[mem]			56/57op reg		mem----------									1		1		0
				 0011	regl,[regh+mem+off]	56/57op reg/reg	mem----------	off----------					2		2		0
				 0100	regl,[regh+reg2]	56/57op reg/reg	reg												3		0		0
				 0101	[regl],regh			56/57op reg/reg													2		0		0
				 0110	[regl],imm			56/57op reg		imm-- ???????									1		0		1
				 0111	[regl+off],regh		56/57op reg/reg	off----------									2		0		1
				 1000	[regl+off],imm		56/57op reg		off----------	imm-- ???????					1		1		1
				 1001	[mem],regh			56/57op reg		mem----------									1		1		0
				 1010	[mem],imm			56/57op mem----------	imm-- ???????							0		1		1
				 1011	[regl+mem+off],regh	56/57op reg/reg	mem----------	off----------					2		2		0
				 1100	[regl+mem+off],imm	56/57op reg		mem----------	off----------	imm-- ???????	1		2		1
				 1101	[regl+regh],reg2	56/57op reg/reg	reg												3		0		0
				 1110	[regl+regh],imm		56/57op reg/reg imm-- ???????									2		0		1
				 1111	LEA/PEA
				 */
				w = GETMEMW(PC);		// in all but two cases, there's a reg/reg word following the opcode, so get it
				PC++;					// we'll skip back in the two cases where reg/reg isn't there
				dr = (opcode & 0x0180)>>7;	// get chunk size (0=byte, 1=word, 2=dword)
				switch( opcode & 0x000F ) {
					case 0:	//regl,[regh]			56/57op reg/reg
						if( !dr ) b = GETMEMB(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF));
						else if( dr==1 ) w2 = GETMEMW(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF));
                        else if( dr==2 ) dw2 = GETMEMD(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF));
                        else qw2 = GETMEMQ(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF));
						w &= 0x00FF;
						break;
					case 1:	//regl,[regh+off]		56/57op reg/reg	off----------
						if( !dr ) b = GETMEMB(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
						else if( dr==1 ) w2 = GETMEMW(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
                        else if( dr==2 ) dw2 = GETMEMD(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
                        else qw2 = GETMEMQ(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
/*						else {
							DWORD	dbgW, dbgR28, dbgPC, dbgMEM, dbgADR;
							
							dbgW = w>>8;
							dbgR28 = GETREGD(dbgW);
							dbgPC = (DWORD)((BYTE*)PC-pKode);
							dbgMEM = GETMEMD(PC);
							dbgADR = dbgR28+dbgMEM;
							
							dw2 = (DWORD)((BYTE*)PC-pKode);
							if( dw2==0x4eef8 ) {
								dw2 = GETREGD((w>>8));
								dw2 = GETMEMD(PC);
							}
							dw2 = GETMEMD(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
						}
*/
                        PC += 2;
						w &= 0x00FF;
						break;
					case 2:	//regl,[mem]			56/57op reg		mem----------
						if( !dr ) b = GETMEMB(pKode + (long)(KLONG)GETMEMD(PC));
						else if( dr==1 ) w2 = GETMEMW(pKode + (long)(KLONG)GETMEMD(PC));
                        else if( dr==2 ) dw2 = GETMEMD(pKode + (long)(KLONG)GETMEMD(PC));
                        else qw2 = GETMEMQ(pKode + (long)(KLONG)GETMEMD(PC));
						PC += 2;
						w &= 0x00FF;
						break;
					case 3:	//regl,[mem+regh+off]	56/57op reg/reg	mem----------	off----------
						dw = GETMEMD(PC);
						PC += 2;
						if( !dr ) b = GETMEMB(pKode + (long)(KLONG)dw + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
						else if( dr==1 ) w2 = GETMEMW(pKode + (long)(KLONG)dw + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETMEMD(PC));
                        else if( dr==2 ) dw2 = GETMEMD(pKode + (long)(KLONG)GETMEMD(PC));
                        else qw2 = GETMEMQ(pKode + (long)(KLONG)GETMEMD(PC));
						PC += 2;
						w &= 0x00FF;
						break;
					case 4:	//regl,[regh+reg2]		56/57op reg/reg	reg
						if( !dr ) b = GETMEMB(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(GETMEMW(PC)));
						else if( dr==1 ) w2 = GETMEMW(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(GETMEMW(PC)));
                        else if( dr==2 ) dw2 = GETMEMD(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(GETMEMW(PC)));
                        else qw2 = GETMEMQ(pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(GETMEMW(PC)));
						PC++;
						w &= 0x00FF;
						break;
					case 5:	//[regl],regh			56/57op reg/reg
						if( !dr ) b = GETREGB((w>>8) & 0x00FF);
						else if( dr==1 ) w2 = GETREGW((w>>8) & 0x00FF);
                        else if( dr==2 ) dw2 = GETREGD((w>>8) & 0x00FF);
                        else qw2 = GETREGQ((w>>8) & 0x00FF);
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF);
						break;
					case 6:	//[regl],imm			56/57op reg		imm-- ???????
						if( !dr ) b = (BYTE)GETMEMW(PC);
						else if( dr==1 ) w2 = GETMEMW(PC);
                        else if( dr==2 ) {
                            dw2 = GETMEMD(PC);
                            PC++;
                        } else {
                            qw2 = GETMEMQ(PC);
                            PC += 3;
                        }
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF);
						break;
					case 7:	//[regl+off],regh		56/57op reg/reg	off----------
						if( !dr ) b = GETREGB((w>>8) & 0x00FF);
						else if( dr==1 ) w2 = GETREGW((w>>8) & 0x00FF);
                        else if( dr==2 ) dw2 = GETREGD((w>>8) & 0x00FF);
                        else qw2 = GETREGQ((w>>8) & 0x00FF);
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						break;
					case 8:	//[regl+off],imm		56/57op reg		off----------	imm-- ???????
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						if( !dr ) b = (BYTE)GETMEMW(PC);
						else if( dr==1 ) w2 = GETMEMW(PC);
                        else if( dr==2 ) {
                            dw2 = GETMEMD(PC);
                            PC++;
                        } else {
                            qw2 = GETMEMQ(PC);
                            PC += 3;
                        }
						PC++;
						break;
					case 9:	//[mem],regh			56/57op reg		mem----------
						if( !dr ) b = GETREGB(w & 0x00FF);
						else if( dr==1 ) w2 = GETREGW(w & 0x00FF);
                        else if( dr==2 ) dw2 = GETREGD(w & 0x00FF);
                        else qw2 = GETREGQ(w & 0x00FF);
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						break;
					case 10:	//[mem],imm			56/57op mem----------	imm-- ???????
						--PC;
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						if( !dr ) b = (BYTE)GETMEMW(PC);
						else if( dr==1 ) w2 = GETMEMW(PC);
                        else if( dr==2 ) {
                            dw2 = GETMEMD(PC);
                            PC++;
                        } else {
                            qw2 = GETMEMQ(PC);
                            PC += 3;
                        }
						PC++;
						break;
					case 11:	//[mem+regl+off],regh	56/57op reg/reg	mem----------	off----------
						if( !dr ) b = GETREGB((w>>8) & 0x00FF);
						else if( dr==1 ) w2 = GETREGW((w>>8) & 0x00FF);
                        else if( dr==2 ) dw2 = GETREGD((w>>8) & 0x00FF);
                        else qw2 = GETREGQ((w>>8) & 0x00FF);
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						pMem += (long)(KLONG)GETMEMD(PC);
						PC += 2;
						break;
					case 12:	//[mem+regl+off],imm	56/57op reg		mem----------	off----------	imm-- ???????
						dw = GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw + (long)(KLONG)GETMEMD(PC) + (long)(KLONG)GETREGD(w & 0x00FF);
						PC += 2;
						if( !dr ) b = (BYTE)GETMEMW(PC);
						else if( dr==1 ) w2 = GETMEMW(PC);
                        else if( dr==2 ) {
                            dw2 = GETMEMD(PC);
                            PC++;
                        } else {
                            qw2 = GETMEMQ(PC);
                            PC += 3;
                        }
						PC++;
						break;
					case 13:	//[regl+regh],reg2		56/57op reg/reg	reg
						pMem = pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(w & 0x00FF);
						if( !dr ) b = GETREGB(GETMEMW(PC));
						else if( dr==1 ) w2 = GETREGW(GETMEMW(PC));
                        else if( dr==2 ) dw2 = GETREGD(GETMEMW(PC));
                        else qw2 = GETREGQ(GETMEMW(PC));
						PC++;
						break;
					case 14:	//[regl+regh],imm		56/57op reg/reg	imm-- ???????
						pMem = pKode + (long)(KLONG)GETREGD((w>>8) & 0x00FF) + (long)(KLONG)GETREGD(w & 0x00FF);
						if( !dr ) b = (BYTE)GETMEMW(PC);
						else if( dr==1 ) w2 = GETMEMW(PC);
                        else if( dr==2 ) {
                            dw2 = GETMEMD(PC);
                            PC++;
                        } else {
                            qw2 = GETMEMQ(PC);
                            PC += 3;
                        }
						PC++;
						break;
					case 15:	// LEA d,[aaa]
						/*
						 For LEA, the 0x56/0x57 opcodes are laid out like this:
						 
						 0101 011u uaaa 1111
						 
						 where 'uu' is:
						 00	LEA
						 01	PEA
						 10	reserved for future use
						 11	reserved for future use
						 and 'aaa' is the addressing mode:
						 000	[reg]
						 001	[reg+off]
						 010	[mem]
						 011	[reg+mem+off]
						 100	[reg+reg]
						 101	unused
						 110 unused
						 111 unused
						 */
						dw2 = w2 = b = 0;
						if( (opcode & 0x0180)==0 ) {	// LEA
							switch( opcode & 0x0070 ) {
								case 0x0000:	// LEA d,[regt]			0x560F REGt:REGd
									SETREGD(w & 0x00FF, GETREGD((w>>8) & 0x00FF));
									break;
								case 0x0010:	// LEA d,[regt+off]		0x561F REGt:REGd offDWORD
									SETREGD(w & 0x00FF, GETREGD((w>>8) & 0x00FF) + GETMEMD(PC));
									PC += 2;
									break;
								case 0x0020:	// LEA d,[mem]			0x562F memDWORD
									SETREGD(w & 0x00FF, GETMEMD(PC));
									PC += 2;
									break;
								case 0x0030:	// LEA d,[regt+mem+off]	0x563F REGt:REGd memDWORD offDWORD
									dw2 = GETMEMD(PC);
									PC += 2;
									SETREGD(w & 0x00FF, GETREGD((w>>8) & 0x00FF) + dw2 + GETMEMD(PC));
									PC += 2;
									break;
								case 0x0040:	// LEA d,[regt1+regt2]	0x564F REGt1:REGd REGt2
									SETREGD(w & 0x00FF, GETREGD((w>>8) & 0x00FF) + GETREGD(GETMEMW(PC++)));
									break;
								case 0x0050:	// unused
								case 0x0060:	// unused
								case 0x0070:	// unused
									break;
							}
							goto leabreak;
						} else if( (opcode & 0x0180)==0x0080 ) {	// PEA (push effective address onto C R7 stack)
							switch( opcode & 0x0070 ) {
								case 0x0000:	// PEA [regt]			0x568F 0:REGt
									SETREGD(REG_7, GETREGD(REG_7)-4);
									SETMEMD(pKode + (long)(KLONG)GETREGD(REG_7), GETREGD(w));
									break;
								case 0x0010:	// PEA [regt+off]		0x569F 0:REGt offDWORD
									SETREGD(REG_7, GETREGD(REG_7)-4);
									SETMEMD(pKode + (long)(KLONG)GETREGD(REG_7), GETREGD(w) + GETMEMD(PC));
									PC += 2;
									break;
								case 0x0020:	// PEA [mem]			0x56AF memDWORD
									--PC;	// put back "register" WORD
									SETREGD(REG_7, GETREGD(REG_7)-4);
									SETMEMD(pKode + GETREGD(REG_7), GETMEMD(PC));
									PC += 2;
									break;
								case 0x0030:	// PEA [regt+mem+off]	0x56BF 0:REGt memDWORD offDWORD
									SETREGD(REG_7, GETREGD(REG_7)-4);
									dw2 = GETMEMD(PC);
									PC += 2;
									SETMEMD(pKode + GETREGD(REG_7), GETREGD(w) + dw2 + GETMEMD(PC));
									PC += 2;
									break;
								case 0x0040:	// PEA [regt1+regt2]	0x56CF REGt2:REGt1
									SETREGD(REG_7, GETREGD(REG_7)-4);
									SETMEMD(pKode + GETREGD(REG_7), GETREGD(w & 0x00FF) + GETREGD((w>>8) & 0x00FF));
									break;
								case 0x0050:	// unused
								case 0x0060:	// unused
								case 0x0070:	// unused
									break;
							}
							goto leabreak;
						}
				}
				// w  = target reg OR pMem = target memory
				// b or w2 or dw2 or qw2 = second arg
				dr = (opcode & 0x000F) > 4;	// dr = 0 if reg target, non-0 if mem target
				switch( (opcode >> 4) & 0x001F ) {
					case 0:	//BYTE LDM
						if( dr ) SETMEMB(pMem, b);
						else SETREGB(w, b);
						break;
					case 1:	//BYTE ADD
						if( dr ) SETMEMB(pMem, GETMEMB(pMem)+b);
						else SETREGB(w, GETREGB(w)+b);
						break;
					case 2:	//BYTE SUB
						if( signedmath ) {
							if( dr ) SETMEMB(pMem, (result=(DWORD)((KLONG)(((KLONG)((char)GETMEMB(pMem)))-((KLONG)((char)b))))));
							else SETREGB(w, (result=(DWORD)((KLONG)(((KLONG)((char)GETREGB(w)))-((KLONG)((char)b))))));
						} else {
							if( dr ) SETMEMB(pMem, (result=(DWORD)GETMEMB(pMem)-(DWORD)b));
							else SETREGB(w, (result=(DWORD)GETREGB(w)-(DWORD)b));
						}
						break;
					case 3:	//BYTE CMP
						if( signedmath ) {
							if( dr ) result = (DWORD)((KLONG)(((KLONG)((char)GETMEMB(pMem)))-((KLONG)((char)b))));
							else result = (DWORD)((KLONG)(((KLONG)((char)GETREGB(w)))-((KLONG)((char)b))));
						} else {
							if( dr ) result = (DWORD)GETMEMB(pMem)-(DWORD)b;
							else result = (DWORD)GETREGB(w)-(DWORD)b;
						}
						break;
					case 4:	//BYTE TST
						if( dr ) result = (DWORD)(char)(GETMEMB(pMem) & b);
						else result = (DWORD)(char)(GETREGB(w) & b);
						result |= (result<<24);
						break;
					case 5:	//BYTE AND
						if( dr ) SETMEMB(pMem, (BYTE)(result = (DWORD)((KLONG)(char)GETMEMB(pMem) & (KLONG)(char)b)));
						else SETREGB(w, (BYTE)(result = (DWORD)((KLONG)(char)GETREGB(w) & (KLONG)(char)b)));
						break;
					case 6:	//BYTE OR
						if( dr ) SETMEMB(pMem, GETMEMB(pMem) | b);
						else SETREGB(w, GETREGB(w) | b);
						break;
					case 7:	//BYTE XOR
						if( dr ) SETMEMB(pMem, (BYTE)(result = (DWORD)((KLONG)(char)GETMEMB(pMem) ^ (KLONG)(char)b)));
						else SETREGB(w, (BYTE)(result = (DWORD)((KLONG)(char)GETREGB(w) ^ (KLONG)(char)b)));
						break;
					case 8:	//WORD LDM
						if( dr ) SETMEMW(pMem, w2);
						else SETREGW(w, w2);
						break;
					case 9:	//WORD ADD
						if( dr ) SETMEMW(pMem, GETMEMW(pMem)+w2);
						else SETREGW(w, GETREGW(w)+w2);
						break;
					case 10:	//WORD SUB
						if( signedmath ) {
							if( dr ) SETMEMW(pMem, (WORD)(result=(DWORD)((KLONG)(((KLONG)((KSHORT)GETMEMW(pMem)))-((KLONG)((KSHORT)w2))))));
							else SETREGW(w, (WORD)(result=(DWORD)((KLONG)(((KLONG)((KSHORT)GETREGW(w)))-((KLONG)((KSHORT)w2))))));
						} else {
							if( dr ) SETMEMW(pMem, (WORD)(result=(DWORD)GETMEMW(pMem)-(DWORD)w2));
							else SETREGW(w, (WORD)(result=(DWORD)GETREGW(w)-(DWORD)w2));
						}
						break;
					case 11:	//WORD CMP
						if( signedmath ) {
							if( dr ) result = (DWORD)((KLONG)(((KLONG)((KSHORT)GETMEMW(pMem)))-((KLONG)((KSHORT)w2))));
							else result = (DWORD)((KLONG)(((KLONG)((KSHORT)GETREGW(w)))-((KLONG)((KSHORT)w2))));
						} else {
							if( dr ) result = (DWORD)GETMEMW(pMem)-(DWORD)w2;
							else result = (DWORD)GETREGW(w)-(DWORD)w2;
						}
						break;
					case 12:	//WORD TST
						if( dr ) result = (DWORD)(KSHORT)(GETMEMW(pMem) & w2);
						else result = (DWORD)(KSHORT)(GETREGW(w) & w2);
						result |= (result<<24);
						break;
					case 13:	//WORD AND
						if( dr ) SETMEMW(pMem, (WORD)(result = (DWORD)((KLONG)(KSHORT)GETMEMW(pMem) & (KLONG)(KSHORT)w2)));
						else SETREGW(w, (WORD)(result = (DWORD)((KLONG)(KSHORT)GETREGW(w) & (KLONG)(KSHORT)w2)));
						break;
					case 14:	//WORD OR
						if( dr ) SETMEMW(pMem, GETMEMW(pMem) | w2);
						else SETREGW(w, GETREGW(w) | w2);
						break;
					case 15:	//WORD XOR
						if( dr ) SETMEMW(pMem, (WORD)(result = (DWORD)((KLONG)(KSHORT)GETMEMW(pMem) ^ (KLONG)(KSHORT)w2)));
						else SETREGW(w, (WORD)(result = (DWORD)((KLONG)(KSHORT)GETREGW(w) ^ (KLONG)(KSHORT)w2)));
						break;
					case 16:	//DWORD LDM
						if( dr ) SETMEMD(pMem, dw2);
						else SETREGD(w, dw2);
						break;
					case 17:	//DWORD ADD
						if( dr ) SETMEMD(pMem, GETMEMD(pMem)+dw2);
						else SETREGD(w, GETREGD(w)+dw2);
						break;
					case 18:	//DWORD SUB
						if( signedmath ) {
							if( dr ) SETMEMD(pMem, (result=(DWORD)((KLONG)GETMEMD(pMem)-(KLONG)dw2)));
							else SETREGD(w, (result=(DWORD)((KLONG)GETREGD(w)-(KLONG)dw2)));
						} else {
							if( dr ) SETMEMD(pMem, (result=GETMEMD(pMem)-dw2));
							else SETREGD(w, (result=GETREGD(w)-dw2));
						}
						break;
					case 19:	//DWORD CMP
						if( signedmath ) {
							if( dr ) result = (DWORD)((KLONG)GETMEMD(pMem)-(KLONG)dw2);
							else result = (DWORD)((KLONG)GETREGD(w)-(KLONG)dw2);
						} else {
							if( dr ) result = GETMEMD(pMem)-dw2;
							else result = GETREGD(w)-dw2;
						}
						break;
					case 20:	//DWORD TST
						if( dr ) result = GETMEMD(pMem) & dw2;
						else result = GETREGD(w) & dw2;
						break;
					case 21:	//DWORD AND
						if( dr ) SETMEMD(pMem, (result = GETMEMD(pMem) & dw2));
						else SETREGD(w, (result = GETREGD(w) & dw2));
						break;
					case 22:	//DWORD OR
						if( dr ) SETMEMD(pMem, GETMEMD(pMem) | dw2);
						else SETREGD(w, GETREGD(w) | dw2);
						break;
					case 23:	//DWORD XOR
						if( dr ) SETMEMD(pMem, (result = GETMEMD(pMem) ^ dw2));
						else SETREGD(w, (result = GETREGD(w) ^ dw2));
						break;
                    case 24:    //QWORD LDM
                       if( dr ) SETMEMQ(pMem, qw2);
                       else SETREGQ(w, qw2);
                       break;
                    case 25:    //QWORD ADD
                       if( dr ) SETMEMQ(pMem, (qw = GETMEMQ(pMem)+qw2));
                       else SETREGQ(w, (qw = GETREGQ(w)+qw2));
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 26:    //QWORD SUB
                       if( signedmath ) {
                           if( dr ) SETMEMQ(pMem, (qw = (QWORD)((long long)GETMEMQ(pMem)-(long long)qw2)));
                           else SETREGQ(w, (qw = (QWORD)((long long)GETREGQ(w)-(long long)qw2)));
                       } else {
                           if( dr ) SETMEMQ(pMem, (qw = GETMEMQ(pMem)-qw2));
                           else SETREGQ(w, (qw = GETREGQ(w)-qw2));
                       }
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 27:    //QWORD CMP
                       if( signedmath ) {
                           if( dr ) qw = (QWORD)((long long)GETMEMQ(pMem)-(long long)qw2);
                           else qw = (QWORD)((long long)GETREGQ(w)-(long long)qw2);
                       } else {
                           if( dr ) qw = GETMEMQ(pMem)-qw2;
                           else qw = GETREGQ(w)-qw2;
                       }
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 28:    //QWORD TST
                       if( dr ) qw = GETMEMQ(pMem) & qw2;
                       else qw = GETREGQ(w) & qw2;
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 29:    //QWORD AND
                       if( dr ) SETMEMQ(pMem, (qw = GETMEMQ(pMem) & qw2));
                       else SETREGD(w, (qw = GETREGQ(w) & qw2));
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 30:    //QWORD OR
                       if( dr ) SETMEMQ(pMem, (qw = GETMEMQ(pMem) | qw2));
                       else SETREGQ(w, (qw = GETREGQ(w) | qw2));
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
                    case 31:    //QWORD XOR
                       if( dr ) SETMEMQ(pMem, (qw = GETMEMQ(pMem) ^ qw2));
                       else SETREGQ(w, (qw = GETREGQ(w) ^ qw2));
                       result = (DWORD)((qw>>32)&0x080000000) | (DWORD)(qw & 1);
                       break;
				}
leabreak:
				break;
			case 0x58:	// LD0 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x59:	// LD0 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC))));
				PC += 2;
				break;
			case 0x5A:	// LD0 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x5B:	// LD1 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x5C:	// LD1 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x5D:	// LD1 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x5E:	// LD2 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x5F:	// LD2 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x60:	// LD2 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x61:	// LDS b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x62:	// LDS w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC))));
				PC += 2;
				break;
			case 0x63:	// LDS d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x64:	// ST0 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMB(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x65:	// ST0 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x66:	// ST0 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMD(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x67:	// ST1 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x68:	// ST1 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x69:	// ST1 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETREGD(w) + GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x6A:	// ST2 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(w) + GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x6B:	// ST2 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(w) + GETMEMD(PC), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x6C:	// ST2 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETREGD(w) + GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0x6D:	// STS b,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMB(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw, GETREGB(opcode & 0x00FF));
				break;
			case 0x6E:	// STS w,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw), GETREGW(opcode & 0x00FF));
				break;
			case 0x6F:	// STS d,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMD(pKode + (long)(KLONG)GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw, GETREGD(opcode & 0x00FF));
				break;
			case 0x70:	// PU0 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMB(pKode + (long)(KLONG)GETREGD(w) +(long)(KLONG)dw, GETREGB(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+1);
				break;
			case 0x71:	// PU0 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w) +(long)(KLONG)dw), GETREGW(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+2);
				break;
			case 0x72:	// PU0 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETMEMD(pKode + (long)(KLONG)GETREGD(w) +(long)(KLONG)dw, GETREGD(opcode & 0x00FF));
				SETREGD(w, GETREGD(w)+4);
				break;
			case 0x73:	// PO0 b,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETREGD(w, GETREGD(w)-1);
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw));
				break;
			case 0x74:	// PO0 w,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETREGD(w, GETREGD(w)-2);
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw)));
				break;
			case 0x75:	// PO0 d,d,=dval
				w = GETMEMW(PC);
				PC++;
				dw= GETMEMD(PC);
				PC += 2;
				SETREGD(w, GETREGD(w)-4);
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)dw));
				break;
			case 0x76:	// LSA d,=dval
				SETREGD(opcode & 0x00FF, GETREGD(REG_4) + (KLONG)(KSHORT)(WORD)GETMEMW(PC));
				PC++;
				break;
			case 0x77:	// B2I r,r (byte-2-index dst-cnt-reg,src-bit-test-reg
				dr = opcode & 0x00FF;
				w = GETMEMW(PC);
				PC++;
				if( w<0x0080 ) dw = GETREGD(w);
				else if( w<0x00E0 ) dw = (DWORD)GETREGW(w);
				else dw = (DWORD)GETREGB(w);
				if( dw ) {
					result = 0;
					while( !(dw & 1) ) {
						++result;
						dw >>= 1;
					}
				} else result = 0xFFFFFFFF;
				if( dr<0x0080 ) SETREGD(opcode & 0x00FF, result);
				else if( dr<0x00E0 ) SETREGW(opcode & 0x00FF, (WORD)result);
				else SETREGB(opcode & 0x00FF, (BYTE)result);
				break;
			case 0x78:	// CLR b,b
				dr = opcode & 0x00FF;
				memset(Kregs+dr, 0, GETMEMW(PC)+1-dr);
				PC++;
				break;
			case 0x79:	// PUR b,b
				dr = opcode & 0x00FF;
				w = GETMEMW(PC)+1-dr;
				PC++;
				dw = GETREGD(REG_5);
				memcpy(pKode + dw, Kregs+dr, w);
				SETREGD(REG_5, dw+w);
				break;
			case 0x7A:	// POR b,b
				dr = opcode & 0x00FF;
				w = GETMEMW(PC)+1-dr;
				PC++;
				dw = GETREGD(REG_5)-w;
				SETREGD(REG_5, dw);
				memcpy(Kregs+dr, pKode+dw, w);
				break;
			case 0x7B:	// unused
			case 0x7C:	// unused
				switch( opcode & 0x00FF ) {
					case 0: // GEEKDO: Remove this! (Debug LOGOUT code)
					{
//						Dwrite((char*)AllocPtrs[GETREGD(0)]+GETREGD(REG_15));
						NSString *dbgstr = [NSString stringWithUTF8String:(char *)(BYTE*)AllocPtrs[GETREGD(0)]+ GETREGD(REG_15)];
						NSLog(@"%@", dbgstr);
					}
						break;
					case 1: // ChksumMemlk(r1, r31, r30) handle, offset, len
						dw = GETREGD(REG_1);
						dw2= GETREGD(REG_31);
						dw3= GETREGD(REG_30);
						dw4= 0;
						if( dw==0 ) {
							while( dw3-- ) {
								dw4 = ((dw4 & 0x80000000)?1:0) + (dw4<<1) + (DWORD)pKode[dw2];
								++dw2;
							}
						} else {
							pMem = (BYTE *)(AllocPtrs[GETREGD(REG_1)]) + dw2;
							while( dw3-- ) {
								dw4 = ((dw4 & 0x80000000)?1:0) + (dw4<<1) + (DWORD)GETMEMB(pMem);
								++pMem;
							}
						}
						SETREGD(REG_16, dw4);
						break;
					default:
						break;
				}

				break;
			case 0x7D:	// SAF b
				dw = result & 0x00000001;
				if( result==0 ) dw |= 2;
				if( result & 0x80000000 ) dw |= 4;
				if( signedmath ) dw |= 8;
				SETREGB(opcode & 0x00FF, (BYTE)dw);
				break;
			case 0x7E:	// REF b
				b = GETREGB(opcode & 0x00FF);
				result = ((~b) & 2) | (b & 1);
				if( b & 4 ) result |= 0x80000000;
				signedmath = (b & 8)?1:0;
				break;
			case 0x7F:	// RET b,=0
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGB(opcode & 0x00FF, 0);
				break;
			case 0x80:	// RET w,=0
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGW(opcode & 0x00FF, 0);
				break;
			case 0x81:	// RET d,=0
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGD(opcode & 0x00FF, 0);
				break;
			case 0x82:	// RET b,=1
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGB(opcode & 0x00FF, 1);
				break;
			case 0x83:	// RET w,=1
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGW(opcode & 0x00FF, 1);
				break;
			case 0x84:	// RET d,=1
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				result = (DWORD)SETREGD(opcode & 0x00FF, 1);
				break;
			case 0x85:	// INC b
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr) + 1);
				result |= (result<<24);
				break;
			case 0x86:	// INC w
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, GETREGW(dr) + 1);
				result |= (result<<16);
				break;
			case 0x87:	// INC d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) + 1);
				break;
			case 0x88:	// DEC b
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr) - 1);
				result |= (result<<24);
				break;
			case 0x89:	// DEC w
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, GETREGW(dr) - 1);
				result |= (result<<16);
				break;
			case 0x8A:	// DEC d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) - 1);
				break;
			case 0x8B:	// NOT b
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, ~GETREGB(dr));
				result |= (result<<24);
				break;
			case 0x8C:	// NOT w
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, ~GETREGW(dr));
				result |= (result<<16);
				break;
			case 0x8D:	// NOT d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, ~GETREGD(dr));
				break;
			case 0x8E:	// NEG b
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, -(char)GETREGB(dr));
				result |= (result<<24);
				break;
			case 0x8F:	// NEG w
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, -(KSHORT)GETREGW(dr));
				result |= (result<<16);
				break;
			case 0x90:	// NEG d
				dr = opcode & 0x00FF;
				result = SETREGD(dr, -(KLONG)GETREGD(dr));
				break;
			case 0x91:	// PUSH b	NOTE: we actually push a WORD to keep the stack WORD aligned
				SETMEMW((WORD *)(pKode + GETREGD(REG_5)), (WORD)(BYTE)GETREGB(opcode & 0x00FF));
				SETREGD(REG_5, GETREGD(REG_5)+2);
				break;
			case 0x92:	// PUSH w
				SETMEMW((WORD *)(pKode + GETREGD(REG_5)), GETREGW(opcode & 0x00FF));
				SETREGD(REG_5, GETREGD(REG_5)+2);
				break;
			case 0x93:	// PUSH d
				SETMEMD(pKode + GETREGD(REG_5), GETREGD(opcode & 0x00FF));
				SETREGD(REG_5, GETREGD(REG_5)+4);
				break;
			case 0x94:	// POP b	NOTE: we actually pop a WORD, as that's what was pushed
				SETREGD(REG_5, GETREGD(REG_5)-2);
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_5)));
				break;
			case 0x95:	// POP w
				SETREGD(REG_5, GETREGD(REG_5)-2);
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + GETREGD(REG_5))));
				break;
			case 0x96:	// POP d
				SETREGD(REG_5, GETREGD(REG_5)-4);
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_5)));
				break;
			case 0x97:	// CBW w
				//			w = GETREGW(opcode & 0x00FF);
				//			SETREGW(opcode & 0x00FF, (w & 0x0080)?(w | 0xFF00):(w & 0x00FF));
				SETREGW(opcode & 0x00FF, (WORD)(KSHORT)(char)GETREGB(opcode & 0x00FF));
				break;
			case 0x98:	// LD0 b,=dval
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x99:	// LD0 w,=dval
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETMEMD(PC))));
				PC += 2;
				break;
			case 0x9A:	// LD0 d,=dval
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETMEMD(PC)));
				PC += 2;
				break;
			case 0x9B:	// LD1 b,=dval
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0x9C:	// LD1 w,=dval
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0x9D:	// LD1 d,=dval
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0x9E:	// LD2 b,=dval
				SETREGB(opcode & 0x00FF, GETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0x9F:	// LD2 w,=dval
				SETREGW(opcode & 0x00FF, GETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0xA0:	// LD2 d,=dval
				SETREGD(opcode & 0x00FF, GETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC)));
				PC += 2;
				break;
			case 0xA1:	// LDS b,=dval
				SETREGB(opcode & 0x00FF, GETMEMB(pKode + (long)(KLONG)GETREGD(REG_4) + ((KLONG)((KSHORT)GETMEMW(PC)))));
				PC++;
				break;
			case 0xA2:	// LDS w,=dval
				SETREGW(opcode & 0x00FF, GETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(REG_4) + ((long)(KLONG)((KSHORT)GETMEMW(PC))))));
				PC++;
				break;
			case 0xA3:	// LDS d,=dval
				SETREGD(opcode & 0x00FF, GETMEMD(pKode + (long)(KLONG)GETREGD(REG_4) + (long)((KLONG)((KSHORT)GETMEMW(PC)))));
				PC++;
				break;
			case 0xA4:	// ST0 b,=dval
				SETMEMB(pKode + (long)(KLONG)GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xA5:	// ST0 w,=dval
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETMEMD(PC)), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xA6:	// ST0 d,=dval
				SETMEMD(pKode + (long)(KLONG)GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xA7:	// ST1 b,=dval
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xA8:	// ST1 w,=dval
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xA9:	// ST1 d,=dval
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_1)]) + GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xAA:	// ST2 b,=dval
				SETMEMB((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC), GETREGB(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xAB:	// ST2 w,=dval
				SETMEMW((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC), GETREGW(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xAC:	// ST2 d,=dval
				SETMEMD((BYTE *)(AllocPtrs[GETREGD(REG_2)]) + GETMEMD(PC), GETREGD(opcode & 0x00FF));
				PC += 2;
				break;
			case 0xAD:	// STS b,=dval
				SETMEMB(pKode + (long)(KLONG)GETREGD(REG_4) + (long)((KLONG)((KSHORT)GETMEMW(PC))), GETREGB(opcode & 0x00FF));
				PC++;
				break;
			case 0xAE:	// STS w,=dval
				SETMEMW((WORD *)(pKode + (long)(KLONG)GETREGD(REG_4) + (long)((KLONG)((KSHORT)GETMEMW(PC)))), GETREGW(opcode & 0x00FF));
				PC++;
				break;
			case 0xAF:	// STS d,=dval
				SETMEMD(pKode + (long)(KLONG)GETREGD(REG_4) + (long)((KLONG)((KSHORT)GETMEMW(PC))), GETREGD(opcode & 0x00FF));
				PC++;
				break;
			case 0xB0:	// IJMP b,=dval
				dw2 = (DWORD)GETREGB(opcode & 0x00FF);
				goto ijmp;
			case 0xB1:	// IJMP w,=dval
				dw2 = (DWORD)GETREGW(opcode & 0x00FF);
				goto ijmp;
			case 0xB2:	// IJMP d,=dval
				dw2 = GETREGD(opcode & 0x00FF);
			ijmp:
				pMem = pKode + (long)(KLONG)GETMEMD(PC);	// don't bother to advance PC, we're jumping
				dw = GETMEMD(pMem);	// get # of entries in table
				if( dw2>dw ) dw2 = dw;
				++dw2;	// also skip "number of entries" DWORD
				pMem += 4*dw2;	// skip to the entry (default if past end of table)
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pMem));
				break;
			case 0xB3:	// KJMP b,=dval
				dw2 = (DWORD)GETREGB(opcode & 0x00FF);
				goto kjmp;
			case 0xB4:	// KJMP w,=dval
				dw2 = (DWORD)GETREGW(opcode & 0x00FF);
				goto kjmp;
			case 0xB5:	// KJMP d,=dval
				dw2 = GETREGD(opcode & 0x00FF);
			kjmp:
				pMem = pKode + (long)(KLONG)GETMEMD(PC);	// don't bother to advance PC, we're jumping
				dw = GETMEMD(pMem);	// get # of entries in table
				pMem += 4;
				while( dw-- ) {
					if( dw2==GETMEMD(pMem) ) {
						pMem += 4;	// skip value, point at jmp address
						break;
					}
					pMem += 8;	// skip value & jump address
				}
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pMem));
				break;
			case 0xB6:	// JSR d,=dval
#ifdef GEEK_TRACE2
				if( TBptr2+1 < TBMAX2 ) TraceBuf2[TBptr2++] = (DWORD)((BYTE *)PC-pKode);
#endif
				dw = GETREGD(opcode & 0x00FF) + GETMEMD(PC);
				PC += 2;
				SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
				SETREGD(REG_5, GETREGD(REG_5)+4);
				PC = (WORD *)(pKode + dw);
				break;
			case 0xB7:	// JSI d,=dval
#ifdef GEEK_TRACE2
				if( TBptr2+1 < TBMAX2 ) TraceBuf2[TBptr2++] = (DWORD)((BYTE *)PC-pKode);
#endif
				dw = GETMEMD(pKode + (long)(KLONG)GETREGD(opcode & 0x00FF) + (long)(KLONG)GETMEMD(PC));
				PC += 2;
				SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
				SETREGD(REG_5, GETREGD(REG_5)+4);
				PC = (WORD *)(pKode + dw);
				break;
			case 0xB8:	// CBD d
				//			dw = GETREGD(opcode & 0x00FF);
				//			SETREGD(opcode & 0x00FF, (dw & 0x00000080)?(dw | 0xFFFFFF00):(dw & 0x000000FF));
				SETREGD(opcode & 0x00FF, (DWORD)(KLONG)(char)GETREGB(opcode & 0x00FF));
				break;
			case 0xB9:	// CWD d
				//			dw = GETREGD(opcode & 0x00FF);
				//			SETREGD(opcode & 0x00FF, (dw & 0x00008000)?(dw | 0xFFFF0000):(dw & 0x0000FFFF));
				SETREGD(opcode & 0x00FF, (DWORD)(KLONG)(KSHORT)GETREGW(opcode & 0x00FF));
				break;
			case 0xBA:	// LDI b,=wval
				SETREGB(opcode & 0x00FF, GETMEMW(PC));
				PC++;
				break;
			case 0xBB:	// LDI w,=wval
				SETREGW(opcode & 0x00FF, GETMEMW(PC));
				PC++;
				break;
			case 0xBC:	// LDI d,=dval
				SETREGD(opcode & 0x00FF, GETMEMD(PC));
				PC += 2;
				break;
			case 0xBD:	// SHL b,=wval
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr)<<(BYTE)(GETMEMW(PC) & 0x0007));
				PC++;
				result |= (result<<24);
				break;
			case 0xBE:	// SHL w,=wval
				dr = opcode & 0x00FF;
				result = SETREGW(dr, GETREGW(dr)<<(BYTE)(GETMEMW(PC) & 0x000F));
				PC++;
				result |= (result<<16);
				break;
			case 0xBF:	// SHL d,=wval
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr)<<(BYTE)(GETMEMW(PC) & 0x001F));
				PC++;
				break;
			case 0xC0:	// SHR b,=wval
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)(char)(SETREGB(dr, (char)GETREGB(dr)>>(BYTE)(GETMEMW(PC) & 0x0007)));
				else result = (DWORD)(char)(SETREGB(dr, GETREGB(dr)>>(BYTE)(GETMEMW(PC) & 0x0007)));
				PC++;
				result |= (result<<24);
				break;
			case 0xC1:	// SHR w,=wval
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)(KSHORT)(SETREGW(dr, (KSHORT)(GETREGW(dr))>>(BYTE)(GETMEMW(PC) & 0x000F)));
				else result = (DWORD)(KSHORT)(SETREGW(dr, GETREGW(dr)>>(BYTE)(GETMEMW(PC) & 0x000F)));
				PC++;
				result |= (result<<16);
				break;
			case 0xC2:	// SHR d,=wval
				dr = opcode & 0x00FF;
				if( signedmath ) result = SETREGD(dr, (KLONG)GETREGD(dr)>>(BYTE)(GETMEMW(PC) & 0x001F));
				else result = SETREGD(dr, GETREGD(dr)>>(BYTE)(GETMEMW(PC) & 0x001F));
				PC++;
				break;
			case 0xC3:	// ROL b,=wval
				dr = opcode & 0x00FF;
				w = (GETMEMW(PC) & 0x0007);
				PC++;
				b = GETREGB(dr);
				while( w-- ) {
					if( b & 0x80 ) b = (b << 1) | 1;
					else b <<= 1;
				}
				result = (DWORD)(char)SETREGB(dr, b);
				result |= (result<<24);
				break;
			case 0xC4:	// ROL w,=wval
				dr = opcode & 0x00FF;
				b = (BYTE)GETMEMW(PC) & 0x001F;
				PC++;
				w = GETREGW(dr);
				while( b-- ) {
					if( w & 0x8000 ) w = (w << 1) | 1;
					else w <<= 1;
				}
				result = (DWORD)(KSHORT)SETREGW(dr, w);
				result |= (result<<16);
				break;
			case 0xC5:	// ROL d,=wval
				dr = opcode & 0x00FF;
				dw = GETREGD(dr);
				w = GETMEMW(PC) & 0x1F;
				PC++;
				while( w-- ) {
					dw2 = dw;
					dw <<= 1;
					if( dw2 & 0x80000000 ) dw |= 1;
				}
				result = SETREGD(dr, dw);
				break;
			case 0xC6:	// ROR b,=wval
				dr = opcode & 0x00FF;
				w = GETMEMW(PC) & 0x0007;
				PC++;
				b = GETREGB(dr);
				while( w-- ) {
					if( b & 1 ) b = (b >> 1) | 0x80;
					else b >>= 1;
				}
				result = (DWORD)(char)SETREGB(dr, b);
				result |= (result<<24);
				break;
			case 0xC7:	// ROR w,=wval
				dr = opcode & 0x00FF;
				b = (BYTE)GETMEMW(PC) & 0x001F;
				PC++;
				w = GETREGW(dr);
				while( b-- ) {
					if( w & 1 ) w = (w >> 1) | 0x8000;
					else w >>= 1;
				}
				result = (DWORD)(KSHORT)SETREGW(dr, w);
				result |= (result<<16);
				break;
			case 0xC8:	// ROR d,=wval
				dr = opcode & 0x00FF;
				dw = GETREGD(dr);
				w = GETMEMW(PC) & 0x1F;
				PC++;
				while( w-- ) {
					dw2 = dw;
					dw >>= 1;
					if( dw2 & 1 ) dw |= 0x80000000;
				}
				result = SETREGD(dr, dw);
				break;
			case 0xC9:	// ADD b,=val
				dr = opcode & 0x00FF;
				SETREGB(dr, GETREGB(dr)+(BYTE)GETMEMW(PC));
				PC++;
				break;
			case 0xCA:	// ADD w,=val
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr)+GETMEMW(PC));
				PC++;
				break;
			case 0xCB:	// ADD d,=val
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr) + GETMEMD(PC));
				PC += 2;
				break;
			case 0xCC:	// SUB b,=val
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)((char)GETREGB(dr)))-((KLONG)((char)((BYTE)GETMEMW(PC))))));
				else result = (DWORD)GETREGB(dr)-(DWORD)GETMEMB(PC);
				PC++;
				SETREGB(dr, (BYTE)result);
				break;
			case 0xCD:	// SUB w,=val
				dr = opcode & 0x00FF;
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)((KSHORT)GETREGW(dr)))-((KLONG)((KSHORT)(GETMEMW(PC))))));
				else result = (DWORD)GETREGW(dr)-(DWORD)GETMEMW(PC);
				PC++;
				SETREGW(dr, (WORD)result);
				break;
			case 0xCE:	// SUB d,=val
				dr = opcode & 0x00FF;
				if( signedmath ) result = SETREGD(dr, (KLONG)(GETREGD(dr)) - (KLONG)(GETMEMD(PC)));
				else result = SETREGD(dr, GETREGD(dr) - GETMEMD(PC));
				PC += 2;
				break;
			case 0xCF:	// MUL b,=val
				dr = opcode & 0x00FF;
				b = (BYTE)GETMEMW(PC);
				PC++;
				SETREGB(dr, GETREGB(dr)*b);
				break;
			case 0xD0:	// MUL w,=val
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr)*GETMEMW(PC));
				PC++;
				break;
			case 0xD1:	// MUL d,=val
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr) * GETMEMD(PC));
				PC += 2;
				break;
			case 0xD2:	// DIV b,=val
				dr = opcode & 0x00FF;
				b = (BYTE)GETMEMW(PC);
				PC++;
				if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))/(char)b);
				else SETREGB(dr, GETREGB(dr)/b);
				break;
			case 0xD3:	// DIV w,=val
				dr = opcode & 0x00FF;
				w2 = GETMEMW(PC);
				PC++;
				if( signedmath ) SETREGW(dr, (KSHORT)(GETREGW(dr))/(KSHORT)w2);
				else SETREGW(dr, GETREGW(dr)/w2);
				break;
			case 0xD4:	// DIV d,=val
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGD(dr, (KLONG)(GETREGD(dr)) / (KLONG)(GETMEMD(PC)));
				else SETREGD(dr, GETREGD(dr) / GETMEMD(PC));
				PC += 2;
				break;
			case 0xD5:	// MOD b,=val
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))%(char)((BYTE)GETMEMW(PC)));
				else SETREGB(dr, GETREGB(dr)%((BYTE)GETMEMW(PC)));
				PC++;
				break;
			case 0xD6:	// MOD w,=val
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGW(dr, (KSHORT)(GETREGW(dr))%(KSHORT)((WORD)GETMEMW(PC)));
				else SETREGW(dr, (WORD)GETREGW(dr)%((WORD)GETMEMW(PC)));
				PC++;
				break;
			case 0xD7:	// MOD d,=val
				dr = opcode & 0x00FF;
				if( signedmath ) SETREGD(dr, (KLONG)(GETREGD(dr)) % (KLONG)(GETMEMD(PC)));
				else SETREGD(dr, GETREGD(dr) % GETMEMD(PC));
				PC += 2;
				break;
			case 0xD8:	// OR b,=val
				dr = opcode & 0x00FF;
				SETREGB(dr, GETREGB(dr) | (BYTE)GETMEMW(PC));
				PC++;
				break;
			case 0xD9:	// OR w,=val
				dr = opcode & 0x00FF;
				SETREGW(dr, GETREGW(dr) | GETMEMW(PC));
				PC++;
				break;
			case 0xDA:	// OR d,=val
				dr = opcode & 0x00FF;
				SETREGD(dr, GETREGD(dr) | GETMEMD(PC));
				PC += 2;
				break;
			case 0xDB:	// AND b,=val
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr) & (BYTE)GETMEMW(PC));
				PC++;
				result |= (result<<24);
				break;
			case 0xDC:	// AND w,=val
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, GETREGW(dr) & GETMEMW(PC));
				PC++;
				result |= (result<<16);
				break;
			case 0xDD:	// AND d,=val
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) & GETMEMD(PC));
				PC += 2;
				break;
			case 0xDE:	// XOR b,=val
				dr = opcode & 0x00FF;
				result = (DWORD)(char)SETREGB(dr, GETREGB(dr) ^ (BYTE)GETMEMW(PC));
				PC++;
				result |= (result<<24);
				break;
			case 0xDF:	// XOR w,=val
				dr = opcode & 0x00FF;
				result = (DWORD)(KSHORT)SETREGW(dr, GETREGW(dr) ^ GETMEMW(PC));
				PC++;
				result |= (result<<16);
				break;
			case 0xE0:	// XOR d,=val
				dr = opcode & 0x00FF;
				result = SETREGD(dr, GETREGD(dr) ^ GETMEMD(PC));
				PC += 2;
				break;
			case 0xE1:	// CMP b,=val
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)(char)GETREGB(opcode & 0x00FF))-((KLONG)(char)(GETMEMW(PC)))));
				else result = ((DWORD)(BYTE)GETREGB(opcode & 0x00FF))-((DWORD)(BYTE)GETMEMW(PC));
				PC++;
				break;
			case 0xE2:	// CMP w,=val
				if( signedmath ) result = (DWORD)((KLONG)(((KLONG)(KSHORT)GETREGW(opcode & 0x00FF))-((KLONG)(KSHORT)(GETMEMW(PC)))));
				else result = (DWORD)(WORD)GETREGW(opcode & 0x00FF)-(DWORD)(WORD)GETMEMW(PC);
				PC++;
				break;
			case 0xE3:	// CMP d,=val
				if( signedmath ) result = (KLONG)(GETREGD(opcode & 0x00FF)) - (KLONG)(GETMEMD(PC));
				else result = (DWORD)GETREGD(opcode & 0x00FF) - (DWORD)GETMEMD(PC);
				PC += 2;
				break;
			case 0xE4:	// TST b,=val
				result = (DWORD)(char)(GETREGB(opcode & 0x00FF) & (BYTE)GETMEMW(PC));
				PC++;
				result |= (result<<24);
				break;
			case 0xE5:	// TST w,=val
				result = (DWORD)(KSHORT)(GETREGW(opcode & 0x00FF) & GETMEMW(PC));
				PC++;
				result |= (result<<16);
				break;
			case 0xE6:	// TST d,=val
				result = GETREGD(opcode & 0x00FF) & GETMEMD(PC);
				PC += 2;
				break;
			case 0xE7:	// JMP =val
				dw = GETMEMD(PC);
				PC = (WORD *)(pKode + dw);
				break;
			case 0xE8:	// JE  +offset (0 to 255 WORDS from CP)
				if( !result ) PC += (DWORD)(opcode & 0x01FF);
				break;
			case 0xE9:	// JE  -offset (-256 to -1 WORDS from CP)
				if( !result ) PC += (int)(int16_t)((opcode & 0x01FF) | 0xFE00);
				break;
			case 0xEA:	// JNE +offset
				if( result ) PC += (DWORD)(opcode & 0x01FF);
				break;
			case 0xEB:	// JNE -offset
				if( result ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xEC:	// JL  +offset
				if( result & 0x80000000 ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xED:	// JL  -offset
				if( result & 0x80000000 ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xEE:	// JLE +offset
				if( !result || (result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xEF:	// JLE -offset
				if( !result || (result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xF0:	// JG  +offset
				if( result && !(result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xF1:	// JG  -offset
				if( result && !(result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xF2:	// JGE +offset
				if( !(result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xF3:	// JGE -offset
				if( !(result & 0x80000000) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xF4:	// JEV +offset
				if( !(result & 1) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xF5:	// JEV -offset
				if( !(result & 1) ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xF6:	// JOD +offset
				if( result & 1 ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xF7:	// JOD -offset
				if( result & 1 ) PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xF8:	// JMP +offset
				PC += (long)(KLONG)((DWORD)(opcode & 0x01FF));
				break;
			case 0xF9:	// JMP -offset
				PC += (long)(KLONG)((DWORD)(opcode & 0x01FF) | 0xFFFFFE00);
				break;
			case 0xFA:	// SysFunc
				SysFunc(pKE, (WORD)(opcode & 0x00FF));
				break;
			case 0xFB:	// JSR =dval and LJxx =dval and JMP[]
				rev = GETMEMB(pKode+KHDR_MAJOR_INTERP_REV);
				if( rev & 0x80 ) rev &= 0x7F;
				if( rev<3 ) opcode &= 0xFF00;	// force lower byte to 0 if code assembled for < 3.0 KINT
				dw = GETMEMD(PC);
				PC += 2;
				switch( opcode & 0x00FF ) {
					case 0x00:	// JSR =dval
#ifdef GEEK_TRACE2
						if( TBptr2+1 < TBMAX2 ) TraceBuf2[TBptr2++] = (DWORD)((BYTE *)PC-pKode);
#endif
						SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
						SETREGD(REG_5, GETREGD(REG_5)+4);
						PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x02:	// JMP [reg]
						PC -= 2;
						PC = (WORD *)(pKode + (long)(KLONG)(GETMEMD(pKode + (long)(KLONG)(GETREGD(GETMEMW(PC))))));
						break;
					case 0x03:	// JMP [reg+off]
						PC -= 2;
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						PC = (WORD *)(pKode + (long)(KLONG)(GETMEMD(pKode + (long)(KLONG)(dw2) + (long)(KLONG)(GETMEMD(PC)))));
						break;
					case 0x04:	// JMP [off]
						PC -= 2;
						PC = (WORD *)(pKode + (long)(KLONG)(GETMEMD(pKode + (long)(KLONG)(GETMEMD(PC)))));
						break;
					case 0x05:	// JMP [reg+off+off]
						PC -= 2;
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + (long)(KLONG)(dw2 + (long)(KLONG)GETMEMD(PC))));
						break;
					case 0x06:	// JMP [reg+reg]
						PC -= 2;
						w = GETMEMW(PC);
						PC = (WORD *)(pKode + (long)(KLONG)(GETMEMD(pKode + (long)(KLONG)(GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF)))));
						break;
					case 0x08:	// LJE
						if( !result ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x09:	// LJNE
						if( result ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0A:	// LJL
						if( result & 0x80000000 ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0B:	// LJLE
						if( !result || (result & 0x80000000) ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0C:	// LJG
						if( result && !(result & 0x80000000) ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0D:	// LJGE
						if( !(result & 0x80000000) ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0E:	// LJEV
						if( !(result & 1) ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
					case 0x0F:	// LJOD
						if( result & 1 ) PC = (WORD *)(pKode + (long)(KLONG)(dw));
						break;
				}
				break;
			case 0xFC:	// JSI =dval
#ifdef GEEK_TRACE2
				if( TBptr2+1 < TBMAX2 ) TraceBuf2[TBptr2++] = (DWORD)((BYTE *)PC-pKode);
#endif
				dw = GETMEMD(pKode + (long)(KLONG)(GETMEMD(PC)));
				PC += 2;
				SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
				SETREGD(REG_5, GETREGD(REG_5)+4);
				PC = (WORD *)(pKode + (long)(KLONG)(dw));
				break;
			case 0xFD:	// RET
#ifdef GEEK_TRACE2
				if( TBptr2 ) --TBptr2;
#endif
				SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
				PC = (WORD *)(pKode + (long)(KLONG)GETMEMD(pKode + GETREGD(REG_5)));
				break;
			case 0xFE:	// misc
				switch( opcode & 0x00FF ) {
					case 0x0000:	// BRK
						break;
					case 0x0001:	// NOP
						break;
					case 0x0002:	// TERM
						if( pKE==&mainKE ) killpgm |= 1;	// set flag we need to kill the program, but ONLY "main.ke" can do this! (otherwise, just HALT)
						// NOTE: case TERM falls through into case HALT
					case 0x0003:	// HALT
						--PC;	// move PC back to TERM instruction just for safety
						SETREGD(REG_3, (DWORD)((BYTE *)PC - pKode));	// update PC in Kregs
						goto stopint;
					case 0x0004:	// AFR
						// pu0 R4,R5
						SETMEMD(pKode + GETREGD(REG_5), GETREGD(REG_4));
						SETREGD(REG_5, GETREGD(REG_5)+4);
						// ldr R4,R5
						SETREGD(REG_4, GETREGD(REG_5));
						// add R5,=*PC++
						SETREGD(REG_5, GETREGD(REG_5) + GETMEMW(PC));
						PC++;
						break;
					case 0x0005:	// DFR
						// ldr R5,R4
						SETREGD(REG_5, GETREGD(REG_4)-4);	// minus the "pop" that's coming up
						// po0 R4,R5
						SETREGD(REG_4, GETMEMD(pKode+GETREGD(REG_5)));
						break;
					case 0x0006:	// USM
						signedmath = 0;
						break;
					case 0x0007:	// SM
						signedmath = 1;
						break;
					case 0x0010:	// LDCM b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-1);
						SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
						break;
					case 0x0011:	// LDCM w,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-2);
						SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
						break;
					case 0x0012:	// LDCM d,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-4);
						SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
						break;
					case 0x0013:	// LDC b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
						break;
					case 0x0014:	// LDC w,b
						w = GETMEMW(PC);
						PC++;
						SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
						break;
					case 0x0015:	// LDC d,b
						w = GETMEMW(PC);
						PC++;
						SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
						break;
					case 0x0016:	// LDCP b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+1);
						break;
					case 0x0017:	// LDCP w,b
						w = GETMEMW(PC);
						PC++;
						SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+2);
						break;
					case 0x0018:	// LDCP d,b
						w = GETMEMW(PC);
						PC++;
						SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+4);
						break;
					case 0x0019:	// STCM b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-1);
						SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
						break;
					case 0x001A:	// STCM w,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-2);
						SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
						break;
					case 0x001B:	// STCM d,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-4);
						SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
						break;
					case 0x001C:	// STC b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
						break;
					case 0x001D:	// STC w,b
						w = GETMEMW(PC);
						PC++;
						SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
						break;
					case 0x001E:	// STC d,b
						w = GETMEMW(PC);
						PC++;
						SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
						break;
					case 0x001F:	// STCP b,b
						w = GETMEMW(PC);
						PC++;
						SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+1);
						break;
					case 0x0020:	// STCP w,b
						w = GETMEMW(PC);
						PC++;
						SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+2);
						break;
					case 0x0021:	// STCP d,b
						w = GETMEMW(PC);
						PC++;
						SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
						SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+4);
						break;
						
						
						
					case 0x0022:	// PUB imm,-R7
					case 0x0023:	// PUW imm,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(PC));
						PC++;
						break;
					case 0x0024:	// PUD imm,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(PC));
						PC += 2;
						break;
					case 0x0025:	// PUSH [reg]B,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW(pKode + GETREGD(REG_7), (WORD)(BYTE)GETMEMB(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
						PC++;
						break;
					case 0x0026:	// PUSH [reg+mem/off]B,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMW(pKode + GETREGD(REG_7), (WORD)GETMEMB(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0027:	// PUSH [mem]B,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW(pKode + GETREGD(REG_7), (WORD)GETMEMB(pKode + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0028:	// PUSH [reg+mem+off]B,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMW(pKode + GETREGD(REG_7), (WORD)GETMEMB(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0029:	// PUSH [reg+reg]B,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						w = GETMEMW(PC);
						PC++;
						SETMEMW(pKode + GETREGD(REG_7), (WORD)GETMEMB(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF)));
						break;
					case 0x002A:	// PUSH [reg]W,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
						PC++;
						break;
					case 0x002B:	// PUSH [reg+mem/off]W,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x002C:	// PUSH [mem]W,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(pKode + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x002D:	// PUSH [reg+mem+off]W,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x002E:	// PUSH [reg+reg]W,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						w = GETMEMW(PC);
						PC++;
						SETMEMW(pKode + GETREGD(REG_7), GETMEMW(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF)));
						break;
					case 0x002F:	// PUSH [reg]D,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
						PC++;
						break;
					case 0x0030:	// PUSH [reg+mem/off]D,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0031:	// PUSH [mem]D,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(pKode + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0032:	// PUSH [reg+mem+off]D,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						break;
					case 0x0033:	// PUSH [reg+reg]D,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						w = GETMEMW(PC);
						PC++;
						SETMEMD(pKode + GETREGD(REG_7), GETMEMD(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF)));
						break;
					case 0x0034:	// POP [reg]B,R7+
						SETMEMB(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)), GETMEMB(pKode + GETREGD(REG_7)));
						PC++;
						SETREGD(REG_7, GETREGD(REG_7)-2);
						break;
					case 0x0035:	// POP [reg+mem/off]B,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMB(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMB(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0036:	// POP [mem]B,R7+
						SETMEMB(pKode + (long)(KLONG)GETMEMD(PC), GETMEMB(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0037:	// POP [reg+mem+off]B,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMB(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMB(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0038:	// POP [reg+reg]B,R7+
						w = GETMEMW(PC);
						PC++;
						SETMEMB(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF), GETMEMB(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0039:	// POP [reg]W,R7+
						SETMEMW(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)), GETMEMW(pKode + GETREGD(REG_7)));
						PC++;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x003A:	// POP [reg+mem/off]W,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMW(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMW(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x003B:	// POP [mem]W,R7+
						SETMEMW(pKode + (long)(KLONG)GETMEMD(PC), GETMEMW(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x003C:	// POP [reg+mem+off]W,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMW(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMW(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x003D:	// POP [reg+reg]W,R7+
						w = GETMEMW(PC);
						PC++;
						SETMEMW(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF), GETMEMW(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x003E:	// POP [reg]D,R7+
						SETMEMD(pKode + (long)(KLONG)GETREGD(GETMEMW(PC)), GETMEMD(pKode + GETREGD(REG_7)));
						PC++;
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x003F:	// POP [reg+mem/off]D,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						SETMEMD(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMD(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x0040:	// POP [mem]D,R7+
						SETMEMD(pKode + (long)(KLONG)GETMEMD(PC), GETMEMD(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x0041:	// POP [reg+mem+off]D,R7+
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						SETMEMD(pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC), GETMEMD(pKode + GETREGD(REG_7)));
						PC += 2;
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x0042:	// POP [reg+reg]D,R7+
						w = GETMEMW(PC);
						PC++;
						SETMEMD(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF), GETMEMD(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x0043:	// INC [reg]B
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMB(pMem, GETMEMB(pMem)+1);
						PC++;
						break;
					case 0x0044:	// INC [reg+mem/off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)+1);
						PC += 2;
						break;
					case 0x0045:	// INC [mem]B
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)+1);
						PC += 2;
						break;
					case 0x0046:	// INC [reg+mem+off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)+1);
						PC += 2;
						break;
					case 0x0047:	// INC [reg+reg]B
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMB(pMem, GETMEMB(pMem)+1);
						break;
					case 0x0048:	// INC [reg]W
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMW(pMem, GETMEMW(pMem)+1);
						PC++;
						break;
					case 0x0049:	// INC [reg+mem/off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)+1);
						PC += 2;
						break;
					case 0x004A:	// INC [mem]W
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)+1);
						PC += 2;
						break;
					case 0x004B:	// INC [reg+mem+off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)+1);
						PC += 2;
						break;
					case 0x004C:	// INC [reg+reg]W
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMW(pMem, GETMEMW(pMem)+1);
						break;
					case 0x004D:	// INC [reg]D
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMD(pMem, GETMEMD(pMem)+1);
						PC++;
						break;
					case 0x004E:	// INC [reg+mem/off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)+1);
						PC += 2;
						break;
					case 0x004F:	// INC [mem]D
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)+1);
						PC += 2;
						break;
					case 0x0050:	// INC [reg+mem+off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)+1);
						PC += 2;
						break;
					case 0x0051:	// INC [reg+reg]D
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMD(pMem, GETMEMD(pMem)+1);
						break;
					case 0x0052:	// DEC [reg]B
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMB(pMem, GETMEMB(pMem)-1);
						PC++;
						break;
					case 0x0053:	// DEC [reg+mem/off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)-1);
						PC += 2;
						break;
					case 0x0054:	// DEC [mem]B
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)-1);
						PC += 2;
						break;
					case 0x0055:	// DEC [reg+mem+off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, GETMEMB(pMem)-1);
						PC += 2;
						break;
					case 0x0056:	// DEC [reg+reg]B
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMB(pMem, GETMEMB(pMem)-1);
						break;
					case 0x0057:	// DEC [reg]W
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMW(pMem, GETMEMW(pMem)-1);
						PC++;
						break;
					case 0x0058:	// DEC [reg+mem/off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)-1);
						PC += 2;
						break;
					case 0x0059:	// DEC [mem]W
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)-1);
						PC += 2;
						break;
					case 0x005A:	// DEC [reg+mem+off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, GETMEMW(pMem)-1);
						PC += 2;
						break;
					case 0x005B:	// DEC [reg+reg]W
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMW(pMem, GETMEMW(pMem)-1);
						break;
					case 0x005C:	// DEC [reg]D
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMD(pMem, GETMEMD(pMem)-1);
						PC++;
						break;
					case 0x005D:	// DEC [reg+mem/off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)-1);
						PC += 2;
						break;
					case 0x005E:	// DEC [mem]D
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)-1);
						PC += 2;
						break;
					case 0x005F:	// DEC [reg+mem+off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, GETMEMD(pMem)-1);
						PC += 2;
						break;
					case 0x0060:	// DEC [reg+reg]D
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMD(pMem, GETMEMD(pMem)-1);
						break;
						
						
					case 0x0061:	// PUSC b,-R7	NOTE: we actually push a WORD to keep the stack WORD aligned
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW((WORD *)(pKode + GETREGD(REG_7)), (WORD)(BYTE)GETREGB(GETMEMW(PC++)));
						break;
					case 0x0062:	// PUSC w,-R7
						SETREGD(REG_7, GETREGD(REG_7)-2);
						SETMEMW((WORD *)(pKode + GETREGD(REG_7)), GETREGW(GETMEMW(PC++)));
						break;
					case 0x0063:	// PUSC d,-R7
						SETREGD(REG_7, GETREGD(REG_7)-4);
						SETMEMD(pKode + GETREGD(REG_7), GETREGD(GETMEMW(PC++)));
						break;
					case 0x0064:	// POPC b,R7+	NOTE: we actually pop a WORD, as that's what was pushed
						SETREGB(GETMEMW(PC++), GETMEMB(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0065:	// POPC w,R7+
						SETREGW(GETMEMW(PC++), GETMEMW((WORD *)(pKode + GETREGD(REG_7))));
						SETREGD(REG_7, GETREGD(REG_7)+2);
						break;
					case 0x0066:	// POPC d,R7+
						SETREGD(GETMEMW(PC++), GETMEMD(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+4);
						break;
					case 0x0067:	// JSRC =dval
#ifdef GEEK_TRACE
						if( TBptr+1 < TBMAX ) TraceBuf[TBptr++] = (DWORD)((BYTE *)PC-pKode);
#endif
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw = GETMEMD(PC);
						PC += 2;
						SETMEMD(pKode + GETREGD(REG_7), (DWORD)((BYTE *)PC - pKode));	// push return address
						PC = (WORD *)(pKode + dw);
						break;
					case 0x0068:	// JSIC =dval
#ifdef GEEK_TRACE
						if( TBptr+1 < TBMAX ) TraceBuf[TBptr++] = (DWORD)((BYTE *)PC-pKode);
#endif
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw = GETMEMD(pKode + (long)(KLONG)GETMEMD(PC));
						PC += 2;
						SETMEMD(pKode + GETREGD(REG_7), (DWORD)((BYTE *)PC - pKode));	// push return address
						PC = (WORD *)(pKode + dw);
						break;
					case 0x0069:	// JSRC d,=dval
#ifdef GEEK_TRACE
						if( TBptr+1 < TBMAX ) TraceBuf[TBptr++] = (DWORD)((BYTE *)PC-pKode);
#endif
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw = (DWORD)GETMEMW(PC++);
						dw = GETREGD(dw) + GETMEMD(PC);
						PC += 2;
						SETMEMD(pKode + GETREGD(REG_7), (DWORD)((BYTE *)PC - pKode));	// push return address
						PC = (WORD *)(pKode + dw);
						break;
					case 0x006A:	// JSIC d,=dval
#ifdef GEEK_TRACE
						if( TBptr+1 < TBMAX ) TraceBuf[TBptr++] = (DWORD)((BYTE *)PC-pKode);
#endif
						SETREGD(REG_7, GETREGD(REG_7)-4);
						dw = (DWORD)GETMEMW(PC++);
						dw = GETMEMD(pKode + (long)(KLONG)(GETREGD(dw) + (long)(KLONG)GETMEMD(PC)));
						PC += 2;
						SETMEMD(pKode + GETREGD(REG_7), (DWORD)((BYTE *)PC - pKode));	// push return address
						PC = (WORD *)(pKode + dw);
						break;
					case 0x006B:	// RETC
#ifdef GEEK_TRACE
						if( TBptr ) --TBptr;
#endif
						
						PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_7)));
						SETREGD(REG_7, GETREGD(REG_7)+4);	// pop return address
						break;
					case 0x0070:	// FNEG
						w = GETMEMW(PC++);
						dw = GETREGD(w);
						f1 = *((float *)&dw);
						f1 = -f1;
						SETREGD(w, *((DWORD *)&f1));
						break;
					case 0x0071:	// FADD
						w = GETMEMW(PC++);
						dw = GETREGD(w & 0x00FF);
						f1 = *((float *)&dw);
						dw = GETREGD((w>>8) & 0x00FF);
						f1 += *((float *)&dw);
						SETREGD(w & 0x00FF, *((DWORD *)&f1));
						break;
					case 0x0072:	// FSUB
						w = GETMEMW(PC++);
						dw = GETREGD(w & 0x00FF);
						f1 = *((float *)&dw);
						dw = GETREGD((w>>8) & 0x00FF);
						f1 -= *((float *)&dw);
						SETREGD(w & 0x00FF, *((DWORD *)&f1));
						break;
					case 0x0073:	// FMUL
						w = GETMEMW(PC++);
						dw = GETREGD(w & 0x00FF);
						f1 = *((float *)&dw);
						dw = GETREGD((w>>8) & 0x00FF);
						f1 *= *((float *)&dw);
						SETREGD(w & 0x00FF, *((DWORD *)&f1));
						break;
					case 0x0074:	// FDIV
						w = GETMEMW(PC++);
						dw = GETREGD(w & 0x00FF);
						f1 = *((float *)&dw);
						dw = GETREGD((w>>8) & 0x00FF);
						f1 /= *((float *)&dw);
						SETREGD(w & 0x00FF, *((DWORD *)&f1));
						break;
					case 0x0075:	// FF2I
						w = GETMEMW(PC++);
						dw = GETREGD(w);
						f1 = *((float *)&dw);
						SETREGD(w, (DWORD)(KLONG)f1);
						break;
					case 0x0076:	// FI2F
						w = GETMEMW(PC++);
						f1 = (float)(KLONG)GETREGD(w);
						SETREGD(w, *((DWORD *)&f1));
						break;
					case 0x0077:	// FCMP
						w = GETMEMW(PC++);
						dw = GETREGD(w & 0x00FF);
						f1 = *((float *)&dw);
						dw = GETREGD((w>>8) & 0x00FF);
						f1 -= *((float *)&dw);
						// NOTE: FCMP only sets GT, GE, LT, LE, EQ, NEQ.  "Even/odd" is undefined for FCMP.
						result = (f1<0) ? 0x80000000 : ((f1==0) ? 0 : 1);
						break;
						
						
					case 0x0078:	// PUF [r]
					case 0x0079:	// PUF [r][off]
					case 0x007A:	// PUF [mem]
					case 0x007B:	// PUF [r][mem][off]
					case 0x007C:	// PUF [r][r]
					case 0x007D:	// PUF r
					case 0x007E:	// PUF imm
					case 0x007F:	// PUF 0
						*(DWORD *)(pKode+20) = *(DWORD *)(pKode+24);
						*(DWORD *)(pKode+24) = *(DWORD *)(pKode+28);
						*(DWORD *)(pKode+28) = *(DWORD *)(pKode+32);
						*(DWORD *)(pKode+32) = *(DWORD *)(pKode+36);
						*(DWORD *)(pKode+36) = *(DWORD *)(pKode+40);
						*(DWORD *)(pKode+40) = *(DWORD *)(pKode+44);
						*(DWORD *)(pKode+44) = *(DWORD *)(pKode+48);
						switch( opcode & 0x00FF ) {
							case 0x0078:	// PUF [r]
								pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
								PC++;
								*(DWORD *)(pKode+48) = GETMEMD(pMem);
								break;
							case 0x0079:	// PUF [r][off]
								dw2 = GETREGD(GETMEMW(PC));
								PC++;
								pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								*(DWORD *)(pKode+48) = GETMEMD(pMem);
								break;
							case 0x007A:	// PUF [mem]
								pMem = pKode + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								*(DWORD *)(pKode+48) = GETMEMD(pMem);
								break;
							case 0x007B:	// PUF [r][mem][off]
								dw2 = GETREGD(GETMEMW(PC));
								PC++;
								dw2 += GETMEMD(PC);
								PC += 2;
								pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								*(DWORD *)(pKode+48) = GETMEMD(pMem);
								break;
							case 0x007C:	// PUF [r][r]
								w = GETMEMW(PC);
								PC++;
								pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
								*(DWORD *)(pKode+48) = GETMEMD(pMem);
								break;
							case 0x007D:	// PUF r
								*(DWORD *)(pKode+48) = GETREGD(GETMEMW(PC));
								PC++;
								break;
							case 0x007E:	// PUF imm
								*(DWORD *)(pKode+48) = GETMEMD(PC);
								PC += 2;
								break;
							case 0x007F:	// PUF 0
								*(float *)(pKode+48) = 0.0;
								break;
						}
						break;
					case 0x0080:	// POF [r]
					case 0x0081:	// POF [r][off]
					case 0x0082:	// POF [mem]
					case 0x0083:	// POF [r][mem][off]
					case 0x0084:	// POF [r][r]
					case 0x0085:	// POF r
					case 0x0086:	// POF once (discard)
					case 0x0087:	// POF twice (discard)
						dw = *(DWORD *)(pKode+48);
						*(DWORD *)(pKode+48) = *(DWORD *)(pKode+44);
						*(DWORD *)(pKode+44) = *(DWORD *)(pKode+40);
						*(DWORD *)(pKode+40) = *(DWORD *)(pKode+36);
						*(DWORD *)(pKode+36) = *(DWORD *)(pKode+32);
						*(DWORD *)(pKode+32) = *(DWORD *)(pKode+28);
						*(DWORD *)(pKode+28) = *(DWORD *)(pKode+24);
						*(DWORD *)(pKode+24) = *(DWORD *)(pKode+20);
						switch( opcode & 0x00FF ) {
							case 0x0080:	// POF [r]
								pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
								PC++;
								SETMEMD(pMem, dw);
								break;
							case 0x0081:	// POF [r][off]
								dw2 = GETREGD(GETMEMW(PC));
								PC++;
								pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								SETMEMD(pMem, dw);
								break;
							case 0x0082:	// POF [mem]
								pMem = pKode + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								SETMEMD(pMem, dw);
								break;
							case 0x0083:	// POF [r][mem][off]
								dw2 = GETREGD(GETMEMW(PC));
								PC++;
								dw2 += GETMEMD(PC);
								PC += 2;
								pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
								PC += 2;
								SETMEMD(pMem, dw);
								break;
							case 0x0084:	// POF [r][r]
								w = GETMEMW(PC);
								PC++;
								pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
								SETMEMD(pMem, dw);
								break;
							case 0x0085:	// POF r
								SETREGD(GETMEMW(PC), dw);
								PC++;
								break;
							case 0x0086:	// POF ONCE (DISCARD)
								break;
							case 0x0087:	// POF TWICE (DISCARD)
							fpop:
								*(DWORD *)(pKode+48) = *(DWORD *)(pKode+44);
								*(DWORD *)(pKode+44) = *(DWORD *)(pKode+40);
								*(DWORD *)(pKode+40) = *(DWORD *)(pKode+36);
								*(DWORD *)(pKode+36) = *(DWORD *)(pKode+32);
								*(DWORD *)(pKode+32) = *(DWORD *)(pKode+28);
								*(DWORD *)(pKode+28) = *(DWORD *)(pKode+24);
								*(DWORD *)(pKode+24) = *(DWORD *)(pKode+20);
								break;
						}
						break;
					case 0x0088:	// ADF
						*(float *)(pKode+44) += *(float *)(pKode+48);
						goto fpop;
					case 0x0089:	// SUF
						*(float *)(pKode+44) -= *(float *)(pKode+48);
						goto fpop;
					case 0x008A:	// MUF
						*(float *)(pKode+44) *= *(float *)(pKode+48);
						goto fpop;
					case 0x008B:	// DIF
						*(float *)(pKode+44) /= *(float *)(pKode+48);
						goto fpop;
					case 0x008C:	// CMF
						f1 = *((float *)(pKode+44)) - *((float *)(pKode+48));
						// NOTE: CMF only sets GT, GE, LT, LE, EQ, NEQ.  "Even/odd" is undefined for CMF.
						result = (f1<0) ? 0x80000000 : ((f1==0) ? 0 : 1);
						*(DWORD *)(pKode+48) = *(DWORD *)(pKode+40);	// pop (discard) 2 values
						*(DWORD *)(pKode+44) = *(DWORD *)(pKode+36);
						*(DWORD *)(pKode+40) = *(DWORD *)(pKode+32);
						*(DWORD *)(pKode+36) = *(DWORD *)(pKode+28);
						*(DWORD *)(pKode+32) = *(DWORD *)(pKode+24);
						*(DWORD *)(pKode+28) = *(DWORD *)(pKode+20);
						break;
					case 0x008D:	// NEF
						*(float *)(pKode+48) = -*(float *)(pKode+48);
						break;
					case 0x008E:	// I2F
						*(DWORD *)(pKode+20) = *(DWORD *)(pKode+24);
						*(DWORD *)(pKode+24) = *(DWORD *)(pKode+28);
						*(DWORD *)(pKode+28) = *(DWORD *)(pKode+32);
						*(DWORD *)(pKode+32) = *(DWORD *)(pKode+36);
						*(DWORD *)(pKode+36) = *(DWORD *)(pKode+40);
						*(DWORD *)(pKode+40) = *(DWORD *)(pKode+44);
						*(DWORD *)(pKode+44) = *(DWORD *)(pKode+48);
						*(float *)(pKode+48) = (float)(KLONG)GETREGD(GETMEMW(PC++));
						break;
					case 0x008F:	// F2I
						w = GETMEMW(PC++);
						f1 = *(float *)(pKode+48);
						SETREGD(w, (DWORD)(KLONG)f1);
						goto fpop;
						
						
                    case 0x0090:    // R16 = KFADD(R15, R14)
                        dw = GETREGD(REG_15);
                        dw2= GETREGD(REG_14);
                        kf = KFadd(*((KFf4*)&dw), *((KFf4*)&dw2));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;
                    case 0x0091:    // R16 = KFSUB(R15, R14)
                        dw = GETREGD(REG_15);
                        dw2= GETREGD(REG_14);
                        kf = KFsub(*((KFf4*)&dw), *((KFf4*)&dw2));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;
                    case 0x0092:    // R16 = KFMUL(R15, R14)
                        dw = GETREGD(REG_15);
                        dw2= GETREGD(REG_14);
       //        kf = *((KFf4*)&dw);
       //        kf = *((KFf4*)&dw2);
                        kf = KFmul(*((KFf4*)&dw), *((KFf4*)&dw2));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;
                    case 0x0093:    // R16 = KFDIV(R15, R14)
                        dw = GETREGD(REG_15);
                        dw2= GETREGD(REG_14);
                        kf = KFdiv(*((KFf4*)&dw), *((KFf4*)&dw2));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;
                    case 0x0094:    // R16 = KFNEG(R15)
                        dw = GETREGD(REG_15);
                        kf = KFneg(*((KFf4*)&dw));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;
                    case 0x0095:    // R16 = KFSQR(R15)
                        dw = GETREGD(REG_15);
                        kf = KFsqr(*((KFf4*)&dw));
                       SETREGD(REG_16, *((DWORD*)&kf));
                       break;




						
						
					case 0x00AA:	// NEG [reg]B
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMB(pMem, -(char)GETMEMB(pMem));
						PC++;
						break;
					case 0x00AB:	// NEG [reg+mem/off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, -(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00AC:	// NEG [mem]B
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, -(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00AD:	// NEG [reg+mem+off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, -(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00AE:	// NEG [reg+reg]B
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMB(pMem, -(char)GETMEMB(pMem));
						break;
					case 0x00AF:	// NEG [reg]W
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMW(pMem, -(KSHORT)GETMEMW(pMem));
						PC++;
						break;
					case 0x00B0:	// NEG [reg+mem/off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, -(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00B1:	// NEG [mem]W
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, -(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00B2:	// NEG [reg+mem+off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, -(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00B3:	// NEG [reg+reg]W
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMW(pMem, -(KSHORT)GETMEMW(pMem));
						break;
					case 0x00B4:	// NEG [reg]D
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMD(pMem, -(KLONG)GETMEMD(pMem));
						PC++;
						break;
					case 0x00B5:	// NEG [reg+mem/off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, -(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00B6:	// NEG [mem]D
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, -(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00B7:	// NEG [reg+mem+off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, -(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00B8:	// NEG [reg+reg]D
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMD(pMem, -(KLONG)GETMEMD(pMem));
						break;
					case 0x00B9:	// NOT [reg]B
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMB(pMem, ~(char)GETMEMB(pMem));
						PC++;
						break;
					case 0x00BA:	// NOT [reg+mem/off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, ~(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00BB:	// NOT [mem]B
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, ~(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00BC:	// NOT [reg+mem+off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMB(pMem, ~(char)GETMEMB(pMem));
						PC += 2;
						break;
					case 0x00BD:	// NOT [reg+reg]B
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMB(pMem, ~(char)GETMEMB(pMem));
						break;
					case 0x00BE:	// NOT [reg]W
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMW(pMem, ~(KSHORT)GETMEMW(pMem));
						PC++;
						break;
					case 0x00BF:	// NOT [reg+mem/off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, ~(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00C0:	// NOT [mem]W
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, ~(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00C1:	// NOT [reg+mem+off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMW(pMem, ~(KSHORT)GETMEMW(pMem));
						PC += 2;
						break;
					case 0x00C2:	// NOT [reg+reg]W
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMW(pMem, ~(KSHORT)GETMEMW(pMem));
						break;
					case 0x00C3:	// NOT [reg]D
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						SETMEMD(pMem, ~(KLONG)GETMEMD(pMem));
						PC++;
						break;
					case 0x00C4:	// NOT [reg+mem/off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, ~(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00C5:	// NOT [mem]D
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, ~(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00C6:	// NOT [reg+mem+off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						SETMEMD(pMem, ~(KLONG)GETMEMD(pMem));
						PC += 2;
						break;
					case 0x00C7:	// NOT [reg+reg]D
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						SETMEMD(pMem, ~(KLONG)GETMEMD(pMem));
						break;
						
					case 0x00C8:	// TST [reg]B
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						PC++;
						result = (DWORD)(char)GETMEMB(pMem);
						result |= (result<<24);
						break;
					case 0x00C9:	// TST [reg+mem/off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(char)GETMEMB(pMem);
						result |= (result<<24);
						break;
					case 0x00CA:	// TST [mem]B
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(char)GETMEMB(pMem);
						result |= (result<<24);
						break;
					case 0x00CB:	// TST [reg+mem+off]B
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(char)GETMEMB(pMem);
						result |= (result<<24);
						break;
					case 0x00CC:	// TST [reg+reg]B
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						result = (DWORD)(char)GETMEMB(pMem);
						result |= (result<<24);
						break;
					case 0x00CD:	// TST [reg]W
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						PC++;
						result = (DWORD)(KSHORT)GETMEMW(pMem);
						result |= (result<<16);
						break;
					case 0x00CE:	// TST [reg+mem/off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(KSHORT)GETMEMW(pMem);
						result |= (result<<16);
						break;
					case 0x00CF:	// TST [mem]W
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(KSHORT)GETMEMW(pMem);
						result |= (result<<16);
						break;
					case 0x00D0:	// TST [reg+mem+off]W
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = (DWORD)(KSHORT)GETMEMW(pMem);
						result |= (result<<16);
						break;
					case 0x00D1:	// TST [reg+reg]W
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						result = (DWORD)(KSHORT)GETMEMW(pMem);
						result |= (result<<16);
						break;
					case 0x00D2:	// TST [reg]D
						pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
						PC++;
						result = GETMEMD(pMem);
						break;
					case 0x00D3:	// TST [reg+mem/off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = GETMEMD(pMem);
						break;
					case 0x00D4:	// TST [mem]D
						pMem = pKode + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = GETMEMD(pMem);
						break;
					case 0x00D5:	// TST [reg+mem+off]D
						dw2 = GETREGD(GETMEMW(PC));
						PC++;
						dw2 += GETMEMD(PC);
						PC += 2;
						pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
						PC += 2;
						result = GETMEMD(pMem);
						break;
					case 0x00D6:	// TST [reg+reg]D
						w = GETMEMW(PC);
						PC++;
						pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
						result = GETMEMD(pMem);
						break;
// D8-DF PU8    (push double on the double math stack
// E0-E7 PO8    (pop double)
// E8 AD8        (add double)
// E9 SU8        (sub double)
// EA MU8        (mul double)
// EB DI8        (div double)
// EC CM8        (cmp double)
// ED NE8        (neg double)
// EE I2D        (convert long long to double)
// EF D2I        (convert double to long long)
                     case 0x00D8:    // PU8 [r]
                     case 0x00D9:    // PU8 [r][off]
                     case 0x00DA:    // PU8 [mem]
                     case 0x00DB:    // PU8 [r][mem][off]
                     case 0x00DC:    // PU8 [r][r]
                     case 0x00DD:    // PU8 r
                     case 0x00DE:    // PU8 imm
                     case 0x00DF:    // PU8 0
                        *(QWORD volatile *)(pKode+128) = *(QWORD volatile *)(pKode+136);
                        *(QWORD volatile *)(pKode+136) = *(QWORD volatile *)(pKode+144);
                        *(QWORD volatile *)(pKode+144) = *(QWORD volatile *)(pKode+152);
                        *(QWORD volatile *)(pKode+152) = *(QWORD volatile *)(pKode+160);
                        *(QWORD volatile *)(pKode+160) = *(QWORD volatile *)(pKode+168);
                        *(QWORD volatile *)(pKode+168) = *(QWORD volatile *)(pKode+176);
                        *(QWORD volatile *)(pKode+176) = *(QWORD volatile *)(pKode+184);
                        switch( opcode & 0x00FF ) {
                         case 0x00D8:    // PU8 [r]
                            pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                            ++PC;
                            *(QWORD volatile *)(pKode+184) = GETMEMQ(pMem);
                            break;
                         case 0x00D9:    // PU8 [r][off]
                            dw2 = GETREGD(GETMEMW(PC));
                            PC++;
                            pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            *(QWORD volatile *)(pKode+184) = GETMEMQ(pMem);
                            break;
                         case 0x00DA:    // PU8 [mem]
                            pMem = pKode + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            qw = GETMEMQ(pMem);
                            *(QWORD volatile *)(pKode+184) = qw;
                            break;
                         case 0x00DB:    // PU8 [r][mem][off]
                            dw2 = GETREGD(GETMEMW(PC));
                            PC++;
                            dw2 += GETMEMD(PC);
                            PC += 2;
                            pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            *(QWORD volatile *)(pKode+184) = GETMEMQ(pMem);
                            break;
                         case 0x00DC:    // PU8 [r][r]
                            w = GETMEMW(PC);
                            PC++;
                            pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                            *(QWORD volatile *)(pKode+184) = GETMEMQ(pMem);
                            break;
                         case 0x00DD:    // PU8 r
                            *(QWORD volatile *)(pKode+184) = GETREGQ(GETMEMW(PC));
                            PC++;
                            break;
                         case 0x00DE:    // PU8 imm
                            *(QWORD volatile *)(pKode+184) = GETMEMQ(PC);
                            PC += 2;
                            break;
                         case 0x00DF:    // PU8 0
                            *(double volatile *)(pKode+184) = 0.0;
                            break;
                        }
                        break;
                     case 0x00E0:    // PO8 [r]
                     case 0x00E1:    // PO8 [r][off]
                     case 0x00E2:    // PO8 [mem]
                     case 0x00E3:    // PO8 [r][mem][off]
                     case 0x00E4:    // PO8 [r][r]
                     case 0x00E5:    // PO8 r
                     case 0x00E6:    // PO8 once (discard)
                     case 0x00E7:    // PO8 twice (discard)
                        qw = *(QWORD *)(pKode+184);
                        *(QWORD volatile *)(pKode+184) = *(QWORD volatile *)(pKode+176);
                        *(QWORD volatile *)(pKode+176) = *(QWORD volatile *)(pKode+168);
                        *(QWORD volatile *)(pKode+168) = *(QWORD volatile *)(pKode+160);
                        *(QWORD volatile *)(pKode+160) = *(QWORD volatile *)(pKode+152);
                        *(QWORD volatile *)(pKode+152) = *(QWORD volatile *)(pKode+144);
                        *(QWORD volatile *)(pKode+144) = *(QWORD volatile *)(pKode+136);
                        *(QWORD volatile *)(pKode+136) = *(QWORD volatile *)(pKode+128);
                        switch( opcode & 0x00FF ) {
                         case 0x00E0:    // PO8 [r]
                            pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                            PC++;
                            SETMEMQ(pMem, qw);
                            break;
                         case 0x00E1:    // PO8 [r][off]
                            dw2 = GETREGD(GETMEMW(PC));
                            PC++;
                            pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            SETMEMQ(pMem, qw);
                            break;
                         case 0x00E2:    // PO8 [mem]
                            pMem = pKode + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            SETMEMQ(pMem, qw);
                            break;
                         case 0x00E3:    // PO8 [r][mem][off]
                            dw2 = GETREGD(GETMEMW(PC));
                            PC++;
                            dw2 += (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                            PC += 2;
                            SETMEMQ(pMem, qw);
                            break;
                         case 0x00E4:    // PO8 [r][r]
                            w = GETMEMW(PC);
                            PC++;
                            pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                            SETMEMQ(pMem, qw);
                            break;
                         case 0x00E5:    // PO8 r
                            SETREGQ(GETMEMW(PC), qw);
                            PC++;
                            break;
                         case 0x00E6:    // POF ONCE (DISCARD)
                            break;
                         case 0x00E7:    // POF TWICE (DISCARD)
        pop8:
                            *(QWORD volatile *)(pKode+184) = *(QWORD volatile *)(pKode+176);
                            *(QWORD volatile *)(pKode+176) = *(QWORD volatile *)(pKode+168);
                            *(QWORD volatile *)(pKode+168) = *(QWORD volatile *)(pKode+160);
                            *(QWORD volatile *)(pKode+160) = *(QWORD volatile *)(pKode+152);
                            *(QWORD volatile *)(pKode+152) = *(QWORD volatile *)(pKode+144);
                            *(QWORD volatile *)(pKode+144) = *(QWORD volatile *)(pKode+136);
                            *(QWORD volatile *)(pKode+136) = *(QWORD volatile *)(pKode+128);
                            break;
                        }
                        break;
                     case 0x00E8:    // AD8
                        *(double volatile *)(pKode+176) += *(double volatile *)(pKode+184);
                        goto pop8;
                     case 0x00E9:    // SU8
                        *(double volatile *)(pKode+176) -= *(double volatile *)(pKode+184);
                        goto pop8;
                     case 0x00EA:    // MU8
                        *(double volatile *)(pKode+176) *= *(double volatile *)(pKode+184);
                        goto pop8;
                     case 0x00EB:    // DI8
                        *(double volatile *)(pKode+176) /= *(double volatile *)(pKode+184);
                        goto pop8;
                     case 0x00EC:    // CM8
                        d1 = *((double volatile *)(pKode+176)) - *((double volatile *)(pKode+184));
                        // NOTE: CMF only sets GT, GE, LT, LE, EQ, NEQ.  "Even/odd" is undefined for CM8.
                        result = (d1<0.0) ? 0x80000000 : ((d1==0.0) ? 0 : 1);
                        *(QWORD volatile *)(pKode+184) = *(QWORD volatile *)(pKode+168);    // pop (discard) 2 values
                        *(QWORD volatile *)(pKode+176) = *(QWORD volatile *)(pKode+160);
                        *(QWORD volatile *)(pKode+168) = *(QWORD volatile *)(pKode+152);
                        *(QWORD volatile *)(pKode+160) = *(QWORD volatile *)(pKode+144);
                        *(QWORD volatile *)(pKode+152) = *(QWORD volatile *)(pKode+136);
                        *(QWORD volatile *)(pKode+144) = *(QWORD volatile *)(pKode+128);
                        break;
                     case 0x00ED:    // NE8
                        *(double volatile *)(pKode+184) = -*(double volatile *)(pKode+184);
                        break;
                     case 0x00EE:    // I2D
                        *(QWORD volatile *)(pKode+128) = *(QWORD volatile *)(pKode+136);
                        *(QWORD volatile *)(pKode+136) = *(QWORD volatile *)(pKode+144);
                        *(QWORD volatile *)(pKode+144) = *(QWORD volatile *)(pKode+152);
                        *(QWORD volatile *)(pKode+152) = *(QWORD volatile *)(pKode+160);
                        *(QWORD volatile *)(pKode+160) = *(QWORD volatile *)(pKode+168);
                        *(QWORD volatile *)(pKode+168) = *(QWORD volatile *)(pKode+176);
                        *(QWORD volatile *)(pKode+176) = *(QWORD volatile *)(pKode+184);
                        qw = GETREGQ(GETMEMW(PC));
                        *(double volatile *)(pKode+184) = (double)(long long)qw;
                        ++PC;
                        break;
                     case 0x00EF:    // D2I
                        w = GETMEMW(PC);
                        ++PC;
                        d1 = *(double volatile *)(pKode+184);
                        SETREGQ(w, (QWORD)(long long)d1);
                        goto pop8;
                     case 0x00F0:    // F2D
                        w = GETMEMW(PC);
                        ++PC;
                        dw = GETREGD(w);
                        f1 = *((float*)&dw);
                        d1 = (double)f1;
                        SETREGQ(w, *((QWORD*)&d1));
                        break;
                     case 0x00F1:    // D2F
                        w = GETMEMW(PC);
                        ++PC;
                        qw = GETREGQ(w);
                        d1 = *((double*)&qw);
                        f1 = (float)d1;
                        SETREGD(w, *((DWORD*)&f1));
                        break;
				}
				break;
            case 0xFF:    // mostly QWORD ops
               switch( opcode & 0x00FF ) {
                case 0x0000:    // SHL q,b
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = (dr>>8)&0x0FF;
                   dr &= 0x0FF;
                   qw = SETREGQ(w, GETREGQ(w)<<(BYTE)(GETREGB(dr) & 0x003F));
                   goto setqwresult;
                case 0x0001:    // SHR q,b
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = (dr>>8)&0x0FF;
                   dr &= 0x0FF;
                   if( signedmath ) qw = SETREGQ(w, ((long long)GETREGQ(w))>>((BYTE)GETREGB(dr) & 0x003F));
                   else qw = SETREGQ(w, GETREGQ(w)>>((BYTE)GETREGB(dr) & 0x003F));
   setqwresult:
                   if( qw==0 ) result = 0;
                   else if( qw<0 ) result = 0x80000000 | ((DWORD)qw & 1);
                   else result = 2 | ((DWORD)qw & 1);
                   break;
                case 0x0002:    // ROL q,b
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = GETREGB(dr & 0x0FF) & 0x3F;
                   dr = (dr>>8) & 0x0FF;
                   qw = GETREGQ(dr);
                   while( w-- ) {
                       qw2 = qw;
                       qw <<= 1;
                       if( qw2 & 0x8000000000000000 ) qw |= 1;
                   }
                   SETREGQ(dr, qw);
                   goto setqwresult;
                case 0x0003:    // ROR q,b
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = GETREGB(dr & 0x0FF) & 0x3F;
                   dr = (dr>>8) & 0x0FF;
                   qw = GETREGQ(dr);
                   while( w-- ) {
                       qw2 = qw;
                       qw >>= 1;
                       if( qw2 & 1 ) qw |= 0x8000000000000000;
                   }
                   SETREGQ(dr, qw);
                   goto setqwresult;
                case 0x0004:    // ADD q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(dr)+GETREGQ(w));
                   break;
                case 0x0005:    // SUB q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   if( signedmath ) qw = SETREGQ(dr, ((long long)GETREGQ(dr))-((long long)GETREGQ(w)));
                   else qw = SETREGQ(dr, GETREGQ(dr)-GETREGQ(w));
                   goto setqwresult;
                case 0x0006:    // MUL q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(dr)*GETREGQ(w));
                   break;
                case 0x0007:    // DIV q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   if( signedmath ) SETREGQ(dr, ((long long)GETREGQ(dr))/((long long)GETREGQ(w)));
                   else SETREGQ(dr, GETREGQ(dr)/GETREGQ(w));
                   break;
                case 0x0008:    // MOD q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   if( signedmath ) SETREGQ(dr, ((long long)GETREGQ(dr))%((long long)GETREGQ(w)));
                   else SETREGQ(dr, GETREGQ(dr)%GETREGQ(w));
                   break;
                case 0x0009:    // OR q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(dr) | GETREGQ(w));
                   break;
                case 0x000A:    // AND q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(dr) & GETREGQ(w));
                   break;
                case 0x000B:    // XOR q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(dr) ^ GETREGQ(w));
                   break;
                case 0x000C:    // CMP q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   if( signedmath ) qw = ((long long)GETREGQ(dr))-((long long)GETREGQ(w));
                   else qw = GETREGQ(dr)-GETREGQ(w);
                   goto setqwresult;
                case 0x000D:    // TST q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   qw = GETREGQ(dr) & GETREGQ(w);
                   goto setqwresult;
                case 0x000E:    // XCH q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   qw = GETREGQ(dr);
                   SETREGQ(dr, GETREGQ(w));
                   SETREGQ(w, qw);
                   break;
                case 0x000F:    // LDR q,q
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETREGQ(w));
                   break;
                case 0x0010:    // LDE q,b
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, (QWORD)(BYTE)GETREGB(w));
                   break;
                case 0x0011:    // LDE q,w
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, (QWORD)(WORD)GETREGW(w));
                   break;
                case 0x0012:    // LDE q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, (QWORD)(DWORD)GETREGD(w));
                   break;
                case 0x0013:    // LD0 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(w)));
                   break;
                case 0x0014:    // LD1 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                       // not on MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_1) + GETREGD(w)));
                   break;
                case 0x0015:    // LD2 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// not on MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_2) + GETREGD(w)));
                   break;
                case 0x0016:    // LDS q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(REG_4) + GETREGD(w)));
                   break;
                case 0x0017:    // ST0 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + GETREGD(w), GETREGQ(dr));
                   break;
                case 0x0018:    // ST1 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// unsupported on Mac
//                  SETMEMQ(GETREGD(REG_1) + GETREGD(w), GETREGQ(dr));
                   break;
                case 0x0019:    // ST2 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// unsupported on Mac
//                   SETMEMQ(GETREGD(REG_2) + GETREGD(w), GETREGQ(dr));
                   break;
                case 0x001A:    // STS q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + GETREGD(REG_4) + GETREGD(w), GETREGQ(dr));
                   break;
                case 0x001B:    // PU0 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + GETREGD(w), GETREGQ(dr));
                   SETREGD(w, GETREGD(w)+8);
                   break;
                case 0x001C:    // PO0 q,d
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGD(w, GETREGD(w)-8);
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(w)));
                   break;
                case 0x001D:    // LD0 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(w)) + (long)(KLONG)GETMEMD(PC));
                   PC += 2;
                   break;
                case 0x001E:    // LD1 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// no MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x001F:    // LD2 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// no MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0020:    // LDS q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0021:    // ST0 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0022:    // ST1 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// no MAC                   SETMEMQ(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0023:    // ST2 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
// no MAC                   SETMEMQ(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0024:    // STS q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + GETREGD(REG_4) + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0025:    // PU0 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETMEMQ(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   SETREGD(w, GETREGD(w)+8);
                   break;
                case 0x0026:    // PO0 q,d,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = dr & 0x0FF;
                   dr = (dr>>8) & 0x0FF;
                   SETREGD(w, GETREGD(w)-8);
                   SETREGQ(dr, GETMEMQ(pKode + (long)(KLONG)GETREGD(w) + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0027:    // RET q,=0
   #ifdef GEEK_TRACE2
       if( TBptr2 ) --TBptr2;
   #endif
                   SETREGD(REG_5, GETREGD(REG_5)-4);    // pop return address
                   PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
                   dr = GETMEMW(PC);
                   result = (DWORD)SETREGQ(dr, 0);
                   break;
                case 0x0028:    // RET q,=1
   #ifdef GEEK_TRACE2
       if( TBptr2 ) --TBptr2;
   #endif
                   SETREGD(REG_5, GETREGD(REG_5)-4);    // pop return address
                   PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
                   dr = GETMEMW(PC);
                   result = (DWORD)SETREGQ(dr, 1);
                   break;
                case 0x0029:    // INC q
                   dr = GETMEMW(PC);
                   qw = SETREGQ(dr, GETREGQ(dr) + 1);
                   goto setqwresult;
                case 0x002A:    // DEC q
                   dr = GETMEMW(PC);
                   qw = SETREGQ(dr, GETREGQ(dr) - 1);
                   goto setqwresult;
                case 0x002B:    // NOT q
                   dr = GETMEMW(PC);
                   qw = SETREGQ(dr, ~GETREGQ(dr));
                   goto setqwresult;
                case 0x002C:    // NEG q
                   dr = GETMEMW(PC);
                   qw = SETREGQ(dr, (QWORD)(-(long long)GETREGQ(dr)));
                   goto setqwresult;
                case 0x002D:    // PUSH q
                   SETMEMQ(pKode + GETREGD(REG_5), GETREGQ(GETMEMW(PC)));
                   ++PC;
                   SETREGD(REG_5, GETREGD(REG_5)+8);
                   break;
                case 0x002E:    // POP q
                   SETREGD(REG_5, GETREGD(REG_5)-8);
                   SETREGQ(GETMEMW(PC), GETMEMQ(pKode + GETREGD(REG_5)));
                   ++PC;
                   break;
                case 0x002F:    // LD0 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETMEMQ(pKode + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0030:    // LD1 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
// no MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_1) + GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0031:    // LD2 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
// no MAC                   SETREGQ(dr, GETMEMQ(GETREGD(REG_2) + GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0032:    // LDS q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETMEMQ(pKode + GETREGD(REG_4) + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0033:    // ST0 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETMEMQ(pKode + (long)(KLONG)GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0034:    // ST1 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
// no MAC                   SETMEMQ(GETREGD(REG_1) + GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0035:    // ST2 q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
// no MAC                   SETMEMQ(GETREGD(REG_2) + GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0036:    // STS q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETMEMQ(pKode + GETREGD(REG_4) + (long)(KLONG)GETMEMD(PC), GETREGQ(dr));
                   PC += 2;
                   break;
                case 0x0037:    // IJMP q,=dval
                   dw2 = GETREGD(GETMEMW(PC));
                   ++PC;
                   goto ijmp;
                case 0x0038:    // KJMP q,=dval
                   dw2 = GETREGD(GETMEMW(PC));
                   ++PC;
                   goto kjmp;
                case 0x0039:    // CBQ q
                    dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, (QWORD)(long long)(char)GETREGB(dr));
                   break;
                case 0x003A:    // CWQ q
                    dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, (QWORD)(long long)(short)GETREGW(dr));
                   break;
                case 0x003B:    // CDQ q
                    dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, (QWORD)(long long)(long)GETREGD(dr));
                   break;
                case 0x003C:    // LDI q,=qval
                    dr = GETMEMW(PC);
                   ++PC;
                    SETREGQ(dr, GETMEMQ(PC));
                    PC += 4;
                   break;
                case 0x003D:    // SHL q,=wval
                   dr = GETMEMW(PC);
                   ++PC;
                   qw = SETREGQ(dr, GETREGQ(dr)<<(BYTE)(GETMEMW(PC) & 0x003F));
                   ++PC;
                   goto setqwresult;
                case 0x003E:    // SHR q,=wval
                   dr = GETMEMW(PC);
                   ++PC;
                   if( signedmath ) qw = SETREGQ(dr, ((long long)GETREGQ(dr))>>(((BYTE)GETMEMW(PC)) & 0x003F));
                   else qw = SETREGQ(dr, GETREGQ(dr)>>(((BYTE)GETMEMW(PC)) & 0x003F));
                   ++PC;
                   goto setqwresult;
                case 0x003F:    // ROL q,=wval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = GETMEMW(PC) & 0x3F;
                   ++PC;
                   qw = GETREGQ(dr);
                   while( w-- ) {
                       qw2 = qw;
                       qw <<= 1;
                       if( qw2 & 0x8000000000000000 ) qw |= 1;
                   }
                   SETREGQ(dr, qw);
                   goto setqwresult;
                case 0x0040:    // ROR q,=wval
                   dr = GETMEMW(PC);
                   ++PC;
                   w  = GETMEMW(PC) & 0x3F;
                   ++PC;
                   qw = GETREGQ(dr);
                   while( w-- ) {
                       qw2 = qw;
                       qw >>= 1;
                       if( qw2 & 1 ) qw |= 0x8000000000000000;
                   }
                   SETREGQ(dr, qw);
                   goto setqwresult;
                case 0x0041:    // ADD q,=qval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETREGQ(dr)+GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0042:    // SUB q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   if( signedmath ) qw = SETREGQ(dr, ((long long)GETREGQ(dr))-((long long)GETMEMQ(PC)));
                   else qw = SETREGQ(dr, GETREGQ(dr)-GETMEMQ(PC));
                   PC += 4;
                   goto setqwresult;
                case 0x0043:    // MUL q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETREGQ(dr)*GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0044:    // DIV q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   if( signedmath ) qw = SETREGQ(dr, ((long long)GETREGQ(dr))/((long long)GETMEMQ(PC)));
                   else qw = SETREGQ(dr, GETREGQ(dr)/GETMEMQ(PC));
                   PC += 4;
                   goto setqwresult;
                case 0x0045:    // MOD q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   if( signedmath ) SETREGQ(dr, ((long long)GETREGQ(dr))%((long long)GETMEMQ(PC)));
                   else SETREGQ(dr, GETREGQ(dr)%GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0046:    // OR q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETREGQ(dr)|GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0047:    // AND q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETREGQ(dr)&GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0048:    // XOR q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   SETREGQ(dr, GETREGQ(dr)^GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0049:    // CMP q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   if( signedmath ) qw = (QWORD)(((long long)GETREGQ(dr))-((long long)GETMEMQ(PC)));
                   else qw = GETREGQ(dr)-GETMEMQ(PC);
                   PC += 4;
                   goto setqwresult;
                case 0x004A:    // TST q,=dval
                   dr = GETMEMW(PC);
                   ++PC;
                   qw = GETREGQ(dr)&GETMEMQ(PC);
                   PC += 4;
                   goto setqwresult;
                case 0x004B:    // LDCM q,b
                   break;
                case 0x004C:    // LDC q,b
                   break;
                case 0x004D:    // LDCP q,b
                   break;
                case 0x004E:    // STCM q,b
                   break;
                case 0x004F:    // STC q,b
                   break;
                case 0x0050:    // STCP q,b
                   break;
                case 0x0051:    // PUQ qval (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   SETMEMD(pKode + GETREGD(REG_7), GETMEMQ(PC));
                   PC += 4;
                   break;
                case 0x0052:    // PUSH[r]q (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   SETMEMQ(pKode + GETREGD(REG_7), GETMEMQ(pKode + (long)(KLONG)GETREGD(GETMEMW(PC))));
                   PC++;
                   break;
                case 0x0053:    // PUSH [r+x]q (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   SETMEMQ(pKode + GETREGD(REG_7), GETMEMQ(pKode + dw2 + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0054:    // PUSH [x]q (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   SETMEMQ(pKode + GETREGD(REG_7), GETMEMQ(pKode + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0055:    // PUSH [r+x+y]q (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += GETMEMD(PC);
                   PC += 2;
                   SETMEMQ(pKode + GETREGD(REG_7), GETMEMQ(pKode + dw2 + (long)(KLONG)GETMEMD(PC)));
                   PC += 2;
                   break;
                case 0x0056:    // PUSH [r+r]q (r7)
                   SETREGD(REG_7, GETREGD(REG_7)-8);
                   w = GETMEMW(PC);
                   PC++;
                   SETMEMQ(pKode + GETREGD(REG_7), GETMEMQ(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF)));
                   break;
                case 0x0057:    // POP [r]q (r7)
                   SETMEMQ(pKode + GETREGD(GETMEMW(PC)), GETMEMQ(pKode + GETREGD(REG_7)));
                   PC++;
                   SETREGD(REG_7, GETREGD(REG_7)+8);
                   break;
                case 0x0058:    // POP [r+x]q (r7)
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   SETMEMQ(pKode + dw2 + GETMEMD(PC), GETMEMQ(pKode + GETREGD(REG_7)));
                   PC += 2;
                   SETREGD(REG_7, GETREGD(REG_7)+8);
                   break;
                case 0x0059:    // POP [x]q (r7)
                   SETMEMQ(pKode + (long)(KLONG)GETMEMD(PC), GETMEMD(pKode + GETREGD(REG_7)));
                   PC += 2;
                   SETREGD(REG_7, GETREGD(REG_7)+8);
                   break;
                case 0x005A:    // POP [r+x+y]q (r7)
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += GETMEMD(PC);
                   PC += 2;
                   SETMEMQ(pKode + dw2 + (long)(KLONG)GETMEMD(PC), GETMEMQ(pKode + GETREGD(REG_7)));
                   PC += 2;
                   SETREGD(REG_7, GETREGD(REG_7)+8);
                   break;
                case 0x005B:    // POP [r+r]q (r7)
                   w = GETMEMW(PC);
                   PC++;
                   SETMEMQ(pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF), GETMEMQ(pKode + GETREGD(REG_7)));
                   SETREGD(REG_7, GETREGD(REG_7)+8);
                   break;
                case 0x005C:    // INC [r]q
                   pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                   SETMEMQ(pMem, GETMEMQ(pMem)+1);
                   PC++;
                   break;
                case 0x005D:    // INC [r+x]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   pMem = pKode + dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)+1);
                   PC += 2;
                   break;
                case 0x005E:    // INC [x]q
                   pMem = pKode + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)+1);
                   PC += 2;
                   break;
                case 0x005F:    // INC [r+x+y]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += GETMEMD(PC);
                   PC += 2;
                   pMem = pKode + dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)+1);
                   PC += 2;
                   break;
                case 0x0060:    // INC [r+r]q
                   w = GETMEMW(PC);
                   PC++;
                   pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                   SETMEMQ(pMem, GETMEMQ(pMem)+1);
                   break;
                case 0x0061:    // DEC [r]q
                   pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                   SETMEMQ(pMem, GETMEMQ(pMem)-1);
                   PC++;
                   break;
                case 0x0062:    // DEC [r+x]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   pMem = pKode + dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)-1);
                   PC += 2;
                   break;
                case 0x0063:    // DEC [x]q
                   pMem = pKode + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)-1);
                   PC += 2;
                   break;
                case 0x0064:    // DEC [r+x+y]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += (long)(KLONG)GETMEMD(PC);
                   PC += 2;
                   pMem = pKode + dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, GETMEMQ(pMem)-1);
                   PC += 2;
                   break;
                case 0x0065:    // DEC [r+r]q
                   w = GETMEMW(PC);
                   PC++;
                   pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                   SETMEMQ(pMem, GETMEMQ(pMem)-1);
                   break;
                case 0x0066:    // NEG [r]q
                   pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                   SETMEMQ(pMem, -(long long)GETMEMQ(pMem));
                   PC++;
                   break;
                case 0x0067:    // NEG [r+x]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   pMem = pKode + dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, -(long long)GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x0068:    // NEG [x]q
                   pMem = pKode + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, -(long long)GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x0069:    // NEG [r+x+y]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += (long)(KLONG)GETMEMD(PC);
                   PC += 2;
                   pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, -(long long)GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x006A:    // NEG [r+r]q
                   w = GETMEMW(PC);
                   PC++;
                   pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                   SETMEMQ(pMem, -(long long)GETMEMQ(pMem));
                   break;
                case 0x006B:    // NOT [r]q
                   pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                   SETMEMQ(pMem, ~GETMEMQ(pMem));
                   PC++;
                   break;
                case 0x006C:    // NOT [r+x]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, ~GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x006D:    // NOT [x]q
                   pMem = pKode + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, ~GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x006E:    // NOT [r+x+y]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += GETMEMD(PC);
                   PC += 2;
                   pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                   SETMEMQ(pMem, ~GETMEMQ(pMem));
                   PC += 2;
                   break;
                case 0x006F:    // NOT r+r]q
                   w = GETMEMW(PC);
                   PC++;
                   pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                   SETMEMQ(pMem, ~GETMEMQ(pMem));
                   break;
                case 0x0070:    // TST [r]q
                   pMem = pKode + (long)(KLONG)GETREGD(GETMEMW(PC));
                   PC++;
                   qw = GETMEMQ(pMem);
                   goto setqwresult;
                case 0x0071:    // TST [r+x]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                   PC += 2;
                   qw = GETMEMQ(pMem);
                   goto setqwresult;
                case 0x0072:    // TST [x]q
                   pMem = pKode + (long)(KLONG)GETMEMD(PC);
                   PC += 2;
                   qw = GETMEMQ(pMem);
                   goto setqwresult;
                case 0x0073:    // TST [r+x+y]q
                   dw2 = GETREGD(GETMEMW(PC));
                   PC++;
                   dw2 += GETMEMD(PC);
                   PC += 2;
                   pMem = pKode + (long)(KLONG)dw2 + (long)(KLONG)GETMEMD(PC);
                   PC += 2;
                   qw = GETMEMQ(pMem);
                   goto setqwresult;
                case 0x0074:    // TST [r+r]q
                   w = GETMEMW(PC);
                   PC++;
                   pMem = pKode + (long)(KLONG)GETREGD(w & 0x00FF) + (long)(KLONG)GETREGD((w>>8)&0x00FF);
                   qw = GETMEMQ(pMem);
                   goto setqwresult;
               }
				break;
		}
		SETREGD(REG_3, (DWORD)((BYTE *)PC - pKode));	// update PC in Kregs
	}
stopint:
	
	//if( GETREGD(REG_7) != entryr7 ) {
	// entryr7 = entryr7;
	//}
	return GETREGW(REG_39);
}


//************************************************************
//************************************************************
//************************************************************
KLONG	INIsize, INIused;
char	*INIbuf=NULL;
char	*pINIsec, *pINIvar;
	
#define	INIwhite(s)			(*(s)==' ' || *(s)=='\t')
#define	INIeol(s)			(*(s)=='\r' || *(s)=='\n')
#define	INIwhiteALL(s)		(INIwhite(s) || INIeol(s))
#define	INIskipwhite(s) 	{while( INIwhite(s) ) ++s;}
#define	INIskipline(s)  	{while( *s && !INIeol(s) ) ++s; while( *s && INIeol(s) ) ++s; }
#define	INIskipwhiteALL(s)	{\
do {\
INIskipwhite(s);\
if( INIeol(s) ) INIskipline(s);\
} while( INIwhiteALL(s) );\
}
	
#define	INIchunkslop		2048
	
	/*************************************************
	 If SECTION is found:
	 pINIsec = ptr to '[' at beginning of section
	 pINIvar = NULL
	 returns TRUE
	 Else
	 pINIsec = ptr to NUL at EOF
	 pINIvar = NULL
	 returns FALSE
	 *************************************************/
	BOOL FindSec(char *srchname) {
		
		size_t	len;
		char	*pSecName=NULL;
		
		//char tttbuf[128];
		if( INIbuf != NULL ) {
			len = strlen(srchname);
			pINIvar = NULL;	// need a iniFindVariable before iniGetNum or iniGetStr
			for(pINIsec=INIbuf; *pINIsec; ) {
				//sprintf(tttbuf,"@sec(len=%d) %16.16s", *((long*)(pINIsec+1)), pINIsec+12);
				//SystemMessage(tttbuf);
				if( *pINIsec=='[' ) {
					//SystemMessage("FS got [");
					pSecName = pINIsec + 12;	// skip [0xCHECKSUM=
					//sprintf(tttbuf,"FS pSecName=%16.16s", pSecName);
					//SystemMessage(tttbuf);
					if( !strncmp(pSecName, srchname, len) ) {	// if found a section with the same STARTING characters
						//SystemMessage("FS name FIRST match");
						if( *(pSecName+len)==']' ) return TRUE;	// if found matching name doesn't have any additional trailing characters, SUCCESS!
						//SystemMessage("FS name DID NOT match");
					}
				}
				pINIsec += *((KLONG*)(pINIsec+1));
			}
		}
		// leave pINIsec pointing at End Of File
		return FALSE;
	}
	
	/********************************************************
	 Finds a variable in the current SECTION, leaving pINIvar
	 EITHER pointing to start of variable name AND returning length of old variable line (including end-of-line sequence),
	 OR pointing to the '[' that starts the next section OR pointing to the EOF NUL char AND returning 0.
	 ********************************************************/
	long FindVar(char *srchvar) {
		
		long	oldlen, namlen;
		char	*pINItmp;
		
		if( INIbuf!=NULL && pINIsec!=NULL ) {
			namlen = strlen(srchvar);
			pINIvar = pINIsec + 13;
			INIskipline(pINIvar);	// skip the section line
			while( *pINIvar ) {
				INIskipwhiteALL(pINIvar);
				if( *pINIvar=='[' ) return 0;	// all new variable, didn't exist before
				if( !strncmp(pINIvar, (char*)srchvar, namlen) ) {	// found beginning of old variable (at least)
					pINItmp = pINIvar + namlen;
					if( *pINItmp=='=' ) {	// found old variable
						INIskipline(pINItmp);
						oldlen = pINItmp - pINIvar;
						// pINIvar now points to the beginning of the VAR line and oldlen is its length
						return oldlen;
					}
					// start of found variable matched, but is longer than what we're looking for
				}
				INIskipline(pINIvar);
			}
		}
		return 0;
	}
	
	/************************************/
	BOOL INIopenHole(char *s, long len) {
		
		char	*m;
		
		if( INIused+(len) > INIsize ) {	// need to increase size of buffer
			if( NULL==(m=(char*)malloc(INIsize+len+INIchunkslop)) ) return FALSE;
			memcpy(m, INIbuf, s-INIbuf);	// copy stuff above the hole
			memcpy(m+(s-INIbuf)+len, s, INIbuf+INIused-s);	// copy stuff below the hole
			if( pINIsec!=NULL ) pINIsec = m + (pINIsec-INIbuf) + ((pINIsec>s) ? len : 0);
			if( pINIvar!=NULL ) pINIvar = m + (pINIvar-INIbuf) + ((pINIvar>s) ? len : 0);
			free(INIbuf);
			INIbuf = m;
			INIsize += len+INIchunkslop;
			INIused += len;
		} else {
			memmove(s+len, s, INIused - (s-INIbuf));
			if( pINIsec!=NULL && pINIsec>s ) pINIsec += len;
			if( pINIvar!=NULL && pINIvar>s ) pINIvar += len;
			INIused += len;
		}
		return TRUE;
	}
	
	/************************************/
	void INIcloseHole(char *s, long len) {
		
		if( INIused-(s-INIbuf) < len ) len = INIused-(s-INIbuf);
		memcpy(s, s+len, INIused-(s-INIbuf+len));
		if( pINIsec!=NULL && pINIsec>=s+len ) pINIsec -= len;
		if( pINIvar!=NULL && pINIvar>=s+len ) pINIvar -= len;
		INIused -= len;
	}
	
	/***************************************/
	BOOL GetOctDecHex(char **pp, DWORD *pv) {
		
		char	*p, c;
		BOOL	retval;
		DWORD	val;
		
		p = *pp;
		retval = FALSE;
		val = 0;
		
		c = *p;
		if( c>='0' && c<='9' ) {
			if( c=='0' ) {	// OCT or HEX
				++p;
				c = *p;
				if( toupper(c)=='X' ) {	// HEX
					++p;
					c = toupper(*p);
					while( (c>='0' && c<='9') || (c>='A' && c<='F') ) {
						++p;
						c -= '0';
						if( c>9 ) c -= 7;
						val = 16*val + (DWORD)c;
						c = toupper(*p);
					}
				} else {	// OCT
					--p;
					c = '0';
					do {
						++p;
						val = 8*val + (c-'0');
						c = *p;
					} while( c>='0' && c<='7' );
				}
			} else {	// DEC
				do {
					++p;
					val = 10*val + (c-'0');
					c = *p;
				} while( c>='0' && c<='9' );
			}
			retval = TRUE;
		}
		*pv = val;
		*pp = p;
		return retval;
	}
	
	/*****************************************************************
	 Calculate a checksum for buffer p (length 'len') and return it. */
	DWORD GetChksum(char *p, DWORD len) {
		
		DWORD	dw, sum;
		
		dw = len>>2;	// number of whole DWORDS
		sum = 0;
		while( dw-- ) {
			sum += *((DWORD*)p);
			p += 4;
		}
		len &= 3;
		dw = 0;
		while( len-- ) dw = (dw<<8) | *p++;
		sum += dw;
		
		return sum;
	}
	
	//************************************************************
	//* In general, DWORD arguments start with R15 and progress downwards
	//*		WORD arguments start with R39 and progress downwards
	//*		BYTE arguments start with R95 and progress downwards
	//*		return values are generally in R15, R39, and/or R95
	//*		of course, there's always exceptions to every rule...
	void SysFunc(KE *pKE, WORD opcode) {
		
		BYTE		*pMem, *pMem2, a, b;
		WORD		w, h, *pW, asc;
		DWORD		dw, dh, dasc, a3;
		KLONG		i, x, y;
		PBMB		pKB;
        QWORD       qw;
        double      d1;
		
		switch( opcode ) {
				//**** SYS_FILESIZE
				// NOTE: Maximum filesize usable by the KASM/KINT system is 2^32-1 bytes (ie, length will fit in DWORD)
				// inputs:
				//	R15 = offset of filename in memblk [R95]
				//	R95 = memblk of filename
				//	R94 = folder, as in:
				//		0 = folder the game is installed in (home)
				//		1 = UI (folder containing UI image files)
				//		2 = IMAGES (folder containing game image set files)
				//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
				//		4 = MUSIC (folder containing music files)
				//		5 = SOUNDS (folder containing sound effect files)
				// outputs:
				//	R13 = size of file (-1 if error)
				//	R14 = time of last modification (seconds since Jan 1 1970)
				//	R93 = 0 if doesn't exist, 1 if normal file, 2 if directory (KINT 4.0)
			case SYS_FILESIZE:
				dw = Kstat((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), GETREGB(REG_94), &dh, &a);
				if( dw==0xFFFFFFFF ) {
					SETREGD(REG_13, 0xFFFFFFFF);
					SETREGD(REG_14, 0);
					SETREGB(REG_93, 0);
				} else {
					SETREGD(REG_13, dw);
					SETREGD(REG_14, dh);
					SETREGB(REG_93, a);
				}
				return;
				//**** SYS_MAKEFOLDER
				// inputs:
				//	R15 = offset of filename in memory block specified by R95
				//	R95 = memory block # of filename (0 if R0 based, 1 if R1 based, 2 if R2 based)
				// outputs:
				//	R15 = 0 if fail, non-0 if succeed
			case	SYS_MAKEFOLDER:
				SETREGD(REG_15, KMakeFolder((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15)));
				return;
				//**** SYS_DELFOLDER
				// NOTE: Folder MUST be empty or the DelFolder will fail!
				// inputs:
				//	R15 = offset of filename in memory block specified by R95
				//	R95 = memory block # of filename (0 if R0 based, 1 if R1 based, 2 if R2 based)
			case	SYS_DELFOLDER:
				KDelFolder((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15));
				return;
				//**** SYS_OPEN:
				// inputs:
				//	R15 = offset of filename in memory block specified by R95
				//	R95 = memory block # of filename (0 if R0 based, 1 if R1 based, 2 if R2 based)
				//	R94 = folder, as in:
				//		0 = folder the game is installed in (home)
				//		1 = UI (folder containing UI image files)
				//		2 = IMAGES (folder containing game image set files)
				//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
				//		4 = MUSIC (folder containing music files)
				//		5 = SOUNDS (folder containing sound effect files)
				//
				//	In the Windows implementation, UI, IMAGES, and BACKGNDS files
				//	are all .BMP files, and get that file extendsion added to them
				//	automatically.  MUSIC files are assumed to be MIDI files and get .MID.
				//	SOUNDS are assumed to be .WAV files, and get .WAV appended.
				//	When porting this to another platform (or even under Windows),
				//	there's absolutely NO reason why you can't use other graphics file
				//	formats (so long as they're 'lossless', so as to preserve the exact
				//	colors of the pixels, so that transparency works right), or other
				//	music (like MP3) or sound file formats.  As long as you modify
				//	the KINT code that opens and loads/plays the files, you can use
				//	whatever file formats you want.  The KASM code doesn't care.
				//
				//	R39 = open flags.  Supported bits are:
				//		O_RDBIN:		EQU	0x8000	; read only, binary
				//		O_WRBIN:		EQU	0x8002	; read/write, binary
				//		O_WRBINNEW:		EQU	0x8302	; read/write, binary, create or truncate
				// outputs:
				//	R15 = file handle if successful, -1 if failed
			case	SYS_FOPEN:
				// WARN: PTR4
				SETREGD(REG_15, Kopen((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), (KLONG)(DWORD)GETREGW(REG_39), GETREGB(REG_94)));
				return;
				//**** SYS_EOF
				// Inputs:
				//	R15 = file handle
				// Outputs:
				//	R12 = 1 if at EOF, 0 if not at EOF, -1 if error
			case	SYS_FEOF:
				SETREGD(REG_12, (DWORD)Keof(GETREGD(REG_15)));
				return;
				//**** SYS_READ
				// Inputs:
				//	R15 = file handle
				//	R14 = buffer offset
				//	R13 = max number of bytes to read
				//	R95 = memory block # (0 if R0 based, 1 if R1 based, 2 if R2 based)
				// Outputs:
				//	R12 = number of bytes read.  (-1 if error)
			case	SYS_FREAD:
				// WARN: PTR4
				SETREGD(REG_12, Kread(GETREGD(REG_15), (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_14), GETREGD(REG_13)));
				return;
				//**** SYS_WRITE
				// Inputs:
				//	R15 = file handle
				//	R14 = buffer offset in memory block specified by R95
				//	R13 = number of bytes to write
				//	R95 = memory block # (0 if R0 based, 1 if R1 based, 2 if R2 based)
				// Outputs:
				//	R12 = number of bytes written.  (-1 if error)
			case	SYS_FWRITE:
				// WARN: PTR4
				SETREGD(REG_12, Kwrite(GETREGD(REG_15), (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_14), GETREGD(REG_13)));
				return;
				//**** SYS_SEEK
				// Inputs:
				//	R15 = file handle
				//	R14 = offset to move the file pointer to
				//	R95 = origin:
				//		SEEK_SET    0
				//		SEEK_CUR    1
				//		SEEK_END    2
				// Outputs:
				//	R12 = offset in bytes from the beginning of the file of the new position (-1 if error)
			case	SYS_FSEEK:
				SETREGD(REG_12, Kseek(GETREGD(REG_15), GETREGD(REG_14), (KLONG)GETREGB(REG_95)));
				return;
				//**** SYS_CLOSE
				// Inputs:
				//	R15 = file handle to close
			case	SYS_FCLOSE:
				Kclose(GETREGD(REG_15));
				return;
				//**** SYS_FINDFIRST
				// Inputs:
				//	R15 = offset in memblock[R95] of filespec to search for
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblock[R94] of buffer for found filename (must be minimum of 260 bytes)
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				//	R93 = folder, as in:
				//		0 = folder the game is installed in (home)
				//		1 = UI (folder containing UI image files)
				//		2 = IMAGES (folder containing game image set files)
				//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
				//		4 = MUSIC (folder containing music files)
				//		5 = SOUNDS (folder containing sound effect files)
				//		6 = game folder, but GRAPHICS file
				// Outputs:
				//	R15 = "findfirst/next" handle of first file found, or -1 if none found or error
				//	  and found file name in buffer [R94]:R14
				//  R95 = file attributes:
				//		 attributes of file found: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR
				//		NOTE: the returned file name must be JUST the filename, not
				//			including any part of the full path or the file extension.
				// ADDED: KINT 4.8
				//  R96 = TRUE if file is AppData, else FALSE if PgmData
				//
			case	SYS_FFINDFIRST:
				// WARN: PTR4
				SETREGD(REG_15, Kfindfirst((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14), &a, &b, GETREGB(REG_93)));
				SETREGB(REG_95, a);
				SETREGB(REG_96, b);
				return;
				//**** SYS_FINDNEXT
				// Inputs:
				//	R15 = "find first/next" handle returned by previous FINDFIRST
				//	R14 = offset in memblock[R94] of buffer for found filename (must be minimum of 260 bytes)
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				// Outputs:
				//	R15 = 0 if next file found, or -1 if none found or error
				//	  and found file name in buffer [R94]:R14
				//  R95 = file attributes:
				//		 attributes of file found: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR
				// ADDED: KINT 4.8
				//  R96 = TRUE if file is AppData, else FALSE if PgmData
				//
			case	SYS_FFINDNEXT:
				// WARN: PTR4
				SETREGD(REG_15, Kfindnext((BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14), &a, &b));
				SETREGB(REG_95, a);
				SETREGB(REG_96, b);
				return;
				//**** SYS_FINDCLOSE
				// Inputs:
				//	R15 = "find first/next" handle returned by previous FINDFIRST
			case	SYS_FFINDCLOSE:
				Kfindclose();
				return;
				//**** SYS_DELETE
				// Inputs:
				//	R15 = address of filename
				//	R95 = memblk of filename
				//		KINT 3.0 - if R95 > 2, then memblk==0 and R95-3 = 'dir'
			case	SYS_DELETE:
				dw = (DWORD)GETREGB(REG_95);
				if( (long)dw >= 3 ) Kdelete((char *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_15), (BYTE)(dw-3));
				else Kdelete((char *)AllocPtrs[GETREGD(4*dw)] + GETREGD(REG_15), DIR_GAME);
				return;
				//**** SYS_MAKESHARECOPY
				// Inputs:
				//	R95 = command
				//	R15 = address of buffer in memblk R0 for next filename
				//	R13 = address of buffer in memblk R0 containing destination path (if R95 != 0)
				// Outputs:
				//	R14 = -1 if error
				//		else total size of files to be copied if cmd==0 (and filename of FIRST file in R15 buffer)
				//		else 0 (and filename of NEXT file in R15 buffer, or NUL if last file)
				//
				// NOTE: KASM code always calls this FIRST with command=0 (read file names, return first file name in buffer R15 and total size in R14)
				//  followed by a sequential series of calls with R95 != 0 to copy successive files.
				//  each successive call returns the NEXT filename to be copied in buffer R15.
				//	When no more files to be copied, buffer R15 starts with a NUL.
				
				// NOTE: this function is FAR more complex than ordinarily need be, but I didn't want
				// the KASM code to know anything about the files being copied for the shareware version
				// (as their names and/or formats will undoubtedly be different on different OS's), and
				// I didn't want the KINT interpreter (this code) to know anything about the specific
				// list of files being copied either, as I want the KINT code to remain generic across
				// all of the games.  So, the files to be copied are in an ASCII file call "swlist",
				// as a comma-separated list of src/dst pairs.  For example:
				//	eks0.ebk,setup.exe,eksctlss.ebk,setup.ctl,ekssubk.ebk,setupbk.ebk,shs.epk,shs.epk
				// all on one line, no CRs, no LFs, source filename first, comma, destination filename
				// second, comma, next pair....etc.
				//
			case	SYS_MAKESHARECOPY:
#if KINT_MAKE_COPY
				//if( GETREGB(REG_95)==0 )
				SystemMessage("Sorry, the \"Make a copy\" feature is not supported on\nthe Mac version of this game.\n\nPlease pass this website to your friends: www.kaser.com");
				//SystemMessage("X");
				SETREGD(REG_14, 0xFFFFFFFE);
				/*
				 pMem = (BYTE *)GETREGD(REG_0) + GETREGD(REG_15);
				 if( GETREGB(REG_95)==0 ) {	// get total sizes and return first filename in R15 buffer
				 DWORD	hFile, fsize, dw2;
				 
				 pMSCmem = NULL;
				 
				 hFile = Kopen("swlist", O_RDBIN, DIR_GAME);
				 if( hFile==-1 ) {
				 MSCerr:
				 if( pMSCmem!=NULL ) free(pMSCmem);
				 if( hFile!=-1 ) Kclose(hFile);
				 SETREGD(REG_14, 0xFFFFFFFF);
				 } else {
				 fsize = Kseek(hFile, 0, 2);
				 Kseek(hFile, 0, 0);
				 pMSCmem = (char *)malloc(fsize+1);
				 if( pMSCmem==NULL ) goto MSCerr;
				 swllen = Kread(hFile, (BYTE *)pMSCmem, fsize);
				 Kclose(hFile);
				 for(pMem2=(BYTE *)pMSCmem; (char *)pMem2<(pMSCmem+swllen); pMem2++) if( *pMem2==',' ) *pMem2 = 0;
				 *pMem2 = 0;	// make sure last entry is nul-terminated
				 dw = dw2 = 0;
				 for(pMem2=(BYTE *)pMSCmem; (char *)pMem2<(pMSCmem+swllen); ) {
				 hFile = Kopen((char *)pMem2, O_RDBIN, DIR_GAME);
				 if( hFile==-1 ) goto MSCerr;
				 dw += (Kseek(hFile, 0, 2)+8191) & 0xFFFFE000;	// add it's size to the running total (rounded up to nearest 8Kbytes for estimated file system slop)
				 Kclose(hFile);
				 
				 while( *pMem2!=0 && (char *)pMem2<(pMSCmem+swllen) ) ++pMem2;	// skip the src name
				 ++pMem2;	// skip the nul
				 if( !dw2 ) strcpy((char *)pMem, (const char *)pMem2);	// if first one, copy dst name to R15 buffer
				 ++dw2;	// no longer on first one
				 while( *pMem2!=0 && (char *)pMem2<(pMSCmem+swllen) ) ++pMem2;	// skip the dst name
				 ++pMem2;	// skip the nul
				 }
				 pNxtSrc = pMSCmem;		// set us ready to copy first file
				 SETREGD(REG_14, dw);	// return total size of files
				 }
				 } else { // copying file
				 char	target[256], *pSRC;
				 
				 pSRC = pNxtSrc;
				 strcpy(target, (char *)(GETREGD(REG_0)) + GETREGD(REG_13));
				 // make sure target path is properly terminated for this OS for concatenating the filename
				 // onto the end of it.  For MS Windows, that means making sure it ends in a '\'
				 i = strlen(target);
				 if( !i || target[i-1]!='\\' ) strcat(target, "\\");
				 
				 while( *pNxtSrc!=0 && pNxtSrc<(pMSCmem+swllen) ) ++pNxtSrc;	// skip src filename
				 ++pNxtSrc;	// skip the nul
				 strcat(target, pNxtSrc);	// finish building target filename
				 while( *pNxtSrc!=0 && pNxtSrc<(pMSCmem+swllen) ) ++pNxtSrc;	// skip dst filename
				 if( pNxtSrc<(pMSCmem+swllen) ) ++pNxtSrc;	// skip the nul if not the last one
				 
				 pMem2 = (BYTE *)pNxtSrc;
				 while( *pMem2!=0 && (char *)pMem2<(pMSCmem+swllen) ) ++pMem2;	// skip the dst name
				 if( (char *)pMem2<(pMSCmem+swllen) ) ++pMem2;	// skip the nul if this isn't the last one
				 strcpy((char *)pMem, (const char *)pMem2);	// copy the next filename into R15 buffer (or nul if no more)
				 
				 if( KCopyFile(pSRC, target) ) {	// if copy successful
				 SETREGD(REG_14, 0);
				 if( *pMem==0 ) free(pMSCmem);	// free memory if this was the last file to copy
				 } else {
				 SETREGD(REG_14, -1);
				 free(pMSCmem);
				 }
				 }
				 */
#endif
				return;
				//**** SYS_PRINTSTART
				// Inputs:
				// Outputs:
				//	R15 = 0 if fail, otherwise non-0
				//	R14 = page size in X direction (total pixels in printable area)
				//	R13 = page size in Y direction (total pixels in printable area)
				//	R12 = printer resolution in X direction (pixels per inch)
				//	R11 = printer resolution in Y direction (pixels per inch)
			case	SYS_PRINTSTART:
			{
				DWORD	success, pw, ph, Xppi, Yppi;
				
				success = KPrintStart(&pw, &ph, &Xppi, &Yppi);
				SETREGD(REG_14, pw);
				SETREGD(REG_13, ph);
				SETREGD(REG_12, Xppi);
				SETREGD(REG_11, Yppi);
				SETREGD(REG_15, success);
			}
				return;
				//**** SYS_PRINTBEGPAGE
				// Output:
				//	R95 = 1 if successful, 0 if fail
			case	SYS_PRINTBEGPAGE:
				SETREGB(REG_95, KPrintBeginPage());
				return;
				//**** SYS_PRINTENDPAGE
				// Output:
				//	R95 = 1 if successful, 0 if fail
			case	SYS_PRINTENDPAGE:
				SETREGB(REG_95, KPrintEndPage());
				return;
				//**** SYS_PRINTSTOP
			case	SYS_PRINTSTOP:
				KPrintStop();
				return;
				//**** SYS_PRINTTEXT
				// inputs:
				//	R14 = string buffer address in memblk[R95]
				//	R13 = textcolor
				//	R95 = memblk for string buffer
				//	R12 = x-coordinate of point
				//	R11 = y-coordinate of point
				// On KINT versions prior to 2.0, R37 was the font handle.
				// Starting on version 2.0, there is a SysPrintFont opcode which
				// creates/selects the printer font.  This font is automatically
				// deleted (if necessary by the SysPrintStop opcode).
			case	SYS_PRINTTEXT:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + (KLONG)GETREGD(REG_14);
				KPrintText((KLONG)((long)GETREGD(REG_12)), (KLONG)((long)GETREGD(REG_11)), (char*)pMem, GETREGD(REG_13));
				return;
				//**** SYS_PRINTFONT
				// NOTE!!!!!!!!!!!!!!!!!NOTE!!!!!!!!!!!!!NOTE!!!!!!!!!!!!!!!!!!!NOTE!
				//		This function was added at rev 2.0 of KINT, and replaces the
				//		use of SysGetFont when printing.
				// inputs:
				//	R38 = pixel height of font
				//	R37 =
				//		BIT 0 =0 for variable pitch, =1 for fixed pitch
				//		BIT 1 =0 for normal, =1 for BOLD
				//		BIT 2 =0 for normal, =1 for ITALIC
				// outputs:
				//	R38 = height of font
				//	R37 = ascending height of font
			case	SYS_PRINTFONT:
				KPrintFont(GETREGW(REG_38), GETREGW(REG_37), &h, &asc);
				SETREGW(REG_38, (WORD)h);
				SETREGW(REG_37, (WORD)asc);
				return;
				//**** SYS_PRINTBM
				// inputs:
				//	R14 = handle of source bitmap
				//	R19 = x-coord of destination
				//	R18 = y-coord of destination
				//	R17 = x-coord of source
				//	R16 = y-coord of source
				//	R13 = width of source
				//	R12 = height of source
				//	R11 = width of destination
				//	R10 = height of destination
			case	SYS_PRINTBM:
				KPrintBMB((KLONG)GETREGD(REG_19), (KLONG)GETREGD(REG_18), (KLONG)GETREGD(REG_11), (KLONG)GETREGD(REG_10),
						  &kBM[GETREGD(REG_14)], (KLONG)GETREGD(REG_17), (KLONG)GETREGD(REG_16), (KLONG)GETREGD(REG_13), (KLONG)GETREGD(REG_12));
				return;
				//**** SYS_PRINTRECT (KINT 2.2)
				// inputs:
				//	R19 = X
				//	R18 = Y
				//	R17 = W
				//	R16 = H
				//	R14 = FILL color
				//	R13 = EDGE color
			case SYS_PRINTRECT:
				KPrintRect((KLONG)GETREGD(REG_19), (KLONG)GETREGD(REG_18), (KLONG)GETREGD(REG_17), (KLONG)GETREGD(REG_16), GETREGD(REG_14), GETREGD(REG_13));
				return;
				//**** SYS_PRINTLINES
				// inputs:
				//	R14 = offset in memblock[R95] of (x,y) DWORD coordinate array
				//	R13 = RGB color of lines
				//	R39 = # of points in the [R14] array
				//	R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset
			case	SYS_PRINTLINES:
				// WARN: PTR4
				KPrintLines((DWORD *)((BYTE*)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + GETREGD(REG_14)), GETREGW(REG_39), GETREGD(REG_13));
				return;
				//**** SYS_PRINTGETTEXTSIZE
				// assumes a SYS_PRINTSTART and SYS_BEGTEXT have already been done
				// inputs:
				//	R14 = string buffer address in memblk[R95]
				//	R95 = memblk for string buffer
				//	 On KINT versions prior to 1.6, R37 was the font handle.
				//	 Starting on version 1.6, there is a SysPrintFont opcode which
				//	 creates/selects the printer font.  This font is automatically
				//	 deleted (if necessary by the SysPrintStop opcode).
				// outputs:
				//	R15 = width
				//	R14 = height
				//	R13 = height of non-descending portion of font
			case SYS_PRINTGETTEXTSIZE:
				KPrintGetTextSize((char *)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + GETREGD(REG_14), &dw, &dh, &dasc);
				SETREGD(REG_15, dw);	// width of font
				SETREGD(REG_14, dh);	// height of font
				SETREGD(REG_13, dasc);	// height of non-descending portion of font
				return;
				//**** SYS_ALLOCMEM
				// Inputs:
				//	R15 = size of memory block to allocate
				// Outputs:
				//	R15 = "handle" of memory block allocated (must be non-zero for valid memory allocation)
				//		this "handle" must be place in R1 or R2 to access the memory (using LD1/LD2 type instructions)
				//		==NULL if error
			case	SYS_ALLOCMEM:
				// WARN: PTR4
				dw = GETREGD(REG_15);
				//NSLog(@"ALLOCMEM dw=%d AllocCnt=%d", dw, AllocCnt);
				if( AllocCnt < MAXALLOCCNT ) {
					for(i=1; i<AllocCnt && AllocPtrs[i]!=0; i++) ;
					AllocPtrs[i] = malloc((size_t)dw);
					if( i>=AllocCnt ) ++AllocCnt;
				} else {
					i = 0;
					SystemMessage("AllCnt");
				}
				//NSLog(@"ALLOCMEM i=%d", i);
				SETREGD(REG_15, (DWORD)i);
				return;
				//**** SYS_FREEMEM
				// Inputs:
				//	R15 = "handle" of memory block to free
			case	SYS_FREEMEM:
				dw = GETREGD(REG_15);
				//NSLog(@"FREEMEM dw=%d AllocCnt=%d AllocPtrs[]=%d", dw, AllocCnt, AllocPtrs[dw]);
				if( dw!=0 && dw<AllocCnt && AllocPtrs[dw] ) {
					free(AllocPtrs[dw]);
					AllocPtrs[dw] = 0;
					if( dw+1==AllocCnt ) {
						do {
							--AllocCnt;
						} while( AllocCnt>0 && AllocPtrs[AllocCnt-1]==0 );
					}
				}
				//NSLog(@"FREEMEM");
				return;
				//**** SYS_TIMEREG
				// Outputs:
				//	R39 = year (ie, 2001)
				//	R38 = month (0-11, Jan=0, Feb=1, Mar=2, etc)
				//	R37 = day (1-31)
				//	R36 = hour (0-23)
				//	R35 = minute (0-59)
				//	R34 = second (0-59)
			case	SYS_TIMEREG:
			{
				KTIME	t;
				
				KGetTime(&t);
				SETREGW(REG_39, t.y);
				SETREGW(REG_38, t.mo);
				SETREGW(REG_37, t.d);
				SETREGW(REG_36, t.h);
				SETREGW(REG_35, t.mi);
				SETREGW(REG_34, t.s);
			}
				return;
				//**** SYS_TIMESTR
				// Inputs:
				//	R15 = offset in memblock[R95] of buffer for TIME string
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				// Outputs:
				//		buffer [R95]:R15 filled with string representing current time
			case	SYS_TIMERSTR:
				// WARN: PTR4
				KTimeToString((BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15));
				return;
				//**** SYS_TIMEMILLI
				// Outputs:
				//	R15 = "progressive" time count in milliseconds.
			case	SYS_TIMEMILLI:
				SETREGD(REG_15, KGetTicks());
				return;
				//**** SYS_SLEEP
				// Outputs:
				//	R15 = time count in milliseconds.
			case	SYS_SLEEP:
				KSleep(GETREGD(REG_15));
				return;
				//**** SYS_STRCAT
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of source buffer
				//	R94 = 0 for R0, 1 for R1, 2 for R2
			case	SYS_STRCAT:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				strcat((char *)pMem, (char *)pMem2);
				return;
				//**** SYS_MEMCPY
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of source buffer
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				//	R13 = number of bytes to copy
			case	SYS_MEMCPY:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				memcpy(pMem, pMem2, (size_t)GETREGD(REG_13));
				return;
				//**** SYS_MEMMOVE
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of source buffer
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				//	R13 = number of bytes to move
			case	SYS_MEMMOVE:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				memmove(pMem, pMem2, (size_t)GETREGD(REG_13));
				return;
				//**** SYS_STRCPY
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of source buffer
				//	R94 = 0 for R0, 1 for R1, 2 for R2
			case	SYS_STRCPY:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				dw = (DWORD)strlen((char*)pMem2);
				memmove((char *)pMem, pMem2, dw+1);
				return;
				//**** SYS_STRCMP
				// Inputs:
				//	R15 = offset in memblk[R95] of buffer1
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of buffer2
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_STRCMP:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				SETREGB(REG_95, (BYTE)(char)strcmp((char *)pMem, (char *)pMem2));
				return;
				//**** SYS_STRICMP
				// Inputs:
				//	R15 = offset in memblk[R95] of buffer1
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of buffer2
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_STRICMP:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				SETREGB(REG_95, (BYTE)(char)_stricmp((char *)pMem, (char *)pMem2));
				return;
				//**** SYS_STRLEN
				// Inputs:
				//	R15 = offset in memblk[R95] of buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				// Outputs:
				//	R39 = length of nul-terminated string in buffer
			case	SYS_STRLEN:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				SETREGW(REG_39, (WORD)strlen((char *)pMem));
				return;
				//**** SYS_MEMSET
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R94 = value to write into bytes of buffer
				//	R13 = number of bytes to set
			case	SYS_MEMSET:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				memset(pMem, GETREGB(REG_94), (size_t)GETREGD(REG_13));
				return;
				//**** SYS_MEMSET2
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = WORD value to write into bytes of buffer
				//	R13 = number of WORDs to set
			case	SYS_MEMSET2:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				WORD *pWord = (WORD *)pMem;
				for(dw=GETREGD(REG_13), w=GETREGW(REG_14); dw--; ) *pWord++ = w;
				return;
				//**** SYS_MEMSET4
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = DWORD value to write into bytes of buffer
				//	R13 = number of DWORDs to set
			case	SYS_MEMSET4:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				DWORD *pDword = (DWORD *)pMem;
				for(dw=GETREGD(REG_13), dh=GETREGW(REG_14); dw--; ) *pDword++ = dh;
				//**** SYS_MEMCMP
				// Inputs:
				//	R15 = offset in memblk[R95] of buffer1
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of buffer2
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				//	R13 = number of bytes to compare
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_MEMCMP:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				pMem2= (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_94))] + GETREGD(REG_14);
				i = memcmp(pMem, pMem2, (size_t)GETREGD(REG_13));
				SETREGB(REG_95, (BYTE)((i==0)?0:((i & 0x8000000)?0xFF:0x01)));
				return;
				//**** SYS_STRCAT00
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R14 = offset in memblk R0 of source buffer
			case	SYS_STRCAT00:
				strcat((char *)(pKE->pKode + GETREGD(REG_15)), (char *)(pKE->pKode + GETREGD(REG_14)));
				return;
				//**** SYS_MEMCPY00
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R14 = offset in memblk R0 of source buffer
				//	R13 = number of bytes to copy
			case	SYS_MEMCPY00:
				memcpy(pKE->pKode + GETREGD(REG_15), pKE->pKode + GETREGD(REG_14), (size_t)GETREGD(REG_13));
				return;
				//**** SYS_MEMMOVE00
				// Inputs:
				//	R15 = offset in memblk[R95] of destination buffer
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R14 = offset in memblk[R94] of source buffer
				//	R94 = 0 for R0, 1 for R1, 2 for R2
				//	R13 = number of bytes to move
			case	SYS_MEMMOVE00:
				memmove(pKE->pKode + GETREGD(REG_15), pKE->pKode + GETREGD(REG_14), (size_t)GETREGD(REG_13));
				return;
				//**** SYS_STRCPY00
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R14 = offset in memblk R0 of source buffer
			case	SYS_STRCPY00:
				// we have to use memmove instead of strcpy because Apple, in their stupidity, won't allow
				// a strcpy(dst, src) where dst and src overlap and dst<src (ie, yes, they overlap, but
				// in a fashion that the source will be read before the destination is written, so the
				// destination will never catch up to the source.  Sigh.
				pMem = (BYTE *)(pKE->pKode + GETREGD(REG_14));
				dw = (DWORD)strlen((char*)pMem);
				memmove((char *)(pKE->pKode + GETREGD(REG_15)), pMem, dw+1);
				return;
				//**** SYS_STRCMP00
				// Inputs:
				//	R15 = offset in memblk R0 of buffer1
				//	R14 = offset in memblk R0 of buffer2
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_STRCMP00:
				SETREGB(REG_95, (BYTE)(char)strcmp((char *)(pKE->pKode + GETREGD(REG_15)), (char *)(pKE->pKode + GETREGD(REG_14))));
				return;
				//**** SYS_STRICMP00
				// Inputs:
				//	R15 = offset in memblk R0 of buffer1
				//	R14 = offset in memblk R0 of buffer2
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_STRICMP00:
				SETREGB(REG_95, (BYTE)(char)_stricmp((char *)(pKE->pKode + GETREGD(REG_15)), (char *)(pKE->pKode + GETREGD(REG_14))));
				return;
				//**** SYS_STRLEN00
				// Inputs:
				//	R15 = offset in memblk R0 of buffer
				// Outputs:
				//	R39 = length of nul-terminated string in buffer
			case	SYS_STRLEN00:
				SETREGW(REG_39, (WORD)strlen((char *)(pKE->pKode + GETREGD(REG_15))));
				return;
				//**** SYS_MEMSET00
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R94 = value to write into bytes of buffer
				//	R13 = number of bytes to set
			case	SYS_MEMSET00:
				memset(pKE->pKode + GETREGD(REG_15), GETREGB(REG_94), (size_t)GETREGD(REG_13));
				return;
				//**** SYS_MEMSET200
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R14 = WORD value to write into bytes of buffer
				//	R13 = number of WORDs to set
			case	SYS_MEMSET200:
				pMem = pKE->pKode + GETREGD(REG_15);
				pWord = (WORD *)pMem;
				for(dw=GETREGD(REG_13), w=GETREGW(REG_14); dw--; ) *pWord++ = w;
				return;
				//**** SYS_MEMSET400
				// Inputs:
				//	R15 = offset in memblk R0 of destination buffer
				//	R14 = DWORD value to write into bytes of buffer
				//	R13 = number of DWORDs to set
			case	SYS_MEMSET400:
				pMem = pKE->pKode + GETREGD(REG_15);
				pDword = (DWORD *)pMem;
				for(dw=GETREGD(REG_13), dh=GETREGW(REG_14); dw--; ) *pDword++ = dh;
				return;
				//**** SYS_MEMCMP00
				// Inputs:
				//	R15 = offset in memblk R0 of buffer1
				//	R14 = offset in memblk R0 of buffer2
				//	R13 = number of bytes to compare
				// Outputs:
				//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
			case	SYS_MEMCMP00:
				i = memcmp(pKE->pKode + GETREGD(REG_15), pKE->pKode + GETREGD(REG_14), (size_t)GETREGD(REG_13));
				SETREGB(REG_95, (BYTE)((i==0)?0:((i & 0x8000000)?0xFF:0x01)));
				return;
				
				//**** SYS_SETCURSOR
				// Inputs:
				//	R39 = 0 for normal, 1 for wait, 2 for hotlink
			case SYS_SETCURSOR:
				KSetCursor(GETREGW(REG_39));
				return;
				//**** SYS_SETPGMTITLE
				// Inputs:
				//	R15 = offset in memblock[R95] of buffer containing program title
				//	R95 = 0 for R0, 1 for R1, 2 for R2
			case	SYS_SETPGMTITLE:
				// WARN: PTR4
				pMem = (BYTE *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15);
				if( strlen((char*)pMem)<sizeof(titleBuf) ) strcpy(titleBuf, (char*)pMem);
				else strcpy(titleBuf, "Title too long");
				KSetTitle();
				return;
				//**** SYS_CAPTUREMOUSE
				// Inputs:
				//	R15 = 1 to capture mouse, 0 to release mouse capture
			case	SYS_CAPTUREMOUSE:
				KCaptureMouse(GETREGD(REG_15));
				return;
				
				
				/*************************************************************
				 * BITMAP functions: the KINT program creates a "display bitmap"
				 * automatically.  A PBMB handle of NULL refers to this
				 * "display bitmap".  The NULL PBMB cannot be deleted, but
				 * it can be written to by all of the drawing functions.  The
				 * KINT program is also responsible for any "cleanup" work that
				 * needs to be done regarding the NULL "display bitmap" when
				 * the program is exiting, as well as deleting all resources
				 * allocated for other bitmaps created by SysCreBM, SysLoadBM,
				 * and SysLoadMBM.
				 *************************************************************/
				
				
				
				//**** SYS_CREBM
				// inputs:
				//	R39 = width
				//	R38 = height
				// outputs:
				//	R15 = bitmap handle of created bitmap, NULL if creation error.
			case	SYS_CREBM:
				SETREGD(REG_15, (DWORD)CreateBMB((KSHORT)GETREGW(REG_39), (KSHORT)GETREGW(REG_38), FALSE));	// return to the Kode the pointer to the bitmap structure as the "bitmap handle"
				return;
				//**** SYS_DELBM
				// inputs:
				//	R15 = handle (PBMB) of bitmap to delete
			case	SYS_DELBM:
				DeleteBMB(GETREGD(REG_15));
				return;
				//**** SYS_SIZEBM
				// Inputs:
				//	R15 = handle of bitmap
				// Outputs:
				//	R39 = width of bitmap
				//	R38 = height of bitmap
				//	R37 = orientation of bitmap (BMO_ value) (KINT 2.0)
			case SYS_SIZEBM:
				pKB = &kBM[GETREGD(REG_15)];
//				if( pKB==NULL ) pKB = &kBM[0];
				SETREGW(REG_39, pKB->logW);
				SETREGW(REG_38, pKB->logH);
				SETREGW(REG_37, pKB->orient);
				return;
				//**** SYS_ALPHABLEND (KINT 2.0)
				// Inputs:
				//	R95 = Dst ratio (0 to disable alpha-blending, 255 to disable src writing)
				// Outputs:
				//	R95 = previous setting
			case SYS_ALPHABLEND:
				w = (WORD)GETREGB(REG_95);
				SETREGB(REG_95, (BYTE)AlphaBlendD);
				AlphaBlendD = w;
				AlphaBlendS = 255-AlphaBlendD;
				return;
				//**** SYS_DRAWBM
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
			case	SYS_DRAWBM:
				BltBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
					   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
					   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), (KLONG)FALSE, (KLONG)FALSE);
				return;
				//**** SYS_INVDRAWBM (Added KINT rev 2.3)
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
			case	SYS_INVDRAWBM:
				BltBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
					   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
					   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), (KLONG)FALSE, (KLONG)TRUE);
				return;
				//**** SYS_STRETCHBM
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
				//	R33 = width of destination
				//	R32 = height of destination
			case	SYS_STRETCHBM:	// CHANGED at REV 2.3 of KINT!!!! (To add CompressBMB when R35 is negative)
				w = GETREGW(REG_35);
				h = GETREGW(REG_34);
				if( w & 0x8000 )
					CompressBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
								(KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32),
								&kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
								-((KLONG)(KSHORT)w), (KLONG)(KSHORT)h, FALSE);
				
				else
					StretchBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
							   (KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32),
							   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
							   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), FALSE);
				return;
				//**** SYS_STRETCHMBM	NOTE: This opcode added at REV 2.0 of KINT!!!!
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
				//	R33 = width of destination
				//	R32 = height of destination
			case	SYS_STRETCHMBM:	// modified at KINT 3.2: added smooth stretching of MASKED bitmaps (when R35 is negative)
				w = GETREGW(REG_35);
				h = GETREGW(REG_34);
				if( w & 0x8000 )
					CompressBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
								(KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32),
								&kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
								-((KLONG)(KSHORT)w), (KLONG)(KSHORT)h, TRUE);
				else
					StretchBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
							   (KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32),
							   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
							   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), TRUE);
				return;
				//**** SYS_DRAWMBM (Masked BitMap)
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
			case	SYS_DRAWMBM:
				BltBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
					   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
					   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), (KLONG)TRUE, (KLONG)FALSE);
				return;
				//**** SYS_INVDRAWMBM (Masked BitMap) (Added KINT rev 2.3)
				// Inputs:
				//	R15 = handle of destination bitmap
				//	R14 = handle of source bitmap
				//	R39 = x-coord of destination
				//	R38 = y-coord of destination
				//	R37 = x-coord of source
				//	R36 = y-coord of source
				//	R35 = width of source
				//	R34 = height of source
			case	SYS_INVDRAWMBM:
				BltBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
					   &kBM[GETREGD(REG_14)], (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36),
					   (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), (KLONG)TRUE, (KLONG)TRUE);
				return;
				//**** SYS_LOADBM
				// Inputs:
				//	R15 = offset into memory block[R95] of filename
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R94 =
				//		0 file is in GAME folder
				//		1 file is in UI folder
				//		2 file is in IMAGES folder
				//		3 file is in BACKGNDS folder
				// Outputs:
				//	R15 = handle of loaded bitmap, or NULL if failed
			case	SYS_LOADBM:
				// WARN: PTR4
				SETREGD(REG_15, LoadBMB((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), -1, -1, -1, -1, GETREGB(REG_94)));
				return;
				//**** SYS_LOADMBM (Masked BitMap)
				// Inputs:
				//	R15 = offset into memory block[R95] of filename
				//	R39 = x-coord of transparent pixel 1
				//	R38 = y-coord of transparent pixel 1
				//	R37 = x-coord of transparent pixel 2 (KINT 2.0)
				//	R36 = y-ccord of transparent pixel 2 (KINT 2.0)
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R94 =
				//		0 file is in GAME folder
				//		1 file is in UI folder
				//		2 file is in IMAGES folder
				//		3 file is in BACKGNDS folder
				// Outputs:
				//	R15 = handle of loaded bitmap, or NULL if failed
			case	SYS_LOADMBM:
				// WARN: PTR4
				if( (pKE->pKode[KHDR_MAJOR_INTERP_REV] & 0x7F) < 2 ) x = y = -1;
				else {
					x = (KLONG)(KSHORT)GETREGW(REG_37);
					y = (KLONG)(KSHORT)GETREGW(REG_36);
				}
				SETREGD(REG_15, LoadBMB((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), GETREGW(REG_39), GETREGW(REG_38), x, y, GETREGB(REG_94)));
				return;
				//**** SYS_SAVEBM (Masked BitMap) (KINT 2.0)
				// Inputs:
				//	R15 = bitmap handle
				//	R14 = offset into memory block[R95] of filename
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R94 =
				//		0 file is in GAME folder
				//		1 file is in UI folder
				//		2 file is in IMAGES folder
				//		3 file is in BACKGNDS folder
				// Outputs:
				//	R94 = 0 if failed, non-0 if succeeded
			case	SYS_SAVEBM:	// KINT 2.0
				// WARN: PTR4
				SETREGB(REG_94, (BYTE)SaveBMB(&kBM[GETREGD(REG_15)], (char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_14), GETREGB(REG_94)));
				return;
				//**** SYS_MASKBM (Create new masked for BitMap)
				// Inputs:
				//	R15 = bitmap handle
				//	R14 = transparent color 1
				//	R13 = transparent color 2
			case	SYS_MASKBM:
				CreateMaskBMB(&kBM[GETREGD(REG_15)], GETREGD(REG_14), GETREGD(REG_13));
				return;
				//**** SYS_LOADLOSSYBM
				// Inputs:
				//	R15 = offset into memory block[R95] of filename
				//	R95 = 0 for R0, 1 for R1, 2 for R2
				//	R94 =
				//		0 file is in GAME folder
				//		1 file is in UI folder
				//		2 file is in IMAGES folder
				//		3 file is in BACKGNDS folder
				// Outputs:
				//	R15 = handle of loaded bitmap, or NULL if failed
			case	SYS_LOADLOSSYBM:
				// WARN: PTR4
				// try to load lossless (BMP) first
				// if that fails, try to load lossy (JPG)
				if( 0==(dw = (DWORD)LoadBMB((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), -1, -1, -1, -1, GETREGB(REG_94))) )
					dw =(DWORD)LoadJPEG((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15), GETREGB(REG_94));
				SETREGD(REG_15, dw);
				return;
				//**** SYS_GETSYSTEMDATA
				// Inputs:
				//	R15 = offset of buffer in memblk R0; buffer MUST be at least 4 Kbytes
				//	R95 = what to get
				//		= 0 for AppDataPath (NOTE: *ONLY* to be used for Help output!!!)
				//		= 1 for PgmPath  (NOTE: *ONLY* to be used for Help output!!!)
				//		= 2 for OS version
				//		= 3 for "CTL" or "CMD" (depending upon Windows or Mac OS X) NOTE: *ONLY* to be used for Help output!!!)
				
				
				//		= 30 for void InitSysFonts(int w, int h); (called one time at start-up)
				//			IN:
				//				R32 = WIDTH of tmp drawing bitmap
				//				R33 = HEIGHT of tmp drawing bitmap and fonts
				//		= 31 for void FreeSysFonts();	(called one time at shut-down)
				//		= 32 for DWORD DrawSysFontText(PBMB pBM, char *pText, int x, int y, int w, int h, int flags, COLORREF bclr, COLORREF fclr, long *pW)
				//			IN:
				//				R14 = pBM handle/ptr
				//				R15 = *pText (offset in R0 memblk)
				//				R32 = x
				//				R33 = y
				//				R34 = w
				//				R35 = flags
				//				R17 = bclr
				//				R18 = fclr
				//				R19 = pW
				//			OUT: # chars used up from pText
				
				// KINT >= 4.6
				
				//		= 90 discard PointerMove and Timer MSGs, return first KEY message
				//			IN: R15 = offset of buffer in memblk R0 which will be filled with DWORD keycode, DWORD arg1, DWORD arg2, DWORD arg3
				//			OUT: R16 = 0 if no keycode message available, non-0 if keycode and arguments in R15 buffer
				//		= 91 discard PointerMove and Timer and first KEY msg, return first KEY message
				//			IN: R15 = offset of buffer in memblk R0 which will be filled with DWORD keycode, DWORD arg1, DWORD arg2, DWORD arg3
				//			OUT: R16 = 0 if no keycode message available, non-0 if keycode and arguments in R15 buffer
				// I'm overloading this system call so that I don't have to revise KASM and KLCC, etc, to add .ini file handling,
				// which I want to be done in actual machine compiled code to make it faster and easier.
				//		= 100 iniRead (read the R0:R15 ini-file into a memory block, set 'read' pointer to beginning of buffer
				//			IN:	R15 = offset of filename in memblk R0
				//			OUT: R16 = -1 if error opening and buffer created, 0 if opened and read and checksum matched, 1 if read error, 2 if read but chksum mismatch
				//		= 101 iniWrite (write buffer to R0:R15 ini-file)
				//			IN:	R15 = offset of filename in memblk R0
				//			OUT: R16 = TRUE if successful, FALSE if error
				//		= 102 iniClose (free the current buffer)
				//		= 103 iniFindSec (finds an already existing section)
				//			IN: R15 = offset of section name in memblk R0 (without []'s)
				//			OUT: R16 = TRUE if successful, FALSE if not
				//		= 104 iniFindVar (starts at pINIsec and tries o find a line beginning with the specified variable name
				//			IN: R15 = offset of variable name in memblk R0
				//			OUT: R16 = TRUE if variable found in current section and pINIvar pointing after it's '=', FALSE if not found
				//		= 105 iniGetNum (chews up a number at pINIvar, advancing pINIvar past the number)
				//			IN: R15 = offset of 4-byte variable in memblk R0 where number is to be stored
				//			OUT: R16 = TRUE if number found, FALSE if not
				//		= 106 iniGetStr (chews up a quoted string at pINIvar, advancing pINIvar past the string)
				//			IN: R15 = offset of string buffer in memblk R0
				//				R39 = max bytes in R15 buffer
				//			OUT: R16 = -1 if found string is longer than buffer, 0 if no string found, else len of found string
				//		= 107 iniSetSec (edits a new or updated section into INIbuf)
				//			IN: R15 = offset of section name in memblk R0 (without []'s)
				//			OUT: R16 = TRUE if successful, FALSE if not
				//		= 108 iniSetVar (edits a new or updated variable into INIbuf within the current (pINIsec) section
				//			IN: R15 = offset of variable name in memblk R0
				//				R14 = offset of formatted string of values (numbers and/or quoted strings) to associate with the variable
				//			OUT: R16 = TRUE if successful, FALSE if not
				//		= 109 iniDelSec (deletes the current SECTION and all its variables)
				//			IN: iniFindSec has been previously called to set the pINIsec
				//			OUT: nothing
				//		= 110 iniDelVariable (deletes the named variable and its value)
				//			IN: R15 = offset of variable name in memblk R0
				//			OUT: R16 = TRUE if successful, FALSE if not
				//		= 111 iniNextSection (moves to next section, returns its name)
				//			IN: R15 = offset of string buffer in memblk R0 to be filled with NUL-terminated section name
				//				R39 = max bytes in R15 buffer
				//				iniFindSection MUST already have been called to set pINIsec
				//			OUT: R16 = -1 if found section name is longer than buffer, 0 if no next section (EOF), else len of 'next' (new current) section name
				//		= 112 iniDelAll (delete entire INI contents, leave buffer allocated)
				//		= 113 iniEndVar (returns TRUE if no more data before EOL of current variable, else returns FALSE)
				//		= 114 iniBadSum (returns TRUE if the current SECTION had a bad checksum when read from file, otherwise returns FALSE
				//		= 115 iniAddVarNum (add a numeric value to the end of the variable)
				//			IN: R15 = offset of variable name in memblk R0
				//				R14 = value to format out
				//				R13 = 0 for decimal, 1 for octal, 2 for hex
				//		= 116 iniAddVarStr (add a quoted string to the end of the variable)
				//			IN: R15 = offset of variable name in memblk R0
				//				R14 = offset of unquoted, nul-terminated string in memblk R0
			case SYS_GETSYSTEMDATA:
				pMem = (BYTE *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_15);
				switch( GETREGB(REG_95) ) {
					case 0:
						strcpy((char *)pMem, AppDataPath);
						break;
					case 1:
						strcpy((char *)pMem, AppPath);
						break;
					case 2:
						KGetOSversion(pMem);
						break;
					case 3:
						strcpy((char *)pMem, OSctrl_cmdSTR);
						break;
                        
                        
/*
 case 8:    // write pMem (0:R15) to ClipBoard as TEXT
    /// 0:R15 = source text (NUL-terminated) to write as TEXT to the clipboard
    {
        size_t    len = strlen((char*)pMem)+1;
        HGLOBAL    hGlobal = GlobalAlloc(GMEM_MOVEABLE, len);
        memcpy(GlobalLock(hGlobal), pMem, len);
        GlobalUnlock(hGlobal);
        OpenClipboard(hWndMain);
        EmptyClipboard();
        SetClipboardData(CF_TEXT, hGlobal);
        CloseClipboard();
    }
    break;
 case 9:    // read Clipboard TEXT to pMem (empty string if not available)
     /// 0:R15 = target buffer
     /// R39 = max len of target buffer (including terminating NUL)
    if( IsClipboardFormatAvailable(CF_TEXT) ) {
        HGLOBAL    hGlobal;
        LPVOID    pGlobal;
        char    *pText;
        int        siz;

        *pMem = 0;    // if nothing on clipboard, return empty string
        OpenClipboard(hWndMain);
        hGlobal = GetClipboardData(CF_TEXT);
        if( NULL != hGlobal ) {
            siz = GlobalSize(hGlobal);
            if( siz>=GETREGW(REG_39) ) *pMem = 0;
            else {
                pGlobal = GlobalLock(hGlobal);
                strcpy(pMem, (const char *)pGlobal);
                GlobalUnlock(hGlobal);
            }
        }
        CloseClipboard();
    }
    break;


*/
                        
                    case 10:    // sprintf("%f")
                       ///    R15 = offset of output buffer in memblk R0; buffer is assumed to be long enough
                       /// R40 = width
                       /// r41 = precision
                       /// R52q= double value
                       /// R95 = 10
           /*
                       {
                           int    wid=GETREGW(REG_40), prec=GETREGW(REG_41);

                           if( wid ) {
                               if( prec ) sprintf((char*)pMem, "%*.*f", wid, prec, *((double*)(&Kregs[REG_52])));
                               else sprintf((char*)pMem, "%*f", wid, *((double*)(&Kregs[REG_52])));
                           } else {
                               if( prec ) sprintf((char*)pMem, "%.*f", prec, *((double*)(&Kregs[REG_52])));
                               else sprintf((char*)pMem, "%f", *((double*)(&Kregs[REG_52])));
                           }
                       }
           */
                       sprintf((char*)pMem, "%.16f", *((double*)(&Kregs[REG_52])));
                       break;
                    case 20:    // sin
                        /// R52q = double angle
                        /// returns R52q = sin(angle)
                        qw = GETREGQ(REG_52);
                        d1 = *((double*)&qw);
                        d1 = sin(d1);
                        qw = *((QWORD*)&d1);
                        SETREGQ(REG_52, qw);
                        break;
                    case 21:    // cos
                        /// R52q = double angle
                        /// returns R52q = cos(angle)
                        qw = GETREGQ(REG_52);
                        d1 = *((double*)&qw);
                        d1 = cos(d1);
                        qw = *((QWORD*)&d1);
                        SETREGQ(REG_52, qw);
                        break;
                    case 22:    // arcsin
                        /// R52q = double angle
                        /// returns R52q = arcsin(angle)
                        qw = GETREGQ(REG_52);
                        d1 = *((double*)&qw);
                        d1 = asin(d1);
                        qw = *((QWORD*)&d1);
                        SETREGQ(REG_52, qw);
                        break;
                    case 23:    // arccos
                        /// R52q = double angle
                        /// returns R52q = arccos(angle)
                        qw = GETREGQ(REG_52);
                        d1 = *((double*)&qw);
                        d1 = acos(d1);
                        qw = *((QWORD*)&d1);
                        SETREGQ(REG_52, qw);
                       break;


                        
						//**************************
						//==========================
						
						//		= 30 for SetBBFonts(WIDTH, HEIGHT) (set FIXED, VARNORM, VARBOLD, VARITALIC, and VARBOLDITALIC fonts for BB)
						//			IN:
						//				R32 = WIDTH of tmp drawing bitmap
						//				R33 = HEIGHT of tmp drawing bitmap and fonts
					case 30:	// InitSysFonts()
						InitSysFonts(GETREGW(REG_32), GETREGW(REG_33));
						break;
					case 31:	// FreeSysFonts()
						FreeSysFonts();
						break;
						//		= 32 for DWORD DrawSysFontText(PBMB pBM, char *pText, int x, int y, int w, int h, int flags, COLORREF bclr, COLORREF fclr, long *pW)
						//			IN:
						//				R14 = pBM handle/ptr
						//				R15 = *pText (offset in R0 memblk)
						//				R32 = x
						//				R33 = y
						//				R34 = w
						//				R35 = h
						//				R36 = flags
						//				R17 = bclr
						//				R18 = fclr
						//				R19 = pW
						//			OUT: # chars used up from pText
					case 32:	// DWORD DrawSysFontText(PBMB pBM, char *pText, int x, int y, int w, int h, int flags, COLORREF bclr, COLORREF fclr, long *pW)
						SETREGD(REG_16, DrawSysFontText(&kBM[GETREGD(REG_14)], (char*)pMem, GETREGW(REG_32), GETREGW(REG_33), GETREGW(REG_34), GETREGW(REG_35), GETREGW(REG_36), (DWORD)GETREGD(REG_17), (DWORD)GETREGD(REG_18), (long*)(GETREGD(REG_19)==0?NULL:((BYTE *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_19)))));
						break;
						
						//**************************
						//==========================
						
						//		= 90 discard PointerMove and Timer MSGs, return first KEY message
						//			IN: R15 = offset of buffer in memblk R0 which will be filled with DWORD keycode, DWORD arg1, DWORD arg2, DWORD arg3
						//			OUT: R16 = 0 if no keycode message available, non-0 if keycode and arguments in R15 buffer
					case 90:	// get first KEY msg
						SETREGD(REG_16, KeyMsg(&dw, &dh, &dasc, &a3));
						 *((DWORD*)(pMem)) = dw;
						 *((DWORD*)(pMem+4)) = dh;
						 *((DWORD*)(pMem+8)) = dasc;
						 *((DWORD*)(pMem+12)) = a3;
						break;
						/***
						 INI FILE HANDLING
						 
						 NOTE: Do NOT use '[', ']', or '=' in NAMES of sections or variables.
						 There is no checking for that situation, and if you do, hell will rain down upon your head with a liquified garlic and harbenero sautee...
						 When formatting the string value of a variable for the iniSetVariable call, you MUST surround all string-values with double-quotes (ASCII 34)
						 and any 'internal' double-quotes within a string value must be preceded by a backslash (ASCII 92), ie, "abc\"def".  A backslash character may be included
						 by preceding it with another backslash (ie, \\).  Numbers may be formatted at octal (leading ZERO 0 character), decimal (NO leading 0 character),
						 or hex (leading 0x or 0X, ie, 0xFFFF).  Numbers are NOT quoted.
						 
						 When in the disk file, the iniSec header looks like     [0xCHECKSUM=HeaderName]
						 When in memory, in INIbuf, the iniSec header looks like [LLLLfxxxxx=HeaderName]
						 where:
						 LLLL is a DWORD value that is the LENGTH of this section, from the '[' at the start of this section to the '[' at the start of the next section OR the EOF.
						 f is a byte of FLAGS, currently which is all 0 bits except the LSbit which MAY be 1 if the checksum didn't match for this section when it was read in.
						 xxxxx are unused bytes, currently undefined.
						 This transformation takes place when the file is read into INIbuf from disk,
						 and the LLLL length is maintained as the buffer is manipulated.
						 ALL of these values get overwritten by the checksum during iniWrite, although they are saved and restored so that iniWrite doesn't destroy the information in INIbuf (for continued use).
						 The application has NO idea that this is going on behind the scenes.
						 ***/
						
						//		= 100 iniRead (read the R0:R15 ini-file into a memory block, set 'read' pointer to beginning of buffer
						//			IN:	R15 = offset of filename in memblk R0
						//			OUT: R16 =
						//				-1 if error opening file, empty buffer created
						//				 0 if opened and read and checksum matched
						//				 1 if read error
						//				 2 if read okay, but there's a chksum mismatch
					case 100:	// iniRead
					{
						DWORD	hFile, chksum, seclen, filsum;
						BYTE	*pWsec, *pSum, *pNext;
						BOOL	gotchksum;
						KLONG	s, r;
						
						SETREGD(REG_16, 1);	// preset to FAIL
						
						if( INIbuf!=NULL ) {
							free(INIbuf);
							INIbuf = NULL;
						}
						pINIvar = pINIsec = NULL;
						if( -1!=(hFile=Kopen((char*)pMem, O_RDBIN, DIR_GAME)) ) {
							s = Kseek(hFile, 0, SEEK_END);
							// one extra for NUL on last line in case no CR/LF on last line
							INIsize = s+1;
							if( s>=10 ) {
								if( NULL != (INIbuf = (char *)malloc(INIsize)) ) {
									Kseek(hFile, 0, SEEK_SET);
									r = Kread(hFile, (BYTE*)INIbuf, s);
									
									INIbuf[r] = 0;	// place NUL at end of buffer
									INIused = r+1;
									SETREGD(REG_16, 0);	// success
									
									pINIsec =INIbuf;
									INIskipwhiteALL(pINIsec);
									pWsec = (BYTE*)pINIsec;
									while( *pWsec && *pWsec=='[' ) {
										// find len of each section, checksum it, and compare it to the stored checksum.
										// if checksums don't match, set flag bit0.
										pNext = pWsec+1;	// skip the '[', point at the file's checksum for this section
										gotchksum = GetOctDecHex((char**)&pNext, &filsum);	// get the checksum from the file
										pNext = pWsec+1;				// point at the start of the checksum area
										for(i=0; i<10; i++) *pNext++ = 0;	// zero-out the checksum area
										for(pSum=pWsec; pSum<pNext || (*pSum && *pSum!='['); pSum++) ;	// find the end of this SECTION
										seclen = (DWORD)(pSum - pWsec);			// get the length of this SECTION
										chksum = GetChksum((char*)pWsec, seclen);	// get the actual checksum of this SECTION
										*((DWORD*)(pWsec+1)) = seclen;	// set the length of this SECTION
										/*{
										 char	tttbuf[64];
										 
										 sprintf(tttbuf,"%16.16s got=%d cs=%8.8lX fs=%8.8lX", pWsec+12, gotchksum, chksum, filsum);
										 SystemMessage(tttbuf);
										 }*/
										if( !gotchksum || chksum != filsum ) {
											pWsec[5] = 1;	// if the checksums don't match, set the BAD CHECKSUM flag
											SETREGD(REG_16, 2);
										}
										pWsec += seclen;
									}
									if( r!=s && GETREGD(REG_16)==0 ) SETREGD(REG_16, 1);
								}
								Kclose(hFile);
								/*if( -1!=(hFile=Kopen("groddi", O_WRBINNEW, DIR_GAME)) ) {
								 Kwrite(hFile, (BYTE*)INIbuf, INIused);
								 Kclose(hFile);
								 }*/
								
								break;
							}
						}
						INIbuf = (char*)malloc(INIchunkslop);
						*INIbuf = 0;
						INIsize = INIchunkslop;
						INIused = 1;
						SETREGD(REG_16, -1);	// set to didn't exist, buffer created
					}
						break;
						//		= 101 iniWrite (write buffer to R0:R15 ini-file)
						//			IN:	R15 = offset of filename in memblk R0
						//			OUT: R16 = TRUE if successful, FALSE if error
					case 101:	// iniWrite
					{
						DWORD	hFile, chksum, seclen;
						char	chkbuf[16], savbuf[10];
						BYTE	*pWsec, *pSum;
						
						SETREGD(REG_16, 1);	// preset to SUCCESS
						if( INIbuf != NULL ) {
							if( -1!=(hFile=Kopen((char*)pMem, O_WRBINNEW, DIR_GAME)) ) {
								pWsec = (BYTE*)INIbuf;
								// for each section
								while( *pWsec=='[' ) {
									// save the checksum "internal variables" area and zero it out
									pSum = pWsec+1;
									for(i=0; i<10; i++) {
										savbuf[i] = pSum[i];
										pSum[i] = 0;
									}
									// now checksum the section
									seclen = *((DWORD*)savbuf);	// get the section LENGTH
									chksum = GetChksum((char*)pWsec, seclen);
									sprintf(chkbuf, "0x%8.8X", chksum);
									// and write its ASCII value into the checksum area for this section
									for(i=0; i<10; i++) pSum[i] = chkbuf[i];
									// and write this section to the file
									if( seclen!=Kwrite(hFile, (BYTE*)pWsec, seclen) ) SETREGD(REG_16, 0);	// failed
									// now restore "internal variables" area
									for(i=0; i<10; i++) pSum[i] = savbuf[i];
									// now move to the next section
									pWsec += seclen;
								}
								Kclose(hFile);
							} else SETREGD(REG_16, 0);	// failed
						} else SETREGD(REG_16, 0);	// failed
					}
						break;
						//		= 102 iniClose (free the current buffer)
					case 102:	// iniClose
						if( INIbuf!=NULL ) {
							free(INIbuf);
							INIbuf = NULL;
						}
						break;
						//		= 103 iniFindSection (finds a section if it already exists)
						//			IN: R15 = offset of section name in memblk R0 (without []'s)
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 103:	// iniFindSec
						// pMem = ptr to string in Memblk 0 of Section name to find
					{
						size_t	len;
						char	*pSec;
						
						len = strlen((char*)pMem);
						SETREGD(REG_16, 0);	// preset to FAIL
						pINIvar = NULL;	// need a iniFindVar before iniGetNum or iniGetStr
						if( INIbuf!=NULL ) {
							for(pINIsec=INIbuf; *pINIsec; ) {
								if( *pINIsec=='[' ) {
									pSec = pINIsec + 12;	// skip [0xCHECKSUM=
									if( !strncmp(pSec, (char*)pMem, len) && *(pSec+len)==']' ) {
										SETREGD(REG_16, 1);	// signal success
										break;
									}
								}
								pINIsec += *((KLONG*)(pINIsec+1));
							}
						}
						//SystemMessage(pINIsec+12);
					}
						break;
						//		= 104 iniFindVar (starts at pINIsec and tries o find a line beginning with the specified variable name
						//			IN: R15 = offset of variable name in memblk R0
						//			OUT: R16 = TRUE if variable found in current section and pINIvar pointing after it's '=', FALSE if not found
					case 104:	// iniFindVar
						// pMem = ptr to string in Memblk 0 of variable name to find in this section
					{
						size_t	len;
						
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							len = FindVar((char*)pMem);
							if( len ) {
								while( *pINIvar && *pINIvar++ != '=' ) ;	// skip to variable values, now ready for iniGetNum and/or iniGetStr
								SETREGD(REG_16, 1);	// signal success
							}
						}
					}
						break;
					case 105:	// iniGetNum
						// pMem = ptr to 4-byte variable in Memblk 0 into which to store value
					{
						long	sign;
						DWORD	val;
						
						SETREGD(REG_16, 0);	// preset to FAIL
						sign = 0;
						if( INIbuf!=NULL && pINIvar!=NULL ) {	// must have a buffer, must have done an iniFindVariable
							INIskipwhite(pINIvar);
							if( *pINIvar=='-' ) {
								sign = 1;
								++pINIvar;
							}
							if( GetOctDecHex(&pINIvar, &val) ) SETREGD(REG_16, 1);
							if( sign ) val = (DWORD)(-(KLONG)val);
							*((DWORD *)pMem) = val;
						}
					}
						break;
						//		= 106 iniGetStr (chews up a quoted string at pINIvar, advancing pINIvar past the string)
						//			IN: R15 = offset of string buffer in memblk R0
						//				R39 = max bytes in R15 buffer
						//			OUT: R16 = -1 if found string is longer than buffer, 0 if no string found, else len of found string
					case 106:	// iniGetStr
						// pMem = ptr to 4-byte variable in Memblk 0 into which to store value
					{
						long	maxlen, i;
						char	c;
						
						maxlen = GETREGW(REG_39);
						i = 0;
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIvar!=NULL ) {	// must have a buffer, must have done an iniFindVariable
							INIskipwhite(pINIvar);
							c = *pINIvar;
							if( c=='"' ) {
								++pINIvar;
								while( i+1<maxlen && (c = *pINIvar) && c!='"' && c>=' ' && (c!='\\' || *(pINIvar+1)) ) {
									if( c=='\\' ) c = *(++pINIvar);
									pMem[i++] = c;
									++pINIvar;
								}
								if( *pINIvar=='"' ) {
									++pINIvar;
									SETREGD(REG_16, i);	// set to length of found string
								} else SETREGD(REG_16, -1);	// set to string longer than buffer
							}
						}
						pMem[i] = 0;
					}
						break;
						//		= 107 iniSetSection (edits a new or updated section into INIbuf)
						//			IN: R15 = offset of section name in memblk R0 (without []'s)
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 107:
						// pMem = ptr to string in Memblk 0 of Section name to find
					{
						KLONG	len, eollen, seclen;
						char	eolbuf[8], *pSec;
						//char tttbuf[128];
						
						len = (KLONG)strlen((char*)pMem);
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL ) {
							//sprintf(tttbuf, "FindSec(%s)", pMem);
							//SystemMessage(tttbuf);
							if( !FindSec((char*)pMem) ) {	// if didn't find section, add it (pINIsec is pointing to the end of file
								//SystemMessage("didn't find it");
								eollen = GetSystemEOL((BYTE*)eolbuf);
								seclen = len+eollen+13;	// +2 for []'s, +1 for '=', +10 for 0xCHECKSUM
								if( INIopenHole(pINIsec, seclen) ) {
									pSec = pINIsec;
									*pSec++ = '[';
									for(i=0; i<10; i++) *pSec++ = 0;	// clear checksum area for LENGTH and FLAGS
									*pSec++ = '=';
									strcpy(pSec, (char*)pMem);
									pSec += len;
									*pSec++ = ']';
									for(i=0; i<eollen; i++) *pSec++ = eolbuf[i];
									*((DWORD*)(pINIsec+1)) = seclen;
									pINIvar = NULL;
									SETREGD(REG_16, 1);	// signal success
								}
							} else SETREGD(REG_16, 1);	// signal success
							//SystemMessage("end of it");
						}
					}
						break;
						//		= 108 iniSetVariable (edits a new or updated variable into INIbuf within the current (pINIsec) section
						//			IN: R15 = offset of variable name in memblk R0
						//				R14 = offset of formatted string of values (numbers and/or quoted strings) to associate with the variable
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 108:	// iniSetVariable
						// pMem = variable name to update in pINIsec
						// pMem2= formatted string of values to save
						pMem2 = (BYTE *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_14);
					{
						long	oldlen, namlen, newlen, varlen, eollen;
						BYTE	eolbuf[8];
						KLONG		i;
						
						eollen = GetSystemEOL(eolbuf);
						namlen = strlen((char*)pMem);
						varlen = strlen((char*)pMem2);
						newlen = namlen+1+varlen+ eollen;// +1 for '='
						
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							oldlen = FindVar((char*)pMem);
							if( oldlen>newlen ) INIcloseHole(pINIvar, oldlen-newlen);
							else if( newlen!=oldlen ) INIopenHole(pINIvar, newlen-oldlen);
							*((KLONG*)(pINIsec+1)) += (newlen-oldlen);	// adjust SECTION LENGTH
							
							strcpy(pINIvar, (char*)pMem);	// copy name
							pINIvar += namlen;
							*pINIvar++ = '=';
							strcpy(pINIvar, (char*)pMem2);
							pINIvar += varlen;
							for(i=0; i<eollen; i++) *pINIvar++ = eolbuf[i];
							SETREGD(REG_16, 1);	// signal success
						}
					}
						break;
						//		= 109 iniDelSection (deletes the current SECTION and all its variables)
						//			IN: iniFindSection has been previously called to set the pINIsec
						//			OUT: nothing
					case 109:	// iniDelSection
					{
						long	len;
						
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							len = *((KLONG*)(pINIsec+1));
							INIcloseHole(pINIsec, len);
							pINIsec = INIbuf;
							pINIvar = NULL;
						}
					}
						break;
						//		= 110 iniDelVariable (deletes the named variable and its value)
						//			IN: R15 = offset of variable name in memblk R0
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 110:	// iniDelVariable
						// pMem = variable name to delete from pINIsec
					{
						long	oldlen;
						
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							oldlen = FindVar((char*)pMem);
							if( oldlen ) {
								INIcloseHole(pINIvar, oldlen);
								*((KLONG*)(pINIsec+1)) -= oldlen;
								SETREGD(REG_16, 1);	// signal success
							}
						}
					}
						break;
						//		= 111 iniNextSection (moves to next section, returns its name)
						//			IN: R15 = offset of string buffer in memblk R0 to be filled with NUL-terminated section name
						//				R39 = max bytes in R15 buffer
						//				iniFindSection MUST already have been called to set pINIsec
						//			OUT: R16 = -1 if found section name is longer than buffer, 0 if no next section (EOF), else len of 'next' (new current) section name
					case 111:
						// pMem = ptr to R15 buffer
					{
						long	i, maxlen;
						char	*pSec;
						
						maxlen = GETREGW(REG_39);
						i = 0;
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							pINIsec += *((KLONG*)(pINIsec+1));
							if( *pINIsec ) {
								pSec = pINIsec+12;	// skip [0xCHECKSUM=
								while( i+1<maxlen && *pSec!=']' ) pMem[i++] = *pSec++;
								SETREGD(REG_16, (i+1<maxlen || *pSec==']')?i:(DWORD)(-1));
							} else pINIsec = NULL;
							pINIvar = NULL;
						}
						pMem[i] = 0;
					}
						break;
						//		= 112 iniDelAll (delete entire INI contents)
						//			IN: nothing
						//			OUT: all contents of ini buffer deleted
					case 112:	// iniDelAll
						if( INIbuf!=NULL ) {
							INIused = 1;
							*INIbuf = 0;
							pINIsec = pINIvar = NULL;
						}
						break;
					case 113:	// iniEndVar
						SETREGD(REG_16, 1);	// preset to TRUE, is end of cur var data
						if( INIbuf!=NULL && pINIvar!=NULL ) {
							INIskipwhite(pINIvar);
							SETREGD(REG_16, (*pINIvar==0 || INIeol(pINIvar)) ? 1 : 0);
						}
						break;
					case 114:	// iniBadSum
						if( INIbuf!=NULL && pINIsec!=NULL ) SETREGD(REG_16, (pINIsec[5] & 1)? 1 : 0);
						else SETREGD(REG_16, 0);
						break;
						//		= 115 iniAddVarNum (add a numeric value to the end of the variable)
						//			IN: R15 = offset of variable name in memblk R0
						//				R14 = value to format out (interpreted as a DWORD if hex or oct output, else signed_long if dec)
						//				R13 = 0 for decimal, 1 for octal, 2 for hex
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 115:	// iniAddVarNum
						// pMem = variable name to update in pINIsec
					{
						long	val, form, oldlen, namlen, newlen, varlen, eollen;
						BYTE	eolbuf[8], numbuff[16];
						KLONG		i;
						
						val = (long)GETREGD(REG_14);
						form= (long)GETREGD(REG_13);
						
						if( form==2 ) sprintf((char*)numbuff, " 0x%8.8X", (DWORD)val);
						else if( form==1 ) sprintf((char*)numbuff, " 0%o", (DWORD)val);
						else sprintf((char*)numbuff, " %ld", val);
						pMem2 = numbuff;
						
						eollen = GetSystemEOL(eolbuf);
						namlen = strlen((char*)pMem);
						varlen = strlen((char*)pMem2);
						newlen = namlen+1+varlen+ eollen;// +1 for '='
						
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							oldlen = FindVar((char*)pMem);
							if( oldlen ) {	// var existed before
								INIopenHole(pINIvar+oldlen, varlen);
								*((KLONG*)(pINIsec+1)) += varlen;	// adjust SECTION LENGTH
								pINIvar += oldlen-eollen;	// setup to overwrite previous EOL and new hole with formatted string
							} else {	// else variable didn't exist before, create it
                                INIopenHole(pINIvar, newlen);//namlen);
								*((KLONG*)(pINIsec+1)) += newlen;	// adjust SECTION LENGTH
								strcpy(pINIvar, (char*)pMem);	// copy name
								pINIvar += namlen;
								*pINIvar++ = '=';
							}
							for(i=0; i<varlen; i++) {	// copy out the formatted number string
								*pINIvar++ = pMem2[i];
							}
							
							for(i=0; i<eollen; i++) *pINIvar++ = eolbuf[i];	// output eol sequence
							pINIvar = NULL;	// NULL pointer since it's not in a good place to do further work
							SETREGD(REG_16, 1);	// signal success
						}
					}
						break;
						//		= 116 iniAddVarStr (add a quoted string to the end of the variable)
						//			IN: R15 = offset of variable name in memblk R0
						//				R14 = offset of unquoted, nul-terminated string in memblk R0
						//			OUT: R16 = TRUE if successful, FALSE if not
					case 116:	// iniAddVarStr
						// pMem = variable name to update in pINIsec
						// pMem2= unquoted, nul-term'd string
						pMem2 = (BYTE *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_14);
					{
						long	oldlen, namlen, newlen, varlen, eollen, formvlen;
						BYTE	eolbuf[8];
						KLONG		i;
						
						eollen = GetSystemEOL(eolbuf);
						namlen = strlen((char*)pMem);
						varlen = strlen((char*)pMem2);
						newlen = namlen+1+varlen+ eollen;// +1 for '='
						
						SETREGD(REG_16, 0);	// preset to FAIL
						if( INIbuf!=NULL && pINIsec!=NULL ) {
							// add 3 to target length, 1 for leading space, 2 for surrounding quotes.
							// then add one more for every '\' or '"' in the string, which will need to be escaped by a '\'
							for(i=0, formvlen=varlen+3; i<varlen; i++) if( pMem2[i]=='\\' || pMem2[i]=='"' ) ++formvlen;
							oldlen = FindVar((char*)pMem);
							if( oldlen ) {	// var existed before
								INIopenHole(pINIvar+oldlen, formvlen);
								*((KLONG*)(pINIsec+1)) += formvlen;	// adjust SECTION LENGTH
								pINIvar += oldlen-eollen;	// setup to overwrite previous EOL and new hole with formatted string
							} else {	// else variable didn't exist before, create it
								INIopenHole(pINIvar, namlen + (formvlen-varlen));
								*((KLONG*)(pINIsec+1)) += (newlen+(formvlen-varlen));	// adjust SECTION LENGTH
								strcpy(pINIvar, (char*)pMem);	// copy name
								pINIvar += namlen;
								*pINIvar++ = '=';
							}
							*pINIvar++ = ' ';	// output leading space
							*pINIvar++ = '"';	// output leading quote
							for(i=0; i<varlen; i++) {	// copy out the unquoted string, escaping any '\' or '"' chars with a '\' char.
								if( pMem2[i]=='\\' || pMem2[i]=='"' ) *pINIvar++ = '\\';
								*pINIvar++ = pMem2[i];
							}
							*pINIvar++ = '"';
							
							for(i=0; i<eollen; i++) *pINIvar++ = eolbuf[i];	// output eol sequence
							pINIvar = NULL;	// NULL pointer since it's not in a good place to do further work
							SETREGD(REG_16, 1);	// signal success
						}
					}
						break;
						//		= 117 iniNextVar (copies the ENTIRE next variable line to R0:R15 buffer)
						//			IN: R15 = offset of string buffer in memblk R0
						//				R39 = max bytes in R15 buffer
						//			OUT: R16 = -1 if found string is longer than buffer, 0 if no string found, else len of found string
					case 117:	// iniNextVar
						// pMem = ptr to 4-byte variable in Memblk 0 into which to store value
					{
						long	maxlen, i;
						char	c;
						
						maxlen = GETREGW(REG_39);
						i = 0;
						SETREGD(REG_16, 0);	// preset to FAIL (no string found)
						if( INIbuf!=NULL && pINIsec!=NULL ) {	// must have a buffer, must have done an iniFindSec
							if( pINIvar==NULL ) {	// first time after iniFindSec
								pINIvar = pINIsec+12;
								INIskipline(pINIvar);
								if( *pINIvar==0 || *pINIvar=='[' ) {
									pINIvar = pINIsec = NULL;
									break;
								}
							}
							while( i+1<maxlen && (c = *pINIvar) && !INIeol(pINIvar) ) {
								if( c=='\\' ) c = *(++pINIvar);
								pMem[i++] = c;
								++pINIvar;
							}
							if( *pINIvar && INIeol(pINIvar) ) {
								SETREGD(REG_16, i);	// set to length of found string
							} else {
								SETREGD(REG_16, -1);	// set to string longer than buffer
							}
							INIskipline(pINIvar);
							if( *pINIvar==0 || *pINIvar=='[' ) pINIvar = pINIsec = NULL;
						}
						pMem[i] = 0;
					}
						break;
					default:
						break;
				}
				return;
				//**** SYS_GETWINDOWPOS
				// outputs:
				//	R15 = X
				//	R14 = Y
				//	R13 = W
				//	R12 = H
				//	R11 = system dependant
			case	SYS_GETWINDOWPOS:
			{
				DWORD	r11, r12, r13, r14, r15;
//                Dwrite("SYS_GETWINDOWPOS");
				KINT_GetWindowPos(&r11, &r12, &r13, &r14, &r15);
				SETREGD(REG_15, r15);
				SETREGD(REG_14, r14);
				SETREGD(REG_13, r13);
				SETREGD(REG_12, r12);
				SETREGD(REG_11, r11);
/*                {
                    char dBuf[128];
                    sprintf(dBuf,"SYS_GETWINDOWPOS: %d,%d,%d,%d",r15,r14,r13,r12);
                    Dwrite(dBuf);
                }
*/			}
				return;
				//**** SYS_SETWINDOWPOS
				// inputs:
				//	R15 = X
				//	R14 = Y
				//	R13 = W
				//	R12 = H
				//	R11 = system dependant
				// NOTE!!!!! This MUST NOT generate an OnWindowSize or OnWindowMove event!
				// It is assumed that the Kode knows that it has changed the window and will
				// take appropriate action, if need be.  This is handled by the 'InSetWindowPos'
				// variable in MAIN.C that 'gates' activity in the WM_SIZE code.
			case	SYS_SETWINDOWPOS:
				if( GETREGD(REG_10)==0 || (pKE->pKode[KHDR_MAJOR_INTERP_REV] & 0x7F) < 2 || ((pKE->pKode[KHDR_MAJOR_INTERP_REV] & 0x7F)==2 && pKE->pKode[KHDR_MINOR_INTERP_REV]<2) ) {
                    
					SWPopt = GETREGD(REG_11);
					if( SWPw<20 || SWPh<20 ) SWPw = SWPh = 300;
					SWPh = GETREGD(REG_12);
					SWPw = GETREGD(REG_13);
					SWPy = GETREGD(REG_14);
					SWPx = GETREGD(REG_15);
/*                    {
                        char dBuf[128];
                        sprintf(dBuf,"SYS_SETWINDOWPOS: %d,%d,%d,%d",SWPx,SWPy,SWPw,SWPh);
                        Dwrite(dBuf);
                    }
*/
                    KINT_SetWindowPos();
					SetSWPv();
				} else {
					SWPopt = GETREGD(REG_10);
					SWPh = 0;
					SWPw = 0;
					SWPy = 0;
					SWPx = 0;
					KINT_SetWindowPos();
				}
				return;
				//**** SYS_GETVIEWSIZE
				// outputs:
				//	R39 = window width
				//	R38 = window height
			case	SYS_GETVIEWSIZE:
/*            {
                char dBuf[128];
                sprintf(dBuf,"SYS_GETVIEWSIZE: %d x %d",SWPvw,SWPvh);
                Dwrite(dBuf);
            }
*/
				SETREGW(REG_39, SWPvw);
				SETREGW(REG_38, SWPvh);
				return;
				//**** SYS_ORIENTBM
				// inputs:
				//	R15 = Bitmap handle
				//	R39 = BMO_ value
			case	SYS_ORIENTBM:
				pKB = &kBM[GETREGD(REG_15)];
//				if( pKB==0 ) pKB = kBuf;
				w = GETREGW(REG_39);
				if( (w ^ pKB->orient) & BMO_90 ) {	// if rotating multiple of 90 degrees
					h = pKB->logH;					//	swap logW and logH
					pKB->logH = pKB->logW;
					pKB->logW = h;
				}
				pKB->orient = w;
				pKB->clipX1 = pKB->clipY1 = 0;
				if( w & BMO_90 ) {	// if rotated multiple of 90 degrees
					pKB->clipX2 = pKB->physH;
					pKB->clipY2 = pKB->physW;
				} else {
					pKB->clipX2 = pKB->physW;
					pKB->clipY2 = pKB->physH;
				}
				return;
				//**** SYS_GETSCREENSIZE
				// outputs:
				//	R39 = screen width
				//	R38 = screen height
				//	R37 = size index (0=640, 1=800, 2=1024, 3=1152, 4=1280, 5=1600)
				
				// NOTE: PhysicalScreenW, PhysicalScreenH, and PhysicalScreenSizeIndex *MUST* be initialized by code in 'main' module
				
			case	SYS_GETSCREENSIZE:
/*			{
				char dBuf[128];
				sprintf(dBuf,"SCREEN: %d x %d",PhysicalScreenW, PhysicalScreenH);
				Dwrite(dBuf);
			}
*/
				SETREGW(REG_39, PhysicalScreenW);
				SETREGW(REG_38, PhysicalScreenH);
				SETREGW(REG_37, PhysicalScreenSizeIndex);
				return;
				//**** SYS_FLUSHSCREEN
				// force display update/refresh RIGHT NOW (rather than waiting a later "update message" or whatever)
			case	SYS_FLUSHSCREEN:
				FlushScreen();
				return;
				//**** SYS_RECTEDGED: (Kept for backwards compatibility only.  Use SysRect.)
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = RGB color fill
				//	R13 = RGB color frame
				//	R39 = left
				//	R38 = top
				//	R37 = right
				//	R36 = bottom
			case	SYS_RECTEDGED:
				// WARN: PTR4
				RectBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
						(KLONG)(KSHORT)GETREGW(REG_37)+1, (KLONG)(KSHORT)GETREGW(REG_36)+1,
						GETREGD(REG_14), GETREGD(REG_13));
				return;
				//**** SYS_RECTSOLID: (Kept for backwards compatibility only.  Use SysRect.)
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = RGB color fill
				//	R39 = left
				//	R38 = top
				//	R37 = right
				//	R36 = bottom
			case	SYS_RECTSOLID:
				// WARN: PTR4
				RectBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
						(KLONG)(KSHORT)GETREGW(REG_37)+1, (KLONG)(KSHORT)GETREGW(REG_36)+1,
						GETREGD(REG_14), -1);
				return;
				//**** SYS_RECTOUTLINE: (Kept for backwards compatibility only.  Use SysRect.)
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R13 = RGB color frame
				//	R39 = left
				//	R38 = top
				//	R37 = right
				//	R36 = bottom
			case	SYS_RECTOUTLINE:
				// WARN: PTR4
				RectBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
						(KLONG)(KSHORT)GETREGW(REG_37)+1, (KLONG)(KSHORT)GETREGW(REG_36)+1,
						-1, GETREGD(REG_13));
				return;
				//**** SYS_RECT (KINT 2.0) and (KINT 2.2)
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = RGB color fill (-1 if none) (on 2.2, if -2, then R90-R94 are valid and do Gradient fill)
				//	R13 = RGB color frame (-1 if none) (on 2.2, if -2, then R90-R94 are valid and do Gradient fill)
				//	R39 = left
				//	R38 = top
				//	R37 = w
				//	R36 = h
				//	R90 = hue (if R13 or R14 == -2)
				//	R91 = satmin (if R13 or R14 == -2)
				//	R92 = satmax (if R13 or R14 == -2)
				//	R93 = lightmin (if R13 or R14 == -2)
				//	R94 = lightmax (if R13 or R14 == -2)
			case	SYS_RECT:
				// WARN: PTR4
				x = (KLONG)(KSHORT)GETREGW(REG_39);
				y = (KLONG)(KSHORT)GETREGW(REG_38);
				dw= GETREGD(REG_13);
				dh= GETREGD(REG_14);
				if( dw==-2 || dh==-2 ) RectGradientBMB(&kBM[GETREGD(REG_15)], x, y, (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36), GETREGB(REG_90), GETREGB(REG_91), GETREGB(REG_92), GETREGB(REG_93), GETREGB(REG_94));
				else RectBMB(&kBM[GETREGD(REG_15)], x, y, x+(KLONG)(KSHORT)GETREGW(REG_37), y+(KLONG)(KSHORT)GETREGW(REG_36), dh, dw);
				return;
				//**** SYS_OVAL (KINT 2.0)
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = RGB color fill (-1 if no fill)
				//	R13 = RGB color frame (-1 if no fill)
				//	R39 = left
				//	R38 = top
				//	R37 = w
				//	R36 = h
			case	SYS_OVAL:
				// WARN: PTR4
				x = (KLONG)(KSHORT)GETREGW(REG_39);
				y = (KLONG)(KSHORT)GETREGW(REG_38);
				OvalBMB(&kBM[GETREGD(REG_15)], x, y,
						x+(KLONG)(KSHORT)GETREGW(REG_37), y+(KLONG)(KSHORT)GETREGW(REG_36),
						GETREGD(REG_14), GETREGD(REG_13));
				return;
				//**** SYS_LINES
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = offset in memblock[R95] of (x,y) WORD coordinate array
				//	R13 = RGB color of lines
				//	R39 = # of points in the [R14] array
				//	R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset
			case	SYS_LINES:
				// WARN: PTR4
				pW = (WORD *)((BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_95))]  + GETREGD(REG_14));
				pKB = &kBM[GETREGD(REG_15)];
				
				w = GETREGW(REG_39);
				while( --w ) {
					x = (KLONG)(KSHORT)(GETMEMW(pW)); pW++;
					y = (KLONG)(KSHORT)(GETMEMW(pW)); pW++;
					LineBMB(pKB, x, y, (KLONG)(KSHORT)(GETMEMW(pW)), (KLONG)(KSHORT)(GETMEMW(pW+1)), GETREGD(REG_13));
				}
				return;
				//**** SYS_LINESTYLE
				// inputs:
				//	R12 = new linestyle (DWORD, bits set where points plotted, gets rotated as used by LineBMB and RectBMB)
				// outputs
				//	R12 = old linestyle
			case	SYS_LINESTYLE:
				dw = GETREGD(REG_12);
				SETREGD(REG_12, LineStyle);
				LineStyle = dw;
				return;
				//**** SYS_SETPIXEL:
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = RGB color of point
				//	R39 = x-coordinate of point
				//	R38 = y-coordinate of point
			case	SYS_SETPIXEL:
				PointBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38), GETREGD(REG_14));
				return;
				//**** SYS_GETPIXEL:
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R39 = x-coordinate of point
				//	R38 = y-coordinate of point
				// outputs:
				//	R14 = RGB color of point
			case	SYS_GETPIXEL:
				pKB = &kBM[GETREGD(REG_15)];
//				if( pKB==0 ) pKB = kBuf;
				x = (KLONG)(KSHORT)GETREGW(REG_39);
				y = (KLONG)(KSHORT)GETREGW(REG_38);
				
				SetupBMD(&pKB);
				if( x>=0 && x<pKB->logW && y>=0 && y<pKB->logH ) {
					pMem = Tdpbits + TdstepY*y + TdstepX*x;
#if MAC_PORT // LBO 09/16/07
					BYTE r, g, b;
					GetPixelFromSource (&r, &g, &b, pMem);
					dw = (b << 16) + (g << 8) + r;
#else
					dw = ((DWORD)(*pMem++))<<16;
					dw |=((DWORD)(*pMem++))<<8;
					dw |=((DWORD)(*pMem++));
#endif
					SETREGD(REG_14, dw);
				}
				return;
				//**** SYS_CLIP (KINT 2.0)
				// inputs:
				//	R15 = bitmap handle (NULL==screen) to clip
				//	R39 = left-coordinate of clip region (nothing is drawn < R39)
				//	R38 = top-coordinate of clip region  (nothing is drawn < R38)
				//	R37 = width of clip region (nothing is drawn >= R39+R37)
				//	R36 = height of clip region (nothing is drawn >= R38+R36)
				// NOTE: if R39==-1, then UNclip (set clip to bitmap's size)
			case	SYS_CLIP:
				pKB = &kBM[GETREGD(REG_15)];
//				if( pKB==0 ) pKB = kBuf;
				if( -1==((KLONG)(KSHORT)GETREGW(REG_39)) ) {
					pKB->clipX1 = 0;
					pKB->clipY1 = 0;
					pKB->clipX2 = pKB->logW;
					pKB->clipY2 = pKB->logH;
				} else {
					KLONG	x1, y1, x2, y2;
					
					x1 = Kmax((KLONG)(KSHORT)GETREGW(REG_39), 0);
					y1 = Kmax((KLONG)(KSHORT)GETREGW(REG_38), 0);
					x2 = Kmin((KLONG)(KSHORT)GETREGW(REG_39) + (KLONG)(KSHORT)GETREGW(REG_37), pKB->logW);
					y2 = Kmin((KLONG)(KSHORT)GETREGW(REG_38) + (KLONG)(KSHORT)GETREGW(REG_36), pKB->logH);
					
					if( x2 > pKB->physW ) x2 = pKB->physW;
					if( y2 > pKB->physH ) y2 = pKB->physH;
					
					pKB->clipX1 = x1;
					pKB->clipY1 = y1;
					pKB->clipX2 = x2;
					pKB->clipY2 = y2;
				}
				return;
				//**** SYS_GETCLIP (KINT 2.0)
				// inputs:
				//	R15 = bitmap handle (NULL==screen) to get clip settings from
				// outputs:
				//	R39 = left-coordinate of clip region (nothing is drawn < R39)
				//	R38 = top-coordinate of clip region  (nothing is drawn < R38)
				//	R37 = width of clip region (nothing is drawn >= R39+R37)
				//	R36 = height of clip region (nothing is drawn >= R38+R36)
			case	SYS_GETCLIP:
				pKB = &kBM[GETREGD(REG_15)];
//				if( pKB==0 ) pKB = kBuf;
				SETREGW(REG_39, (WORD)(KSHORT)pKB->clipX1);
				SETREGW(REG_38, (WORD)(KSHORT)pKB->clipY1);
				SETREGW(REG_37, (WORD)(KSHORT)(pKB->clipX2-pKB->clipX1));
				SETREGW(REG_36, (WORD)(KSHORT)(pKB->clipY2-pKB->clipY1));
				return;
				//**** SYS_COLORIZE
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R39 = x
				//	R38 = y
				//	R37 = w
				//	R36 = h
				//	R35 = hue (-1==no change, else 0-255)
				//	R34 = saturation (-1==no change, 0=convert to grayscale, else 1-255.  When sat==0, hue is ignored.)
				//	R33 = brightness (-255 to 255, 0==no change, else newbrightness = oldbrightness + R33 (result limited to 0-255)
				//	R32 = contrast (-255 to 255, 0==no change, else
			case	SYS_COLORIZE:
				if( GETREGW(REG_34)==0 ) GrayBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
												 (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36), (KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32));
				else ColorizeBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38),
								 (KLONG)(KSHORT)GETREGW(REG_37), (KLONG)(KSHORT)GETREGW(REG_36), (KLONG)(KSHORT)GETREGW(REG_35), (KLONG)(KSHORT)GETREGW(REG_34), (KLONG)(KSHORT)GETREGW(REG_33), (KLONG)(KSHORT)GETREGW(REG_32));
				return;
				//**** SYS_RGB2HSL	(KINT 2.0.  Changed input registers from 83-85 to 84-86 on 2.2)
				// inputs:
				//	R80 = r
				//	R81 = g
				//	R82 = b
				// outputs:
				//	R84 = hue
				//	R85 = saturation
				//	R86 = lightness
				//	R87 = 0
			case	SYS_RGB2HSL:
			{
				KLONG		r, g, b, mx, mn, v, s, hh, rc, gc, bc;
				
				r = (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_80);
				g = (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_81);
				b = (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_82);
				mx= (r>=g && r>=b)?r:((g>=r && g>=b)?g:b);
				mn= (r<=g && r<=b)?r:((g<=r && g<=b)?g:b);
				v = mx;	// lightness
				if( mx ) s = 255*(mx-mn)/mx;	// saturation
				else s = 0;
				if( s==0 ) hh = 0;
				else {
					rc = 255*(mx-r)/(mx-mn);
					gc = 255*(mx-g)/(mx-mn);
					bc = 255*(mx-b)/(mx-mn);
					
					if( r==mx ) hh = (bc-gc)/6;
					else if( g==mx ) hh = 85+(rc-bc)/6;
					else hh = 171+(gc-rc)/6;
				}
				if( s<0 ) s = 0;
				else if( s>255 ) s = 255;
				
				if( v>255 ) v = 255;
				else if( v<0 ) v = 0;
				
				SETREGB(REG_84, (BYTE)hh);
				SETREGB(REG_85, (BYTE)s);
				SETREGB(REG_86, (BYTE)v);
				SETREGB(REG_87, 0);
			}
				return;
				//**** SYS_HSL2RGB	(KINT 2.0.  Changed input registers from 83-85 to 84-86 on 2.2)
				// inputs:
				//	R84 = hue
				//	R85 = saturation
				//	R86 = lightness
				// outputs:
				//	R80 = r
				//	R81 = g
				//	R82 = b
				//	R83 = 0
			case	SYS_HSL2RGB:
			{
				KLONG		r=0, g=0, b=0, v, s, hh, f, p, q, t;
				
				v = (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_86);
				s = (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_85);
				hh= (KLONG)(KSHORT)(WORD)(BYTE)GETREGB(REG_84);
				if( s==0 ) {
					r = g = b = v;
				} else {
					if( hh>255 ) hh = 0;
					f = 6*(hh % 42);
					hh= hh / 42;
					if( hh>5 || hh<0 ) hh = 0;
					p = (v*(255-s))/255;
					q = (f==0)?v:((v*(255*246-s*f))/(255*246));
					t = (v*(255*246-(s*(246-f))))/(255*266);
					switch( hh ) {
						case 0:
							r = (BYTE)v;
							g = (BYTE)t;
							b = (BYTE)p;
							break;
						case 1:
							r = (BYTE)q;
							g = (BYTE)v;
							b = (BYTE)p;
							break;
						case 2:
							r = (BYTE)p;
							g = (BYTE)v;
							b = (BYTE)t;
							break;
						case 3:
							r = (BYTE)p;
							g = (BYTE)q;
							b = (BYTE)v;
							break;
						case 4:
							r = (BYTE)t;
							g = (BYTE)p;
							b = (BYTE)v;
							break;
						case 5:
							r = (BYTE)v;
							g = (BYTE)p;
							b = (BYTE)q;
							break;
					}
				}
				SETREGB(REG_80, (BYTE)r);
				SETREGB(REG_81, (BYTE)g);
				SETREGB(REG_82, (BYTE)b);
				SETREGB(REG_83, 0);
			}
				return;
				//**** SYS_GETTEXTSIZE
				// assumes a SYS_BEGTEXT and a SYS_TEXTSTYLE have already been done
				// inputs:
				//	R14 = string buffer address in memblk[R95]
				//	R95 = memblk for string buffer
				// outputs:
				//	R39 = width
				//	R38 = height
				//	R37 = height of non-descending portion of font
			case SYS_GETTEXTSIZE:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + GETREGD(REG_14);
				TextSizeBMB(pMem, &w, &h, &asc);
				
				SETREGW(REG_39, w);	// width of font
				SETREGW(REG_38, h);	// height of font
				SETREGW(REG_37, asc);	// height of non-descending portion of font
				return;
				//**** SYS_GETFONT
				// inputs:
				//	R13 = ptr to 512-byte buffer in memblk [R94] to be filled with font-widths (KINT 2.0)
				//			set to NULL if not needed
				//			256 byte-pairs, 1st byte is base-width, 2nd byte is cell-width
				//			those are the same for non-italic fonts, different for italic fonts (usually)
				//	R94 = memblk for font-width buffer (KINT 2.0)
				//	R14 = string buffer address in memblk[R95]
				//	R95 = memblk for string buffer
				//	R39 = pixel width of rectangle to fit text into
				//	R38 = pixel height of rectangle to fit text into
				//	R37 =
				//		BIT 0 =0 for variable pitch, =1 for fixed pitch
				//		BIT 1 =0 for normal, =1 for BOLD
				//		BIT 2 =0 for normal, =1 for ITALIC
				// outputs:
				//	R39 = handle of font
				//	R38 = height of font
				//	R37 = ascending height of font
			case	SYS_GETFONT:
				if( GETREGW(REG_38)==0 ) w = GETREGW(REG_39);	// getting info for specific font # in R39 ADDED: KINT 3.0
				else {
					pMem = (BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + GETREGD(REG_14);
					w = FitFont(pMem, GETREGW(REG_39), GETREGW(REG_38), GETREGW(REG_37));
				}
				SETREGW(REG_39, w);
				SETREGW(REG_38, *((WORD *)(pFontData + pFontTable[w])));	// height of font
				SETREGW(REG_37, *((WORD *)(pFontData + pFontTable[w] + 2)));// height of non-descending portion of font
				if( (pKE->pKode[KHDR_MAJOR_INTERP_REV] & 0x7F) >= 2 && GETREGD(REG_13)!=0 ) {	// KINT 2.0
					pMem = (BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_94))] + GETREGD(REG_13);	// where to put widths
					a = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(127-32)))));
					b = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(127-32)))+3));
					for(i=0; i<32; i++) {
						*pMem++ = b;
						*pMem++ = a;
					}
					for(; i<128; i++) {
						*pMem++ = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(i-32)))+3));
						*pMem++ = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(i-32)))));
					}
					for(; i<160; i++) {
						*pMem++ = b;
						*pMem++ = a;
					}
					for(; i<256; i++) {
						*pMem++ = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(i-64)))+3));
						*pMem++ = *((BYTE *)(pFontData + *((DWORD *)(pFontData+pFontTable[w]+6+4*(i-64)))));
					}
				}
				return;
				//**** SYS_RELFONT
				// inputs:
				//	R39 = handle of font to release
			case	SYS_RELFONT:
				// nothing to do with "internal" fonts
				return;
				//**** SYS_BEGTEXT
				// save whatever states you need to save regarding text output for "ddc"
				// so that they can be restored at SYS_ENDTEXT
				// inputs:
			case	SYS_BEGTEXT:
				Moldie[MoldCnt].BkColor = CurBackColor;
				Moldie[MoldCnt].TextColor = CurTextColor;
				Moldie[MoldCnt].Font = CurFont;
				if( MoldCnt < MAXFONTS-1 ) ++MoldCnt;
				return;
				//**** SYS_ENDTEXT
				// restore whatever states need to be restored that were saved at SYS_BEGTEXT
			case	SYS_ENDTEXT:
				if( MoldCnt ) --MoldCnt;
				CurBackColor = Moldie[MoldCnt].BkColor;
				CurTextColor = Moldie[MoldCnt].TextColor;
				CurFont = Moldie[MoldCnt].Font;
				return;
				//**** SYS_TEXTSTYLE
				//TextStyle(backcolor=-1 if transparent, textcolor, font, align)
				// inputs:
				//	R14 = backcolor (-1 if transparent, otherwise opaque)
				//	R13 = textcolor
				//	R39 = font handle
			case	SYS_TEXTSTYLE:
				CurBackColor = GETREGD(REG_14);
				CurTextColor = GETREGD(REG_13);
				CurFont = GETREGW(REG_39);
				return;
				//**** SYS_TEXT
				// inputs:
				//	R15 = bitmap handle (NULL==screen)
				//	R14 = string buffer address in memblk[R95]
				//	R95 = memblk for string buffer
				//	R39 = x-coordinate of point
				//	R38 = y-coordinate of point
			case	SYS_TEXT:
				pMem = (BYTE *)AllocPtrs[GETREGD(4*GETREGB(REG_95))] + GETREGD(REG_14);
				TextBMB(&kBM[GETREGD(REG_15)], (KLONG)(KSHORT)GETREGW(REG_39), (KLONG)(KSHORT)GETREGW(REG_38), pMem);
				return;
				//**** SYS_PLAYSOUND
				// Inputs:
				//	R15 = offset to memblock[R95] memory containing filename of sound to play
				//	R95 = 0 if R0, 1 if R1, 2 if R2
				// KINT 3.4
				//	if R95 > 2, then:
				//		R95 = 3 + (0,1,2 memblock), ie, 3,4, or 5
				//		R15 = offset to memblock[R95-3] memory containing RAM copy of a ".wav" file to play
				//		R96 = non-0 if don't play if another sound is already playing
				//	KINT 4.8
				//		R14 = len of R15 (mem sound buffer)
			case	SYS_PLAYSOUND:
				if( GETREGB(REG_95)<3 ) KPlaySound((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15));
				else KPlaySoundMem((char *)AllocPtrs[GETREGD(4*(DWORD)(GETREGB(REG_95)-3))] + GETREGD(REG_15), GETREGD(REG_14), (DWORD)GETREGB(REG_96));
				return;
				//**** SYS_PLAYMUSIC
				// Inputs:
				//	R15 = offset to memblock[R95] memory containing filename of music to play
				//	R95 = 0 if R0, 1 if R1, 2 if R2
			case	SYS_PLAYMUSIC:
				PlayMusic((char *)AllocPtrs[GETREGD(4*(DWORD)GETREGB(REG_95))] + GETREGD(REG_15));
				return;
				//**** SYS_STOPMUSIC
			case	SYS_STOPMUSIC:
				StopMusic();
				return;
				//**** SYS_ISMUSIC
				// outputs:
				//	R39=0 if no music playing, !=0 if music IS playing
			case	SYS_ISMUSIC:
				SETREGW(REG_39, MusicPlaying);
				return;
				//**** SYS_STARTTIMER
				// inputs:
				//	R15 = timer period in milliseconds
			case	SYS_STARTTIMER:
				StartTimer(GETREGD(REG_15));
				return;
				//**** SYS_STOPTIMER
			case	SYS_STOPTIMER:
				StopTimer(GETREGD(REG_15));
				return;
				//**** SYS_GETEOL
				// inputs:
				//	R15 = address in memblk R0 of where to push the OS's EOL sequence.
				//		On MS-Windows, this pushes out a 13 and a 10.
				//		On Linux and other OS's, it probably will push out ONLY a 10.
			case SYS_GETEOL:
				SETREGD(REG_15, GETREGD(REG_15) + GetSystemEOL((BYTE *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_15)));
				return;
				//**** SYS_SYSTEMMSG
				// inputs:
				//	R15 = address of string in memblk R0
			case SYS_SYSTEMMSG:
				SystemMessage((char *)AllocPtrs[GETREGD(REG_0)] + GETREGD(REG_15));
				return;
				//**** SYS_LOADKE (KINT 2.0)
				// inputs:
				//	R15 = address in memblk R95 of filename (WITHOUT the .KE extension) (may include sub-folder path)
				//	R95 = memblk for R15
			case SYS_LOADKE:
/*
 				if( NULL!=(pMem2=(BYTE *)malloc(sizeof(KE))) ) {
					memset(pMem2, 0, sizeof(KE));
					
					strcpy(BPname, (char *)(GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15)));
					for(pMem=(BYTE *)BPname; *pMem; pMem++) if( *pMem==(BYTE)'\\' ) *pMem = (BYTE)'/';
					strcat(BPname, ".ke");
					if( LoadKEFile(BPname, (KE *)pMem2) ) {
						SETREGD(REG_12, (DWORD)pMem2);
						SendMsgKE((KE *)pMem2, pKE, KHDR_BEGIN_STARTUP, (DWORD)&mainKE, 0, 0);
					} else {
						free(pMem2);
						pMem2 = NULL;
					}
				}
				SETREGD(REG_12, (DWORD)pMem2);
*/
				return;
				//**** SYS_UNLOADKE (KINT 2.0)
				// inputs:
				//	R12 = KE handle to unload
			case SYS_UNLOADKE:
/*				pMem = (BYTE *)GETREGD(REG_12);
				SendMsgKE((KE *)pMem, pKE, KHDR_SHUT_DOWN, (DWORD)&mainKE, 0, 0);
				free(((KE *)pMem)->pKode);
				free(pMem);
*/
				return;
				//**** SYS_MSGKE (KINT 2.0)
				// inputs:
				//	R32 = msg (@vector table offset, see KHDR_ #defines in PMACH.H)
				//	R15 = arg1
				//	R14 = arg2
				//	R13 = arg3
				//	R12 = target KE handle (as returned by SysLoadKE, etc)
				// Sets R11 = caller's KE handle before passing the message to the target.
				// outputs:
				//	all registers as left by target .KE (R39=value set before target's HALT instruction, etc)
			case SYS_MSGKE:
/*				a = signedmath;	// save state of SM flag
				SendMsgKE((KE *)GETREGD(REG_12), pKE, GETREGW(REG_32), GETREGD(REG_15), GETREGD(REG_14), GETREGD(REG_13));
				signedmath = a;	// restore state of SM flag
*/
				return;
		}
	}
	

//***************************************************************
// Find a font such that "text" will be as large as possible
// while still fitting within a rectangle of width "w" and height "h"
// and matching the attribute "flags".
WORD FitFont(BYTE *text, WORD w, WORD h, WORD flags) {

	int		i, i1;
	WORD	*pF;
	WORD	tw, th, ta, saveCF;
	
	if( NumFonts==0 ) return 0;	// to save us from checking all the time within this function
	
#if MAC_PORT
	MacStringFixup (text);
#endif
	saveCF = CurFont;
	for(i=0; i<NumFonts; i++) {
		FontSelection[i] = 0;
		pF = (WORD *)(pFontData + pFontTable[i]);
		th = pF[0];
		if( th <= h ) {	// if height is <= desired height
			th = pF[2];
			if( !((th ^ flags) & 1) ) {// if correct FIXED/VARIABLE pitch
				FontSelection[i] = 1;	// mark as a possibility
				if( !((th ^ flags) & 6) ) FontSelection[i] = 2; // if correct BOLD/ITALICS
				CurFont = (WORD)i;
				TextSizeBMB(text, &tw, &th, &ta);
				if( tw>w ) FontSelection[i] = 0;
			}
		}
	}
	CurFont = saveCF;
	
	// Fonts are assumed to be sorted in order of increasing height within the FONTS file.
	// So, start at the top and work down, remembering the first file with FontSelection[]==1
	// and TAKING the first file with FontSelection[]==2.  If no 2's are found, and there was
	// a '1', take the first '1'.  Else, select font 0 as the default, which should be the
	// smallest, variable pitch font.
	
	i1 = 0;
	for(i=NumFonts-1; i>=0; i--) {
		if( FontSelection[i]==2 ) return i;
		if( i1==0 && FontSelection[i]==1 ) i1 = i;
	}
	if( i1 ) return i1;
	
	return 0;
}

//*********************************************************
void RelFont(KLONG w) {

	kFONT[w].used = false;
	while( FONTcnt && !kFONT[FONTcnt-1].used ) --FONTcnt;
}
//**********************************************************
void SetTextStyle(KLONG font, DWORD textClr, DWORD backClr) {

	CurFont = font;
	CurTextColor = textClr;
	CurBackColor = backClr;
}
///***************************************************************

#define	SYSFONTMINW	2048
#define	SYSFONTNUM	18

#define	NUMSYSFONT	100
typedef struct {
	int		h, flags;
	int		prevFont, nextFont;
} SYSFONT;
//CTFontRef	hFont[NUMSYSFONT];
NSFont	*hFont[NUMSYSFONT];

SYSFONT	SysFonts[NUMSYSFONT];
long	firstSysFont;

int		SysFontW;	// width of SysFontBuf
int		SysFontH;	// height of SysFontBuf
KLONG	SysFontBMB=-1;
//NSString	*SysFontWbuf=NULL;
//int		SysFontWbufLen;
//HDC		SFdc=NULL;
//HBITMAP	oldSFbm;

///***************************************************************
/// Creates a font such that "text" will be as large as possible
/// while still fitting within a rectangle of width "w" and height "h".
NSFont *FitSysFont(NSString *text, int w, int h, WORD flags, CGSize *fontsize) {
	
	int		i, len;
	NSFont	*fnt;
	CGSize	size;
	NSString *fntName;//, *fntAttribs;

//	Dwrite("FitSysFont ENTER");
	i = h;
/*	if( flags & ST_FIXED ) {
		fntName = @"Courier";
		if( flags & ST_BOLD ) {
			if( flags & ST_ITALIC ) fntAttribs = @"Bold Oblique";
			else fntAttribs = @"Bold";
		} else if( flags & ST_ITALIC ) fntAttribs = @"Oblique";
		else fntAttribs = @"";
	} else {
		if( flags & ST_SERIF ) {
			fntName = @"Times";
			if( flags & ST_BOLD ) {
				if( flags & ST_ITALIC ) fntAttribs = @"Bold Italic";
				else fntAttribs = @"Bold";
			} else if( flags & ST_ITALIC ) {
				fntAttribs = @"Italic";
			} else fntAttribs = @"";
		} else {
			fntName = @"Helvetica";
			if( flags & ST_BOLD ) {
				if( flags & ST_ITALIC ) fntAttribs = @"Bold Oblique";
				else fntAttribs = @"Bold";
			} else if( flags & ST_ITALIC ) fntAttribs = @"Oblique";
			else fntAttribs = @"";
		}
	}

	 // create the font
	NSDictionary *fontAttribs = [NSDictionary dictionaryWithObjectsAndKeys:
								fntName, (NSString*)kCTFontFamilyNameAttribute,
								fntAttribs, (NSString *)kCTFontStyleNameAttribute,
								[NSNumber numberWithFloat:160.0],
								(NSString *)kCTFontSizeAttribute,
								nil];
	CTFontDescriptorRef descriptor = CTFontDescriptorCreateWithAttributes((CFDictionaryRef)fontAttribs);
	CTFontRef font = CTFontCreateWithFontDescriptor(descriptor, 0.0, NULL);
	CFRelease(descriptor);
	 
	// set up the string and font
	CFStringRef str = CFStringCreateWithCString(NULL, "\320\215y", kCFStringEncodingUTF8);
	CFStringRef keys[] = {kCTFontAttributeName};
	CFTypeRef values[] = {font};
	CFDictionaryRef attributes = CFDictionaryCreate(kCFAllocatorDefault, (const void**)&keys,
												(const void**)&values, sizeof(keys)/sizeof(keys[0]),
												&kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	CFAttributedStringRef attrString = CFAttributedStringCreate(kCFAllocatorDefault, str, attributes);
	CFRelease(attributes);
	 
	// set the text position and draw it
	CTLineRef line = CTLineCreateWithAttributedString(attrString);
	CGFloat	ah, dh, lh;
	CTLineGetTypographicBounds(line, &ah, &dh, &lh);
	
	CGContextSetTextPosition(kBM[0].bitmapContext, 100,  kBM[0].physH-200);// [self frame].size.height);
	CTLineDraw(line, kBM[0].bitmapContext);
	//	CGContextSetTextPosition(context, 100, 100);
	//	CTLineDraw(line, context);
	CFRelease(line);
	 
	//	NSString *tmpWbuf = [[NSString alloc] initWithUTF8String:"\320\215y"];
	 
//	NSRange	strRange = [text rangeOfString:text];

	i = 8;
*/
	len = (int)[text length];
	if( flags & ST_FIXED ) {
		if( flags & ST_BOLD ) {
			if( flags & ST_ITALIC ) fntName = @"Courier-BoldOblique";
			else fntName = @"Courier-Bold";
		} else if( flags & ST_ITALIC ) {
			fntName = @"Courier-Oblique";
		} else fntName = @"Courier";
	} else {
		if( flags & ST_SERIF ) {
			if( flags & ST_BOLD ) {
				if( flags & ST_ITALIC ) fntName = @"Times-BoldItalic";
				else fntName = @"Times-Bold";
			} else if( flags & ST_ITALIC ) {
				fntName = @"Times-Italic";
			} else fntName = @"Times";
		} else {
			if( flags & ST_BOLD ) {
				if( flags & ST_ITALIC ) fntName = @"Helvetica-BoldOblique";
				else fntName = @"Helvetica-Bold";
			} else if( flags & ST_ITALIC ) {
				fntName = @"Helvetica-Oblique";
			} else fntName = @"Helvetica";
		}
	}
	NSMutableDictionary *attribs = [[NSMutableDictionary alloc] init];
//	NSMutableAttributedString *string = [[NSMutableAttributedString alloc] initWithString:text];
//	[string beginEditing];
	do {
		++i;
//		[string addAttribute:NSFontAttributeName value:[NSFont fontWithName:fntName size:(CGFloat)i] range:strRange];
		fnt = [NSFont fontWithName:fntName size:i];
		[attribs setObject:fnt forKey:NSFontAttributeName];
		size = [text sizeWithAttributes:attribs];
	} while( len && size.width<w && size.height<h );
	do {
		--i;
//		[string beginEditing];
//		[string addAttribute:NSFontAttributeName value:[NSFont fontWithName:fntName size:(CGFloat)i] range:strRange];
//		[string endEditing];
		fnt = [NSFont fontWithName:fntName size:i];
		[attribs setObject:fnt forKey:NSFontAttributeName];
		size = [text sizeWithAttributes:attribs];

		if( i<=4 || (size.width<w && size.height<h) ) break;
	} while( i>4 && (size.width>=w || size.height>=h) );

	if( len==0 || size.width>=w || size.height>=h ) {
//		[string beginEditing];
//		[string addAttribute:NSFontAttributeName value:[NSFont fontWithName:fntName size:(CGFloat)i] range:strRange];
//		[string endEditing];
		fnt = [NSFont fontWithName:fntName size:i];
		[attribs setObject:fnt forKey:NSFontAttributeName];
		size = [text sizeWithAttributes:attribs];
	}
//	[string endEditing];

//	NSFont *somefont;
//	CGSize	size = [text sizeWithAttributes:@{NSFontAttributeName: somefont}];
/*	NSString * hws = @"Hello world!";
	 NSPoint	p;
	 
	 p = NSMakePoint(10,100);
	 [hws drawAtPoint:p withAttributes:attribs];
*/
/*
 	hdc = GetDC(NULL);
	len = wcslen(text);
	i = 8;
	logpixy = GetDeviceCaps(hdc, LOGPIXELSY);
	sFont.lfWidth = 0;
	sFont.lfEscapement = sFont.lfOrientation = 0;
	sFont.lfStrikeOut = FALSE;
	sFont.lfCharSet = ANSI_CHARSET;
	sFont.lfOutPrecision = OUT_TT_PRECIS;
	sFont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	sFont.lfQuality = PROOF_QUALITY;// DEFAULT_QUALITY;//2=PROOF_QUALITY;

	
	
	sFont.lfWeight = (flags & ST_BOLD)?FW_BOLD:FW_NORMAL;
	sFont.lfItalic = (flags & ST_ITALIC)?TRUE:FALSE;
	sFont.lfUnderline = (flags & ST_UNDERLINED)?TRUE:FALSE;
	if( flags & ST_FIXED ) {
		sFont.lfPitchAndFamily = FIXED_PITCH | FF_MODERN;
		wcscpy(sFont.lfFaceName, L"Courier New");
	} else {
		if( flags & ST_SERIF ) {
			sFont.lfPitchAndFamily = VARIABLE_PITCH | FF_ROMAN;
			wcscpy(sFont.lfFaceName, L"Times New Roman");
		} else {
			sFont.lfPitchAndFamily = VARIABLE_PITCH | FF_SWISS;
			wcscpy(sFont.lfFaceName, L"Tahoma");
		}
	}
 */
	if( fontsize!=NULL ) {
		fontsize->width = size.width;
		fontsize->height = size.height;
	}
//	Dwrite("FitSysFont EXIT");
	return fnt;
}

//**************************************
void InitSysFonts(int w, int h) {
	
//	HDC		hDC;
	
	firstSysFont = -1;
	memset(&SysFonts[0], 0xFF, sizeof(SysFonts));
	if( w<SYSFONTMINW ) w = SYSFONTMINW;
	SysFontW = w;
	SysFontH = h;
	SysFontBMB = CreateBMB(SysFontW, SysFontH, TRUE);
//	SysFontWbufLen = 2048;
//	SysFontWbuf = (WCHAR*)malloc(SysFontWbufLen * sizeof(WCHAR));
	
//	hDC = GetDC(NULL);
//	if( NULL!=(SFdc=CreateCompatibleDC(hDC)) ) {
//		SetBkMode(SFdc, TRANSPARENT);
//		SetTextAlign(SFdc, TA_LEFT | TA_TOP);
//		oldSFbm = (HBITMAP)SelectObject(SFdc, SysFontBMB->hBitmap);
//	}
//	ReleaseDC(NULL, hDC);
}

//**************************************
void FreeSysFonts() {
	
	int		i;
	
//	SelectObject(SFdc, oldSFbm);
//	DeleteDC(SFdc);
//	SFdc = NULL;
	
	while( firstSysFont >= 0 ) {
		i = SysFonts[firstSysFont].nextFont;
		hFont[firstSysFont] = nil;//DeleteObject(SysFonts[firstSysFont].hFont);
		SysFonts[firstSysFont].prevFont = SysFonts[firstSysFont].nextFont = -1;
		firstSysFont = i;
	}
	
	if( SysFontBMB>=0 ) DeleteBMB(SysFontBMB);
	SysFontBMB = -1;
//	if( SysFontWbuf!=NULL ) {
//		free(SysFontWbuf);
//		SysFontWbuf = NULL;
//	}
}

/// flags:
///	BIT 0 ( 1) = 1 if FIXED font
///	BIT 1 ( 2) = 1 if BOLD
///	BIT 2 ( 4) = 1 if ITALIC
///	BIT 3 ( 8) = 1 if UNDERLINED
///	BIT 4 (16) = 1 if SERIF
///	BIT 5 (32) = 1 if CENTERED
///	BIT 6 (64) = 1 if RIGHT-JUSTIFIED (if both BIT5 and BIT6 are set, text is CENTERED)
///	BIT 7 (128)= 1 of ROTATE counter-clockwise into target bitmap (vertical text) so the "top-left corner" of the text becomes the bottom-left corner in the target
///		NOTE: ROTATE can NOT be used if bclr==CLR_NONE.  ROTATE is forced OFF in that case.
///	BIT 8 (256)= 1 if break text on any character boundary if the text overflows (x+w).  BIT8 takes precedence over BIT9 if both are set.
///	BIT 9 (512)= 1 if break text on a SPACE (word) boundary if the text overflows (x+w).  If neither BIT8 nor BIT9 are set, then a partial character may appear at (x+w) if the text overflows.

/// BIT 15(32768) = 1 if pBM is invalid and do NOT draw the text, just return the WIDTH of the text

/// RETURN # of CHARs consumed from pText in lower WORD, pixel-width of characters drawn in upper WORD

DWORD DrawSysFontText(PBMB pBM, char *pText, int x, int y, int w, int h, int flags, DWORD bclr, DWORD fclr, long *pW) {
	
	char		c;
	int			wcnt=0, n, p, fontix=0, dx, dy;
	int			i, j, ww, savend, curend;
//	HBRUSH		hBrush;
//	RECT		rect;
//	CGSize		tsize;
//	WCHAR		tmpWbuf[8];
//	NSSize		tsize;
	NSString	*tmpWbuf;
	WORD		th;
	SYSFONT		*pSF;
	
//	Dwrite("DrawSysFontText ENTER");
	/// bail if nothing to do
	if( SysFontBMB<0 || w==0 || h==0 || *pText==0 ) return 0;
	if( bclr==0xFFFFFFFF ) flags &= ~128;	// don't allow rotate if we're not filling the background
	
	/// find or create the desired font
	for(i=(int)firstSysFont, j=0; i>=0 && (SysFonts[i].h!=h || SysFonts[i].flags!=(flags & ST_FONT_CHARS)); j++) {
		fontix = i;	// save last 'good' SysFont found (the least-used one, in case we're out of slots)
		i = SysFonts[i].nextFont;
	}
	/// if the desired font hasn't been created yet, create it
	if( i<0 ) {	/// font doesn't already exist
		if( j >= NUMSYSFONT ) {	/// there's NO room for another
			pSF = &SysFonts[fontix];	/// point to the least-used font
			if( pSF->nextFont >= 0 ) SysFonts[pSF->nextFont].prevFont = pSF->prevFont;	/// unlink it from the list
			if( pSF->prevFont >= 0 ) SysFonts[pSF->prevFont].nextFont = pSF->nextFont;
			hFont[fontix] = nil;// DeleteObject(SysFonts[fontix].hFont);	/// delete it
		} else {	/// there's room for another font, somewhere
			/// find a free slot
			for(fontix=0; fontix<NUMSYSFONT; fontix++) if( SysFonts[fontix].h<0 ) break;
		}
		/// fontix now indexes to an empty SysFonts[] slot, create the font there
		
/*		NSPoint	p;
		NSMutableDictionary *attribs;
		NSColor	*c;
		NSFont	*fnt;
		fnt = FitSysFont(hws, 1000, 100, FSFflags, NULL);
		p = NSMakePoint(10,100);
		attribs = [[NSMutableDictionary alloc] init];
		c = [NSColor redColor];
		[attribs setObject:c forKey:NSForegroundColorAttributeName];
		[attribs setObject:fnt forKey:NSFontAttributeName];
*/
		
		pSF = &SysFonts[fontix];
		
		tmpWbuf = [[NSString alloc] initWithUTF8String:"\320\215y"];
		hFont[fontix] = FitSysFont(tmpWbuf, 2048, h, flags, NULL);
		
		/// link the new font into the front of the list, as it is now most-recent
		pSF->h = h;
		pSF->flags = flags & ST_FONT_CHARS;
		pSF->prevFont = -1;
		pSF->nextFont = (int)firstSysFont;
		if( pSF->nextFont >= 0 ) SysFonts[pSF->nextFont].prevFont = fontix;	/// point previous 'first' font at this new 'first' font
		firstSysFont = fontix;
	} else {	/// font exists, move it to the front of the list (most-recently used) if it's not already there
		fontix = i;
		if( fontix != firstSysFont ) {	/// it's not first
			pSF = &SysFonts[fontix];
			n = pSF->nextFont;
			p = pSF->prevFont;
			if( n >= 0 ) SysFonts[n].prevFont = p;	/// unlink it from the list
			if( p >= 0 ) SysFonts[p].nextFont = n;
			/// we know firstSysFont != -1, because THIS font already exists and was found
			/// so just link this font into the front of the list (now most-recently used)
			pSF->prevFont = -1;
			if( firstSysFont >= 0 ) SysFonts[firstSysFont].prevFont = fontix;
			pSF->nextFont = (int)firstSysFont;
			firstSysFont = fontix;
		}
	}
/*
	CTFontRef	font=CFBridgingRetain(hFont[fontix]);
	
	// set up the string and font
	CFStringRef str = CFStringCreateWithCString(NULL, pText, kCFStringEncodingUTF8);
	CFStringRef keys[] = {kCTFontAttributeName};
	CFTypeRef values[] = {font};
	CFDictionaryRef attributes = CFDictionaryCreate(kCFAllocatorDefault, (const void**)&keys,
													(const void**)&values, sizeof(keys)/sizeof(keys[0]),
													&kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	CFAttributedStringRef attrString = CFAttributedStringCreate(kCFAllocatorDefault, str, attributes);
	CFRelease(attributes);
*/
	tmpWbuf = [[NSString alloc] initWithUTF8String:pText];
	CGSize	size;// = [tmpWbuf sizeWithAttributes:@{NSFontAttributeName: hFont[fontix]}];
	NSPoint	point;
	 
	point = NSMakePoint(0,0);
	NSMutableDictionary *attribs = [[NSMutableDictionary alloc] init];
	[attribs setObject:hFont[fontix] forKey:NSFontAttributeName];

	[attribs setObject:[NSColor colorWithRed:(CGFloat)(fclr&0xFF)/255.0 green:(CGFloat)((fclr>>8)&0xFF)/255.0 blue:(CGFloat)((fclr>>16)&0xFF)/255.0 alpha:1.0] forKey:NSForegroundColorAttributeName];
	if( flags & 8 ) [attribs setObject:[NSNumber numberWithInt:NSUnderlineStyleSingle] forKey:NSUnderlineStyleAttributeName];

	size = [tmpWbuf sizeWithAttributes:attribs];

//	[tmpWbuf drawAtPoint:point withAttributes:attribs];

	// set the text position and draw it
/*
 CTLineRef line = CTLineCreateWithAttributedString(attrString);
	CGFloat	ah, dh, lh, toth;
	CTLineGetTypographicBounds(line, &ah, &dh, &lh);
	
	CGContextSetTextPosition(kBM[SysFontBMB].bitmapContext, 0,  kBM[SysFontBMB].physH-toth);// [self frame].size.height);
*/
	/// set up for text drawing to SysFontBMB
//	SetTextColor(SFdc, fclr);
//	SelectObject(SFdc, SysFonts[firstSysFont].hFont);
	
//	SetupBMD(&SysFontBMB);
	
	/// convert UTF8 string to WCHAR
	// first, make sure the buffer is big enough
//	i = MultiByteToWideChar(CP_UTF8, 0, pText, -1, 0, 0);
//	if( i>SysFontWbufLen ) {
//		free(SysFontWbuf);
//		SysFontWbufLen = i+512;
//		SysFontWbuf = (WCHAR*)malloc(SysFontWbufLen * sizeof(WCHAR));
//	}
	// second, convert it
//	wcnt = MultiByteToWideChar(CP_UTF8, 0/*MB_PRECOMPOSED*/, pText, -1, SysFontWbuf, SysFontWbufLen);
	wcnt = (int)[tmpWbuf length];
	
//	CTLineDraw(line, kBM[SysFontBMB].bitmapContext);
//	CFRelease(line);
	/*
	 if( wcnt == 0 ) {
	 char* msg = "Unknown error";
	 
	 switch( GetLastError() ) {
	 case ERROR_INSUFFICIENT_BUFFER:
	 msg = "Buffer size not large enough";
	 break;
	 case ERROR_INVALID_FLAGS:
	 msg = "Invalid flags";
	 break;
	 case ERROR_INVALID_PARAMETER:
	 msg = "Invalid parameter";
	 break;
	 case ERROR_NO_UNICODE_TRANSLATION:
	 msg = "Invalid Unicode in string";
	 break;
	 };
	 MessageBoxA(NULL, msg, "MultiByteToWideChar error", MB_OK);
	 }
	 */
	curend = wcnt;	/// because wcnt includes the NUL char, apparently
	/// now draw the string (possibly multiple times if it's too long for the space)
	ww = -1;	/// flag this is the first time through
	
	dy = SysFontH /*kBM[SysFontBMB].physH*/ - size.height-[hFont[fontix] descender];
	[NSGraphicsContext saveGraphicsState];
	[NSGraphicsContext setCurrentContext:[NSGraphicsContext graphicsContextWithCGContext:kBM[SysFontBMB].bitmapContext flipped:FALSE]];

	if( bclr==0xFFFFFFFF ) {
		BltBMB(&kBM[SysFontBMB], 0, 0, pBM, x, y, w, h, (int)FALSE, (int)FALSE);
	} else {
		[[NSColor colorWithCalibratedRed:(CGFloat)(bclr & 0xFF)/255.0 green:(CGFloat)((bclr>>8)&0xFF)/255.0 blue:(CGFloat)((bclr>>16)&0xFF)/255.0 alpha:1.0] set];
		NSRectFill(NSMakeRect(0, kBM[SysFontBMB].physH-h, (flags & 4) ? min(w+h, 2048) : w, h));
	}
	NSMutableString *frag = [@"" mutableCopy];
	
	if( size.width>w ) {
		if( (flags & 512) ) {	// if breaking on SPACE
			curend = 0;
			int lastSpace=0;
			for(curend=0; curend<wcnt; curend++) {
				[frag appendString:[tmpWbuf substringWithRange:NSMakeRange(curend, 1)]];
				if( [frag characterAtIndex:curend]==' ' ) lastSpace = curend;
				if( [frag sizeWithAttributes:attribs].width > w ) {
					if( curend==0 ) return 0;	// if not room to output even a single char
					if( lastSpace==0 && (flags & 256) ) {
						frag = [@"" mutableCopy];
						goto breakany;
					}
	// the following line should be commented out!  It causes breaks in the middle of words
	//				if( lastSpace==0 ) lastSpace = curend-1;
					while( curend && lastSpace>=0 && curend > lastSpace ) {
						[frag deleteCharactersInRange:NSMakeRange(curend--, 1)];
					}
					if( [frag length]==1 && [frag characterAtIndex:0]==' ' ) return 0;
					break;
				}
			}
		} else if( (flags & 256) ) {	// if breaking on ANY CHAR
breakany:
			for(curend=0; curend<wcnt; curend++) {
				[frag appendString:[tmpWbuf substringWithRange:NSMakeRange(curend, 1)]];
				if( [frag sizeWithAttributes:attribs].width > w ) {
					if( curend==0 ) return 0;	// if not room to output even a single char
					[frag deleteCharactersInRange:NSMakeRange(curend, 1)];
					break;
				}
			}
		} else {	// no break
			frag = [tmpWbuf mutableCopy];
		}
	} else frag = [tmpWbuf mutableCopy];
	dx = 0;
//	[frag characterAtIndex:0]==' ' ? h/6 : 0;	// if starts with space, give previous char a little more room so we don't clip it
	ww = [frag sizeWithAttributes:attribs].width;
    [frag drawWithRect:NSMakeRect(dx, dy, kBM[SysFontBMB].physW, min(size.height, (float)(kBM[SysFontBMB].physH-dy))) options:0 attributes:attribs];
//                                  size.height) options:0 attributes:attribs];
	
	if( flags & 4 ) {	// italics
		ww += h/6;	// hack! a little slop for last char of italics
	}
	
	if( !(flags & 32768) ) {	// if not just doing a "get text width," copy the drawn text
		/// flags & 32 = CENTERED
		/// flags & 64 = RIGHT-JUSTIFIED
		/// flags & 128= ROTATED counterclockwise
		/// copy SysFontBMB to pBM
		if( flags & 128 ) {	// rotate our WORK bitmap
			th = kBM[SysFontBMB].logH;					//	swap logW and logH
			kBM[SysFontBMB].logH = kBM[SysFontBMB].logW;
			kBM[SysFontBMB].logW = th;
			kBM[SysFontBMB].orient = BMO_270;
			kBM[SysFontBMB].clipX1 = kBM[SysFontBMB].clipY1 = 0;
			kBM[SysFontBMB].clipX2 = kBM[SysFontBMB].physH;
			kBM[SysFontBMB].clipY2 = kBM[SysFontBMB].physW;
		}
		if( flags & 32 ) {			/// CENTERED
			/// if transparent (no background color) we have to redo it, to get the PROPER background image
			if( bclr==0xFFFFFFFF ) {
				BltBMB(&kBM[SysFontBMB], 0, 0, pBM, x+(w-ww)/2, y, ww, h, (int)FALSE, (int)FALSE);
				ww = [frag sizeWithAttributes:attribs].width;
//				dx = [frag characterAtIndex:0]==' ' ? h/6 : 0;	// if starts with space, give previous char a little more room so we don't clip it
				[frag drawWithRect:NSMakeRect(dx, dy, kBM[SysFontBMB].physW, size.height) options:0 attributes:attribs];
				if( flags & 4 ) {	// italics
					ww += h/6;	// hack! a little slop for last char of italics
				}
			}
			if( flags & 128 ) BltBMB(pBM, x, y+(w-ww)/2, &kBM[SysFontBMB], 0, kBM[SysFontBMB].physW-ww, h, ww, (int)FALSE, (int)FALSE);
			else {
				BltBMB(pBM, x+(w-ww)/2, y, &kBM[SysFontBMB], 0, 0, ww, h, (int)FALSE, (int)FALSE);
			}
		} else if( flags & 64 ) {	/// RIGHT JUSTIFIED
			/// if transparent (no background color) we have to redo it, to get the PROPER background image
			if( bclr==0xFFFFFFFF ) {
				BltBMB(&kBM[SysFontBMB], 0, 0, pBM, x+w-ww, y, ww, h, (int)FALSE, (int)FALSE);
				ww = [frag sizeWithAttributes:attribs].width;
//				dx = [frag characterAtIndex:0]==' ' ? h/6 : 0;	// if starts with space, give previous char a little more room so we don't clip it
				[frag drawWithRect:NSMakeRect(dx, dy, kBM[SysFontBMB].physW, size.height) options:0 attributes:attribs];
				if( flags & 4 ) {	// italics
					ww += h/6;	// hack! a little slop for last char of italics
				}
			}
			if( flags & 128 ) BltBMB(pBM, x, y, &kBM[SysFontBMB], 0, kBM[SysFontBMB].physW-ww, h, ww, (int)FALSE, (int)FALSE);
			else BltBMB(pBM, x+w-ww, y, &kBM[SysFontBMB], 0, 0, ww, h, (int)FALSE, (int)FALSE);
		} else {					/// LEFT JUSTIFIED
			if( flags & 128 ) BltBMB(pBM, x, y+w-ww, &kBM[SysFontBMB], 0, kBM[SysFontBMB].physW-min(w,ww), h, min(w,ww), (int)FALSE, (int)FALSE);
			else BltBMB(pBM, x, y, &kBM[SysFontBMB], 0, 0, min(w,ww), h, (int)FALSE, (int)FALSE);
		}
		if( flags & 128 ) {	// rotate our WORK bitmap back
			th = kBM[SysFontBMB].logH;					//	swap logW and logH
			kBM[SysFontBMB].logH = kBM[SysFontBMB].logW;
			kBM[SysFontBMB].logW = th;
			kBM[SysFontBMB].orient = BMO_NORMAL;
			kBM[SysFontBMB].clipX1 = kBM[SysFontBMB].clipY1 = 0;
			kBM[SysFontBMB].clipX2 = kBM[SysFontBMB].physW;
			kBM[SysFontBMB].clipY2 = kBM[SysFontBMB].physH;
		}
	}
	[NSGraphicsContext restoreGraphicsState];

	for(i=0, savend=curend; savend && pText[i]; --savend) {
		c = pText[i];
		if( c & 0x80 ) {
			if( (c & 0xE0)==0xC0 ) {		/// 2-byte code
				c = pText[++i];
				if( (c & 0xC0)==0x80 ) ++i;
			} else if( (c & 0xF0)==0xE0 ) {	/// 3-byte code
				c = pText[++i];
				if( (c & 0xC0)==0x80 ) {
					c = pText[++i];
					if( (c & 0xC0)==0x80 ) ++i;
				}
			} else if( (c & 0xF8)==0xF0 ) {	/// 4-byte code
				c = pText[++i];
				if( (c & 0xC0)==0x80 ) {
					c = pText[++i];
					if( (c & 0xC0)==0x80 ) {
						c = pText[++i];
						if( (c & 0xC0)==0x80 ) ++i;
					}
				}
			}
		} else ++i;							/// 1-byte code
	}
	/// return length drawn (in lower word) & pixel-width of drawn text (in upper word)
//	Dwrite("DrawSysFontText EXIT");
	return (ww<<16)+i;
}

//***********************************************************
//* SysStartup() is called from MAIN during startup, and is *
//* a chance to do any one-time initialization stuff.       *
//***********************************************************
void SysStartup() {
	
	DWORD	FileSize, BytesRead, hFile;
	
//	Dwrite("SysStartUp ENTER");
	KINTmsgTimer = [NSTimer scheduledTimerWithTimeInterval:5.0/1000.0 target:pKV selector: @selector(KINTtimer:) userInfo:nil repeats:YES];
	
	NumFonts = 0;
	// Open the file: read access, prohibit write access
	hFile = Kopen("fonts", O_RDBIN, DIR_GAME);
	if( -1 != hFile ) {
		//		SEEK_SET    0
		//		SEEK_CUR    1
		//		SEEK_END    2
		FileSize = Kseek(hFile, 0, 2);	// get the file size
		Kseek(hFile, 0, 0);	// move back to start of file
		if( NULL!=(pFontData=(BYTE *)malloc(FileSize)) ) {
			BytesRead = Kread(hFile, pFontData, FileSize);
			if( BytesRead==FileSize ) {
				NumFonts = *((WORD *)pFontData);
				pFontTable = (DWORD *)(pFontData+2);
			} else {
				free(pFontData);
			}
		}
		Kclose(hFile);
	} else {
		SystemMessage("FONTS file missing");
		exit(0);
	}
//	Dwrite("SysStartUp EXIT");
}

//***********************************************************
//* SysCleanup() is called when KINT is about to exit.		*
//* It is responsible for deleting/freeing all resources	*
//* which were allocated by SYS_ opcodes.					*
//***********************************************************
void SysCleanup() {
	
	long i;

	if( KINTmsgTimer!=NULL ) {
		[KINTmsgTimer invalidate];
		KINTmsgTimer = NULL;
	}
	for(i=0; i<MAXTIMERS; i++) {
		if( Timers[i]!=NULL ) StopTimer((DWORD)i);
	}
	for(i=1; i<BMcnt; i++) {
		if( kBM[i].bitmapContext!=nil ) CGContextRelease(kBM[i].bitmapContext);
		kBM[i].bitmapContext = nil;
		if( kBM[i].cgImage!=nil ) CGImageRelease(kBM[i].cgImage);
		kBM[i].cgImage = nil;
	}
	DeleteBMB(0);
	if( pFontData!=NULL ) free(pFontData);
}

//***********************************************************
void SysExit() {
	
	while( runKEmsg || inKE ) ;
	ForceMsg(KHDR_SHUT_DOWN, 0, 0, 0);
	SysCleanup();
}

//***********************************************************
// arg2==1 indicates to game code that we're coming back from being hidden
// arg3==OS, 0==Android, 1==iOS
void SysMinimize(Boolean isMin) {
	
	while( runKEmsg ) ;
	if( KEloaded>1 ) AddMsg(KHDR_WINDOW_HIDDEN, isMin, 0, 1);
//	if( KEloaded ) AddMsg(KHDR_WINDOW_HIDDEN, isMin, 0, 1);
}
